(*
 * Copyright  2003 Cadence Berkeley Labs. Cadence Design Systems.
 *
 *)

(* Read in two predicates p1 and p2.

   Verify that they are inconsistent, and output an interpolant
   for (p1,p2).
*)

open FociDag
module Linear = FociLinear
module Simplify = FociSimplify
module TheoremProver = FociTheoremProver
module Interpolate = FociInterpolate2
		  
let verbose = false

module PredHash = Predicate.Hash

module IntHashStruct = struct
  type t = int
  let equal x y = (x = y)
  let hash x = x
end

module IntHash = Hashtbl.Make(IntHashStruct)

module ExprHash = Expression.Hash

module ConcreteHashStruct = struct
  type t = Concrete.concrete
  let equal x y = (x = y)
  let hash = Hashtbl.hash
end

module ConcreteHash = Hashtbl.Make(ConcreteHashStruct)


type cnfInfo = {
  mutable nAtoms : int;
  atomEnum : int PredHash.t;        (* enumerates atoms *)
  aMap : predicate IntHash.t;       (* atoms by number *)
  mutable splits : int list;        (* atoms we have split cases on *)
  theory_clauses : int PredHash.t; (* clauses added from theory *)
  mutable theory_disjoint : bool;   (* no common theory symbols *)
  mutable symmetric : bool;         (* do symmetric interpolation *)
  sat_cache : unit PredHash.t;      (* theory problems found SAT *)
  mutable nParts : int;                     (* number of partitions *)
  mutable theory : theory;
}

let cnf = {
  nAtoms = 0;
  atomEnum = PredHash.create 251;    (* enumerates atoms *)
  aMap = IntHash.create 251;         (* inverse of atomEnum *)
  splits = [];
  theory_clauses = PredHash.create 251;
  theory_disjoint = false;
  symmetric = false;
  sat_cache = PredHash.create 251;
  nParts = 0;
  theory = IntegerSeparation;
}


(* Utility function -- return the integers [lo,hi) as a list *)

let rec range lo hi =
  if lo < hi then lo :: (range (lo+1) hi)
  else []


(* convert a formula to CNF form, adding fresh variables *)

let freshCtr = ref 0;;

let freshVar _ = 
  let res = "$" ^ (string_of_int !freshCtr) in
    freshCtr := !freshCtr + 1;
    pwr (Atom(res));;

let litNot r =
  match puw r with
      Not(z) -> z
    | z -> pwr(Not(r));;

let litAtom r =
  match puw r with
      Not(z) -> z
    | z -> r;;

(* Convert a boolean formula to CNF. Formula may be a DAG *)

let toCNFlist p =
  let cnf = ref [] in
  let tab = PredHash.create 251 in
  let rec f p =
    (match puw p with
	 Not(q) -> (litNot (f q))
       | And(l) -> 
	   (try PredHash.find tab p with
		Not_found ->
		  let args = List.map f l in
		  let r = freshVar () in
		  let _ = if false then (Predicate.print Format.std_formatter r;
					 Format.printf ":";
					 Predicate.print Format.std_formatter p;
					 Format.print_newline ();
					) in
		  let cls =
		    (pwr(Or(r :: (List.map litNot args))))
		    :: (List.map (fun s -> pwr(Or[(litNot r); s])) args) in
		    begin
		      cnf := cls @ !cnf;
		      PredHash.replace tab p r;
		      r
		    end)
       | Or(l) -> 
	   (try PredHash.find tab p with
		Not_found ->
		  let args = List.map f l in
		  let r = freshVar () in
		  let cls =
		    (pwr(Or((litNot r) :: args)))
		    :: (List.map (fun s -> pwr(Or[r; (litNot s)])) args) in
		    begin
		      cnf := cls @ !cnf;
		      PredHash.replace tab p r;
		      r
		    end)
       | True -> 
	   let r = freshVar () in
	   let cls = [pwr(Or([r]))] in
	     begin
	       cnf := cls @ !cnf;
	       r
	     end
       | False -> 
	   let r = freshVar () in
	   let cls = [pwr(Or([litNot r]))] in
	     begin
	       cnf := cls @ !cnf;
	       r
	     end
       | _ -> p)
  in let r = List.map f p in
    (r, !cnf);;

let toCNF p =
  let ([r],l) = toCNFlist [p] in
    (r,l)
    
(* A SAT solver clause is a list of numeric literals. Return the dual
   cube of a sat solver clause, expressed as a Predicate, removing
   uninterpreted predicates. *)
	    
let caseSplitsToCube satInst =
  let caseSplitLit a =
    let value = Sat.value satInst a in
    let pred = IntHash.find cnf.aMap a in
      if value then pred else pNot pred
  in
    List.map caseSplitLit cnf.splits

let satClauseToCube satInst cls = 
  (* Printf.printf "clause length: %d\n" (List.length cls); *)
  if false then (
    let atoms =
      (List.map
	 (fun lit ->
	    if lit > 0 then pwr(Not(IntHash.find cnf.aMap lit))
	    else (IntHash.find cnf.aMap (-lit)))
	 cls) in
      
    let atoms = (caseSplitsToCube satInst) @ atoms in
      
      (* filter out the uninterpreted predicates *)
      
      pwr(And
	    (List.filter
	       (fun p -> match puw p with
		    Atom(_) -> false
		  | Not(q) -> (match puw q with Atom(_) -> false | _ -> true)
		  | _ -> true)
	       atoms)))
  else (
    let atoms = ref [] in
    let l = ref cls in
      while !l != [] do
	let lit = List.hd !l in
	let _ = l := List.tl !l in
	let atomNum = if lit < 0 then -lit else lit in
	let atom = IntHash.find cnf.aMap atomNum in
	  match atom with
	      Atom(_),_ -> ()
	    | _ -> 
		let symlit = if lit > 0 then pNot atom else atom in
		  atoms := symlit :: !atoms
      done;
      atoms := (caseSplitsToCube satInst) @ !atoms;
      (* Printf.printf "atoms: %d\n" (List.length !atoms); *)
      pAnd(List.rev !atoms)
  )

;;


(* Same as above, removing *interpreted* predicates *)

let satClauseToAPCube cls = 

  let atoms =
    (List.map
       (fun lit ->
	  if lit > 0 then pwr(Not(IntHash.find cnf.aMap lit))
	  else (IntHash.find cnf.aMap (-lit)))
       cls) in
    
    (* filter out the interpreted predicates *)
    
    (List.filter
       (fun p -> match puw p with
	    Atom(_) -> true
	  | Not(q) -> (match puw q with Atom(_) -> true | _ -> false)
	  | _ -> false)
       atoms);;


(* Convert SAT clause to predicate *)

let satClauseToClause cls = 

  let atoms =
    (List.map
       (fun lit ->
	  if lit < 0 then pwr(Not(IntHash.find cnf.aMap (-lit)))
	  else (IntHash.find cnf.aMap (lit)))
       cls) in
    pwr(Or atoms)

(* make a theory model fully concrete *)

let concretizeModel a = 
  let h = ExprHash.create 251 in
  let k = ConcreteHash.create 251 in
  let rec eval x =
    match x with
	Constant(Constant.Int c),_ -> (Num.num_of_int c)
      | Sum l,_ ->
	  List.fold_left Num.add_num (Num.num_of_int 0) (List.map eval l)
      | Coeff (Constant.Int(c),e),_ -> Num.mult_num (Num.num_of_int c) (eval e)
      | x -> 
	  try ExprHash.find h x 
	  with Not_found -> (failwith "no value for term")
  in
  let res = ref [] in
    begin
      List.iter (function (e,v) ->
		   ExprHash.add h e v) a;
      List.iter
	(function (e,v) ->
	   let lhs = match e with
	       Application(s,l),_ ->
		 Concrete.Application(s,(List.map eval l))
	     | Variable(s),_ -> Concrete.Variable(s) in
	     try ConcreteHash.find k lhs with
		 Not_found ->
		   begin
		     res := (lhs,v) :: !res;
		     ConcreteHash.add k lhs ()
		   end) a;
      !res
    end



exception TheorySAT of (expression * Num.num) list
exception SAT of (predicate list) * ((Concrete.concrete * Num.num) list)
exception UNSAT
exception Inconsistent of FociProof.proof
  
(* Try to refute a cube, returning a theory proof *)
  
let refuteCube cube nParts part_fun opts =
  let treeAtomPartition = PredHash.create 251 in
  let _ =
    List.iter (fun t -> let d = t in
		 PredHash.add treeAtomPartition t (try part_fun d with Not_found -> failwith "refuteCube: unknown atom"))
      (match cube with And l,_ -> l) in
  let f = pImplies (cube, pFalse) in
    try (FociBstats.time "query" (TheoremProver.query f (nParts,treeAtomPartition)) opts)
    with
	TheoremProver.Failed ->
	  raise (TheorySAT (TheoremProver.counterexample ()))
      | Not_found -> failwith "Not_found in THeoremProver.query"
      
      
(* Given a cube, eliminate the negative inequalities
   by converting to positive inequalities, thus:

   ~(x <= y)  iff  y+1 <= x

   (which is true over the integers).

   Returns both a new cube and a map from new literals
   to old literals *)

let elimNegLeqInt cube =
  let h = PredHash.create 251 in
    match puw cube with
	And(l) ->
	  let bar =
	    List.map
	      (fun lit -> 
		 let foo =
		   match puw lit with
		       Not(atom) -> 
			 (match puw atom with
			      Leq(x,y) ->
				let term = FociBstats.time "elimNegLeq:comb" (Linear.comb 1 (y) 1 ) (eConstant(Constant.Int(1))) in
				  pwr(Leq(term,x))
			    | _ -> lit)
		     | x -> lit
		 in
		   begin
		     PredHash.replace h foo lit;
		     foo
		   end)
	      l
	  in
	    (h,pwr(And(bar)));;

(* Given a cube, eliminate the negative inequalities
   by converting to positive inequalities, thus:

   ~(x <= y)  iff  y <= x & x ~= y

   (which is sufficient for the rationals).

   Returns both a new cube and a map from new literals
   to old literals *)

let elimNegLeqRat cube =
  let rlits = ref [] in
  let h = PredHash.create 251 in
  let add_lit newlit oldlit =
    PredHash.replace h newlit oldlit;
    rlits := newlit :: !rlits in
    
    List.iter (function
		   (Not(Leq(x,y),_),_) as p  ->
		     add_lit (pLeq(y,x)) p;
		     add_lit (pNot(pEquality(x,y))) p
		 | p -> add_lit p p) (match cube with And lits,_ -> lits);
    (h,pwr(And(List.rev !rlits)));;


let elimNegLeq cube =
  if cnf.theory = RationalLinear then
    elimNegLeqRat cube
  else elimNegLeqInt cube

(* Add a clause c to partition i, updating the atom parition.
   return clause number. Returns clause number. *)

let addClause satInst c atomPartition i =
  begin
    (* for every atom in this clause... *)
    
    (match puw c with
	 Or(l) ->
	   List.iter (fun lit ->
			let atom = litAtom lit in
			  
			  (* update partition of this atom *)
			  if not(PredHash.mem atomPartition atom)
			    || (PredHash.find atomPartition atom) < i
			  then PredHash.replace atomPartition atom i;
			  
			  (* number this atom *)
			  if not(PredHash.mem cnf.atomEnum atom)
			  then begin
			    PredHash.add cnf.atomEnum atom (cnf.nAtoms + 1);
			    IntHash.add cnf.aMap (cnf.nAtoms + 1) atom;
			    cnf.nAtoms <- cnf.nAtoms + 1
			  end)
	   l);
    
    (* add this clauses to the SAT instance, translating atoms to literal numbers *)
    
    let clits = 
      (match puw c with
	   Or(l) ->
	     List.map (fun lit -> 
			 match puw lit with
			     Not(p) -> -(PredHash.find cnf.atomEnum p)
			   | _ -> PredHash.find cnf.atomEnum lit)
	     l) in
    let cnum = Sat.add_clause satInst clits in
      
      Sat.set_clause_partition satInst cnum i;
      
      (* Format.printf "clause %d in %d: " cnum i;
	 Predicate.print Format.std_formatter c;
	 Format.print_newline (); *)

      cnum

  end


(* Add a list of clauses to partition i, updating the atom parition *)
  
let addClauses satInst cls atomPartition i =
  List.iter (fun c -> ignore(addClause satInst c atomPartition i)) cls

(* protect a bare literal with a gate *)

let protect p atomPartition i =
  let foo = litAtom p in
    if (match puw foo with And _ -> false | Or _ -> false | False -> false | True -> false | _ -> true)
    then pwr(And [p]) else p
      
(* Given a proposed blocking clause, reduce it to a stronger blocking clause that
   is provable (in the ground theory), interpolate this blocking clause,
   and add the interpolated resulting clauses to the SAT solver.

   We also also maintains a map itpMap that gives, for each interpolated
   blocking clause, a pair (i,itp), where i is the partition of the clause,
   and itp is its interpolant.

*)

(* Here, we take a refutation and derive from it a set of clauses, on
   for each partition. This is used in the case when we need theory
   interpolants. Here, rmap maps proof hypotheses back to the original
   literals that generated them.
*)

let queryCtr = ref 0

let pf_to_clauses1 satInst nParts atomPartition itpMap rmap pf=
  (* compute the theory interpolants *)
    
  let clits =
    try
      let h = PredHash.create 251 in
	begin
	  let rec f p =
	    match p with
	      | _,FociProof.Hypothesis(q) -> (PredHash.replace h (q) ())
	      | _,FociProof.EqRefl(_) -> ()
	      | _,FociProof.EqSymm(q) -> f q
	      | _,FociProof.EqTrans(q1,q2) -> begin f q1; f q2 end
	      | _,FociProof.EqCong(_,l) -> List.iter (fun q -> f q) l
	      | _,FociProof.Contra(q1,q2) -> begin f q1; f q2 end
	      | _,FociProof.LeqRefl(_) -> ()
	      | _,FociProof.LeqLinear(_,q1,_,q2) -> begin f q1; f q2 end
	      | _,FociProof.LeqRewrite(_,q) -> (f q)
	      | _,FociProof.EqLeq(q1,q2) -> begin f q1; f q2 end
	      | _,FociProof.ContraLeq(p) -> f p
	      | _,FociProof.LeqEq(p) -> f p
	  in
	    f pf;
	    PredHash.fold (fun k v l -> (k::l)) h []
	end
    with Not_found -> failwith "Not_found in Foci.queryClause:clits"
  in
    
(*  let clits = match puw cube with And(flas) -> flas in *)
  let itps =
    try
      List.map
	(fun j ->
	   let (localLits,globalLits) =
	     try List.partition
	       (fun a ->
		  let v =
		    try
		      litAtom (PredHash.find rmap a)
		    with Not_found ->
		      Printf.printf "bad literal: ";
		      Format.print_newline ();
		      Predicate.print Format.std_formatter a;
		      Format.print_newline ();
		      Printf.printf "all literals: ";
		      Format.print_newline ();
		      Predicate.print Format.std_formatter (pwr (And clits));
		      Format.print_newline ();
		      failwith "Not_found in Foci.queryClause:itps:localLits:v"
		  in
		    ( try (PredHash.find atomPartition v) <= j
		      with Not_found ->
		      Printf.printf "bad literal: ";
		      Format.print_newline ();
		      Predicate.print Format.std_formatter v;
		      Format.print_newline ();
		      Printf.printf "all literals: ";
		      Format.print_newline ();
		      Predicate.print Format.std_formatter (pwr (And clits));
		      Format.print_newline ();
		      failwith "Not_found in Foci.queryClause:itps:localLits: <= j ")
	       )
	       clits
	     with Not_found -> failwith "Not_found in Foci.queryClause:itps:localLits"
	   in
	   (* let _ = if !queryCtr = 285 then
	     Printf.printf "foo!\n" in *)
	   let locals = List.length localLits in
	   let lits = localLits @ globalLits in
	   let itp = Interpolate.create lits pf in
	   let foo = Interpolate.interpolate (itp,(0,(locals - 1))) in
	   let foo = Simplify.simplify foo in
	     (* Not(And([foo; (satClauseToCube commonLits)])) *)
	   let _ =
 	     if false (* j = 2 *) then
	       (Printf.printf "interpolant %d:------------\n" !queryCtr;
		queryCtr := !queryCtr + 1;
		List.iter (fun p -> Predicate.print Format.std_formatter p; Format.printf " ") localLits;
		Format.print_newline ();
		List.iter (fun p -> Predicate.print Format.std_formatter p; Format.printf " ") globalLits;
		Format.print_newline ();
		Predicate.print Format.std_formatter foo;
		Format.print_newline ();
		flush stdout) in
	     foo) (range 0 (nParts - 1))
    with Not_found ->
      Printf.printf "all literals: ";
      Format.print_newline ();
      Predicate.print Format.std_formatter (pwr (And clits));
      Format.print_newline ();
      failwith "Not_found in Foci.queryClause:itps"
  in


  let itps = itps @ [pwr False] in  (* last interpolant always false *)
    
  (* build a reduced blocking clause from hypotheses in the proof *)
    
  let clits = List.map (PredHash.find rmap) clits in
  let block_lits = List.map litNot clits in
    
  (* compute interpolated blocking clauses *)
    
  let trips = List.map2
		(fun itp i -> (i, itp, (List.filter (fun a -> (PredHash.find atomPartition (litAtom a)) = i) block_lits))) 
		itps (range 0 nParts) in
    
  let prev_itp = ref (pwr True) in
    List.iter
      (fun (i,itp,lits) ->

	 let _ = if false then (
	   let cls = itp :: lits in
	   let cls = if !prev_itp != (pwr True) then (litNot !prev_itp) :: cls else cls in
	     Predicate.print Format.std_formatter (pwr (And cls));
	     Format.print_newline ();
	 ) in
	   
	 let r = if itp = (pwr False) then itp else
	   let pitp = protect itp atomPartition i in
	   let (r,l) = toCNF pitp in
	   let _ = addClauses satInst l atomPartition (i+1) in r in
	 let clits = if r = (pwr False) then lits else r :: lits in
	 let clits = if !prev_itp = (pwr True) then clits else (litNot !prev_itp) :: clits in
	 let clsid = addClause satInst (pwr (Or clits)) atomPartition i in
	   (* Format.printf "adding clause to %d: " i;
	      Predicate.print Format.std_formatter (pOr clits);
	      Format.print_newline (); *)
	   Sat.set_clause_blocking satInst clsid;
	   IntHash.replace itpMap clsid (i,itp,lits);
	   prev_itp := r)
      trips

(* Here, we generate one clause for every proof rule instance in the
   proof.  This is only used if theory interpolants are not needed (i.e.,
   if every theory symbol is local to one partition, so all the
   interpolants are propositional) *)

let pf_to_clauses2 satInst nParts atomPartition itpMap rmap n pf =
  
  let memo = FociProof.ProofHash.create 251 in (* remember subproofs already done *)

  let add_clause lits = 
    let cl = pOr lits in
      if not(PredHash.mem cnf.theory_clauses cl) then
	(PredHash.add cnf.theory_clauses cl n;
	 let cn = addClause satInst cl atomPartition n in
	   (* Format.printf "adding clause %d to %d: " cn n;
	   Predicate.print Format.std_formatter cl;
	   Format.print_newline (); *) ()) 
  in

  let rec f p =
    try FociProof.ProofHash.find memo p with
	Not_found -> let r = f1 p in FociProof.ProofHash.add memo p r; r
  and f1 p =
    match p with
	  | _,FociProof.Hypothesis(q) ->
	      (try
		 let r = q in
		 let r' = PredHash.find rmap r in
		   (if r != r' then add_clause [litNot r'; r]); r
	       with Not_found -> failwith "pf_to_clauses2")
	  | _,FociProof.EqRefl(e) -> 
	      let e' = e in
	      let r = pEquality (e',e') in
		add_clause [r]; r
	  | _,FociProof.EqSymm(q) ->
	      let (Equality(x,y),_) as r' = f q in
	      let r = pEquality(y,x) in
		add_clause [litNot(r'); r]; r
	  | _,FociProof.EqTrans(q1,q2) ->
	      let (Equality(x,y),_) as r1 = f q1 in
	      let (Equality(y',z),_) as r2 = f q2 in
		if y != y' then failwith "proof error"
		else let r = pEquality(x,z) in
		  add_clause [litNot r1; litNot r2; r]; r
	  | _,FociProof.EqCong(s,l) -> 
	      let antes = List.map f l in
	      let xbar = List.map (function Equality(x,_),_ -> x) antes in
	      let ybar = List.map (function Equality(_,y),_ -> y) antes in
	      let r = pEquality(eApplication(s,xbar),eApplication(s,ybar)) in
		add_clause (r :: (List.map litNot antes)); r
	  | _,FociProof.Contra(q,_) -> ignore(f q); pFalse
	  | _,FociProof.LeqRefl(_) ->
	      let r = pLeq(zero,zero) in
		add_clause [r]; r
	  | _,FociProof.LeqLinear(c1,q1,c2,q2) ->
	      let p1 = f q1 and p2 = f q2 in
	      (match p1,p2 with
		   ((Leq(x1,y1),_) as r1), ((Leq(x2,y2),_) as r2) ->
		     let r = pLeq(zero,(Linear.comb
					  1 (Linear.comb (-1) (x1) 1 (y1))
					  1 (Linear.comb (-1) (x2) 1 (y2)))) in
		       add_clause [litNot r1; litNot r2; r]; r
		 | _ -> failwith "foo!")
	  | _,FociProof.LeqRewrite(p,q) ->
	      let p' = f q in
		(if (p' != p) then add_clause [litNot p'; p]); p
	  | _,FociProof.EqLeq(q1,q2) ->
	      let (Leq(x1,y1),_) as r1 = f q1 in
	      let (Leq(x2,y2),_) as r2 = f q2 in
	      let r = pEquality(y1,x1) in
		add_clause [litNot r1; litNot r2; r]; r
	  | _,FociProof.ContraLeq(p) ->
	      let q = f p in
		add_clause [litNot q]; q
	  | _,FociProof.LeqEq(p) ->
	      let (Equality(x1,y1),_) as r1 = f p in
	      let r = pLeq(x1,y1) in
		add_clause [litNot r1; r]; r;
  in
  let r = f pf in
    match pf with
	_,FociProof.LeqRewrite(_,_) ->       (* prover sometimes forgets ContraLeq *)
	  add_clause [litNot r]
      | _ -> ()
	  
exception SplitPart of predicate * int
	  
let queryClause satInst cls nParts atomPartition itpMap = 
  let get_cube () = FociBstats.time "elimNegLeq" elimNegLeq (FociBstats.time "satClauseToCube" (satClauseToCube satInst) cls) in

  let (rmap,cube) =  FociBstats.time "get_cube" get_cube () in
    try
      if not cnf.theory_disjoint
      then
	let pf =
	  try refuteCube cube nParts (fun d -> (PredHash.find atomPartition (litAtom (PredHash.find rmap d)))) {TheoremProver.reduce_constants = false;
														TheoremProver.gen_counterexample = true;
														TheoremProver.quick = false;
														TheoremProver.theory = cnf.theory}
	  with TheorySAT (a) -> raise (SAT((satClauseToAPCube cls),(concretizeModel a)))
	    | Not_found -> failwith "Not_found in Foci.refuteCube"
	in pf_to_clauses1 satInst nParts atomPartition itpMap rmap pf
      else
	let lits_by_part = Array.make nParts [] in
	let partition_lits () = 
	  List.iter (fun d -> let n = (PredHash.find atomPartition (litAtom (PredHash.find rmap d))) in
		       lits_by_part.(n) <- d :: lits_by_part.(n)) (match cube with (And lits),_ -> lits) in
	let _ = FociBstats.time "partition_lits" partition_lits () in
	let refuted = ref false in
	  for i = 0 to nParts-1 do
	    let canon =  pAnd (List.sort (fun (_,p) (_,q) -> p - q) lits_by_part.(i)) in
	      if PredHash.mem cnf.sat_cache canon
	      then 
		( (* Format.fprintf Format.std_formatter "hit:";
		     Predicate.print Format.std_formatter canon;
		     Format.print_newline () *) )
	      else
		try FociBstats.time "pf_to_clauses" (pf_to_clauses2 satInst nParts atomPartition itpMap rmap i)
		  (refuteCube (pAnd lits_by_part.(i)) 1 (fun d -> 0){TheoremProver.reduce_constants = false; 
								     TheoremProver.gen_counterexample = true;
								     TheoremProver.quick = true;
								     TheoremProver.theory = cnf.theory}  );
		  refuted := true
		with TheorySAT (a) -> PredHash.add cnf.sat_cache canon ()
	  done;
	  if not !refuted then
	    (ignore (try refuteCube cube nParts (fun d -> (PredHash.find atomPartition (litAtom (PredHash.find rmap d)))) {TheoremProver.reduce_constants = false;
															   TheoremProver.gen_counterexample = true;
															   TheoremProver.quick = false;
															   TheoremProver.theory = cnf.theory}
		     with TheorySAT (a) -> raise (SAT((satClauseToAPCube cls),(concretizeModel a))));
	     failwith "should have been consistent")
    with TheoremProver.Split(p,_) ->
      match p with
	  Leq(y,x),_ ->
	    let origAtom = try litAtom (PredHash.find rmap (pLeq(x,y))) with Not_found ->
	      (try raise (SplitPart(p,PredHash.find atomPartition (pEquality(x,y))))
	       with Not_found ->
		 try raise (SplitPart(p,PredHash.find atomPartition (pEquality(y,x))))
		 with Not_found -> failwith "case split on unknown atom") in
	    let n = try PredHash.find atomPartition origAtom with Not_found -> failwith "case split on atom in unknown partition" in
	      raise (SplitPart(p,n))
	       
(* Split cases on an atom. At present, this atom must be of the
   form x <= y, where y <= x or x = y or y = x is an existing atom *)
      
let caseSplit atom i nParts atomPartition =
  if not(PredHash.mem atomPartition atom)
    || (PredHash.find atomPartition atom) < i
  then PredHash.replace atomPartition atom i;
  
  (* number this atom *)
  if not(PredHash.mem cnf.atomEnum atom)
  then begin
    PredHash.add cnf.atomEnum atom (cnf.nAtoms + 1);
    IntHash.add cnf.aMap (cnf.nAtoms + 1) atom;
    cnf.nAtoms <- cnf.nAtoms + 1
  end;
  
  cnf.splits <- (PredHash.find cnf.atomEnum atom) :: cnf.splits

let interpolateCNF cls inps out =
  
  let atomPartition = PredHash.create 251 in (* maps atoms to partitions *)
  let itpMap = IntHash.create 251 in (* maps blocking clause ids to (parition,interpolant) pairs *)
  let i = Sat.make cnf.nAtoms in     (* create a sat instance *)
  let nParts = (List.length cls) in
  let partials = Array.make nParts [] in

  let nthInterp j =
    let tab = IntHash.create 251 in
    let bitp = Sat.pfpart i (if cnf.symmetric then j else 0) j in
    let rec f g' =
      if g' = (Sat.zero i) then (pwr False)
      else if g' = (Sat.one i) then (pwr True)
      else let (g,neg) = if g' < 0 then ((-g'),true) else (g',false) in
	if g <= cnf.nAtoms then 
	  (try
	     let foo = IntHash.find cnf.aMap g in if neg then pNot foo else foo
	   with Not_found -> Printf.printf "Unrecognized variable: %d\n" g; failwith "interpolateCNF")
	  
	else
	  (try IntHash.find tab g' 
	   with Not_found ->
	     let r = 
	       let la = Sat.left_arg i g in
	       let ra = Sat.right_arg i g in
		 if la = 0
		 then  (* This is a blocking clause -- fetch interpolant from map *)
		   let (part,itp,lits) = (try IntHash.find itpMap ra with Not_found -> Printf.printf "Unrecognized blocking clause: %d\n" ra; failwith "interpolateCNF")in
		   let xtra = (List.filter (fun a -> (PredHash.find atomPartition (litAtom a)) > j) lits) in
		   let foo =
		     if part < j then pwr(Or xtra)
		     else if  part = j then pwr(Or (itp :: xtra))
		     else pwr(True) in
		     (* if j = 2 then
			(Printf.printf "interpolant: ------------------\n";
			Predicate.print Format.std_formatter foo;
			Format.print_newline ();
			let cls = Sat.get_clause i ra in
			let cube = satClauseToClause cls in
			(Printf.printf "cube: ------------------\n";
			Predicate.print Format.std_formatter cube;
			Format.print_newline ())); *)
		     if neg then pNot foo else foo
		 else
		   if neg then pAnd [f (-la); f (-ra)]
		   else pOr [f la; f ra]
	     in (IntHash.replace tab g' r; r))
    in (f bitp)
  in
  let setup () = 
    begin
      (* create the clauses, and partition them *)
      
      List.iter2
	(fun l j -> addClauses i l atomPartition j)
	cls (range 0 nParts);
      
      (* set the input atoms *)

      List.iter
	(fun v ->
	   Sat.set_input i (PredHash.find cnf.atomEnum v))
	inps;

      (* set the output atom *)
      
      (let c = Sat.set_pliteral i
		 (match puw out with
		      Not(p) -> -(PredHash.find cnf.atomEnum p)
		    | _ -> PredHash.find cnf.atomEnum out) in
	 Sat.set_clause_partition i c (nParts-1));
    end
  in
    begin
      
      FociBstats.time "setup" setup ();
      
      (* iterate the solver, processing blocking clauses *)
      
      (let n = ref 0 in
       let was_SAT = ref true in
	 (try
	    (
	      while true do
		was_SAT := if !was_SAT then FociBstats.time "solve" Sat.solve i else FociBstats.time "solve" Sat.continue i;
		if !was_SAT then
		  (n := !n + 1;
		   let block_clause = Sat.blocking_clause i in
		     (try
			FociBstats.time "queryClause" (queryClause i block_clause nParts atomPartition) itpMap
		      with Not_found -> failwith "Not_found in Foci.queryClause"
			| SplitPart(p,j) -> caseSplit (p) j nParts atomPartition
		     ))
		  
		else
		  let pcid = Sat.get_proved_clause_id i in
		    if (Sat.get_clause i pcid) == [] then raise UNSAT
		    else 
		      ( (* partial result -- compute a partial interpolant *)
			let j = Sat.get_clause_partition i pcid in
			let itp = nthInterp (j - 1) in
			let itp = Simplify.simplify itp in
			let _ = if true then (Format.printf "partial %d: " (j-1);
					       Predicate.print Format.std_formatter itp;
					       Format.print_newline ()) in
			  partials.(j-1) <- itp :: partials.(j-1);
			  let itpds = match puw itp with Or x -> x | _ -> [itp] in
			  let (cls,l) = toCNFlist itpds in
			    addClauses i ((pOr cls) :: l) atomPartition j
		      )
	      done;
	    )
	  with
	      Not_found -> failwith "Not_found in InterpolateCNF"
	    | UNSAT -> if !FociDebug.debug then (Printf.printf "UNSAT! Prover calls: %d \n" !n;
						 flush stdout)
	    | SAT (x,y) -> (if !FociDebug.debug then (Printf.printf "SAT! Prover calls: %d \n" !n;
						      flush stdout);
			    Sat.destroy i;
			    raise (SAT (x,y)))
	 ));
      
      if !FociDebug.debug then (Printf.printf "case splits: %d\n" (List.length cnf.splits));

      (* compute the interpolants *)
      
      let res =
	(List.map
	   (fun j ->
	      let itp = (FociBstats.time "nthInterp" nthInterp j) in
	      let _ = if false then (Format.printf "total %d: " (j-1);
				     Predicate.print Format.std_formatter (Simplify.simplify itp);
				     Format.print_newline ()) in
	      let res = pAnd (itp :: partials.(j)) in
		(* Printf.printf "Finished computing interpolant. Size = %d\n" (Predicate.size res);
		flush stdout; *)
		(*
		  Predicate.print Format.std_formatter (f bitp);
		  Format.print_newline ();
		*)
		res) (range 0 (if cnf.symmetric then nParts else (nParts-1))))
      in (Sat.destroy i; res)
	   
    end


(* list the atoms occurring in a predicate *)

let listAtoms p =
  let h = PredHash.create 251 in
  let r = ref [] in
  let rec f p =
    match puw p with
	Not(q) -> f q
      | And(l) -> List.iter f l
      | Or(l) -> List.iter f l
      | True -> ()
      | False -> ()
      | q ->
	  if PredHash.mem h p then ()
	  else begin
	    PredHash.add h p ();
	    r := p :: !r
	  end
  in
    ignore(f p); !r;;

(* print a list of predicates *)

let printPredList l = 
  begin
    List.iter
      (fun p ->
	 Predicate.print Format.std_formatter p;
	 Format.print_newline ())
      l;
    Format.print_newline ()
  end;;

(* replace atom r by atom v in predicate p *)

let atomSubst p r v = 
  Predicate.map (fun p -> if p = r then v else p) (fun x -> x) p

(* Returns true if no two predicates in list have a theory symbol
   in common. *)
    
exception No
  
let is_theory_disjoint pl =
  let hash = Hashtbl.create 251 in
  let n = ref 0 in
    try
      List.iter
	(fun p ->
	   let sup = Predicate.support p in
	     List.iter (fun s -> if Hashtbl.mem hash s then 
			  (Printf.printf "Symbol %s occurs formulas %d and %d\n" s (Hashtbl.find hash s) !n;
			   raise No)) sup;
	     List.iter (fun s -> Hashtbl.replace hash s !n) sup;
	     n := !n + 1)
	pl; true
    with No -> false
      


(* Given a list of predicates [p1;...;pn] of length at least one,
   return a list of interpolants for the pairs:

   (p1, p2 /\ ... /\ pn)
   (p2 /\ p2, p3 /\ ... /\ pn)
   ...
   (p1 /\ .../\ p[n-1], pn).

   Raises SAT if predicates are consistent.

  *)

let interpolateN_in_theory ps symm theory =
  try
    begin

      (* clear our state *)

      cnf.nAtoms <- 0;
      PredHash.clear cnf.atomEnum;    (* enumerates atoms *)
      IntHash.clear cnf.aMap;         (* atoms by number *)
      cnf.splits <- [];               (* case splits *)
      PredHash.clear cnf.theory_clauses;
      cnf.theory_disjoint <- is_theory_disjoint ps;
      cnf.symmetric <- symm;
      PredHash.clear cnf.sat_cache;
      cnf.nParts <- List.length ps;
      cnf.theory <- theory;
      
      if symm && not cnf.theory_disjoint then
	failwith "symmetric interpolation allowed only for disjoint formulas";

      let _ = (if false then 
		 List.iter (fun foo ->
			      Printf.printf "formula:--------------------\n";
			      Predicate.print Format.std_formatter foo;
			      Format.print_newline ()) ps) in
	
      let ps = List.map (fun p -> pwr(And([p;pwr True]))) ps in (* outputs can't be an inputs *)
      let (r,ll,rs) =
	List.fold_left
	  (fun (r,ll,rs) p ->
	     let (r',l') = toCNF (pwr(And [r; p])) in
	       (r',l'::ll,r'::rs)) (pwr(True),[],[]) ps in
      let inps = (listAtoms (pwr(And ps))) in
      let out = litNot r in
      let rs = List.rev (List.tl rs) in
      let ll = List.rev ll in
	
      (*
	let _ = printPredList l1 in
	let _ = printPredList l2 in
	let _ = printPredList inps in
	let _ = Predicate.print Format.std_formatter out in
	let _ = Format.print_newline () in
      *)
	
      let fs = FociBstats.time "interpCNF" (interpolateCNF ll inps) out in
      let fs = 
	if cnf.symmetric
	then
	  let fs1 = 
	    List.map2
	      (fun f r ->
		 (match puw r with
		      Not(q) -> (atomSubst f q (pwr False))
		    | _ -> atomSubst f r (pwr True)))
	      fs (rs @ [pTrue]) in
	    List.map2
	      (fun f r ->
		 (match puw r with
		      Not(q) -> (atomSubst f q (pwr False))
		    | _ -> atomSubst f r (pwr True)))
	      fs1 (pTrue :: rs)
	else
	  List.map2
	    (fun f r ->
	       (match puw r with
		    Not(q) -> (atomSubst f q (pwr False))
		  | _ -> atomSubst f r (pwr True)))
	    fs rs
      in
      (* let _ = (Printf.printf "Finished substituting\n";
	       flush stdout) in *)

	(*
	  Predicate.print Format.std_formatter f;
	  Format.print_newline (); *)
	
      let res = List.map Simplify.simplify fs in
	begin
	  (* List.iter (fun p -> Printf.printf "simplified size: %d\n" (Predicate.size p)) res; *)
	  (*	  Predicate.print Format.std_formatter res;
		  Format.print_newline (); *)
	  (* Printf.printf "Finished simplifying\n";
	  flush stdout; *)
	  res
	end
    end
  with Not_found -> failwith "Not_found in Foci.interpolateN"
;;

let interpolateN ps = FociBstats.time "interp" (interpolateN_in_theory ps) false IntegerSeparation
			
let interpolate (p1,p2) =
  List.hd (interpolateN [p1;p2])
    
let interpolateN_symm ps =  FociBstats.time "interp" (interpolateN_in_theory ps) true IntegerSeparation

let theoryClauses () =
  let ca = Array.make cnf.nParts [] in
    PredHash.iter (fun p n -> ca.(n) <- p :: ca.(n)) cnf.theory_clauses;
    Array.to_list ca


		   
