
(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

(* $Id: options.ml,v 1.28 2008/06/17 07:19:16 rupak Exp $
 *)

(**
  Process command line options.
  
   Build database with 
     buildOptionsDatabase : unit -> unit

   Values of flags accessible with 
	isSet : string -> bool
  and
	getValueOf : valueType -> valueType
  and its incarnations.

  Note: values of all types must be added to the database with
   their default values.
   To get the value of an option, query the database with 
   the corresponding command line option, e.g., query "L"
   to get the value of the reachLabel

*)

module M = Message

type valueType = 
  Bool of bool
| String of string
| Int of int
| Float of float	
| StringList of string list
| IntList of int list

let valueToString = function
  | Bool b -> string_of_bool b
  | String s -> s 
  | Int i -> string_of_int i 
  | Float f -> string_of_float f
  | StringList ss -> String.concat "," ss 
  | IntList is -> Misc.string_of_int_list is

exception OptionNotFoundException of string

let safe_msg   = "Program Safe. Error not reachable! :-)"
let unsafe_msg = "Real Error Found! :-("
let abort_msg  = "Ack! The gremlins again!: "

let optionsDatabase = Hashtbl.create 13

let getValueOf arg =
  try
	Hashtbl.find optionsDatabase arg
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfString arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    String s -> s
  | _ -> failwith ("Options: getValueOfString -- non string type: "^arg)
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfString arg str_v =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    String s -> Hashtbl.replace optionsDatabase arg (String str_v)
  | _ -> failwith "Options: setValueOfString -- non string type"
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfInt arg =
  try 
    let a = Hashtbl.find optionsDatabase arg in
    match a with Int s -> s | _ -> failwith "Options: getValueOfInt -- non int type"
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfInt arg int_v =
  try
    let a =   Hashtbl.find optionsDatabase arg in
    match a with Int s -> Hashtbl.replace optionsDatabase arg (Int int_v)
    | _ -> failwith "Options: setValueOfInt -- non int type"
  with Not_found -> raise (OptionNotFoundException arg)


let getValueOfBool arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Bool s -> s
  | _ -> failwith "Options: getValueOfBool -- non bool type"
  with Not_found -> raise (OptionNotFoundException arg)

let setValueOfBool arg bool_v =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    Bool s -> Hashtbl.replace optionsDatabase arg (Bool bool_v)
  | _ -> failwith "Options: setValueOfBool -- non bool type"
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfStringList arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    StringList s -> s
  | _ -> failwith "Options: getValueOfStringList -- non StringList type"
  with Not_found -> raise (OptionNotFoundException arg)

let getValueOfIntList arg =
  try
  let a =   Hashtbl.find optionsDatabase arg in
  match a with 
    IntList s -> s
  | _ -> failwith "Options: getValueOfIntList -- non IntList type"
  with Not_found -> raise (OptionNotFoundException arg)

let check_options () = 
  let rv = not ((getValueOfBool "cf") && (getValueOfBool "fmc")) in
  let _ = M.msg_string M.Normal (Printf.sprintf "checking options 1 : %b" rv) in
  assert rv


let set_options opts = 
  List.iter (fun (x,y) -> Hashtbl.replace optionsDatabase x y) opts

let default_opts = 
  set_options
  [ ("L", (String "ERROR"));
    ("funlabel", (String ""));
    ("interactive", (Bool false));
    ("msvc", (Bool false));
    ("dmod", (Bool false));
    ("O", (Int 0)); 
    (* O stands for optimization level. Default is 0 (no optimizations).
       level 1 : do cone of influence slicing.
       level 2 : do partial evaluation (see also -pe) 
                          and cfg compaction. Note that cfg compaction
                          can add infeasible paths in the program, so
                          you can end up with infeasible error traces.*)
    ("bvcons", (Int 0));
    ("main", (StringList [ "main" ]));
    ("pta", (String "noflow"));
    ("pred", (String ""));
    ("shape-class", (String ""));
    ("init", (String ""));
    ("error", (String ""));
    ("query", (String ""));
    ("ax", (String ""));
    "custom", (String "");
    "subgoal", (String "");
    "spec_fun_pref", (String "__BLAST_");
    "spec_arg_pref", (String "__BLAST_PARAM_");
    "quiet", (Bool false);
    "nofp", (Bool true);
    "initialize", (Bool true);
    "simplemem", (Bool true) ;
    "simplify", (Bool false) ;
    "pe", (Bool false);
    "forget", (Bool false);
    "cfa-dot", (String "");
    "trace-dot", (String "");
    "tree-dot", (String "");
    "callgraph", (String "");
    "nocache", (Bool false);
    "reclaim", (Bool false);
    "cov", (Bool false);
    "cover", (String "");
    "covercheck", (String "single");

    "bcoverage", (Bool false);
    "stop", (Bool false);
    "norefine", (Bool false);
    "bfs", (Bool true);
    "dfs", (Bool false);
    "bnddfs", (Bool false);
    "depth", (Int 1);
    "revDepth", (Int 0); (* reverse call graph *)
    "fdepth", (Int 1);
    "cldepth", (Int 3); (* -1 indicates full unroll - modulo recursion *)

    "enablerecursion", (Bool false);
    "ecov", (Bool false);
    "exit", (Bool false); (* track _EXIT in tproj*)
   (*  "losstrans", (Bool false); *)
    "tg", (Bool false);
    "constpred", (Bool false);
    "lvc", (Bool true) ;

    "gpred", (Bool false); (* use global preds globally *)
    "bddmin", (Bool true) ;
    "bddcov", (Bool true) ;
    "bddsat", (Bool true) ;
    "lvframe", (Bool true);

    "scope", (Bool false);
    "dc", (Bool false);
    "restart", (Bool false);
    "cref", (Bool false);
    "wpsat", (Bool false);
    "auto", (String "");
    "simpquery", (Bool false);
    "uif", (Bool false);

    "cg", (Bool false);
    "mod", (Bool false);
    "modopt", (Bool true);
    "fldunroll", (Bool true);
    "predH", (Int 0); (* 0: preds from proofs, 1: add constant preds, 
                         6 ?, 7: also find_all preds if trace fails *)
    "traces", (Bool false);
    "block", (Bool true);
    "craig", (Int 2);
    "clp", (Bool false); (* Use FOCI as default (use CLP prover in near future). *)
    "bset", (Int 8);
    "localrestart", (Bool false);
    "tsd", (String "");

    "loadabs", (String "");
    "dumpabs", (String "");

    "loadtest", (Bool false);
    "skiptest", (Int 0);
    "selecttest", (Bool false);
    "checktest", (IntList []);
    "qid", (IntList []);
    "edgemode", (Bool true);
    "funmode", (Int 0);
    "timeout", (Int 60000);
    "iter-bound", (Int 0);
    "testfile", (String "test.cfa");

    "scheck", (Bool false);
    "incref", (Bool true);
    "incmod", (Bool true);
    "recompmod", (Bool true);
    "nomod", (Bool false);
    "clock", (Bool false);
    "fociUF", (Bool false);
    "fociminp", (Bool false);
    "multi", (Int 0);
    "efilter", (Bool false);
    "post", (String "slam2");
    "comp", (String "cut");
    "simpres", (Int 0);
    "bccscc", (Bool false);
    "skipfun", (Bool false);
    "bmc", (Bool false);
    "depthoffset", (Int 5);
    "errlv", (Bool false);
    "skipcheck", (Bool false);
    "errorseedfun", (String "") ;            (* internal *)
    "exitDepth", (Int 1);
    "nomp", (Bool false);
    "omega", (Int 0);
    "maxcounter", (Int 1);
    "finitecounter", (Bool false);
    "events", (Bool false);
    "sets", (Bool false);
    "getfunsize", (String "__BLAST_DISPATCH_FUNCTION");      (* internal *)
    "include-lattice", (StringList []) ;     (** lattices: [Const], <more here> *)
    "nolattice", (Bool true);
    "s", (String "simplify") ;               (* theorem provers: [simplify], smt *)
    "smtuf", (Bool false);
    "bits", (Int 4);
    "dotsep", (Bool false);
    "test", (Bool false);
    "ccured", (Bool false);
    "qf", (String "");
    "qcfile", (String "");
    "qc", (Bool true);
    "cf", (Bool false);
    "cfb", (Bool false);
    "reachloc", (Bool false);
    "sumcache", (Bool true);
    "inv", (String "");
    "alias", (String "bdd");
    "talias", (Bool false);
    "aiter", (Bool false);
    "aliasfile", (String "")  ;              (* file with list of alias pairs *)
    "easm", (String "")            ;         (* the env assumptions file *)
    "posp", (Bool false);
    "fmc", (Bool false);
    "focisymm", (Bool false);
    "lazysymm", (Bool false);
    "fmcpa", (Bool false);
    "slicesymm", (Bool false);
    "invbdd", (Bool false); (* only use with focisymm! *)
    "fqdump", (Bool false);
    "multistr", (Bool false);
    "trelc", (Int 1);
    "bddpost", (Bool false);
    "hybrid", (Bool false);
    "allsat", (Bool false);
    "bddprint", (Bool false); (* print bdds during debug etc. *)
    "focimax", (Int (-1));
    "blockfmc", (Bool false);
    "regcons", (Bool false);
    "trunc", (Int 50);
    "checktree", (Bool false);
    "showtree", (Bool false);
    "savetree", (Bool false);
    "exitvar", (String "_EXIT");
    "errorvar", (String "__BLAST_error");
    "tproj", (Bool false) ;
    "dom",  (Bool false) ;
    "tpbug", (Bool false);
    "projfun", (Bool false);
    "itpn", (Bool false);
    "timezero", (Bool false);
    "mccarthy", (Bool false);
    "rp", (Bool false);
    "prop", (Bool true);
    "preqfilter", (Bool false);
    "incdp", (String "simplify");
    "ubdp", (String "simplify")]

let usageMsg = "usage: blast <options> <source files>\n"

let sourceFiles = ref []

let append_to_string_list_option op a =
  match Hashtbl.find optionsDatabase op with
      StringList l ->
	if not (List.mem a l) then
	  Hashtbl.replace optionsDatabase op
	    (StringList (a::l))
    | _ -> failwith "append_to_string_list_option: expected StringList"

(****************************************************************)
(********** Compound Options Settings ***************************)
(****************************************************************)

let set_cref () =
  set_options 
    ["incref",(Bool false); 
     "cldepth",(Int 2)]

(* let set_mccarthy () = 
  set_options
    ["mccarthy", (Bool true);
     "lazysymm", (Bool true)] *)

let set_focimax n = 
  set_options
    ["focimax", (Int n);
     "ubdp", (String "foci")]

let unset_fmcpa () = 
  set_options 
    ["fmc", (Bool false);
     "fmcpa", (Bool false)]
    (*"itpn", (Bool false); "bddpost",(Bool false)*)

let unset_cfb () =
  set_options
    ["cf", (Bool false);
     "cfb", (Bool false) ;
     "ecov", (Bool false);
     "mod", (Bool false);
     "nolattice", (Bool true);
     "sumcache", (Bool false)]

(****************************************************************)
(********** Three top level modes *******************************)
(****************************************************************)

let set_cfb () = 
  unset_fmcpa ();
  set_options
    ["cf", (Bool true);
     "ecov", (Bool true);
     "mod", (Bool true);
     "cfb", (Bool true) ;
     "nolattice", (Bool true);
     "sumcache", (Bool true)]

let set_focisymm () = 
  set_cref ();
  unset_cfb ();
  set_options
    ["fmc", (Bool true);
     "bddpost", (Bool true);
     "itpn", (Bool true);
     "focisymm", (Bool true);
     "ubdp", (String "foci")]

let set_fmcpa () = 
  set_options 
    ["fmc", (Bool true);
     "fmcpa", (Bool true)]
     (* "bddpost", (Bool true);"itpn" (Bool true)*)

(* default mode *)
let _ = set_cfb ()

(****************************************************************)
(********** Command Line Options ********************************)
(****************************************************************)

let arg_spec = [
  ("", Arg.Unit (fun () -> ()), "\n\t\t Model Checking Options\n");  
  ("-main", Arg.String (fun a ->
(* 
	let oldlist = let ol = Hashtbl.find optionsDatabase "main" in
	  match ol with StringList l -> l 
	  | _ -> failwith "Strange type of function list" in
*)
	 Hashtbl.replace optionsDatabase "main" (StringList ([ a ] (*:: oldlist *) ))), 

   "start reachability from this function (default main)");
  ("-L", Arg.String (fun a -> Hashtbl.replace optionsDatabase "L" (String a)), 
   "target label for reachability (default ERROR)");  
  ("-funlabel", Arg.String (fun a -> Hashtbl.replace optionsDatabase "funlabel" (String a)), 
   "target function for reachability: runs each call separately. ");  
  ("-cf", Arg.Unit (fun () -> 
		      Hashtbl.replace optionsDatabase "cf" (Bool true);
		      Hashtbl.replace optionsDatabase "ecov" (Bool true);
		      Hashtbl.replace optionsDatabase "mod" (Bool true) 
		      (*Hashtbl.replace optionsDatabase "nolattice" (Bool true)*)
		   ),
   "use context free reachability");
  ("-cfb", Arg.Unit set_cfb, "use context free reachability [bfs]");
  ("-force", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "qc" (Bool false);), "Force model checking -- avoid mc query caching");
  ("-prop", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "prop" (Bool true);), "Propagate parent preds in -craig 2"); 
  ("-preqfilter", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase
   "preqfilter" (Bool true);), "Only fire queries if some pre_pred holds (-craig 2)"); 
  ("-reachloc", Arg.Unit (fun () -> 
                            Hashtbl.replace optionsDatabase "reachloc" (Bool true);
                            Hashtbl.replace optionsDatabase "norefine" (Bool true)
                         ), "Compute all unreachable locations -- supply initial abs!  ");
  ("-nocf" , Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "cf" (Bool false) ;
			 Hashtbl.replace optionsDatabase "mod" (Bool false)
		      ),
   "use context free reachability");

  ("-modopt", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "modopt" (Bool false)), "Turn OFF mod optimizations for CF checking");
  ("-bfs", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bfs" (Bool true);Hashtbl.replace optionsDatabase "dfs" (Bool false)), "breadth first search (Default)");
  ("-dfs", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dfs" (Bool true);Hashtbl.replace optionsDatabase "bfs" (Bool false)), "depth first search");
  ("-bnddfs", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bnddfs" (Bool true)), "bounded depth first search");
  ("-pred", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pred" (String a)), "seed predicates file");

  ("-qf", Arg.String (fun a -> 
    Hashtbl.replace optionsDatabase "funmode" (Int 5);
    Hashtbl.replace optionsDatabase "qf" (String a)),"query id file");
  ("-qcfile", Arg.String (fun a -> Hashtbl.replace optionsDatabase "qcfile" (String a)), "query cache file");
  ("-init", Arg.String (fun a -> Hashtbl.replace optionsDatabase "init" (String a)), "initial state file");
  ("-include-lattice", Arg.String (fun a -> match Hashtbl.find optionsDatabase "include-lattice" with
    StringList l -> Hashtbl.replace optionsDatabase "include-lattice" (StringList (a::l))
    | _ -> assert false) ,
   "what lattices to include");
  ("-nolattice", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nolattice" (Bool true)), 
   "No lattices");

  ("-lattice", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nolattice"
    (Bool false)), 
    "Activate Lattices");
  ("-enable-recursion", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "enablerecursion" (Bool true)), 
   "treat recursive function calls as skips (UNSOUND)");
(* error predicate file should be moved to spec *)
  ("-error", Arg.String (fun a -> Hashtbl.replace optionsDatabase "error" (String a)), "error predicate file");
(* *)
  ("-query", Arg.String (fun a -> Hashtbl.replace optionsDatabase "query" (String a)), "query file");

  ("-inv",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "inv" (String a)), "invariants file");
  ("-nocache", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nocache"
  (Bool true)), "do not cache theorem prover calls");
  ("-cache", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nocache"
  (Bool false)), 
   "cache theorem prover calls");
  ("-cov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "cov" (Bool true)), 
   "keep only coverable nodes");
  ("-cover", Arg.String (fun a -> 
      Hashtbl.replace optionsDatabase "cov" (Bool true);
      Hashtbl.replace optionsDatabase "cover" (String a)), 
   "Cover node heuristic -- [join,loop]");
("-covercheck", Arg.String (fun a -> 
      Hashtbl.replace optionsDatabase "cov" (Bool true);
      Hashtbl.replace optionsDatabase "covercheck" (String a);
      if a <> "single" then Hashtbl.replace optionsDatabase "ecov" (Bool false)), 
   "Cover node heuristic -- [single,predicated,predicated-delta,all]");

  ("-fqdump", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "fqdump" (Bool true)), 
   "Dump foci queries to file");

  (* ("-losstrans", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "losstrans" (Bool true)), 
   "lossy translation in fmc"); *)

  ("-trelc", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "trelc" (Int i)), 
   "Transition Relation Constraint: 0 -- use full trace, nonzero: use itp, 2 -- build conjunctive trel");
  ("-multistr", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "multistr" (Bool true)), 
   "Multiple foci strengthenings using same trace");

  ("-reclaim", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "rec" (Bool true)), 
   "dont keep whole tree around reclaim on backtrack");

  ("-s", Arg.String (fun a -> Hashtbl.replace optionsDatabase "s" (String a)),       
    "satisfiability solver to be used for post-image computation
    {vampyre,simplify,simplify-bits,cvc,foci}");
  
  ("-smt",  Arg.Unit (fun () ->
    Hashtbl.replace optionsDatabase "incdp" (String "smt");
    Hashtbl.replace optionsDatabase "s" (String "smt")),
   "use SMT solver for sat solving and incremental decision procedures");
  
  ("-smtUF",  Arg.Unit (fun () ->
    Hashtbl.replace optionsDatabase "incdp" (String "smt");
    Hashtbl.replace optionsDatabase "s" (String "smt");
    Hashtbl.replace optionsDatabase "smtuf" (Bool true)),
   "use SMT solver for sat solving and incremental decision procedures,\n\tbut only with uninterpreted function symbols (i.e. no rules like: Select(Address(?x) 0) = ?x)");

  ("-bits", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "bits" (Int a)),
   "bits to be used in bitvector queries when using simplify-bits");

  ("-ax", Arg.String (fun a -> Hashtbl.replace optionsDatabase "ax" (String a)), 
   "additional axioms for simplify to add reasoning about other theories");
  ("-alias",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "alias" (String a)), 
   "alias options\n\tOptions: If file name is bdd, invokes a bdd based alias analysis,\n\totherwise reads the alias relationships from a file. See also -pta");
  ("-aliasfile",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "aliasfile" (String a)), 
   "alias override file: same format as .pred, use only == or != predicates.");
  ("-incdp",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "incdp" (String a)), 
   "incremental decision procedures {simplify,foci}");
  ("-ubdp", Arg.String (fun a -> Hashtbl.replace optionsDatabase "ubdp" (String a)),
   "useful blocks decision procedures {simplify,foci}");
  ("-pta", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pta" (String a)), 
   "Mode for points to analysis. noflow runs a flow insensitive version, flow a 
    flow sensitive version.\n\tThis version of Blast only supports noflow.");  
  ("-nofp", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nofp" (Bool true)), "ignore function pointer calls (Unsound: Blast does not handle function pointers)");
  ("-fp", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "nofp" (Bool false)), "do not ignore function pointer calls");
  ("-noinitialize", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "initialize" (Bool false)), "do not initialize compound globals");
  ("-initialize", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "initialize" (Bool true)), "initialize compound globals (will lead to a really large basic block in initializer!)");
  ("-simplemem", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "simplemem" (Bool true)), "Simplify memory expressions as a prepass (default)");
  ("-nosimplemem", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "simplemem" (Bool false)), "Do not simplify memory expressions");
  ("-simplify", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "simplify" (Bool true)), "Simplify boolean expressions used as integers (default)");
  ("-nosimplify", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "simplify" (Bool false)), "Do not simplify boolean expressions used as integers ");

  (* ("-tsd", Arg.String (fun a -> Hashtbl.replace optionsDatabase "tsd" (String a)), 
   "trace search direction. [zig] goes zigzagging [] restarts from the end");*)
  ("-ecov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "ecov" (Bool true)), 
   "the regions are EQUAL, not just leq");
  ("-no-ecov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "ecov" (Bool false)), 
   "the regions are just LEQ");
  ("-dsq", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dsq" (Bool true)), 
   "Dotted squish");
  ("-csq", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "csq" (Bool true)), 
   "cartes'ing the Dotted squish");
  ("-tc2", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "tc2" (Bool true)), 
   "tree completeness check 2");
  ("-atc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "atc" (Bool true)), 
   "skip atomic check");
  ("-alt", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "alt" (Bool true)), 
   "skip atomic check");
  ("-lvc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "lvc" (Bool true)), 
   "Dont merge if diff. lvals when squishing");

  ("-gpred", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "gpred" (Bool true)), 
   "Use global preds globally");

  ("-bddcov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddcov" (Bool true)), 
   "use bdds when doing cover check");
  ("-nobddcov", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddcov" (Bool false)), 
   "do not use bdds when doing cover check");
  ("-bddmin", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddmin" (Bool true)), 
   "use bdds when building auto from tree");
  ("-nobddmin", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddmin" (Bool false)), 
   "do full cover check");
  ("-scope", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "scope" (Bool true)), "scoping via inferred predicates ");

  ("-wpsat", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "wpsat" (Bool true)), " keep only satisfiable disjuncts in wp ");
  ("-restart", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "restart" (Bool true)), 
   "restart computation (SLAM mode)");
  ("-cg", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "cg" (Bool true)), 
   " Output callgraph info only (and then exit)");
  ("-mod", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "mod" (Bool true)), 
   " Use mod-analysis for globals (useful with cf only)");

  ("-par", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "par" (Bool true)), 
   "Use parallel Blast data structures");
   
 ("-projfun", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "projfun" (Bool true)), 
   " Jump through functions in tproj ");
  ("-dom", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dom" (Bool true)), 
   " Use dominator based always-reach for tproj ");

  
  ("-tpbug", 
   Arg.Unit 
     (fun () -> 
	Hashtbl.replace optionsDatabase "tpbug" (Bool true);
	Hashtbl.replace optionsDatabase "tproj" (Bool true);
	Hashtbl.replace optionsDatabase "mod" (Bool true)
     ), 
     " Use trace-project AND declare an error if the cone becomes empty ");
  ("-tproj", Arg.Unit
     (fun () ->
	Hashtbl.replace optionsDatabase "tproj" (Bool true);
	Hashtbl.replace optionsDatabase "mod" (Bool true)
     ), 
     " Use TRACE-PROJECTION ");
     ("-exit", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "exit" (Bool
     true))," Track exit() during Trace Projection");
  ("-talias", Arg.Unit
     (fun () ->
	Hashtbl.replace optionsDatabase "talias" (Bool true);
     ), 
   " Use lvals in trace during refinement ");
  ("-aiter", Arg.Unit
     (fun () ->
	Hashtbl.replace optionsDatabase "aiter" (Bool true);
     ), 
     " Use alias iterators instead of actual alias queries");

  
(*   
   ("-allPreds", Arg.Unit (fun ()-> Hashtbl.replace optionsDatabase "allPreds" (Bool true)), "do focus on ALL subtraces ");
*) 

  ("-scheck", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "scheck" (Bool true)), "Additional safety checks for aliasing"); 
  ("-incref", Arg.Unit (fun a->
   Hashtbl.replace optionsDatabase "incref" (Bool true);), "incomplete counterex. analysis");
  ("-cref", Arg.Unit set_cref,"complete counterex. analysis");

("-incmod", Arg.Unit 
(fun a-> Hashtbl.replace optionsDatabase "incmod" (Bool true)),
 "Incrementally add mod hooks ");
("-noincmod", Arg.Unit
(fun a->  Hashtbl.replace optionsDatabase "incmod" (Bool false)),
 "Don't incrementally add mod hooks ");
("-dmod", Arg.Unit 
(fun a-> Hashtbl.replace optionsDatabase "dmod" (Bool true)),
 "Demand-driven tproj ");
("-recompmod", Arg.Unit 
(fun a-> Hashtbl.replace optionsDatabase "recompmod" (Bool true)),
 "Recompute mods (set by default)");
("-reloadmod", Arg.Unit 
(fun a-> Hashtbl.replace optionsDatabase "recompmod" (Bool false)),
 "Reload mods from previously saved file ");
 ("-nomod", Arg.Unit 
(fun a-> Hashtbl.replace optionsDatabase "nomod" (Bool true)),
 "No Mod info -- UNSOUND!!");


  ("-clock", Arg.Unit 
    (fun a-> Hashtbl.replace optionsDatabase "clock" (Bool true);
        Hashtbl.replace optionsDatabase "fociUF" (Bool true)
    ), "chlvals using clock"); 
  
  ("-fociUF", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "fociUF" (Bool true)),
   "In foci conversion, convert Chlval using UF ");
  ("-fociminp", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "fociminp" (Bool true)),
   "Find minimum unsat prefix (using binary search) for all traces before interpolating ");
  ("-tg", Arg.Int (fun a-> Hashtbl.replace optionsDatabase "tg" (Bool true)),
   "take all global predicates in every query");
  ("-constpred", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "constpred"
   (Bool true)), "take constant closure of predicates");
  ("-predH", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "predH" (Int a)), 
   "new predicate heuristic level\n\tOptions: 2: add x = c for all constant c, 6, 7: Additional Heuristics. Recommended Level is 7.");
  ("-block", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "block" (Bool true)), "Analyze trace as blocks ");
  ("-noblock", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "block" (Bool false)), "Use the old trace analysis ");
  ("-craig", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "craig" (Int a);
      Hashtbl.replace optionsDatabase "block" (Bool true);
      if a >=2 then Hashtbl.replace optionsDatabase "prop" (Bool true)), 
   "Use craig interpolants (FOCI) to get predicates ([1] use with scope, [2] goes the whole hog)\n\tSee the paper Abstractions from Proofs");
  ("-clp", 
   Arg.Unit (fun () -> 
	       Hashtbl.replace optionsDatabase "clp" (Bool true)
	    ), "Use CLP prover instead of FOCI");

  ("", Arg.Unit (fun () -> ()), "\n\t\t Old Heuristics that are no longer used/supported\n");

  ("-comp", Arg.String (fun a -> Hashtbl.replace optionsDatabase "comp" (String a)), 
   "Completeness level:\n\tOptions = CFB mode {lazy,cut,demand} regular mode
   {cut,path} (default cut)");  
  ("-forget", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "forget" (Bool true)), 
   "forget predicates outside subtree");
(*  ("-post", Arg.String (fun a -> Hashtbl.replace optionsDatabase "post" (String a)), 
   "post algorithm to be used:\n\tOptions are slam and H (default slam)");*)
   ("-dc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dc" (Bool true)), " dont care heuristic for predicates not in scope ");
  ("-refine", Arg.String (fun a -> Hashtbl.replace optionsDatabase "refine" (String a)), 
   "direction of refinement (fwd or bwd), default fwd. No longer used, superceded by -block and -craig.");
  ("-custom", Arg.String (fun a -> Hashtbl.replace optionsDatabase "custom" (String a)), "custom abstraction file");




  ("", Arg.Unit (fun () -> ()), "\n\t\t Program Optimization Options\n");
  ("-pe", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pe" (Bool true)), "partial evaluation (Constant propagation)");
(*  ("-subgoal", Arg.String (fun a -> Hashtbl.replace optionsDatabase "subgoal" (String a)), "subgoals file"); *)
  ("-O", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "O" (Int a)), "Program optimizations"); 
  ("-depth", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "depth" (Int a)), "Unroll Depth"); 
  ("-fdepth", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "fdepth" (Int a)), "Function Unroll Depth"); 
 ("-revDepth", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "revDepth" (Int a)), "Reverse Function Unroll Depth"); 
  ("-cldepth", Arg.Int (fun a -> Hashtbl.replace optionsDatabase "cldepth" (Int a)), "Expression Closure Unroll Depth"); 
  ("", Arg.Unit (fun () -> ()), "\n\t\t Parallel Model Checking and Races\n");  
  ("-events", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "events" (Bool true)),
   "check program with asynchronous function calls");
  ("-getfunsize", Arg.String (fun s -> Hashtbl.replace optionsDatabase
  "getfunsize" (String s)),"print size of prooftree nodes of a particular function, e.g. __BLAST_DISPATCH_FUNCTION");
  ("-sets", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "sets" (Bool true)),
   "compile sets optimizations");
  ("-checkRace", Arg.String (fun s -> Hashtbl.replace optionsDatabase "checkRace" (String s)),
   "check for races on variable accesses");
  ("-racevar", Arg.String (fun s -> Hashtbl.replace optionsDatabase "racevar" (String s)),
   "variable to check for races on");
  ("-simpres", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "simpres" (Int i)),
   "reset simplify every k steps");
  ("-skipfun", Arg.Unit (fun () -> 
			 Hashtbl.replace optionsDatabase "skipfun" (Bool true);
			 Hashtbl.replace optionsDatabase "mod" (Bool true)
		      ),
   "skip functions that dont modify any variable in the present pred set");
  ("-skipcheck", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "skipcheck" (Bool true)),
   "Check if skipped funs need unskipping in TPROJ");
  ("-bmc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bmc" (Bool true)),
   "Use bmc heuristic in skipfun");
  ("-errlv", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "errlv" (Bool true)),
   "Add functions modifying error lvals");


 
  ("-exitDepth", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "exitDepth"
    (Int i)), "how far from exit funs should be added in skipfun cone");

  ("-bccscc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bccscc"
    (Bool true)), "Use scc heuristic in skipfun seed. Yes, it is mumbo jumbo."); 

("-multi", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "multi" (Int i)),
   "Max #errors");
("-efilter", Arg.Unit (fun i -> Hashtbl.replace optionsDatabase "efilter" (Bool true)),
   "Filter traces with events");
  ("-maxcounter", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "maxcounter" (Int i)), "Max counter-size");
  ("-finitecounter", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "finitecounter" (Bool true)), "Only count upto maxcounter");
  ("-omega", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "omega" (Int i)),"Max #threads");
  ("-nomp", Arg.Unit (fun i -> Hashtbl.replace optionsDatabase "nomp" (Bool true)),
   "Dont merge parents ...");
  ("-par", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "par" (Bool true)),
   "use data structures for parallel blast");
  ("-dotsep", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "dotsep" (Bool false)),
   "squish dotted states as well...");
  ("-easm",  Arg.String (fun a -> Hashtbl.replace optionsDatabase "easm" (String a)), "environment assumptions file");

  ("", Arg.Unit (fun () -> ()), "\n\t\t Saved Abstractions and Summarization\n");  
  ("-loadabs", Arg.String (fun a-> Hashtbl.replace optionsDatabase "loadabs" (String a)), " Load Abstraction from fname.abs "); 
  ("-dumpabs", Arg.String (fun a-> Hashtbl.replace optionsDatabase "dumpabs" (String a)), " Dump Abstraction to fname.abs "); 
  ("-loadtest", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "loadtest" (Bool true)), " Load tests from fname.tst "); 
  ("-skiptest", Arg.Int (fun a-> Hashtbl.replace optionsDatabase "skiptest" (Int a)), " Skip the next n tests "); 
  ("-edgemode", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "edgemode" (Bool true)), "In funlabel mode, replace each edge individually"); 
  ("-funmode", Arg.Int (fun a-> Hashtbl.replace optionsDatabase "edgemode" (Bool false); 
                                Hashtbl.replace optionsDatabase "funmode" (Int a)), 
   "In funlabel mode, 0: replace all edges in the same function together , 1: partition according to variable"); 
  ("-selecttest", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "selecttest" (Bool true)), "Test selectively, use checktest option to register checks to be done "); 
  ("-checktest", Arg.Int (fun a-> let oldl = Hashtbl.find optionsDatabase "checktest" in match oldl with
		IntList x -> Hashtbl.replace optionsDatabase "checktest" (IntList (a::x) )
                | _ -> failwith "checktest: Unexpected failure"), "Check only these tests"); 
  ("-timeout", Arg.Int (fun a-> Hashtbl.replace optionsDatabase "timeout" (Int a)), " Time out the model checker "); 
  ("-iter-bound", Arg.Int (fun a-> Hashtbl.replace optionsDatabase "iter-bound" (Int a)), "Limit the number of big-while-loop iterations "); 
  ("-interface", Arg.String (fun a -> Hashtbl.replace optionsDatabase "interface" (String a)), "name of a function to take as describing a component interface");
  ("-component", Arg.String (fun a -> Hashtbl.replace optionsDatabase "component" (String a)), "name of a function to check for satisfying the interface");

  ("", Arg.Unit (fun () -> ()), "\n\t\t Proof Generation Options\n");
  ("-pf", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pf" (Bool true)), "generate proof tree");
  ("-pfgen",Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "pfgen" (Bool true)), "spit out vampyre proofs");
  ("-pffile", Arg.String (fun a -> Hashtbl.replace optionsDatabase "pffile" (String a)), 
   "file to write vampyre proofs (default /tmp/lf.pfs)");

  ("", Arg.Unit (fun () -> ()), "\n\t\t General Options\n");

  ("-quiet", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "quiet" (Bool true); M.set_verbosity M.Quiet), "quiet mode");
  ("-nodebug", Arg.Unit (fun () -> M.set_verbosity M.Default), "normal verbosity mode");
  ("-verbose", Arg.Unit (fun () -> M.set_verbosity M.Verbose), "verbose mode");
  ("-debug", Arg.Unit (fun () -> M.set_verbosity M.Debugging), "debug mode");
  ("-log", Arg.String (fun s -> M.enable_logging_normal s),
   "turn on logging for specified type");
  ("-logdbg", Arg.String (fun s -> M.enable_logging_debug s),
   "turn on debug logging for specified type");
  ("-loghelp", Arg.Unit (fun () -> M.print_loghelp_msg ()), "print help message for logging");
  ("-test", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "test" (Bool true)), "test mode");
  ("-testfile", Arg.String (fun s -> Hashtbl.replace optionsDatabase "testfile" (String s)), "file to which tested CFA is dumped");
  ("-cfa-dot", Arg.String (fun a -> Hashtbl.replace optionsDatabase "cfa-dot" (String a)), "output the CFA of every function in ATT dot format");
  ("-trace-dot", Arg.String (fun a -> Hashtbl.replace optionsDatabase "trace-dot" (String a)), "output the graph of the error trace in ATT dot format");
  ("-tree-dot", Arg.String (fun a -> Hashtbl.replace optionsDatabase "tree-dot" (String a)), "output the final ART in ATT dot format");
  ("-callgraph", Arg.String (fun a -> Hashtbl.replace optionsDatabase "callgraph" (String a)), "output the callgraph in ATT dot format");
  ("-auto", Arg.String (fun a -> Hashtbl.replace optionsDatabase "auto" (String a)), "output the auto of the tree in ATT dot format");
  ("-simpquery", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "simpquery" (Bool true)), "dump simplify queries to file");
  ("-uif", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "uif" (Bool true)), "arithmetic uninterpreted for refinement");

  ("-I", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "interactive" (Bool true)), "interactive shell mode");
  ("-msvc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "msvc" (Bool true) ), "parse MS VC");

  ("-norefine", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "norefine" (Bool true)), "don't refine -- only with CFB"); 
  ("-stop", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "stop" (Bool true)), "stop when it hits first counterex"); 
  ("-bcoverage", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bcoverage" (Bool true)), "stop when it hits first counterex"); 
  ("-traces", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "traces" (Bool true)), "dump bogus traces");
  ("-tracefile", Arg.String (fun a-> Hashtbl.replace optionsDatabase "tracefile" (String a)), " file containing trace information -- for trace viewer");
  ("-plaintracefile", Arg.String (fun a-> Hashtbl.replace optionsDatabase "plaintracefile" (String a)), " file containing trace information -- for Eclipse viewer");
  ("-file", Arg.String (fun a-> Hashtbl.replace optionsDatabase "file" (String a)), " file containing trace information -- for trace viewer"); 
  ("-demo", Arg.Unit (fun a-> Hashtbl.replace optionsDatabase "demo" (Bool true)), "Demo Mode for the GUI"); 
  ("-ti", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "ti" (Bool true)), "Build Transition Invariant (!)");
  ("-posp", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "posp" (Bool true)), "Use predicates positively -- i.e. pretend they are closed under negation");
  ("-bddpost", 
   Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "fmc" (Bool true);
                      Hashtbl.replace optionsDatabase "bddpost" (Bool true)), 
                      "Use BDD-based post-images, requires fmc");
 
  ("-slicesymm", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "slicesymm" (Bool true)), 
   "use predicates just discovered for refining trel");
  ("-focisymm", Arg.Unit set_focisymm, "Use BDD-based post-images, find trel via symm itp, requires fmc");
  ("-fmcpa", Arg.Unit set_fmcpa, "Use fmc instead of lmc");
  ("-timezero", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "timezero" (Bool true)), 
   "Debugging option only! Use 0 as last chlval id in trace formulas");
  ("-mccarthy", Arg.Unit (fun () -> set_options ["mccarthy", Bool true]), 
   "Use select/store in refinement. Only with (split) foci, set using focimax, without cf.");
  ("-rp", Arg.Int (fun n -> set_focimax n; set_options ["rp", Bool true; "mccarthy", Bool true; "lazysymm", Bool true]), 
   "Use range predicates. n is paramater used for focimax. Use with -mccarthy , without cf.");
  ("-itpn", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "itpn"
  (Bool true)), "Use single foci call for refinement");
  ("-invbdd", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "invbdd" (Bool true)), "Globally useful facts. Only use with -focisymm!");
  ("-bddprint", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "bddprint" (Bool true)), "DNF bddprint");
   
  ("-hybrid", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "hybrid" (Bool true)), "Hybrid post");
  ("-allsat", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "allsat" (Bool true)), "Implement Lahiri's allsat-based precise predicate abstraction in post");
  ("-fmcc", Arg.Unit (fun () -> 
     Hashtbl.replace optionsDatabase "fmcc" (Bool true); 
     Hashtbl.replace optionsDatabase "incref" (Bool false);
   ), "foci model checker concrete");
  ("-focimax", Arg.Int set_focimax, "max constant allowed in interpolant ");
  ("-blockfmc", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "blockfmc"
   (Bool true)), "use get useful blocks in fmc");
  ("-regcons", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "regcons"
   (Bool true)), "use region constraints when refining in fmc");
 
  ("-lockset", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "lockset" (Bool true)), "Do a lock set based analysis for deadlock");

 
  ("-showtree", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "showtree" (Bool true)), "Print Reachability Tree");
  ("-savetree", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "savetree" (Bool true)), "Save abstractions by iterating over Reachability Tree --use with focisymm to find the actual localization info");
 ("-checktree", Arg.Unit (fun () -> Hashtbl.replace optionsDatabase "checktree" (Bool true)), "Check Reachability Tree");
 ("-qid", Arg.Rest (fun s -> let oldq = Hashtbl.find optionsDatabase "qid" in
    match oldq with IntList x -> Hashtbl.replace optionsDatabase "qid" (IntList
    (x@[int_of_string s])) | _ -> failwith "options qid: error!"), "Run specific query -- put at
END of options! after filename");
  ("-bvcons", Arg.Int (fun i -> Hashtbl.replace optionsDatabase "bvcons" (Int i)), 
   "bitv: 0 -- all, 1 -- drop BOT, 2 -- drop IeQ, 3 -- drop triv eq")
] 

let buildOptionsDatabase () = 
  Arg.parse arg_spec (fun s -> (sourceFiles := s :: (!sourceFiles))) usageMsg ;
  sourceFiles := List.rev !sourceFiles;
  let filename_prefix =
    if !sourceFiles = [] then "__foo" else
      let fname = List.hd !sourceFiles in
      match Misc.chop fname "[.]" with 
	| [] -> fname
	| e -> List.hd e in
  Hashtbl.replace optionsDatabase "mainsourcename" (String filename_prefix)

let dumpOptionsDatabase () =
  Hashtbl.iter 
    (fun k d -> M.msg_string M.Normal (Printf.sprintf "%s :: %s" k (valueToString d)))
    optionsDatabase;
  check_options ()

let getSourceFiles () = !sourceFiles

let isSet arg =
  try
    let a = Hashtbl.find optionsDatabase arg in
    match a with 
      Bool b -> b
    | _ -> failwith "Options: isSet -- non boolean type"
  with Not_found -> 
    (M.msg_string M.Major ("\nOptions: isSet: arg '" ^ arg ^ "' not found in database!\n");
     false)

let errorseed_table : (string,bool) Hashtbl.t = Hashtbl.create  31
let set_errorseed fn_list = 
  Hashtbl.clear errorseed_table;
  List.iter (fun fname -> Hashtbl.replace errorseed_table fname true) fn_list

let get_errorseed () = Misc.hashtbl_keys errorseed_table
let is_errorseed fname = Hashtbl.mem errorseed_table fname
