/* A simplified memory allocator based on the description in
     Scalable lock-free dynamic memory allocation
     M. Michael
     PLDI 2004

   We make the following simplifying assumptions:
   1. Memory is modeled as (type-safe) arrays, so pointers are integer indices
      into array.
   2. Bit-packed fields are expanded as structures.
   3. Initially, we assume atomic operations of the allocation/free routines
   4. There is just one sizeclass
   5. there are no requests for "big" blocks that get sent to the OS 

   Author: Rupak Majumdar rupak@cs.ucla.edu
   v0.1 05/22/2008
*/

#define MEMSIZE 1024
#define BLOCKS 32
#define BLOCKSIZE 128

// states:
#define ACTIVE 	0
#define FULL   	1
#define PARTIAL	2
#define EMPTY	3

typedef struct _anchor {
  unsigned int avail;
  unsigned int count;
  unsigned int state; // active, full, partial, empty
  unsigned int tag; 
} anchor;

typedef struct _descriptor descriptor;
typedef struct _procheap   procheap;
typedef struct _sizeclass  sizeclass;
typedef struct _superblock  superblock;

struct _descriptor {
  anchor anc;
  struct _descriptor *next;
  void *sb;	// pointer to superblock
  procheap *heap; // pointer to owner procheap
  unsigned sz;
  unsigned maxcount; // = superblock size / sz
};

//UNUSED
typedef struct _active {
  superblock * ptr;
  unsigned int credits;
} active;

struct _procheap {
  active act; // initially NULL;
  descriptor *partial; // initially NULL 
  sizeclass *sc; // pointer to parent sizeclass
};

//UNUSED
struct _sizeclass {
  descriptor *partial; // pointer to list of descriptors
  unsigned int sz; // block size
  unsigned int sbsize ; // superblock size
};

/************************************************************/
// memory model
//block : data + ptr to superblock's descriptor
typedef struct _block {
 	descriptor *desc;
	int data[BLOCKSIZE];
} block;

//superblock : a descriptor and an array of blocks
struct _superblock {
	descriptor desc;
	block	   blocks[BLOCKS];
} ; 

//heap : an array of procheaps
procheap	Heap[MEMSIZE];

/*
// descriptor allocation
descriptor *desc_avail; // initially NULL

descriptor *desc_alloc() {
  descriptor *desc, *next;
  while(1) {
		desc = desc_avail;
		if (desc) {
			next = desc->next;
			// XXX this will change to SafeCAS for correct concurrent behavior
			// with lock free structures
			desc_avail = desc->next;
		} else {
			desc = allocNewSB(DESCSBSIZE);
			// organize descriptors in a linked list TODO
			desc_avail = desc->next;
			//free desc ???
  }
  return desc;
}

//assume atomic
void desc_retire(descriptor *desc) {
		descriptor *oldhead = desc_avail;
		desc->next = oldhead;
		desc_avail = desc;
}
*/

void update_active(heap, descriptor *desc, unsigned int morecredits) {
  newactive = desc;
  newactive.credits = morecredits - 1;
  heap->active = newactive;
}

void *malloc_from_active(void *heap) {
	//first step: reserve block
	newactive = heap->act;
	oldactive = heap->act;
	if (oldactive == 0) return NULL;
	if (oldactive.credits == 0) {
		newactive = NULL;
	} else {
		newactive.credits --;
	}
	heap->active = newactive;
	desc = heap->active.ptr;
	//state may be ACTIVE, PARTIAL, or FULL
	newanchor = desc->anchor;
	oldanchor = desc->anchor;
	addr = desc->sb+oldanchor.avail * desc->sz;
	next = *(unsigned *)addr;
	newanchor.avail = next;
	newanchor.tag++;
	if (oldactive.credits == 0) {
		// state must be ACTIVE
		if (oldanchor.count == 0)
			newanchor.state = FULL;
		else {
			morecredits = min(oldanchor.count, MAXCREDITS);
			newanchor.count -=morecredits;
		}
	}
	desc->anchor = newanchor;
	if (oldactive.credits == 0 && oldanchor.count > 0) {
		update_active(heap, desc, morecredits);
	*addr = desc; 
	return addr+EIGHTBYTES;
}
void *my_malloc(unsigned int sz) {
	while(1) {
		addr = malloc_from_active();
		if(addr) return addr;
	}
}
