(* $Id: smtlib_euf.ml 1788 2007-09-28 18:23:28Z jgrundy $ *)
(** EUF solver executable. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

let log_level = ref 1

let file = ref ""

let lib = ref "."

let arg_spec = Arg.align
  [("-v", Arg.Set_int log_level,"n Verbosity, default is 1");
   ("-l", Arg.Set_string lib, 
    "lib Path to SMT-LIB libraries, default is \"" ^ !lib ^ "\"")]

let usage_msg = String.concat "\n"
  ["DPLL(T)-based EUF solver for SMT-LIB files.";
   "";
   "Usage: "^(Filename.basename Sys.executable_name)^" [OPTION]... FILE";
   "Exit status 0 if satisfiable, 1 if unsatisfiable";
   "";
   "OPTIONS:"]

let set_file x = 
  file := x

IFDEF FAST THEN
let fast_flag = true
ELSE
let fast_flag = false
ENDIF

let _ = 
  Arg.parse arg_spec set_file usage_msg;
  if !file = "" then begin
    Arg.usage arg_spec usage_msg;
    raise (Arg.Bad "missing file")
  end;
  Log.set_log_level !log_level;
  Log.log_string 2 "EXECUTABLE = ";
  Log.log_string 2 Sys.executable_name;
  Log.log_newline 2 ();
  Log.log_string 2 "BENCHMARK = ";
  Log.log_string 2 !file;
  Log.log_newline 2 ();
  Log.log_string 2 "FAST = ";
  Log.log_bool 2 fast_flag;
  Log.log_newline 2 ();
  Smtlib_parser.initialize !lib Smtlib_lexer.token 
    Smtlib_grammar.theory Smtlib_grammar.logic Smtlib_grammar.benchmark;
  let attributes = Smtlib_parser.parse_benchmark !file in
  let assumptions = Smtlib_attribute.assumptions attributes in
  let formula = Smtlib_attribute.formula attributes in
  let status = Smtlib_attribute.status attributes in
  let tm =
    List.fold_left Term_logic.Standard.conjunction Term_constant.truth 
      (formula::assumptions)
  in
  let solver = new Solver_dpllt.make (new Dpll.make) (new Cc.make) in
  let satisfiable = 
    solver#add_clause [tm];
    solver#solve
  in
  let error = 
    ((status = Smtlib_attribute.Sat) && not satisfiable)
    || ((status = Smtlib_attribute.Unsat) && satisfiable)
  in
    if error then begin
      Log.log_string 0 "Error: unexpected status";
      Log.log_newline 0 ();
    end;
    if satisfiable then begin
      Log.log_string 0 "Satisfiable";
      Log.log_newline 0 ();
      exit 1
    end
    else begin
      Log.log_string 0 "Unsatisfiable";
      Log.log_newline 0 ();
      exit 0
    end;

