(* $Id: index.ml 1788 2007-09-28 18:23:28Z jgrundy $ *)
(* Bounded natural subranges and fast access containers indexed by them. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

module type IndexedType = sig
  type t
  val index: t -> int
  val inverse: int -> t
end

(*****************************************************************************)
(* Map *)

module type Map = sig
  type key
  type 'a t

  val size: 'a t -> int
  val create: 'a -> 'a t
  val is_empty: 'a t -> bool
  val mem: 'a t -> key -> bool
  val find: 'a t -> key -> 'a
  val add: 'a t -> key -> 'a -> unit
  val replace: 'a t -> key -> 'a -> unit
  val iter: (key -> 'a -> unit) -> 'a t -> unit
  val iter_const: ('a -> key -> 'b -> unit) -> 'a -> 'b t -> unit
  val fold: ('a -> key -> 'b -> 'a) -> 'a -> 'b t -> 'a
  val fold_const : ('a -> 'b -> key -> 'c -> 'b) -> 'a -> 'b -> 'c t -> 'b
  val modify: ('a -> 'a) -> 'a t -> unit
  val modify_const: ('a -> 'b ->'b) -> 'a -> 'b t -> unit
  val set_all: 'a t -> 'a -> unit
end

(* A thinly disguised Vec. *)
module Nat_map = struct

  type key = int

  type 'a t = 'a Vec.t

  let size = Vec.length

  let create x = Vec.create 0 x

  let is_empty m = Vec.length m = 0

  let mem m n = 
    assert (0 <= n); 
    n < size m

  let find = Vec.get

  let add m n x =
    if n < size m then
      Vec.set m n x
    else
      Vec.grow m (n+1) x;
    assert (find m n = x)

  let replace = Vec.set

  let iter = Vec.iteri

  let iter_const = Vec.iteri_const

  let fold = Vec.foldi_left

  let fold_const = Vec.foldi_left_const

  let modify = Vec.modify

  let modify_const = Vec.modify_const

  let set_all m x = Vec.set_all m 0 (size m) x

end

(* A thinly disguised Nat_map that first maps keys using the index
   function before calling the appropriate Nat_map function. *)
module Make_map(I: IndexedType) = struct
  
  type key = I.t

  type 'a t = 'a Nat_map.t

  let size = Nat_map.size

  let create = Nat_map.create

  let is_empty = Nat_map.is_empty

  let add m k x = Nat_map.add m (I.index k) x

  let find m k = Nat_map.find m (I.index k)

  let mem m k = Nat_map.mem m (I.index k)

  let replace m k x = Nat_map.replace m (I.index k) x

  let iter c m = 
    let c_o_inverse n d = c (I.inverse n) d in
      Nat_map.iter c_o_inverse m

  let iter_const c x m =
    let c_o_inverse x n d = c x (I.inverse n) d in
      Nat_map.iter_const c_o_inverse x m

  let fold f a m =
    let f_o_inverse a n d = f a (I.inverse n) d in
      Nat_map.fold f_o_inverse a m

  let fold_const f c a m =
    let f_o_inverse c a n d = f c a (I.inverse n) d in
      Nat_map.fold_const f_o_inverse c a m

  let modify = Nat_map.modify

  let modify_const = Nat_map.modify_const

  let set_all = Nat_map.set_all

end

(*****************************************************************************)
(* Two dimensional map *)

module type Map2 = sig
  type key0
  type key1
  type 'a t

  val create: int -> int -> 'a t
  val mem: 'a t -> key0 -> key1 -> bool
  val find: 'a t -> key0 -> key1 -> 'a
  val set: 'a t -> key0 -> key1 -> 'a -> unit
  val remove: 'a t -> key0 -> key1 -> unit
end


module Nat_map2 (J: Hashtbl.HashedType) = struct
  type key0 = int
  type key1 = J.t
      
  module J_hash = Hashtbl.Make (J)

  type 'a t = 'a J_hash.t Vec.t

  let create i_size j_size =
    Vec.init i_size (fun _ -> J_hash.create j_size)

  let mem m i j =
    i < Vec.length m && J_hash.mem (Vec.get m i) j

  let find m i j =
    if i < Vec.length m then
      J_hash.find (Vec.get m i) j
    else
      raise Not_found

  let set m i j x =
    if i >= Vec.length m then
      Vec.grow_init m (i+1) (fun _ -> J_hash.create 0);
    assert (i < Vec.length m);
    J_hash.replace (Vec.get m i) j x;
    assert (find m i j = x)
    
  let remove m i j = 
    if i < Vec.length m then
      J_hash.remove (Vec.get m i) j;
    assert (not (mem m i j))
end

module Make_map2 (I: IndexedType) (J: Hashtbl.HashedType) = struct
  type key0 = I.t
  type key1 = J.t
      
  module Nat_J_Map2 = Nat_map2 (J)

  type 'a t = 'a Nat_J_Map2.t

  let create = Nat_J_Map2.create

  let mem m i j = Nat_J_Map2.mem m (I.index i) j

  let find m i j = Nat_J_Map2.find m (I.index i) j

  let set m i j x = Nat_J_Map2.set m (I.index i) j x
    
  let remove m i j = Nat_J_Map2.remove m (I.index i) j
end

(*****************************************************************************)
(* Set *)

module type Set = sig
  type elt
  type t

  val create: unit -> t
  val size: t -> int
  val is_empty: t -> bool
  val mem: t -> elt -> bool
  val add: t -> elt -> unit
  val remove: t -> elt -> unit
  val clear: t -> unit
  val iter: (elt -> unit) -> t -> unit
  val iter_const: ('a -> elt -> unit) -> 'a -> t -> unit
  val fold: ('a -> elt -> 'a) -> 'a -> t -> 'a
  val fold_const: ('a -> 'b -> elt -> 'b) -> 'a -> 'b -> t -> 'b
end

(* A thinly disguised Vec.t. *)
module Nat_set = struct

  type elt = int

  type t = {
    mutable member: int;
    mutable size: int;
    status: int Vec.t
  }

  let invariant s = s.member > 0 && s.size >= 0

  let size s = s.size

  let is_empty s = s.size = 0

  let create () = 
    let empty = 
      {
	member = 1;
	size = 0;
	status = Vec.create 0 0
      }
    in
      assert (invariant empty);
      assert (is_empty empty);
      empty

  let mem s n = 
    assert (invariant s);
    assert (0 <= n);
    n < Vec.length s.status && Vec.get s.status n = s.member

  let add s n =
    assert (invariant s);
    assert (0 <= n);
    if n >= Vec.length s.status then Vec.grow s.status (n+1) 0;
    assert (n < Vec.length s.status);
    if Vec.get s.status n != s.member then begin
      Vec.set s.status n s.member;
      s.size <- s.size + 1
    end;
    assert (invariant s);
    assert (not (is_empty s));
    assert (mem s n)

  let remove s n =
    assert (invariant s);
    assert (0 <= n);
    if n < Vec.length s.status && Vec.get s.status n = s.member then begin
      Vec.set s.status n 0;
      s.size <- s.size - 1
    end;
    assert (invariant s);
    assert (not (mem s n))

  let clear s = 
    assert (invariant s);
    if not (is_empty s) then begin
      if s.member = max_int then begin
	Vec.set_all s.status 0 (Vec.length s.status) 0;
	s.member <- 1
      end else
	s.member <- s.member + 1;
      s.size <- 0
    end;
    assert (invariant s);
    assert (is_empty s)

  let iter c s = 
    assert (invariant s);
    let if_mem_c el status = if status = s.member then c el in
      Vec.iteri if_mem_c s.status;
      assert (invariant s)

  let iter_const c x s =
    assert (invariant s);
    let if_mem_c x el status = if status = s.member then c x el in
      Vec.iteri_const if_mem_c x s.status;
      assert (invariant s)
	
  let fold f a s =
    assert (invariant s);
    let if_mem_f a el status = if status = s.member then f a el else a in
    let result = Vec.foldi_left if_mem_f a s.status in
      assert (invariant s);
      result
	
  let fold_const f x a s =
    assert (invariant s);
    let if_mem_f x a el status = if status = s.member then f x a el else a in
    let result = Vec.foldi_left_const if_mem_f x a s.status in
      assert (invariant s);
      result

end

module Make_set(I: IndexedType) = struct

  type elt = I.t

  type t = Nat_set.t

  let create = Nat_set.create
    
  let size = Nat_set.size

  let is_empty = Nat_set.is_empty

  let add s e = Nat_set.add s (I.index e)

  let mem s e = Nat_set.mem s (I.index e)

  let remove s e = Nat_set.remove s (I.index e)

  let clear = Nat_set.clear

  let iter c s = 
    let c_o_inverse n = c (I.inverse n) in
      Nat_set.iter c_o_inverse s

  let iter_const c x s =
    let c_o_inverse x n = c x (I.inverse n) in
      Nat_set.iter_const c_o_inverse x s

  let fold f a s =
    let f_o_inverse a n = f a (I.inverse n) in
      Nat_set.fold f_o_inverse a s

  let fold_const f x a s =
    let f_o_inverse x a n = f x a (I.inverse n) in
      Nat_set.fold_const f_o_inverse x a s

end

(*****************************************************************************)
(* Priority Queue *)

module type PriorityQueue = sig
  type elt
  type t

  val create: elt -> t
  val length: t -> int
  val is_empty: t -> bool
  val mem: t -> elt -> bool
  val head: t -> elt
  val enqueue: t -> elt -> unit
  val dequeue: t -> elt
  val drop: t -> unit
  val priority: t -> elt -> float
  val inc_priority: t-> elt -> float -> unit
  val scale: t -> float -> unit
end

module Nat_priority_queue = struct
  type elt = int

  type t = {
    priority: float Nat_map.t; (* maps elements to their priority *)
    position: int Nat_map.t;   (* maps elements to queue position *)
    heap: elt Vec.t;           (* elements in the priority queue *)
  }

  (* Elements in the queue are organized as a heap : the priority of
     an element is no less than that of either of its children. *)

  (* The position map takes elements to their position in the heap
     where they may be found. We maintain the invariant that the
     position of an element is either a valid index into the queue,
     for elements that are enqueued, or to invalid_position, for
     elements that are not. *)
  let invalid_position = (-1)

  (* The root of the heap is the highest priority element and is
     stored at position 0. *)
  let root_position = 0

  (* The position of the parent of the element at index i. Note that
     parent_position of root_position is root_position. *)
  let parent_position i = (i - 1)/2

  (* Each heap element has at most 2 children.  For an element stored
     at position i, the left child will be at position 2*i+1 and the
     right child will be at 2*i+2.  An element has a child if the
     position of that would-be child falls within the index bounds of
     the element vector. *)
  let left_child_position i  = 2*i + 1

  (* The position of the right sibling of the node at position i.
     This is a valid position only if the node at i has a right
     sibling. *)
  let right_sibling_position i = i + 1  

  (* Ensures that e is mapped by the priority and position maps.  We
     maintain the invariant that both maps have entries for the same
     set of keys. *)
  let ensure_map_mem q e =
    if not (Nat_map.mem q.priority e) then begin
      Nat_map.add q.priority e 0.0;
      Nat_map.add q.position e invalid_position;
    end

  (* Check if element x has higher priority than element y. *)
  let higher_priority q x y =
    let x_score = Nat_map.find q.priority x in
    let y_score = Nat_map.find q.priority y in
      x_score > y_score || (x_score = y_score && x > y)
	
  (* The element at j may be positioned too low in the heap, move it
     up as required. *)
  let rec percolate_up q j =
    if j > 0 then
      (* j is a valid non-root position in the heap. *)
      let je = Vec.get q.heap j in
      let i = parent_position j in
      let ie = Vec.get q.heap i in
	if higher_priority q je ie then begin
	  (* je is incorrectly positioned under its parent, swap them. *)
          Vec.set q.heap j ie;
          Vec.set q.heap i je;
          Nat_map.replace q.position je i;
	  Nat_map.replace q.position ie j;
          percolate_up q i
	end
	  
  (* The element at i may be position to high in the heap, move it
     down as required. *)
  let rec percolate_down q i =
    let size = Vec.length q.heap in
    let l = left_child_position i in
      if l < size then
	(* element at i has at least one valid child *)
	let r = right_sibling_position l in
	let j = (* position of the higher-priority child *)
	  if 
	    r < size &&
	    higher_priority q 
              (Vec.get q.heap r) (Vec.get q.heap l)
	  then r
	  else l
	in 
	let ie = Vec.get q.heap i in
	let je = Vec.get q.heap j in
	  if higher_priority q je ie then begin
	    (* ie is incorrectly positioned over its parent, swap them. *)
            Vec.set q.heap i je;
            Vec.set q.heap j ie;
	    Nat_map.replace q.position je i;
            Nat_map.replace q.position ie j;
            percolate_down q j
          end

  let create e = {
    priority = Nat_map.create 0.0;
    position = Nat_map.create invalid_position;
    heap     = Vec.create 0 e;
  }    

  let length q = Vec.length q.heap

  let is_empty q = Vec.is_empty q.heap

  let mem q e = 
    Nat_map.mem q.position e && Nat_map.find q.position e != invalid_position

  let head q = 
    assert (not (is_empty q));
    Vec.get q.heap root_position

  (* Remove the highest priority element from the heap.
     
     We fill the gap left by the old root by replacing it with the
     last element in the heap.  We then percolate this element down,
     allowing the highest priority element left in the heap to float
     to the top. *)
  let drop q = 
    assert (not (is_empty q));
    Nat_map.replace q.position (head q) invalid_position;
    Vec.swap_out q.heap root_position;
    if not (is_empty q) then 
      let e = head q in
        Nat_map.replace q.position e root_position;
        percolate_down q root_position

  (* Insert an element e into the heap.

     We place the element at the end of the heap, which - unless the
     heap was empty - will make it the child of some earlier
     positioned element.  The new element might, incorrectly, have a
     higher priority than its parent, so we percolate the child up
     until it reaches the correct hight. *)
  let enqueue q e = 
    ensure_map_mem q e;
    if Nat_map.find q.position e = invalid_position then begin
      let i = Vec.length q.heap in
	Vec.push q.heap e;
	Nat_map.replace q.position e i;
	percolate_up q i
    end

  let dequeue q = 
    assert (not (is_empty q));
    let result = head q in
      drop q;
      result

  let priority q e = 
    if Nat_map.mem q.priority e then
      Nat_map.find q.priority e
    else
      0.0

  let inc_priority q e x = 
    ensure_map_mem q e;
    let old = Nat_map.find q.priority e in
      assert (0.0 < x);
      assert (old <= max_float -. x);
      Nat_map.replace q.priority e (old +. x);
      if mem q e then percolate_up q (Nat_map.find q.position e)
	
  let scale q x = 
    assert (x > 0.0);
    Nat_map.modify_const ( *. ) x q.priority

end

module Make_priority_queue (I: IndexedType) = struct
  type elt = I.t

  type t = Nat_priority_queue.t

  let create e = Nat_priority_queue.create (I.index e)

  let length = Nat_priority_queue.length

  let is_empty = Nat_priority_queue.is_empty

  let mem q e = Nat_priority_queue.mem q (I.index e)

  let head q = I.inverse (Nat_priority_queue.head q)

  let enqueue q e = Nat_priority_queue.enqueue q (I.index e)

  let dequeue q = I.inverse (Nat_priority_queue.dequeue q)

  let drop = Nat_priority_queue.drop

  let priority q e = Nat_priority_queue.priority q (I.index e)

  let inc_priority q e x = Nat_priority_queue.inc_priority q (I.index e) x

  let scale = Nat_priority_queue.scale

end
