(* $Id: index.mli 1793 2007-09-28 19:12:36Z jgrundy $ *)
(** Fast data structures using natural number indexing. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** The implementations use {!Vec} as the underlying
    representation. This allows us to perform most operations in
    constant time but also requires that the type being used have an
    indexing function that assigns unique natural number indices to
    its elements. *)

(** Signature for types with natural number indexing. *)
module type IndexedType = sig
  
  (** The type of map keys. *)
  type t
    
  val index: t -> int
    (** Indexing function for objects of type [t]. 

        - Ensure [index x >= 0]. 
        - Ensure [(index x = index y) = (x = y)]
	- Ensure [inverse (index x) = x] *)

  val inverse: int -> t
    (** [inverse i] is the [x] such that [index x = i].

	- Require [i >= 0]
	- Ensure [(inverse i = inverse j) = (i = j)]
	- Ensure [index (inverse i) = i] *)

end

(** {2 Map} *)

(** Signature of maps over indexed types. See the functor {!Index.Make_map}
    and the implementation {!Index.Nat_map}. 

    - Note : The behavior of {!Index.Map.add} is different than what one
    might expect, for efficiency reasons. *)
module type Map = sig
  
  (** Type of indexed objects. *)
  type key
    
  (** Map from indices [0, ... ,n - 1] (where [n] is the number of keys in
      the map) to ['a]. *)
  type 'a t

  val size: 'a t -> int
    (** [size m] is the number of indices mapped by [m]. 

	- Ensure [size m >= 0] *)

  val create: 'a -> 'a t
    (** [create x] is a new, empty map. [x] can be any element of type ['a]
        and is used only for initializing the underlying representation. *)

  val is_empty: 'a t -> bool
    (** [is_empty m] holds if the map is empty.

	- Ensure [is_empty m = (size m = 0)] *)

  val mem: 'a t -> key -> bool
    (** [mem m k] is [true] iff there is a mapping for [index k] in [m],
        i.e. [index k < ]{!Index.Map.size}[ m]. *)

  val find: 'a t -> key -> 'a
    (** [find m k] is the mapping for the index of [k] in [m].

        - Require {!Index.Map.mem}[ m k] *)

  val add: 'a t -> key -> 'a -> unit
    (** [add m k x] modifies [m] by mapping the index of [k] to [x].
        Any previously unmapped indices in [0, ..., index k] will also
        be mapped to [x]. 

	- Ensure {!Index.Map.find}[ m k = x] *)

  val replace: 'a t -> key -> 'a -> unit
    (** [replace m k x] updates [m] by mapping the index of [k] to [x]. 

        - Require {!Index.Map.mem}[ m k] 
	- Ensure {!Index.Map.find}[ m k = x] *)

  val iter: (key -> 'a -> unit) -> 'a t -> unit
    (** [iter f m] applies [f] to all mappings in [m]. *)

  val iter_const: ('a -> key -> 'b -> unit) -> 'a -> 'b t -> unit
    (** [iter_const f x m] is {!Index.Map.iter}[ (f x) m], only faster
	since applying closures is slow. *)

  val fold: ('a -> key -> 'b -> 'a) -> 'a -> 'b t -> 'a
    (** [fold f a m] applies [f] to an accumulated value and each
	key-value pair in the mapping [m]. *)

  val fold_const : ('a -> 'b -> key -> 'c -> 'b) -> 'a -> 'b -> 'c t -> 'b
  (** [fold_const f x a m] is {!Index.Map.fold}[ (f x) a m], only faster
      since applying closures is slow. *)

  val modify: ('a -> 'a) -> 'a t -> unit
    (** [modify f m] modifies a map so that each index maps to [f] of its
	prior value.  *)

  val modify_const: ('a -> 'b ->'b) -> 'a -> 'b t -> unit
    (** [modify_const f x v] is [modify (f x) v], only faster since applying
	closures is slow. *)

  val set_all: 'a t -> 'a -> unit
    (** [set_all m x] updates [m] by mapping every key to [x]. *)

end

(** Functor building an implementation of maps overs indices. *)
module Make_map (I:IndexedType) : Map with type key = I.t

(** A map for natural numbers. *)
module Nat_map : Map with type key = int

(** {2 Map2} *)

(** Signature of two argument maps where the first argument is an
    indexed type. See the functor {!Index.Make_map2} and the
    implementation {!Index.Nat_map2}. *)
module type Map2 = sig

  (** Type of indexed objects, which serve as the first map key. *)
  type key0
    
  (** Type of the second map key. *)
  type key1

  (** Maps indexed by a key from [key0] and a key from [key1]. *)
  type 'a t

  val create: int -> int -> 'a t
    (** [create n m] creates a new, empty map.  The initial storage
	allocated is sufficient for [n * m] entries. *)

  val mem: 'a t -> key0 -> key1 -> bool
    (** [mem m i j] if [m] maps the keys [i] and [j] to some value. *)

  val find: 'a t -> key0 -> key1 -> 'a
    (** [find m i j] is the element [m] maps [i] and [j] to.

	@raise Not_found if [m] has no value for the keys [i] and [j]. *)

  val set: 'a t -> key0 -> key1 -> 'a -> unit
    (** [set m i j x] modifies [m] by mapping the keys [i] and [j] to [x].

	- Ensure {!Index.Map2.find}[ m i j = x]  *)

  val remove: 'a t -> key0 -> key1 -> unit
    (** [remove m i j]  removes any value associated with [i] and [j] by [m].

	- Ensure [not (]{!Index.Map2.find}[ m i j)] *)
end

(** A two argument map where the first key is a natural number and the
    second is a hashed type. *)
module Nat_map2 (J:Hashtbl.HashedType): Map2 
  with type key0 = int and type key1=J.t

(** A two argument map where the first key is an indexed type and the
    second is a hashed type. *)
module Make_map2 (I:IndexedType) (J:Hashtbl.HashedType): Map2 
  with type key0 =I.t and type key1=J.t

(** {2 Set} *)
(** Signature of sets over indexed types. See the functor
    {!Index.Make_set} and the implementation {!Index.Nat_set}. *)
module type Set = sig

  (** Type of elements. *)
  type elt

  (** Type of set of elements. *)
  type t

  val create: unit -> t
    (** Create an empty set in constant time. *)

  val size: t -> int
    (** Cardinality of the set.

	- Ensure [size s >= 0] *)

  val is_empty: t -> bool
    (** Is a set empty?

	- Ensure [is_empty s = (]{!Index.Set.size}[ s = 0)] *)
    
  val mem: t -> elt -> bool
    (** Determine whether element is a member of the set in constant time. *)

  val add: t -> elt -> unit
    (** [add s e] adds element [e] to the set [s] in (amortized)
	constant time.

	- Ensure {!Index.Set.mem} [s n] 
	- Ensure [not (]{!Index.Set.is_empty}[ s)] *)

  val remove: t -> elt -> unit
    (** [remove s e] removes element [e] from the set [s] in constant time. 

	- Ensure [not (]{!Index.Set.mem}[ s n)] *)

  val clear: t -> unit  
    (** Makes the set empty in (amortized) constant time. 

	- Ensure {!Index.Set.is_empty}[ s] *)

  val iter: (elt -> unit) -> t -> unit
    (** [iter c s] applies [c] to all elements of [s]. *)

  val iter_const: ('a -> elt -> unit) -> 'a -> t -> unit
    (** [iter_const c x s] is {!Index.Set.iter}[ (c x) s], only faster
	since applying closures is slow. *)

  val fold: ('a -> elt -> 'a) -> 'a -> t -> 'a
    (** [fold f a s] applies [f] to an accumulated value and each
	element in the set [s]. *)

  val fold_const: ('a -> 'b -> elt -> 'b) -> 'a -> 'b -> t -> 'b
  (** [fold_const f x a s] is {!Index.Set.fold}[ (f x) a s], only faster
      since applying closures is slow. *)

end

(** Functor building an implementation of sets over indices. *)
module Make_set (I:IndexedType) : Set with type elt = I.t

(** A set implementation for natural numbers. *)
module Nat_set : Set with type elt = int

(** {2 Priority Queue} *)

(** Signature of priority queues over indexed types. See the functor
    {!Index.Make_priority_queue} and the implementation
    {!Index.Nat_priority_queue}. *)
module type PriorityQueue = sig

  (** Type of elements. *)
  type elt

  (** Type of priority queues of elements. *)
  type t

  val create: elt -> t
    (** An empty queue. *)

  val length: t -> int
    (** [length q] is the number of elements in the queue. 

	- Ensure [length q >= 0] *)
    
  val is_empty: t -> bool
    (** [is_empty q] is [true] iff there are no elements enqueued in [q].

	- Ensure [is_empty q = (length q = 0)] *)

  val mem: t -> elt -> bool
    (** [mem q e] is [true] iff [e] is enqueued in [q]. *)

  val head: t -> elt
    (** [head q] is the highest priority element of [q]. 
	
	- Require [not (is_empty q)] *)
    
  val enqueue: t -> elt -> unit
    (** [enqueue q e] enqueues the element [e]. The position of [e] in [q] is
        determined by its priority.

        Note: The default priority of elements is [0.0]. *)

  val dequeue: t -> elt
    (** [dequeue q] removes the highest priority element from [q].
      
	@return The element removed.
	
	- Require [not (is_empty q)] *)
 
  val drop: t -> unit
    (** [drop q] removes the highest priority element from [q]. 
      
	- Require [not (is_empty q)] *)

  val priority: t -> elt -> float
    (** The priority of an element.

        Note: The default priority of elements is [0.0]. *)

  val inc_priority: t-> elt -> float -> unit
    (** [inc_priority q e x] adds [x] to the priority of element [e].
        If [e] is enqueued, it will move up the queue as necessary.

        - Require [0.0 < x]
        - Require [priority q e <= max_float -. x]*)

  val scale: t -> float -> unit
    (** [scale q x] multiplies the priority of every element by the constant
        factor [x]. The ordering of elements in the queue is unchanged.

        - Require [x > 0.0]. *)

end

(** Functor building an implementation of priority queues for indexed types. *)
module Make_priority_queue (I:IndexedType) : PriorityQueue with type elt = I.t

(** An implementation of priority queue with natural number elements. *)
module Nat_priority_queue : PriorityQueue with type elt = int
