(* $Id: log.mli 1788 2007-09-28 18:23:28Z jgrundy $ *)
(** Message logging. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** A mechanism for logging messages.  Messages are logged with a {i
    level}, and the module has a notion of the current {i log level}.
    If a message is logged at the current current log level, or lower,
    then it will be output.

    The default log level is 0, only messages logged at level 0 are
    output.

    The default output destination is [Format.std_formatter]. *)

val set_log_formatter: Format.formatter -> unit
  (** Set the output destination for messages *)

val set_log_level: int -> unit
  (** Set current log level

      @raise Invalid_argument "negative level" if you attempt to set
      the current log level to a negative value *)

val log: (Format.formatter -> 'a -> unit) -> int -> 'a -> unit
  (** [Log.log print level x] outputs [x] using [print], if
      [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_int: int -> int -> unit
  (** [Log.log_int level i] logs [i], if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_float: int -> float -> unit
  (** [Log.log_float level j] logs [j], if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_string: int -> string -> unit
  (** [Log.log_string level s] logs [s], if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_bool: int -> bool -> unit
  (** [Log.log_bool level b] logs [b], if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_time: int -> Time.time -> unit
  (** [Log.log_time level t] logs seconds of processor time since
      [t], if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_flush: int -> unit -> unit
  (** [Log.log_flush level ()] flushes the log output, if [level]
      <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)

val log_newline: int -> unit -> unit
  (** [Log.log_newline level ()] outputs a new line and flushes
      the log output, if [level] <= current log level

      @raise Invalid_argument "negative level" if [level] < 0 *)
