(* $Id: subarray.ml 1788 2007-09-28 18:23:28Z jgrundy $ *)
(* Array subregions. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Array

let swap a i j =
  assert (0 <= i && i < length a);
  assert (0 <= j && j < length a);
  if i != j then
    let tmp = a.(i) in
      a.(i) <- a.(j);
      a.(j) <- tmp      
	
let rec rev a i j =
  assert (0 <= i && i <= j + 1 && j <= length a);
  if i < j then begin
    swap a i (j-1);
    rev a (i+1) (j-1)
  end

let set_all a i j x =
  assert (0 <= i && i <= j && j <= length a);
  Array.fill a i (j-i) x

let rec set_fun a i j f =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    a.(i) <- f i;
    set_fun a (i+1) j f
  end

let rec mem x a i j =
  assert (0 <= i && i <= j && j <= length a);
  i != j && ((x = a.(i)) || mem x a (i+1) j);;

let rec exists p a i j =
  assert (0 <= i && i <= j && j <= length a);
  i != j && (p a.(i) || exists p a (i+1) j)

let rec exists_const p x a i j =
  assert (0 <= i && i <= j && j <= length a);
  i != j && (p x a.(i) || exists_const p x a (i+1) j)

let rec for_all p a i j =
  assert (0 <= i && i <= j && j <= length a);
  i = j || (p a.(i) && for_all p a (i + 1) j)

let rec for_all_const p x a i j =
  assert (0 <= i && i <= j && j <= length a);
  i = j || (p x a.(i) && for_all_const p x a (i + 1) j)

let rec first p a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then j 
  else if p a.(i) then i
  else first p a (i+1) j

let rec first_const p x a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then j 
  else if p x a.(i) then i 
  else first_const p x a (i+1) j

let rec fold_left f acc a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else fold_left f (f acc a.(i)) a (i+1) j

let rec fold_right f a i j acc =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else fold_right f a i (j-1) (f a.(j-1) acc)

let rec foldi_left f acc a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else foldi_left f (f acc i a.(i)) a (i+1) j

let rec foldi_right f a i j acc =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else foldi_right f a i (j-1) (f (j-1) a.(j-1) acc)

let rec fold_left_const f x acc a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else fold_left_const f x (f x acc a.(i)) a (i+1) j

let rec fold_right_const f x a i j acc =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else fold_right_const f x a i (j-1) (f x a.(j-1) acc)

let rec foldi_left_const f x acc a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else foldi_left_const f x (f x acc i a.(i)) a (i+1) j

let rec foldi_right_const f x a i j acc =
  assert (0 <= i && i <= j && j <= length a);
  if i = j then acc 
  else foldi_right_const f x a i (j-1) (f x (j-1) a.(j-1) acc)

let rec iter c a i j = 
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    c a.(i);
    iter c a (i+1) j
  end
    
let rec iteri c a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    c i a.(i);
    iteri c a (i+1) j
  end

let rec iter_const c x a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    c x a.(i);
    iter_const c x a (i+1) j
  end

let rec iteri_const c x a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    c x i a.(i);
    iteri_const c x a (i+1) j
  end

let rec modify f a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    a.(i) <- f a.(i);
    modify f a (i+1) j
  end

let rec modify_const f x a i j =
  assert (0 <= i && i <= j && j <= length a);
  if i != j then begin
    a.(i) <- f x a.(i);
    modify_const f x a (i+1) j
  end

let rec merge cmp b n m i a j k =
  assert (0 <= n && n <= m && m <= length b);
  assert (0 <= i && i <= j && j <= k && k <= length a);
  if n != m then
    if j = k || cmp b.(n) a.(j) <= 0 then begin
      a.(i) <- b.(n);
      merge cmp b (n+1) m (i+1) a j k
    end
    else begin
      a.(i) <- a.(j);
      merge cmp b n m (i+1) a (j+1) k
    end

let rec msort cmp a i j tmp =
  assert (0 <= i && i <= j && j <= length a);
  if i + 1 < j then 
    let mid = i + (j - i)/2 in 
      assert (mid - i <= length tmp);
      msort cmp a i mid tmp;
      msort cmp a mid j tmp;
      Array.blit a i tmp 0 (mid - i);
      merge cmp tmp 0 (mid - i) i a mid j
      
let merge_sort cmp a i j =
  assert (0 <= i && i <= j && j <= length a);
  let tmp = Array.make ((j - i)/2) a.(i) in
    msort cmp a i j tmp

let sort = merge_sort

let stable_sort = merge_sort
	
