/* $Id: smtlib_grammar.mly 1788 2007-09-28 18:23:28Z jgrundy $ */
%{
  (** Parser for SMT-LIB theory, logic, and benchmark files. *)
  
  (* Copyright 2007 Intel Corporation 
     
     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
     You may obtain a copy of the License at
     
     http://www.apache.org/licenses/LICENSE-2.0
     
     Unless required by applicable law or agreed to in writing,
     software distributed under the License is distributed on an "AS
     IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
     express or implied.  See the License for the specific language
     governing permissions and limitations under the License. *)

  (** {2 Parser API } *)

  (** This grammar recognizes SMT-LIB syntax as described in
      {{:http://combination.cs.uiowa.edu/smtlib/papers.html} The
      SMT-LIB Standard} Version 1.2 (30 August 2006). *)

  open Smtlib_attribute
  open Smtlib_syntax

  module Var_hashtbl = Hashtbl.Make (Variable)

  let rec subst tbl term = 
    match term with
      | Term.Variable (_, v) -> 
          begin
            try Var_hashtbl.find tbl v
            with Not_found -> term
          end
      | Term.Constant _ -> term
      | Term.Application (_, t1, t2) ->
          Term.application (subst tbl t1) (subst tbl t2)
      | _ -> (* SMT-LIB format never constructs abstractions. *)
	  assert false

  let bind_one tbl (var, term) = 
    Var_hashtbl.add tbl var (subst tbl term)

  let bind_local (bindings,body) =
    if bindings = [] then 
      body
    else 
      let tbl = Var_hashtbl.create 0 in
        List.iter (bind_one tbl) bindings;
        subst tbl body
%}

/* Atomic tokens */
%token LEFT_PAREN RIGHT_PAREN 

/* Atomic language keywords */
%token BENCHMARK FALSE FLET ITE LET LOGIC THEORY TRUE 

/* Language token classes */
%token <string> CONNECTIVE	QUANT_SYMB	STRING		USER_VALUE
%token <string>	NUMERAL		RATIONAL	IDENTIFIER	VAR
%token <string>	FVAR		ATTRIBUTE	ARITH_SYMB

/* Attribute keywords */
%token <Smtlib_attribute.status> STATUS

%token ATTRIBUTE_ASSUMPTION	ATTRIBUTE_AXIOMS	ATTRIBUTE_DEFINITION
%token ATTRIBUTE_EXTENSIONS	ATTRIBUTE_FORMULA	ATTRIBUTE_FUNS
%token ATTRIBUTE_EXTRAFUNS	ATTRIBUTE_EXTRASORTS	ATTRIBUTE_EXTRAPREDS
%token ATTRIBUTE_LANGUAGE	ATTRIBUTE_LOGIC		ATTRIBUTE_NOTES
%token ATTRIBUTE_PREDS		ATTRIBUTE_SORTS		ATTRIBUTE_STATUS
%token ATTRIBUTE_THEORY

/* Terms */

%type <string> fun_symb pred_symb sort_symb
%type <string list> rev_sort_symb_plus sort_symb_plus sort_symb_star
%type <string * (string option)> annotation
%type <(string * (string option)) list> rev_annotation_plus annotation_plus
%type <(string * (string option)) list> annotation_star
%type <Term.term> base_term
%type <Term.term list> rev_an_term_plus an_term_plus
%type <Term.term> an_term

/* Formulas */

%type <Term.term> prop_atom an_atom
%type <Term.term list> rev_an_formula_plus an_formula_plus
%type <Variable.t> quant_var 
%type <Variable.t list * Term.term> quant_var_plus_an_formula
%type <(Variable.t * Term.term) list * Term.term> an_formula

/* Theories */

%type <string> theory_name
%type <Constant.t> fun_symb_decl pred_symb_decl
%type <Constant.t list> rev_fun_symb_decl_plus fun_symb_decl_plus 
%type <Constant.t list> rev_pred_symb_decl_plus pred_symb_decl_plus
%type <Smtlib_attribute.attribute> theory_attribute
%type <Smtlib_attribute.attribute list> rev_theory_attribute_plus
%type <Smtlib_attribute.attribute list> theory_attribute_plus theory

%start theory

/* Logics */

%type <string> logic_name
%type <Smtlib_attribute.attribute> logic_attribute
%type <Smtlib_attribute.attribute list> rev_logic_attribute_plus 
%type <Smtlib_attribute.attribute list> logic_attribute_plus logic

%start logic

/* Benchmarks */

%type <string> bench_name
%type <Smtlib_attribute.attribute> bench_attribute
%type <Smtlib_attribute.attribute list> rev_bench_attribute_plus 
%type <Smtlib_attribute.attribute list> bench_attribute_plus benchmark

%start benchmark

%%

/* Terms */

fun_symb:
| IDENTIFIER { $1 }
| ARITH_SYMB { $1 }  
;

pred_symb: 
| IDENTIFIER { $1 }
| ARITH_SYMB { $1 }
;

sort_symb: 
| IDENTIFIER { $1 }
;

rev_sort_symb_plus:
| sort_symb { [$1] }
| rev_sort_symb_plus sort_symb { $2::$1 }
;

sort_symb_plus: 
| rev_sort_symb_plus { List.rev $1 }
;

sort_symb_star:
| { [] }
| sort_symb_plus { $1 }
;

annotation:
| ATTRIBUTE { ($1, None) }
| ATTRIBUTE USER_VALUE { ($1, Some $2) }
;

rev_annotation_plus:
| annotation { [$1] }
| rev_annotation_plus annotation { $2::$1 }
;

annotation_plus: 
| rev_annotation_plus { List.rev $1 }
;

annotation_star:
| { [] }
| annotation_plus { $1 }
;

base_term: 
| VAR { Term.variable (variable $1) }
| NUMERAL { Term.constant (integer $1) }
| RATIONAL { Term.constant (rational $1) }
| IDENTIFIER { Term.constant (constant $1) }
;

rev_an_term_plus:
| an_term { [$1] }
| rev_an_term_plus an_term { $2::$1 }
;

an_term_plus:
| rev_an_term_plus { List.rev $1 }
;

an_term:
| base_term { $1 }
| LEFT_PAREN base_term annotation_plus RIGHT_PAREN { $2 }
| LEFT_PAREN fun_symb an_term_plus annotation_star RIGHT_PAREN { 
    application $2 $3}
| LEFT_PAREN ITE an_formula an_term an_term annotation_star RIGHT_PAREN { 
    Term_logic.Standard.ite (bind_local $3) $4 $5 }
;

/* Formulas */

prop_atom:
| TRUE       { Term_constant.truth }
| FALSE      { Term_constant.falsity }
| FVAR       { Term.variable (variable $1) }
| IDENTIFIER { Term.constant (constant $1) }
;

an_atom:
| prop_atom { $1 }
| LEFT_PAREN prop_atom annotation_plus RIGHT_PAREN { $2 }
| LEFT_PAREN pred_symb an_term_plus annotation_star RIGHT_PAREN { 
    predicate $2 $3 }
;

rev_an_formula_plus:
| an_formula { [bind_local $1] }
| rev_an_formula_plus an_formula { (bind_local $2)::$1 }
;

an_formula_plus:
| rev_an_formula_plus { List.rev $1 }
;

quant_var:
| LEFT_PAREN VAR sort_symb RIGHT_PAREN { Variable.named $2 }
;

/* These must be parsed together to disambiguate the grammar. */
quant_var_plus_an_formula:
| quant_var an_formula { ([$1], bind_local $2) }
| quant_var quant_var_plus_an_formula { let (vs,f) = $2 in ($1::vs, f) }
;

an_formula:
| an_atom { ([],$1) }
| LEFT_PAREN CONNECTIVE an_formula_plus annotation_star RIGHT_PAREN { 
    ([], connection $2 $3) }
| LEFT_PAREN QUANT_SYMB quant_var_plus_an_formula annotation_star RIGHT_PAREN {
    let (vars,body) = $3 in ([], quantification $2 vars body) }
| LEFT_PAREN 
      LET LEFT_PAREN VAR an_term RIGHT_PAREN an_formula annotation_star
  RIGHT_PAREN {
    let (bindings,body) = $7 in
      ((Variable.named $4, $5)::bindings, body) }
| LEFT_PAREN 
    FLET LEFT_PAREN FVAR an_formula RIGHT_PAREN an_formula annotation_star 
  RIGHT_PAREN { 
    let (bindings,body) = $7 in
      ((Variable.named $4, bind_local $5)::bindings, body) }
;

/* Theories */

fun_symb_decl: 
| LEFT_PAREN fun_symb sort_symb_plus annotation_star RIGHT_PAREN { 
    declare_constant $2 $3 }
| LEFT_PAREN NUMERAL sort_symb annotation_star RIGHT_PAREN { 
    (integer $2) }
| LEFT_PAREN RATIONAL sort_symb annotation_star RIGHT_PAREN { 
    (rational $2) }
;

pred_symb_decl:
| LEFT_PAREN pred_symb sort_symb_star annotation_star RIGHT_PAREN { 
    declare_predicate $2 $3 }
;

theory_name:
| IDENTIFIER { $1 }
;

rev_fun_symb_decl_plus:
| fun_symb_decl { [$1] }
| rev_fun_symb_decl_plus fun_symb_decl { $2::$1 }
;

fun_symb_decl_plus:
| rev_fun_symb_decl_plus { List.rev $1 }
;

rev_pred_symb_decl_plus:
| pred_symb_decl { [$1] }
| rev_pred_symb_decl_plus pred_symb_decl { $2::$1 }
;

pred_symb_decl_plus:
| rev_pred_symb_decl_plus { List.rev $1 }
;

theory_attribute:
| ATTRIBUTE_SORTS LEFT_PAREN sort_symb_plus RIGHT_PAREN { Sorts $3 }
| ATTRIBUTE_FUNS LEFT_PAREN fun_symb_decl_plus RIGHT_PAREN { Funs $3 }
| ATTRIBUTE_PREDS LEFT_PAREN pred_symb_decl_plus RIGHT_PAREN { Preds $3 }
| ATTRIBUTE_DEFINITION STRING { Definition $2 }
| ATTRIBUTE_AXIOMS LEFT_PAREN an_formula_plus RIGHT_PAREN { Axioms $3 }
| ATTRIBUTE_NOTES STRING { Notes $2 }
| annotation { Annotation $1 }
;

rev_theory_attribute_plus:
| theory_attribute { [$1] }
| rev_theory_attribute_plus theory_attribute { $2::$1 }
;

theory_attribute_plus:
| rev_theory_attribute_plus { List.rev $1 }
;

theory:
| LEFT_PAREN THEORY theory_name theory_attribute_plus RIGHT_PAREN 
    { (Name $3)::$4 }
;

/* Logics */

logic_name: 
| IDENTIFIER { $1 }
;

logic_attribute:
| ATTRIBUTE_THEORY theory_name { 
    let file = Smtlib_parser.smtlib_theory_file $2 in
      Theory (Smtlib_parser.parse_theory file) }
| ATTRIBUTE_LANGUAGE STRING { Language $2 }
| ATTRIBUTE_EXTENSIONS STRING { Extensions $2 }
| ATTRIBUTE_NOTES STRING { Notes $2 }
| annotation { Annotation $1 }
;

rev_logic_attribute_plus:
| logic_attribute { [$1] }
| rev_logic_attribute_plus logic_attribute { $2::$1 }
;

logic_attribute_plus:
| rev_logic_attribute_plus { List.rev $1 }
;

logic:
| LEFT_PAREN LOGIC logic_name logic_attribute_plus RIGHT_PAREN 
    { (Name $3)::$4 }
;

/* Benchmarks */

bench_name:
| IDENTIFIER { $1 }
;

bench_attribute:
| ATTRIBUTE_LOGIC logic_name {
    let file = Smtlib_parser.smtlib_logic_file $2 in
      Logic (Smtlib_parser.parse_logic file) }
| ATTRIBUTE_ASSUMPTION an_formula { 
    Assumption (bind_local $2) }
| ATTRIBUTE_FORMULA an_formula { 
    Formula (bind_local $2) }
| ATTRIBUTE_STATUS STATUS { 
    Status $2 }
| ATTRIBUTE_EXTRASORTS LEFT_PAREN sort_symb_plus RIGHT_PAREN { Sorts $3 }
| ATTRIBUTE_EXTRAFUNS LEFT_PAREN fun_symb_decl_plus RIGHT_PAREN { 
    Funs $3 }
| ATTRIBUTE_EXTRAPREDS LEFT_PAREN pred_symb_decl_plus RIGHT_PAREN { 
    Preds $3 }
| ATTRIBUTE_NOTES STRING { 
    Notes $2 }
| annotation {
    Annotation $1 }
;

rev_bench_attribute_plus:
| bench_attribute { [$1] }
| rev_bench_attribute_plus bench_attribute { $2::$1 }
;

bench_attribute_plus:
| rev_bench_attribute_plus { List.rev $1 }
;

benchmark: 
| LEFT_PAREN BENCHMARK bench_name bench_attribute_plus RIGHT_PAREN { 
    (Name $3)::$4 }
;
