(* $Id: smtlib_syntax.ml 1793 2007-09-28 19:12:36Z jgrundy $ *)
(* Compound syntactic constructs occurring in SMT-LIB terms and formulas. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

module T = Term
module L = Term_logic.Standard

module String_map = Map.Make (Util.String)

module String_set = Set.Make (Util.String)

exception Syntax_error

(*** constants ***)

let declared_constants = ref String_set.empty

let declare_symbol name types =
  let arity = List.length types - 1 in
  assert (arity >= 0);
  if String_set.mem name !declared_constants then
    (* We have already been asked to define this. *)
    raise Syntax_error
  else 
    begin
      declared_constants := String_set.add name !declared_constants;
      if Symbol.name_declared name then
	(* There is a built-in symbol with this name. *)
	let s = Symbol.find name in
	  if Symbol.arity s = arity then (* The arities match - great *)
	    Constant.symbolic s
	  else (* Oh no, the referenced symbol has a different arity. *)
	    begin
	      Log.log_string 0 "Wrong arity: ";
	      Log.log_string 0 name;
	      Log.log_newline 0 ();
	      Log.log_string 0 "Requested: ";
	      Log.log_int 0 arity;
	      Log.log_newline 0 ();
	      Log.log_string 0 "Actual: ";
	      Log.log_int 0 (Symbol.arity s);
	      Log.log_newline 0 ();
	      raise Syntax_error
	    end
      else begin
	(* Give them a new uninterpreted symbol. *)
        let symbol = Symbol.declare_uninterpreted name arity in
          Constant.symbolic symbol
      end
    end

let declare_constant name sorts =
  declare_symbol name sorts

let declare_predicate name sorts =
  declare_symbol name (sorts @ ["Bool"])

let constant s =
  if not (Symbol.name_declared s) then begin
    Log.log_string 0 "Undeclared symbol: ";
    Log.log_string 0 s;
    Log.log_newline 0 ();
  end;
  Constant.symbolic (Symbol.find s)
  
let rational s = 
  Constant.rational (Num.num_of_string s)

let integer s =
  Constant.integer (Big_int.big_int_of_string s)

let variable s = Variable.named s

(*** terms ***)

let map_from_assoc l =
  List.fold_left (fun t (k,v) -> String_map.add k v t) String_map.empty l

let variadic_application f u args =
  List.fold_left f u args

let variadic1_application f = function
  | [] -> assert false;
  | h::t ->  variadic_application f h t

let term_disequality x y = 
  if T.term_equal x y then
    Term_constant.falsity
  else
    L.negation (L.equality x y)

let rec continuing_application c = function
  | [] | [_] -> raise Syntax_error
  | [x0;x1]  -> c x0 x1
  | (x0::(x1::_ as xs)) -> L.conjunction (c x0 x1) (continuing_application c xs)

let rec pairwise c acc = function
  | [] -> variadic1_application L.conjunction acc
  | h::t -> pairwise c (acc @ (List.map (c h) t)) t

let apply_pairwise c = function
  | [] | [_] -> raise Syntax_error
  | l -> pairwise c [] l

let fixed1_application f = function
  | [x] -> f x
  | _ -> raise Syntax_error

let fixed2_application f = function
  | [x;y] -> f x y
  | _ -> raise Syntax_error

let fixed3_application f = function
  | [x;y;z] -> f x y z
  | _ -> raise Syntax_error
	
let function_map =
  let fs =
    [("~", fixed1_application (T.application Term_constant.unary_minus));
     ("-", fixed2_application (T.binary_application Term_constant.minus));
     ("+", fixed2_application (T.binary_application Term_constant.plus));
     ("*", fixed2_application (T.binary_application Term_constant.times));
     ("/", fixed2_application (T.binary_application Term_constant.divide))]
  in   
    map_from_assoc fs

let application name args = 
  try
    (* translation fix of unary minus *)
    let name =
      if name = "-" && List.length args == 1 then
	"~"
      else
	name 
    in
      String_map.find name function_map args
  with Not_found -> 
    T.nary_application (T.symbolic_constant (Symbol.find name)) args

let predicate_map =
  let ps =
    [("=",  
      fixed2_application (T.binary_application Term_constant.equality));
     ("<=", 
      fixed2_application (T.binary_application Term_constant.less_or_equal));
     ("<",  
      fixed2_application (T.binary_application Term_constant.less));
     (">=", 
      fixed2_application (T.binary_application Term_constant.greater_or_equal));
     (">",  
      fixed2_application (T.binary_application Term_constant.greater));
     ("distinct", 
      apply_pairwise term_disequality)]
  in
    map_from_assoc ps

let predicate name args = 
  try
    String_map.find name predicate_map args 
  with Not_found -> 
    T.nary_application (T.symbolic_constant (Symbol.find name)) args

let connective_map =
  let cs =
    [("and",          variadic1_application  L.conjunction);
     ("iff",          continuing_application L.equivalence);
     ("if_then_else", fixed3_application     L.conditional);
     ("implies",      fixed2_application     L.implication);
     ("not",          fixed1_application     L.negation);
     ("or",           variadic1_application  L.disjunction);
     ("xor",          variadic1_application  L.difference)]
  in   
    map_from_assoc cs

let connection name args =
  try 
    (String_map.find name connective_map) args
  with Not_found -> raise Syntax_error 
    
let exists_symb = Symbol.declare "exists" 1
let forall_symb = Symbol.declare "forall" 1

let quantifier_map =
  let qs =
    [("exists", T.symbolic_constant exists_symb);
     ("forall", T.symbolic_constant forall_symb)]
  in
    map_from_assoc qs

let quantify q v b =
  T.application q (T.abstraction v b)

let quantification quantifier vars body =
  let q = String_map.find quantifier quantifier_map in
    List.fold_right (quantify q) vars body

let local_let var value body =
  T.beta_normal_form (T.application (T.abstraction var body) value)

let local_flet var value body =
  T.beta_normal_form (T.application (T.abstraction var body) value)

let is_interpreted term =
  match term with
    | T.Constant (_, c) ->
        Symbol.interpreted (Constant.symbolic_value c)
    | T.Variable _ -> false
    | _ -> assert false
