(* $Id: cc.ml 1788 2007-09-28 18:23:28Z jgrundy $ *)
(* Term-level API for {!Cc_core}. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Term

(* A module of canonically ordered pairs representing equations. *)
module Equation = struct
  open Cc_literal

  type t = Variable.t * Variable.t
  let make x y = if Variable.compare x y < 0 then (x,y) else (y,x)
  let equal (x0,y0) (x1,y1) = Variable.equal x0 x1 && Variable.equal y0 y1
  let hash = Hashtbl.hash
end

module Proxy_bij    = Bihashtbl.Make (Term_compared) (Term_compared)
module Equation_tbl = Bihashtbl.Make (Term_compared) (Equation)

class make () = 
  let core = Cc_core.new_solver () in
    (* Underlying EUF solver. *)
object (self)
  val cc_true_var = Cc_core.fresh_variable core
    (* A distinct variable in the EUF solver to represent true. *)

  val proxy_bij = Proxy_bij.create 0
    (* A bijection between "elementary" terms (described below) and
       the proxy variables that represent them. *)

  val elementary_form_cache = Term_hashtbl.create 0
    (* Each term is reduced to its "elementary" form.  This reduction is
       cached in case we encounter shared subterms. *)

  val mutable variables = Term_set.empty
    (* Variables occurring in terms handled by this solver. *)
    
  val cc_variable = Term_hashtbl.create 0
    (* Mapping between term variables and corresponding EUF solver
       variables. *)

  val interface_literal = Equation_tbl.create 0
    (* Mapping between interface variables and corresponding EUF solver
       literals. *)

  (* We have the following state invariants:
     - Every term in the domain of proxy_bij is an elementary term.
     - Every term in the range of elementary_form_cache is an elementary term.
     - Every term in variables is a variable.
     - Every term in the domain of cc_variable is a variable.
     - Proxies not in domain of cc_variable are those that proxy equations. 
     - Every term in the domain of interface_literal is in variables. *)

  initializer
    assert self#invariant;
    assert (self#decision_level = 0);
    assert (not self#conflict_detected);
    assert (Term_set.is_empty self#variables);

  method private invariant =
    (* elementary_form_cache range equals proxy_bij domain, implying *)
    (Term_hashtbl.length elementary_form_cache >= Proxy_bij.length proxy_bij)
    && (* proxy_bij range equals the set variables, implying *)
      (* should check a cardinality constraint, but set cardinality is slow *)
      ((Proxy_bij.length proxy_bij = 0) = Term_set.is_empty variables)
    && (* cc_variable domain is a subset of the set variables, implying *)
      (* should check a cardinality constraint, but set cardinality is slow *)
      (not (Term_set.is_empty variables) 
       || (Term_hashtbl.length cc_variable = 0))
    && (* decision level is always nonnegagive *)
      Cc_core.decision_level core >= 0

  (* Is a term in elementary form? Used only in assertions.
     
     A term is in elementary form when it is one of:
     - a variable
     - the constant true
     - a canonically ordered equation between two proxy variables
       that don't proxy equations, and therefore are in cc_variable
     - an application of one proxy variable to another, where neither
       variable proxies an equation and are therefore in cc_variable *)
  method private is_elementary tm = 
    match tm with 
      | Variable _ -> true
      | Constant _ -> term_equal tm Term_constant.truth
      | Application (_, Application (_, f, x), y) 
	  when term_equal f Term_constant.equality ->
	  Proxy_bij.mem_inv proxy_bij x 
	  && Term_hashtbl.mem cc_variable x 
	  && Proxy_bij.mem_inv proxy_bij y 
	  && Term_hashtbl.mem cc_variable y
	  && term_compare x y < 0
      | Application (_, f, x) ->
	  Proxy_bij.mem_inv proxy_bij f 
	  && Term_hashtbl.mem cc_variable f
	  && Proxy_bij.mem_inv proxy_bij x 
	  && Term_hashtbl.mem cc_variable x
      | _ -> false

  (* The proxy for a term in elementary form. *)
  method private elementary_proxy term =
    assert (self#is_elementary term);
    if Proxy_bij.mem_inv proxy_bij term then 
      term
    else
      try 
	Proxy_bij.find proxy_bij term
      with Not_found ->
	match term with
	  | Variable _ ->
	      let cc_var = Cc_core.fresh_variable core in
		Proxy_bij.add proxy_bij term term;
		variables <- Term_set.add term variables;
		Term_hashtbl.add cc_variable term cc_var;
		term
	  | Application (_, Application (_, f, _x), _y)
	      when term_equal f Term_constant.equality ->
	      let var = variable (Variable.fresh ()) in
		Proxy_bij.add proxy_bij term var;
		variables <- Term_set.add var variables;
		var
	  | Application (_, f, x) ->
	      assert (is_variable f);
	      assert (is_variable x);
	      let cc_var_f = Term_hashtbl.find cc_variable f in
	      let cc_var_x = Term_hashtbl.find cc_variable x in
	      let cc_var = Cc_core.add_definition core cc_var_f cc_var_x in
	      let var = variable (Variable.fresh ()) in
		Proxy_bij.add proxy_bij term var;
		variables <- Term_set.add var variables;
		Term_hashtbl.add cc_variable var cc_var;
		var
	  | _ ->
	      assert (term_equal term Term_constant.truth);
	      term
		    
  (* Reduce a term to its elementary form - (mutually, recursively)
     calls proxy. *)
  method private elementary_form term =
    try
      Term_hashtbl.find elementary_form_cache term 
    with Not_found -> 
      let variable_of = function
	| Variable (_,v) -> v
	| _ -> assert false
      in
      let equation v w = 
	binary_application Term_constant.equality (variable v) (variable w) 
      in
      let new_elementary_form = match term with 
	| Variable _ -> term 
	| Application (_, Application (_, f, x), y)
	    when term_equal f Term_constant.equality ->
	    (* Proxy produces only variables or true. Since x and y
	       are not boolean their proxies must be variables. *)
	    let proxy_x = variable_of (self#proxy x) in
	    let proxy_y = variable_of (self#proxy y) in
	    let comparison = Variable.compare proxy_x proxy_y in
	      if comparison = 0 then 
		Term_constant.truth
	      else if comparison < 0 then
		equation proxy_x proxy_y
	      else (* comparison > 0 *)
		equation proxy_y proxy_x
	| Application (_, f, x) ->
	    application (self#proxy f) (self#proxy x)
	| _ -> assert false
      in
	Term_hashtbl.add elementary_form_cache term new_elementary_form;
	new_elementary_form
	  
  (* The proxy for a term - (mutually, recursively)
     calls elementary_form. *)
  method proxy term = 
    let result = self#elementary_proxy (self#elementary_form term) in
      assert self#invariant;
      result

  (* Does this solver own the constant? *)
  method owns_constant term =
    assert (Term.is_constant term);
    term_equal Term_constant.equality term

  method add_interface_variable v =
    assert (is_variable v);
    assert (Term_set.mem v variables);
    let term = Proxy_bij.find_inv proxy_bij v in
    let (i,j) = match term with
      | Application (_, Application (_, f, x), y)
          when term_equal f Term_constant.equality ->
	  begin 
	    assert (is_variable x);
	    assert (is_variable y);
	    assert (not (Term_hashtbl.mem cc_variable v));
            let cc_x = Term_hashtbl.find cc_variable x in
            let cc_y = Term_hashtbl.find cc_variable y in
	      (cc_x, cc_y)
	  end
      | _ -> 
          let cc_x = Term_hashtbl.find cc_variable v in
          let cc_y = cc_true_var in
	    (cc_x, cc_y) 
    in
      assert (i != j);
      Equation_tbl.add interface_literal v (Equation.make i j);
      Cc_core.add_interface_literal core (i, j);
      assert self#invariant

  method decision_level = 
    Cc_core.decision_level core

  method variables = 
    variables

  method inference_possible = 
    Cc_core.inference_possible core

  method conflict_detected = 
    Cc_core.conflict_detected core

  method increment_level =
    assert (not self#conflict_detected);
    assert (not self#inference_possible);
    assert (not self#get_literal_possible);
    Cc_core.increment_level core;
    assert self#invariant;
    assert (not self#conflict_detected);
    assert (not self#inference_possible);
    assert (not self#get_literal_possible);
    assert (self#decision_level > 0)

  method infer = 
    assert (not self#conflict_detected);
    assert (self#inference_possible);
    Cc_core.infer core;
    assert self#invariant

  (* EUF literal corresponding to a term literal - throws Not_found if
     there is none *)
  method private cc_lit_of term_lit =
    let v = Term_literal.variable term_lit in
    let (i,j) = Equation_tbl.find interface_literal v in
      if Term_literal.is_positive term_lit then
	Cc_literal.Equality (i,j)
      else
        Cc_literal.Disequality (i,j)

  (* term literal corresponding to an EUF literal *)
  method private term_lit_of cc_lit =
    match cc_lit with 
      | Cc_literal.Equality (i,j) ->
          Term_literal.literal 
	    (Equation_tbl.find_inv interface_literal (Equation.make i j))
      | Cc_literal.Disequality (i,j) ->
          Term_literal.negation 
	    (Term_literal.literal
               (Equation_tbl.find_inv interface_literal (Equation.make i j)))
	    
  (* term clause corresponding to an EUF clause *)
  method private term_clause_of cc_clause =
    List.map self#term_lit_of cc_clause

  method is_interface_literal term_lit =
    assert (Term_literal.is_literal term_lit);
    Equation_tbl.mem interface_literal (Term_literal.variable term_lit)

  method is_true term_lit = 
    assert (self#is_interface_literal term_lit);
    Cc_core.is_true core (self#cc_lit_of term_lit)
    
  method private is_false term_lit = 
    assert (self#is_interface_literal term_lit);
    Cc_core.is_true core (Cc_literal.negation (self#cc_lit_of term_lit))
    
  method set_literal term_lit =
    assert (self#is_interface_literal term_lit);
    assert (not self#conflict_detected);
    Cc_core.set_literal core (self#cc_lit_of term_lit);
    assert self#invariant

  method get_literal_possible = 
    Cc_core.get_literal_possible core

  method get_literal =
    assert self#get_literal_possible;
    let result = self#term_lit_of (Cc_core.get_literal core) in
      assert self#invariant;
      assert (self#is_true result);
      result

  method literal_explanation term_lit = 
    let cc_lit = self#cc_lit_of term_lit in
    let cls = self#term_clause_of (Cc_core.literal_explanation core cc_lit) in
      assert (List.for_all self#is_interface_literal cls);
      assert (List.mem term_lit cls);
      assert 
	(List.for_all (fun l -> self#is_false l || term_equal l term_lit) cls);
      assert (List.length cls >= 2);
      cls

  method conflict_clause = 
    assert self#conflict_detected;
    let cls = self#term_clause_of (Cc_core.conflict_clause core) in
      assert (List.for_all self#is_false cls);
      assert (cls != []);
      cls

  method backjump level = 
    assert (0 <= level && level < self#decision_level);
    Cc_core.backjump core level;
    assert self#invariant;
    assert (level = self#decision_level);
    assert (not self#conflict_detected);
    assert (not self#inference_possible);
    assert (not self#get_literal_possible)
      
  method restart = 
    assert (not self#conflict_detected);
    Cc_core.restart core;
    assert self#invariant;
    assert (self#decision_level = 0);
    assert (not self#conflict_detected)
end
