(* $Id: cc_core.ml 1793 2007-09-28 19:12:36Z jgrundy $ *)
(* Congruence-closure based decision procedure for EUF. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Cc_literal

(* Description of (how to undo) a merge event, in which two variables
   are asserted equal. The equivalence class of the source variable is
   merged into the equivalence class of the sink variable. *)
type merge_event = {
  source: Variable.t;                     (* source variable for merge *)
  sink: Variable.t;                       (* sink variable for merge *)
  old_source_rep: Variable.t;             (* represents of old source class *)
  sink_rep: Variable.t;                   (* represents of the sink class *)
  old_sink_members: Variable.t Bunch.t;   (* members of the old sink class *)
  old_sink_fun_uses: Variable.t Bunch.t;  (* old uses of sink as a function *)
  old_sink_arg_uses: Variable.t Bunch.t;  (* old uses of sink as an argument *)
  old_sink_diseq_vars: Variable.t Bunch.t;(* vars disequal to old sink class *) 
  old_sink_interface_lits: (Variable.t * Variable.t) Bunch.t; 
  (* undispatched interface literals for the old sink class *)
}

(* Description of (how to undo) a separate event, in which two
   variables are asserted disequal. *)
type separate_event = {
  a_rep: Variable.t;                    (* rep of first separated class *)
  b_rep: Variable.t;                    (* rep of second separated class *)
  old_a_diseq_vars: Variable.t Bunch.t; (* old vars disequal to first class *)
  old_b_diseq_vars: Variable.t Bunch.t; (* old vars disequal to second class *)
  old_a_interface_lits: (Variable.t * Variable.t) Bunch.t; 
  (* Undispatched interface literals for the first class *)
  old_b_interface_lits: (Variable.t * Variable.t) Bunch.t; 
  (* Undispatched interface literals for the second class *)
}

(* Description of (how to undo / redo) a proxy event, in which a
   variable y is set as the proxy for the application of f to x. *)
type proxy_event = {
  defn: Variable.t;
  fun_rep: Variable.t;
  arg_rep: Variable.t;
  old_fun_uses: Variable.t Bunch.t;
  old_arg_uses: Variable.t Bunch.t;
}

(* Description of (how to undo / redo) an interface event, in which an
   interface (dis)equality between two variables is declared. *)
type interface_event = {
  x: Variable.t;
  y: Variable.t;
  x_rep: Variable.t;
  y_rep: Variable.t;
  old_x_rep_interface_lits: (Variable.t * Variable.t) Bunch.t;
  old_y_rep_interface_lits: (Variable.t * Variable.t) Bunch.t;
}

(* Description of (how to undo) an event. *)
type event = 
  | IncrementLevel
  | Merge of merge_event
  | Separate of separate_event
  | Proxy of proxy_event
  | Interface of interface_event

(* Timestamps

   The solver state includes tables that store data implied by solver
   events. When an event is undone during backjumping, the
   corresponding data in the tables becomes invalid. We do not waste
   time seeking out and removing the data from the table when it
   becomes invalid. Instead, we use a timestamp to determine the
   validity of data on lookup; lazily removing invalid data as we
   encounter it.

   There are two requirements on timestamps : i) each event must have
   a unique timestamp, ii) it should be possible to check the validity
   of a timestamp in constant time.

   When an event is recorded, its position in the event history is
   fixed and remains unique as long as the event remains valid.
   However, if backjumping invalidates an event, then another event
   may be assigned to the position it previously occupied in the event
   history. The important thing is that the position of an event in
   the history is unique for currently valid events.

   We obtain a unique timestamp for all events (valid and invalid) by
   using the pair (position, branch); where branch is the number of
   backjumps prior to the event, and position is the position at which
   it was recorded in the event history.  Uniqueness is guaranteed
   because no two events can share the same position in the event
   history without a backjump separating them, in which case their
   branches will differ.

   This choice of timestamp provides a constant-time validity check.
   Remember that the valid events are those present in the event
   history. The timestamp of an event is clearly invalid if its
   position is greater than the current length of the event
   history. If its position is within bounds, the timestamp is valid
   when the branch of the event stored at that position matches the
   branch in the timestamp.

   As an optimization, we do not store events occurring at decision
   level 0 in the trace, since such events will never be undone. The
   information associated with such events is recorded with a
   timestamp with position -1, and is always valid. *)

type branch   = int
type position = int

(* Reason for an equality assignment *)
type eq_reason = 
  | External      (* external world asserted the literal *)
  | Congruence    (* inferred by congruence in a merge event *)

(* Edges in the proof forest. *)
type edge = Variable.t * eq_reason

let no_edge = (Variable.invalid, External)

let no_conflict = (Variable.invalid, Variable.invalid)

(* Solver state for a congruence-closure-based EUF solver. *)
type solver = {

  vars: Variable.generator;
  (* A pool of variables. *)

  mutable conflict: (Variable.t * Variable.t);
  (* If the conflict pair (x,y) is not no_conflict, then we have
     detected the conflict that both x=y and x!=y *)

  mutable decision_level: int;
  (* Number of IncrementLevel events recorded. *)

  mutable branch: branch;
  (* Number of backjumps performed. *)

  events: event Stack.t;
  (* Event history for back jumping. *)

  event_branch: branch Vec.t;
  (* Maps each position in the event history to the branch when it was
     stored there.  Used for timestamp validity checking. *)

  pending_lits: literal Queue.t;
  (* Externally asserted literals waiting to be processed. *)

  pending_merges: (Variable.t * Variable.t) Queue.t;
  (* Inferred merges waiting to be processed. *)

  pending_definitions: Variable.t Stack.t;
  (* Definitions pending re-processing after backtrack. *)

  pending_interface_lits: (Variable.t * Variable.t) Stack.t;
  (* Interface literal declarations pending re-processing after backtrack. *)


  (* Main Union-find structure. *)

  representative: Variable.t Variable_map.t;
  (* Maps variables to their representatives. *)

  members: (Variable.t Bunch.t) Variable_map.t;
  (* Maps representatives to their class members.  Other variables are
     mapped to their class members when they were last a
     representative. This supports efficient backjumping to the point
     where the variable is, once again, a representative. *)


  (* Disequalities *)

  diseq_vars: (Variable.t Bunch.t) Variable_map.t;
  (* Maps representatives to at least one variable from each class
     known disequal to the class of the representative.  Other
     variables are mapped to what was their disequality information
     when they were last a representative.  This supports efficient
     backjumping to the point where the variable is, once again, a
     representative. *)

  diseq_reason: (Variable.t * Variable.t * position * branch) Variable_map2.t;
  (* Maps a disequality to the reason for the disequality. The reason
     is an interface disequality between variables in the same classes
     as the variables considered.  This information is time-stamped.
     Note the disequalities (x,y) and (y,x) are stored only once, with
     the smaller variable first. *)

  (* Application proxies. *)
  
  definition: (Variable.t * Variable.t) Variable_map.t;
  (* Maps application proxy variables to the function and argument of
     the proxied application.  Other variables are mapped to a pair of
     invalid variables. *)
  
  proxy: (Variable.t * position * branch) Variable_map2.t;
  (* If the application f@x has a proxy and f' and x' are the
     representatives for f and x, then proxy maps (f',x') to the proxy
     for some (g,y) such that g and y are equivalent to f and x
     respectively. *)

  fun_uses: (Variable.t Bunch.t) Variable_map.t;
  (* For every f represented by f' and p such that p proxies f@x for
     some x, we have p in fun_uses of f'. *)

  arg_uses: (Variable.t Bunch.t) Variable_map.t;
  (* For every x represented by x' and p such that p proxies f@x for
     some f, we have p is in arg_uses of x'. *)
  
  (* Dispatches. *)

  dispatches: literal Queue.t;
  (* Literals awaiting dispatch to the external world. *)

  interface_lits: ((Variable.t * Variable.t) Bunch.t) Variable_map.t;
  (* For every x represented by x', interface_lits maps x' to a set of
     (x'',y) interface literals such that x'' and x are in the same
     class.  The set must include all such interface literals that are
     yet to be dispatched.  Previously dispatched literals may be also
     present. *)

  (* Proof Forest *)

  proof_parent: edge Variable_map.t;
  (* Maps variables to their parent in the proof forest, and the
     reason for their equality. Two variables map transitively to a
     common root ancestor if and only if they are in the same class in
     the main union find. *)

  proof_visited: Variable_set.t;
  (* Records which variables in a proof have been visited in an explanation. *)

  (* Explanation Union-Find *)

  exp_representative: Variable.t Variable_map.t;
  (* Maps variables to their representative in the explanation
     union-find. *)

  exp_representative_valid: Variable_set.t;
  (* Records which variables have valid explanation representatives recorded. *)

  exp_members: (Variable.t Bunch.t) Variable_map.t;
  (* Maps representatives in explanation union-find to their class
     members. *)

  exp_highest: Variable.t Variable_map.t;
  (* Maps explanation representatives to the 'highest' variable in the
     equivalence class.  The highest variable is the one closest to
     the root in the corresponding proof-forest tree.  This need not
     be the root of the tree. *)
}

(*****************************************************************************)
(* Solver Status *)

let conflict_detected s = 
  s.conflict != no_conflict

let inference_possible s =
  not (Queue.is_empty s.pending_lits)

let get_literal_possible s =
  not (Queue.is_empty s.dispatches)

let decision_level s = 
  s.decision_level
 
let num_events s = 
  Vec.length s.event_branch
   
(*****************************************************************************)
(* Internal solver state and invariants *)

(* Only used in assertions. *)
let invariant s = 
  (* If there is a conflict, both conflict variables must be valid *)
  (not (conflict_detected s)
   || (let (x,y) = s.conflict in Variable.is_valid x && Variable.is_valid y))
  && (* The decision level is non-negative. *)
    (decision_level s >= 0)
  && (* The trace of events is empty unless decision level > 0 *)
    (decision_level s != 0 || Stack.is_empty s.events)
  && (* The number of backjumps is non-negative *)
    (s.branch >= 0) 
  && (* The number of events is non-negative *)
    (num_events s >= 0)
  && (* We record the branch for every event *)
    (Stack.length s.events = Vec.length s.event_branch)
  && (* Even the most recent event is not in the future *)
    (num_events s = 0 || s.branch >= Vec.top s.event_branch) 
    
(*****************************************************************************)
(* Solver state access and modification *)

let representative s x = Variable_map.find s.representative x
let members        s x = Variable_map.find s.members x
let diseq_vars     s x = Variable_map.find s.diseq_vars x
let fun_uses       s x = Variable_map.find s.fun_uses x
let arg_uses       s x = Variable_map.find s.arg_uses x
let interface_lits s x = Variable_map.find s.interface_lits x
let proof_parent   s x = Variable_map.find s.proof_parent x
let definition     s x = Variable_map.find s.definition x

let set_representative s x r = Variable_map.replace s.representative x r
let set_members        s x c = Variable_map.replace s.members x c
let set_diseq_vars     s x d = Variable_map.replace s.diseq_vars x d
let set_fun_uses       s x l = Variable_map.replace s.fun_uses x l
let set_arg_uses       s x r = Variable_map.replace s.arg_uses x r
let set_interface_lits s x w = Variable_map.replace s.interface_lits x w
let set_proof_parent   s x p = Variable_map.replace s.proof_parent x p
let set_definition     s x d = Variable_map.replace s.definition x d

(* Is a timestamp valid?  See above for a full explanation. *)
let time_stamp_valid s position branch = 
  assert (-1 <= position);
  assert (branch <= s.branch);
  (position = -1) ||
    (position < num_events s && branch = Vec.get s.event_branch position)
    
(* Set a new representative for all the members of a class. *)
let set_class_representative s members rep = 
  Bunch.iter (fun x -> set_representative s x rep) members

(* Only used in assertions. *)
let is_representative s x = representative s x = x

(* Only used in assertions. *)
let is_unique s x = is_representative s x && (members s x) = Bunch.singleton x

(* Are two variables known to be equal? *)
let equal s x y = representative s x = representative s y

(* Has the disequality x!=y been recorded? *)
let disequality_reason_recorded s x y =
  try
    if Variable.compare x y < 0 then 
      let (_, _, position, branch) = Variable_map2.find s.diseq_reason x y in
      let current = time_stamp_valid s position branch in
	if not current then Variable_map2.remove s.diseq_reason x y;
	current
    else
      let (_, _, position, branch) = Variable_map2.find s.diseq_reason y x in
      let current = time_stamp_valid s position branch in
	if not current then Variable_map2.remove s.diseq_reason y x;
 	current
  with Not_found -> false

(* Reason, a!=b, for a recorded disequality x!=y. That is:
   x=a && a!=b && b=y ==> x!=y 
   The reason is an externally set interface literal. *)
let disequality_reason s x y = 
  assert (disequality_reason_recorded s x y);
  if Variable.compare x y < 0 then 
    let (a, b, position, branch) = Variable_map2.find s.diseq_reason x y in
      assert (time_stamp_valid s position branch);
      assert (equal s x a && equal s y b);
      (a,b)
  else
    let (b, a, position, branch) = Variable_map2.find s.diseq_reason y x in
      assert (time_stamp_valid s position branch);
      assert (equal s x a && equal s y b);
      (a,b)
	
(* Record the disequality x!=y for the reason a!=b. That is:
   x=a && a!=b && b=y ==> x!=y 
   The reason must be an interface literal. *)
let record_disequality_reason s a b x y =
  assert (equal s x a && equal s y b);
  assert (not (disequality_reason_recorded s x y));
  let position = num_events s - 1 in
    assert (time_stamp_valid s position s.branch);
    if Variable.compare x y < 0 then 
      Variable_map2.set s.diseq_reason x y (a, b, position, s.branch)
    else
      Variable_map2.set s.diseq_reason y x (b, a, position, s.branch)
	
(* Deduce the disequality x!=y for the same reason as 
   the recorded disequality u!=v. That is:
   x=u && u=a && a!=b && b=v && v=y ==> x!=y *)
let deduce_disequality s u v x y =
  assert (equal s u x && equal s v y);
  assert (disequality_reason_recorded s u v);
  assert (not (disequality_reason_recorded s x y));
  let (a,b) = disequality_reason s u v in
    record_disequality_reason s a b x y

(* Two variables are known disequal when we have a disequality
   recorded between their representatives. *)
let disequal s x y =
  disequality_reason_recorded s (representative s x) (representative s y)

let is_true s = function
  | Equality (a,b) -> equal s a b
  | Disequality (a,b) -> disequal s a b

let is_false s = function
  | Equality (a,b) -> disequal s a b
  | Disequality (a,b) -> equal s a b

let is_assigned s l =
  let result = match l with
    | Equality (a,b) | Disequality (a,b) -> 
	equal s a b || disequal s a b
  in
    assert (result = (is_true s l || is_false s l));
    result

let is_explainable s l = 
  let can_explain = match l with 
    | Equality (a,b) -> equal s a b
    | Disequality (a,b) -> disequality_reason_recorded s a b 
  in
    assert ((not can_explain) || is_true s l);
    can_explain

let were_set_equal s a b =
  assert (equal s a b);
  let equated x y =
    match (proof_parent s x) with
      | (z, External) -> z = y
      | _ -> false
  in
    equated a b || equated b a

let were_set_disequal s a b =
  assert (disequal s a b);
  disequality_reason_recorded s a b 
  && (let reason = disequality_reason s a b in
	(reason = (a,b)) || (reason = (b,a)))
	
let is_inferred s l =
  assert (is_true s l);
  match l with
    | Equality (a,b)    -> not (were_set_equal s a b)
    | Disequality (a,b) -> not (were_set_disequal s a b)

(* Record a conflict between (x,y) because in the current solver
   state, we have determined both x=y and x!=y *)
let set_conflict s x_y =
  assert (not (conflict_detected s));
  assert (let (x,y) = x_y in x != y && equal s x y && disequal s x y);
  s.conflict <- x_y

let no_proxy = Variable.invalid

(* The proxy for an application f@x of representatives f and x, or
   no_proxy if there is none. *)
let proxy s f x = 
  assert (is_representative s f);
  assert (is_representative s x);
  try
    let (proxy, position, branch) = Variable_map2.find s.proxy f x in
      if time_stamp_valid s position branch then
	proxy 
      else begin
	Variable_map2.remove s.proxy f x; (* remove invalid information *)
	no_proxy	  
      end
  with Not_found -> no_proxy

(* Does an application have a proxy? Used only in assertions. *)
let defined s f x = proxy s f x != no_proxy

(* Sets the proxy for the application f@x, of representatives f and x,
   to y. *)
let set_proxy s f x y =
  assert (is_representative s f);
  assert (is_representative s x);
  assert (proxy s f x = no_proxy);
  assert (let (f',x') = definition s y in equal s f f' && equal s x x'); 
  let position = num_events s - 1 in
    assert (time_stamp_valid s position s.branch);
    Variable_map2.set s.proxy f x (y, position, s.branch)

(* Add literal l to the dispatch queue. *)      
let dispatch s l =
  assert (is_true s l);
  Queue.add l s.dispatches

(*****************************************************************************)
(* Solver setup and initialization. *)

let invalid_application = (Variable.invalid,Variable.invalid)

(* Create a new solver instance. *)
let new_solver () =
  let s = {
    vars = Variable.new_generator ();
    
    conflict                 = no_conflict;
    decision_level           = 0;
    branch                   = 0;
    events                   = Stack.create ();
    event_branch             = Vec.create 0 0;
    pending_lits             = Queue.create ();
    pending_merges           = Queue.create ();
    pending_definitions      = Stack.create ();
    pending_interface_lits   = Stack.create ();

    representative           = Variable_map.create Variable.invalid;
    members                  = Variable_map.create Bunch.empty;

    diseq_vars               = Variable_map.create Bunch.empty;
    diseq_reason             = Variable_map2.create 0 0;

    definition               = Variable_map.create invalid_application;
    proxy                    = Variable_map2.create 0 0;
    fun_uses                 = Variable_map.create Bunch.empty;
    arg_uses                 = Variable_map.create Bunch.empty;

    interface_lits           = Variable_map.create Bunch.empty;
    dispatches               = Queue.create ();

    proof_parent             = Variable_map.create no_edge;
    proof_visited            = Variable_set.create ();

    exp_representative       = Variable_map.create Variable.invalid;
    exp_representative_valid = Variable_set.create ();
    exp_members              = Variable_map.create Bunch.empty;
    exp_highest              = Variable_map.create Variable.invalid;
  } in
    assert (invariant s);
    assert (decision_level s = 0);
    s

(* Create a fresh variable in the nominated solver. *)
let fresh_variable s =
  let v = Variable.fresh s.vars in
    Variable_map.add s.representative     v v;
    Variable_map.add s.members            v (Bunch.singleton v);
    Variable_map.add s.diseq_vars         v Bunch.empty;
    Variable_map.add s.definition         v invalid_application;
    Variable_map.add s.fun_uses           v Bunch.empty;
    Variable_map.add s.arg_uses           v Bunch.empty;
    Variable_map.add s.interface_lits     v Bunch.empty;
    Variable_map.add s.proof_parent       v no_edge;
    Variable_map.add s.exp_representative v v;
    Variable_map.add s.exp_members        v Bunch.empty;
    Variable_map.add s.exp_highest        v Variable.invalid;
    assert (invariant s);
    assert (is_unique s v);
    v

(*****************************************************************************)
(* Merge and Separate*)

let push_event s e =
  assert (not (conflict_detected s));
  assert (Stack.length s.events = Vec.length s.event_branch);
  if decision_level s > 0 then begin
    Stack.push e s.events;
    Vec.push s.event_branch s.branch;
  end;
  assert (Stack.length s.events = Vec.length s.event_branch)

let pop_event s =
  assert (Stack.length s.events = Vec.length s.event_branch);
  let _ = Vec.pop s.event_branch in
  let top = Stack.pop s.events in
  assert (Stack.length s.events = Vec.length s.event_branch);
    top

(* Make a variable the representative for its class in the proof forest. *)
let set_proof_representative s a =
  (* Each equivalence class in the proof forest is a tree with the
     class representative at it's root.  If you reverse the direction
     of the links between any variable and the root, then that
     variable becomes the root. *)
  let rec rev_to_root x (parent, reason) =
    if Variable.is_valid parent then begin
      let next_edge = proof_parent s parent in
        (* Flip the edge between x and its parent *)
        set_proof_parent s parent (x, reason);
        (* Continue reversing up to root *)
        rev_to_root parent next_edge;
    end
  in
    rev_to_root a (proof_parent s a)

(* Process an event asserting equality between two variables. *)
let merge_vars s reason a b =
  assert (not (conflict_detected s));
  let a_rep = representative s a in
  let b_rep = representative s b in 
    
    if a_rep != b_rep then begin
      (* a and b are from different equivalence classes. 
	 Merge the two classes. *)
      
      (* Choose the smaller class (source) to merge into the larger
	 one (sink) *)
      let (source, source_rep, source_members, sink, sink_rep, sink_members) =
	let a_members = members s a_rep in
	let b_members = members s b_rep in
	  if Bunch.size a_members > Bunch.size b_members then
            (b, b_rep, b_members, a, a_rep, a_members)
	  else
            (a, a_rep, a_members, b, b_rep, b_members)
      in
	
      (* Get source/sink information *)
      let source_fun_uses       = fun_uses       s source_rep in
      let sink_fun_uses         = fun_uses       s sink_rep   in
      let source_arg_uses       = arg_uses       s source_rep in
      let sink_arg_uses         = arg_uses       s sink_rep   in
      let sink_diseq_vars       = diseq_vars     s sink_rep   in
      let source_diseq_vars     = diseq_vars     s source_rep in
      let source_interface_lits = interface_lits s source_rep in
      let sink_interface_lits   = interface_lits s sink_rep   in
	
	(* Record undo information for this event. *)
	push_event s 
	  (Merge {
	     source                  = source;
	     sink                    = sink;
	     old_source_rep          = source_rep;
	     sink_rep                = sink_rep;
	     old_sink_members        = sink_members;
	     old_sink_fun_uses       = sink_fun_uses;
	     old_sink_arg_uses       = sink_arg_uses;
	     old_sink_diseq_vars     = sink_diseq_vars;
	     old_sink_interface_lits = sink_interface_lits});
	
	(* Merge source and sink equivalence classes in the proof forest. *)
	set_proof_representative s source;
	set_proof_parent s source (sink, reason);
	
	(* Update the main union-find structure *)
        set_class_representative s source_members sink_rep;
        set_members s sink_rep (Bunch.join sink_members source_members);
	
	(* Find equalities implied by congruence, and update the proxy
	   table and variable use information. *)
	begin
          let new_proxy y =
	    (* We are in one of two situations: 
	       1. the proxy y = source@x, for some x
	       --- if some proxy z=sink@x then deduce y=z
	       --- otherwise make y a proxy for sink@x
	       2. the proxy y = f@source, for some f
	       --- if some proxy z=f@sink then deduce y=z
	       --- otherwise make y a proxy for f@sink *)
	    let (f,x) = definition s y in
            let f_rep = representative s f in
            let x_rep = representative s x in
	    let z = proxy s f_rep x_rep in
	    let generate_proxy = (z = no_proxy) in
	      assert (f_rep = sink_rep || x_rep = sink_rep);
	      (* f@x = y && f = f_rep && x = x_rep *)
	      if generate_proxy then
		(* f_rep@x_rep = y *)
		set_proxy s f_rep x_rep y
	      else
		(* f_rep@x_rep = z ==> y = z *)
		Queue.push (y,z) s.pending_merges;
	      generate_proxy
          in
          let new_fun_uses = Bunch.filter new_proxy source_fun_uses in
          let new_arg_uses = Bunch.filter new_proxy source_arg_uses in
            set_fun_uses s sink_rep (Bunch.join sink_fun_uses new_fun_uses);
            set_arg_uses s sink_rep (Bunch.join sink_arg_uses new_arg_uses)
	end;
	
	(* Update disequal variables and table. *)
	begin
	  (* For all x such that source != x, we have sink != x.
	     Deduce and record all such disequalities not previously known. *)
	  let new_sink_diseq x =
            let x_rep = representative s x in
	    let unrecorded = not (disequality_reason_recorded s sink_rep x_rep)
	    in
	      if unrecorded then
		deduce_disequality s source_rep x_rep sink_rep x_rep;
	      unrecorded
	  in
          let new_diseqs = Bunch.filter new_sink_diseq source_diseq_vars 
	  in
	    set_diseq_vars s sink_rep (Bunch.join sink_diseq_vars new_diseqs)
	end;
	
	(* Update interface literals and find literals to dispatch *)
	begin
          (* Filter out equations that have become assigned.  Filtered
	     out (dis)equations are added to the dispatch list. *)
          let undetermined diseq_possible (c,d) =
            assert (representative s c = sink_rep); 
            let d_rep = representative s d in
	    let determined = 
	      sink_rep = d_rep 
	      || (diseq_possible 
		  && disequality_reason_recorded s sink_rep d_rep)
	    in
              (* c = sink_rep && d = d_rep *)
	      if determined then
		if sink_rep = d_rep then 
		  (* sink_rep = d_rep ==> c = d *)
		  dispatch s (Equality (c,d))
		else begin
		  assert (disequality_reason_recorded s sink_rep d_rep);
		  (* sink_rep != d_rep ==> c != d *)
		  if not (disequality_reason_recorded s c d) then 
		    deduce_disequality s sink_rep d_rep c d;
		  dispatch s (Disequality (c,d))
		end;
	      not determined
	  in
          let source_interface_lits' = 
	    Bunch.filter 
	      (undetermined (sink_diseq_vars != Bunch.empty)) 
	      source_interface_lits
	  in
          let sink_interface_lits' =
	    Bunch.filter 
	      (undetermined (source_diseq_vars != Bunch.empty))
	      sink_interface_lits
	  in
            set_interface_lits s sink_rep 
	      (Bunch.join sink_interface_lits' source_interface_lits')
	end;
        
	(* Check for contradiction if source_rep != sink_rep *)
	if (disequality_reason_recorded s sink_rep sink_rep) then
	  set_conflict s (disequality_reason s sink_rep sink_rep)
    end;
    assert (invariant s);
    assert (equal s a b)

let merge s a b =
  assert (not (conflict_detected s));
  merge_vars s External a b;
  while not (conflict_detected s || Queue.is_empty s.pending_merges) do
    let (a,b) = Queue.pop s.pending_merges in
      merge_vars s Congruence a b
  done;
  assert (invariant s);
  assert (equal s a b)

let separate s a b = 
  assert (not (conflict_detected s));
  let a_rep = representative s a in
  let b_rep = representative s b in
    if not (disequality_reason_recorded s a_rep b_rep) then begin
      let a_diseq_vars     = diseq_vars     s a_rep in
      let b_diseq_vars     = diseq_vars     s b_rep in
      let a_interface_lits = interface_lits s a_rep in
      let b_interface_lits = interface_lits s b_rep in

	(* Record undo information for this event. *)
	push_event s 
	  (Separate {
	     a_rep                = a_rep;
	     b_rep                = b_rep;
	     old_a_diseq_vars     = a_diseq_vars;
	     old_b_diseq_vars     = b_diseq_vars;
             old_a_interface_lits = a_interface_lits;
	     old_b_interface_lits = b_interface_lits});
	  
	(* Update disequality information. *)
	begin
	  set_diseq_vars s a_rep (Bunch.add b_rep a_diseq_vars);
          set_diseq_vars s b_rep (Bunch.add a_rep b_diseq_vars);
          record_disequality_reason s a b a_rep b_rep;
	end;
	
	(* Update interface literals and dispatch inferred disequalities. *)
	begin
          let known_diseq (c,d) = 
            let c_rep = representative s c in
            let d_rep = representative s d in
	    let diseq = disequality_reason_recorded s c_rep d_rep in
              assert (c_rep = a_rep || c_rep = b_rep);
	      if diseq then begin
		(* c = b_rep && b_rep != d_rep && d_rep = d ==> c != d *)
		if not (disequality_reason_recorded s c d) then 
		  deduce_disequality s c_rep d_rep c d;
		dispatch s (Disequality (c,d))
	      end;
	      not diseq
          in
            set_interface_lits s a_rep
              (Bunch.filter known_diseq a_interface_lits);
            set_interface_lits s b_rep 
	      (Bunch.filter known_diseq b_interface_lits);
	end;
	
	(* Check for conflict. *)
	if a_rep = b_rep then
	  set_conflict s (a, b)
    end;
    assert (invariant s);
    assert (disequal s a b)
      
(*****************************************************************************)
(* Definitions and Interface_literals *)

(* Only used in assertions. Is v a fresh definition, i.e. : 
   1. It has not been merged / separated with any other variable yet.
   2. It has not been used yet.
   3. It does not occur in an interface literal as yet.
   4. It does, however, have a definition (for a function application). *)
let is_fresh_definition s v =
  is_unique s v
  && Bunch.is_empty (diseq_vars s v)
  && Bunch.is_empty (fun_uses s v)
  && Bunch.is_empty (arg_uses s v)
  && Bunch.is_empty (interface_lits s v)
  && definition s v != invalid_application

(* Process a definition y = f @ x.
   If there exists a definition z = f_rep @ x_rep, then infer y = z.
   Otherwise, set y to be the proxy for f_rep @ x_rep. *)
let process_definition s y =
  assert (not (conflict_detected s));
  assert (is_fresh_definition s y);
  let (f,x) = definition s y in
  let f_rep = representative s f in
  let x_rep = representative s x in
  let f_rep_fun_uses = fun_uses s f_rep in
  let x_rep_arg_uses = arg_uses s x_rep in
  let z = proxy s f_rep x_rep in
  let is_new_proxy = (z = no_proxy) in
    push_event s
      (Proxy {
        defn = y;
        fun_rep = f_rep;
        arg_rep = x_rep;
        old_fun_uses = f_rep_fun_uses;
        old_arg_uses = x_rep_arg_uses});
    (* f@x = y && f = f_rep && x = x_rep *)
    if is_new_proxy then begin
      (* f_rep@x_rep = y *)
      set_proxy s f_rep x_rep y;
      set_fun_uses s f_rep (Bunch.add y f_rep_fun_uses);
      set_arg_uses s x_rep (Bunch.add y x_rep_arg_uses);
    end
    else begin
      (* f_rep@x_rep = z ==> y = z *)
      merge_vars s Congruence y z
    end;
    assert (not (conflict_detected s))

(* Add the definition for an application proxy f @ x = y for some fresh y *)
let add_definition s f x =
  assert (not (conflict_detected s));
  assert (not (defined s f x));
  let y = fresh_variable s in
    set_definition s y (f,x);
    process_definition s y;
    assert (invariant s);
    assert (defined s f x);
    assert (not (conflict_detected s));
    y

(* Add an interface literal. If the interface literal is determined,
   dispatch it. Otherwise, put it in a list of undispatched interface
   literals. *)
let add_interface_literal s (x,y) =
  assert (not (conflict_detected s));
  let x_rep = representative s x in
  let y_rep = representative s y in
  let x_lits = interface_lits s x_rep in
  let y_lits = interface_lits s y_rep in
    push_event s 
      (Interface {
        x = x;
        y = y;
        x_rep = x_rep;
        y_rep = y_rep;
        old_x_rep_interface_lits = x_lits;
        old_y_rep_interface_lits = y_lits});
    (* x = x_rep && y = y_rep *)
    if x_rep = y_rep then
      (* x_rep = y_rep ==> x = y *)
      dispatch s (Equality (x,y))
    else if disequality_reason_recorded s x_rep y_rep then begin
      (* x_rep != y_rep ==> x != y *)
      dispatch s (Disequality (x,y));
      if not (disequality_reason_recorded s x y) then 
	deduce_disequality s x_rep y_rep x y;
    end
    else begin
      (* Record (x,y) in a list of undispatched interface literals. *)
      if (Bunch.size x_lits < Bunch.size y_lits) then
        set_interface_lits s x_rep (Bunch.add_balanced (x,y) x_lits)
      else
        set_interface_lits s y_rep (Bunch.add_balanced (y,x) y_lits);
    end;
    assert (invariant s);
    assert (not (conflict_detected s))


(*****************************************************************************)
(* Search API *)

let get_literal s = 
  assert (get_literal_possible s);
  let lit = Queue.take s.dispatches in
    assert (invariant s);
    assert (is_true s lit);
    lit

let set_literal s lit = 
  assert (not (conflict_detected s));
  Queue.push lit s.pending_lits;
  assert (invariant s)

let infer s = 
  assert (not (conflict_detected s));
  assert (inference_possible s);
  begin
    match Queue.pop s.pending_lits with
      | Equality (a,b)    -> merge s a b
      | Disequality (a,b) -> separate s a b
  end;
  assert (invariant s)

let increment_level s = 
  assert (not (conflict_detected s));
  assert (not (inference_possible s));
  assert (not (get_literal_possible s));
  s.decision_level <- s.decision_level + 1;
  push_event s IncrementLevel;
  assert (invariant s);
  assert (not (conflict_detected s));
  assert (not (inference_possible s));
  assert (not (get_literal_possible s));
  assert (decision_level s > 0)


(* source, sink, old_source_rep, sink_rep *)

(* Undo a merge operation. *)
let undo_merge s e =
  assert (representative s e.source = e.sink_rep);
  assert (representative s e.old_source_rep = e.sink_rep);
  assert (representative s e.sink = e.sink_rep);

  (* remove edge from proof forest *)
  assert ((fst (proof_parent s e.source) = e.sink) || 
	    (fst (proof_parent s e.sink) = e.source));
  if fst (proof_parent s e.source) = e.sink then 
    set_proof_parent s e.source no_edge
  else
    set_proof_parent s e.sink no_edge;
  
  (* de-merge source and sink classes *)
  set_members        s e.sink_rep e.old_sink_members;
  set_fun_uses       s e.sink_rep e.old_sink_fun_uses;
  set_arg_uses       s e.sink_rep e.old_sink_arg_uses;
  set_diseq_vars     s e.sink_rep e.old_sink_diseq_vars;
  set_interface_lits s e.sink_rep e.old_sink_interface_lits;
  set_class_representative s (members s e.old_source_rep) e.old_source_rep;
  
  assert (invariant s);
  assert (representative s e.source = e.old_source_rep);
  assert (representative s e.sink = e.sink_rep);
  assert (not (equal s e.source e.sink))
    
(* Undo a separate operation. *)
let undo_separate s e =
  assert (is_representative s e.a_rep);
  assert (is_representative s e.b_rep);

  (* revert to old disequalities and interface literals *)
  set_diseq_vars     s e.a_rep e.old_a_diseq_vars;
  set_diseq_vars     s e.b_rep e.old_b_diseq_vars;
  set_interface_lits s e.a_rep e.old_a_interface_lits;
  set_interface_lits s e.b_rep e.old_b_interface_lits;
  
  assert (invariant s);
  assert (not (disequality_reason_recorded s e.a_rep e.b_rep))

(* Undo a proxy event *)
let undo_proxy s e =
  assert (is_fresh_definition s e.defn);
  assert (is_representative s e.fun_rep);
  assert (is_representative s e.arg_rep);
  set_fun_uses s e.fun_rep e.old_fun_uses;
  set_arg_uses s e.arg_rep e.old_arg_uses;
  Stack.push e.defn s.pending_definitions;
  assert (invariant s)

(* Undo an interface_literal event *)
let undo_interface s e =
  assert (is_representative s e.x_rep);
  assert (is_representative s e.y_rep);
  assert (representative s e.x = e.x_rep);
  assert (representative s e.y = e.y_rep);
  set_interface_lits s e.x_rep e.old_x_rep_interface_lits;
  set_interface_lits s e.y_rep e.old_y_rep_interface_lits;
  Stack.push (e.x, e.y) s.pending_interface_lits;
  assert (invariant s)

(* backjump to decision level k *)
let backjump s k =
  assert (0 <= k && k < decision_level s);
  while s.decision_level > k do
    match pop_event s with
      | IncrementLevel -> s.decision_level <- s.decision_level - 1
      | Merge e -> undo_merge s e
      | Separate e -> undo_separate s e
      | Proxy e -> undo_proxy s e
      | Interface e -> undo_interface s e
  done;
  s.branch <- s.branch + 1;
  s.conflict <- no_conflict;
  Queue.clear s.dispatches;
  Queue.clear s.pending_lits;
  Queue.clear s.pending_merges;
  Stack.iter (process_definition s) s.pending_definitions;
  Stack.clear s.pending_definitions;
  Stack.iter (add_interface_literal s) s.pending_interface_lits;
  Stack.clear s.pending_interface_lits;
  assert (invariant s);
  assert (decision_level s = k);
  assert (not (conflict_detected s));
  assert (not (inference_possible s))

let restart s =
  assert (not (conflict_detected s));
  if decision_level s > 0 then backjump s 0;
  assert (invariant s);
  assert (decision_level s = 0);
  assert (not (conflict_detected s))
  

(*****************************************************************************)
(* Explanations *)

(* The explanation union find prevents re-explanation of equalities by
   placing variables in the same equivalence class once their equality
   has been explained. If two variables are in the same explanation
   class, then they must be in the same class in both the main union
   find and the proof forest union find, but not necessarily visa
   versa - they may be equal without being the equality having been
   explained. *)

(* The explanation class representative for a variable. *)
let exp_representative s x =
  if Variable_set.mem s.exp_representative_valid x then 
    Variable_map.find s.exp_representative x
  else begin
    (* Yet to explain x equal to anything, initialize required information. *)
    Variable_map.replace s.exp_representative x x;
    Variable_map.replace s.exp_members x (Bunch.singleton x);
    Variable_map.replace s.exp_highest x x;
    Variable_set.add s.exp_representative_valid x;
    x
  end

(* Have two equal variables been explained to be equal? *)
let exp_equal s x y = 
  assert (equal s x y);
  exp_representative s x = exp_representative s y

(* The highest variable in an explanation equivalence class is the
   highest node in the proof forest to which its members have been
   explained equal. *)
let exp_highest s x =
  Variable_map.find s.exp_highest (exp_representative s x)

(* Used only in assertions *)
let is_exp_highest s x = 
  exp_highest s x = x

(* Equal variables occur in the same tree in the proof forest. The
   shortest explanation for why two variables are equal is an
   explanation of why each of the variable is equal to their lowest
   common ancestor in that tree.  We find the lowest common ancestor
   by traversing upward in the tree from both variables until we reach
   a node that (the other) traversal has already visited.

   We may have explained why some variables along these two paths are
   equal before, so instead of moving up the proof tree one step at a
   time, we can use the explanation union-find to leap directly from a
   variable to the highest variable in the tree that it has already
   been explained equal to.  The final variable may not, therefore, be
   the lowest common ancestor in the proof tree, but we will have
   already explained why it and the lowest common ancestor are
   equal. *)

(* Traverse proof forest up until we find a visited variable. *)
let rec visited_ancestor s z = 
  assert (is_exp_highest s z);
  if Variable_set.mem s.proof_visited z then
    z
  else
    visited_ancestor s (exp_highest s (fst (proof_parent s z)))

(* Traverse two paths upwards in the proof forest, taking alternate
   steps along each path, until they encounter a visited variable. *)
let rec visited_common_ancestor s x y =
  assert (equal s x y);
  assert (is_exp_highest s x);
  assert (is_exp_highest s y);
  if Variable_set.mem s.proof_visited x then
    x
  else begin
    Variable_set.add s.proof_visited x;
    let x_parent = fst (proof_parent s x) in
      if x_parent = Variable.invalid then
        visited_ancestor s y
      else
	visited_common_ancestor s y (exp_highest s x_parent)
  end

(* Traverse proof forest up from x and y alternately, till we
   find the lowest common ancestor. *)
let lowest_common_ancestor s x y =
  assert (equal s x y);
  Variable_set.clear s.proof_visited;
  let result = visited_common_ancestor s (exp_highest s x) (exp_highest s y) 
  in
    assert (is_exp_highest s result);
    result

(* Merge a variable and its proof parent in the explanation
   union-find.  The parent is, axiomatically, higher in the proof
   tree, so the highest variable in the explanation class of the
   parent will be the highest variable in the merged explanation
   class. *)
let exp_merge s x parent =
  assert (equal s x parent);
  assert (not (exp_equal s x parent));
  assert (parent = fst (proof_parent s x));
  let x_exp_rep = exp_representative s x in
  let parent_exp_rep = exp_representative s parent in
  let x_exp_members = Variable_map.find s.exp_members x_exp_rep in
  let parent_exp_members = Variable_map.find s.exp_members parent_exp_rep in
  let merged_exp_members = Bunch.join x_exp_members parent_exp_members in
  let set_exp_rep rep v = Variable_map.replace s.exp_representative v rep in
    (* merge smaller class into larger class *)
    if Bunch.size x_exp_members > Bunch.size parent_exp_members then begin
      (* merge explanation class for parent into explanation class for x *)
      let highest_parent = Variable_map.find s.exp_highest parent_exp_rep in
        Variable_map.replace s.exp_highest x_exp_rep highest_parent;
	Bunch.iter_const set_exp_rep x_exp_rep parent_exp_members;
	Variable_map.replace s.exp_members x_exp_rep merged_exp_members
    end
    else begin
      (* merge explanation class for x into explanation class for parent *)
      Bunch.iter_const set_exp_rep parent_exp_rep x_exp_members;
      Variable_map.replace s.exp_members parent_exp_rep merged_exp_members;
    end;
    assert (exp_equal s x parent)
      
(* Adds the path from a -> c to the disjuncts in a negated
   explanation. *)
let rec explain_path_negated s disjuncts unexplained a c = 
  assert (equal s a c);
  assert (is_exp_highest s c);
  let a_high = exp_highest s a in
    assert ((a_high = c) = (exp_equal s a c));
    if a_high != c then begin
      let (a_parent, reason) = proof_parent s a_high in
	begin
	  match reason with
	    | External -> 
		disjuncts := (Disequality (a_high, a_parent)) :: !disjuncts
	    | Congruence -> 
		let (a_f, a_x) = definition s a_high in
		let (p_f, p_x) = definition s a_parent in
		  Queue.add (a_f, p_f) unexplained;
		  Queue.add (a_x, p_x) unexplained
	end;
	exp_merge s a_high a_parent;
	explain_path_negated s disjuncts unexplained a_parent c
    end;
    assert (exp_equal s a c)
      
(* The negation of the explanation for equality between x and y.  For
   example, if the literals (a=b), (b=c), and (c=d) were set, then
   the explanation for the equality between a and d might be (a=b
   && b=c && c=d), the negation of which is (a!=b || b!=c ||
   c!=d). *)
let negated_equality_explanation s x y =
  assert (equal s x y);
  Variable_set.clear s.exp_representative_valid;
  let disjuncts = ref [] in
  let unexplained = Queue.create () in
    Queue.add (x,y) unexplained;
    while not (Queue.is_empty unexplained) do
      let (a,b) = Queue.pop unexplained in
      let c = lowest_common_ancestor s a b in
	explain_path_negated s disjuncts unexplained a c;
	explain_path_negated s disjuncts unexplained b c    
    done;
    assert (exp_equal s x y);
    !disjuncts

(* Explanation for a literal. *)
let literal_explanation s lit = 
  assert (is_inferred s lit);
  let explanation = 
    match lit with 
      | Equality (a,b) ->
	  assert (equal s a b);
	  (* DNF for: equality-explanation a b ==> (a = b) *)
	  lit :: (negated_equality_explanation s a b)
      | Disequality (a,b) ->
	  assert (disequality_reason_recorded s a b);
	  let (x,y) = disequality_reason s a b in
	    (* DNF for:
	       equality-explanation a x && (x != y) &&
	       equality-explanation (y,b) ==> (a != b) *)
	    lit :: Equality (x,y) :: (negated_equality_explanation s a x) 
	    @ (negated_equality_explanation s b y)
  in
    assert (invariant s);
    assert (List.for_all (fun l -> l = lit || is_false s l) explanation);
    assert (List.mem lit explanation);
    assert (List.length explanation >= 2);
    explanation

(* Explanation for the conflict. *)
let conflict_clause s = 
  assert (conflict_detected s);
  let (a,b) = s.conflict in
  let explanation = literal_explanation s (Equality (a,b)) in
    assert (List.for_all (is_false s) explanation);
    assert (explanation != []);
    explanation
