(* $Id: cc_core.mli 1788 2007-09-28 18:23:28Z jgrundy $ *)
(** Congruence-closure solver for EUF. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** Congruence closure solver core for equality and uninterpreted
    functions based on the algorithm described in the paper:

    
    Robert Nieuwenhuis, Albert Oliveras.
    {{:http://dx.doi.org/10.1016/j.ic.2006.08.009} Fast congruence closure
    and extensions}. {i Information and Computation}, 205(4):557-580,
    2007. *)

(** {2 Solver Creation, Problem Setup and Solution} *)

(** Type of Congruence closure-based solver for Equality and
    Uninterpreted Functions. *)
type solver

val new_solver: unit -> solver
  (** A new solver. 

      [let s = new_solver ()]

      - Ensure {!decision_level}[ s = 0]
      - Ensure [not (]{!conflict_detected}[ s)]
      - Ensure [not (]{!get_literal_possible}[ s)]
      - Ensure [not (]{!inference_possible}[ s)]
*)

val fresh_variable: solver -> Cc_literal.Variable.t
  (** A fresh variable in the given solver. *)

val defined: solver -> Cc_literal.Variable.t -> Cc_literal.Variable.t -> bool
  (** [defined s f x] holds when there is a definition for the
      application of [f] to [x]. *)

val add_definition: 
  solver -> Cc_literal.Variable.t -> Cc_literal.Variable.t -> 
  Cc_literal.Variable.t
  (** [add_definition s f x] returns a proxy [y] for the application
      of [f] to [x].

      - Require [not (]{!conflict_detected}[ s)]
      - Require [not (]{!defined}[ s f x)]
      - Ensure [not (]{!conflict_detected}[ s)]
      - Ensure {!defined}[ s f x] *)

val add_interface_literal: 
  solver -> Cc_literal.Variable.t * Cc_literal.Variable.t -> unit
  (** [add_interface_literal s (x, y)] records an interest in the
      literals [x = y] and [x != y]. Only interface literals are
      available via {!get_literal}.
      
      - Require [not (]{!conflict_detected}[ s)]
      - Ensure [not (]{!conflict_detected}[ s)] *)

(** {2 Solver Status} *)

val decision_level: solver -> int
  (** Decision level of the solver. 

      - Ensure [0 <= decision_level s] *)

val inference_possible: solver -> bool
  (** Might an inference be made? *)

val conflict_detected: solver -> bool
  (** Has a conflict been detected in the current solver state? *)


(** {2 Literal Status} *)

val is_true: solver -> Cc_literal.literal -> bool
  (** Is the literal assigned true in the solver state? *)

val is_inferred: solver -> Cc_literal.literal -> bool
  (** Can the solver produce an explanation for a literal?

      - Require {!is_true}[ s l] *)

(** {2 Inference} *)

val infer: solver -> unit
  (** Attempt to make inferences from currently assigned literals. 

      - Require [not (]{!conflict_detected}[ s)]
      - Require {!inference_possible}[ s] *)

val set_literal: solver -> Cc_literal.literal -> unit
  (** Make a literal assignment obtained from another solver. 
      
      - Require [not (]{!conflict_detected}[ s)] *)

val get_literal_possible: solver -> bool
  (** Are there any inferred interface literals waiting to be
      dispatched? *)

val get_literal: solver -> Cc_literal.literal
  (** Choose an inferred interface literal that has not been chosen
      before..
      
      - Require {!get_literal_possible}[ s] 
      - Ensure {!is_true}[ s l] *)

val literal_explanation: solver -> Cc_literal.literal -> Cc_literal.literal list
  (** An explanation clause for an assigned interface literal. The
      explanation clause consists of literals that were
      given to the solver.

      [let cls = literal_explanation l s in]
      - Require {!is_inferred}[ s l] 
      - Ensure 
      {!List.for_all} 
      [(fun m -> m = l || ]{!is_true}[ s (]{!Cc_literal.negation}[ m)) cls]
      - Ensure {!List.mem}[ l cls]
      - Ensure {!List.length}[ cls >= 2] *)

val increment_level: solver -> unit
  (** Increment decision_level.

      - Require [not (]{!conflict_detected}[ s)]
      - Require [not (]{!inference_possible}[ s)]
      - Require [not (]{!get_literal_possible}[ s)]
      - Ensure [not (]{!conflict_detected}[ s)]
      - Ensure [not (]{!inference_possible}[ s)]
      - Ensure [not (]{!get_literal_possible}[ s)]
      - Ensure {!decision_level}[ s > 0] *)

(** {2 Conflict Resolution} *)

val backjump: solver -> int -> unit
  (** Backjumps to the given level. 

      For [backjump s k]:
      - Require [0 <= k && k < ]{!decision_level}[ s]
      - Ensure [k = ]{!decision_level}[ s] 
      - Ensure [not (]{!conflict_detected}[ s)] 
      - Ensure [not (]{!inference_possible}[ s)] *)

val restart: solver -> unit
  (** Backjumps to level 0 if not already there.

      - Require [not (]{!conflict_detected}[ s)]
      - Ensure [not (]{!conflict_detected}[ s)]
      - Ensure {!decision_level}[ s = 0] *)

val conflict_clause: solver -> Cc_literal.literal list
  (** An explanation for a conflict. The explanation clause consists
      of interface literals that have been set in the solver.

      [let cls = conflict_clause s in]
      - Require {!conflict_detected}[ s]
      - Ensure 
      {!List.for_all}[ (fun l -> is_true s (Cc_literal.negation l)) cls]
      - Ensure [cls != []] *)
