(* $Id: dpll.ml 1793 2007-09-28 19:12:36Z jgrundy $ *)
(* A term-level API for {!Dpll_core}. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Term
open Term_logic
open List_util

module Variable_negation    = Hashtbl.Make (Dpll_literal.Variable)
module Variable_definition  = Hashtbl.Make (Hashed_list (Dpll_literal.Literal))
module Variable_translation = 
  Bihashtbl.Make (Dpll_literal.Variable) (Term_compared)

(****************************************************************************)
(* Basic data-structure and functions for multi-input and-invert
   graphs.  These will be used as a semi-canonical elementary form when
   finding proxy variables for terms. *)

module Multi_aig :
sig
  open Dpll_literal
  type t
  val singleton: Literal.t -> t
  val conjunction: Literal.t list -> t
  val negation: t -> t
  val inputs: t -> Literal.t list
  val is_positive: t -> bool
end
  =
struct
  open Dpll_literal

  (* A node in a multi-input and-inverter graph is either an And or Nand. *)
  type t = 
    | And of Literal.t list
    | Nand of Literal.t list

  (* Normalize "And [~l]" to "Nand [l]" *)
  let normalized_and = function
    | [l] when not (Dpll_literal.is_positive l) -> 
	Nand [Dpll_literal.negation l]
    | inputs -> 
	And inputs

  (* A unit multi-input AIG. *)
  let singleton lit = normalized_and [lit]

  (* A multi-input AIG and. *)
  let conjunction inputs = normalized_and inputs

  (* Negate multi-input AIG node. *)
  let negation = function
    | And inputs  -> Nand inputs
    | Nand inputs -> And inputs

  (* Inputs to a multi-input AIG node. *)
  let inputs = function
    | And inputs | Nand inputs -> inputs

  (* Is a multi-input AIG node positive? *)
  let is_positive = function
    | And _  -> true
    | Nand _ -> false
end
      
(****************************************************************************)

class make () = 
  let core                  = Dpll_core.new_solver () in
    (* Underlying DPLL solver. *)
  let true_var              = Dpll_core.fresh_variable core in
    (* DPLL variable to stand for true. *)
  let true_lit              = Dpll_literal.literal true_var in 
    (* DPLL literal standing for true. *)
  let false_lit             = Dpll_literal.negation true_lit in 
    (* DPLL literal standing for false. *)
object (self)
  val var_definition        = Variable_definition.create 0
    (* Mapping from lists of literals to variables defined to be
       equisatisfiable with their conjunction. *)

  val var_translation       = Variable_translation.create 0
    (* Bijection between DPLL and term variables. *)

  val var_negation          = Variable_negation.create 0
    (* Bijection between DPLL variables in a negation relationship. *)

  val mutable variables     = Term_set.empty
    (* Variables occurring in terms handled by this solver. *)

  val recurring_term        = Term_hashtbl.create 0
    (* Is a term a recurring subterm? *)

  val elementary_form_cache = Term_hashtbl.create 0
    (* Cache of and-inverter graph elementary forms for terms. *)
    
  initializer
    Dpll_core.add_clause core [true_lit];
    assert self#invariant;
    assert (self#decision_level = 0);
    assert (not self#conflict_detected);
    assert (Term_set.is_empty self#variables)

  method private invariant =
    (* Truth values understood consistently in the underlying DPLL solver. *)
    Dpll_core.is_true core true_lit
      (* The range of var_translation is the set variables. *)
      (* should check cardinality constraint, but stdlib cardinality is slow *)
    && (Term_set.is_empty variables = 
	(Variable_translation.length var_translation = 0))
      (* Variable negations - which produce 2 var_negation entries -
	 are defined only for proxy variables. *)
    && (Variable_negation.length var_negation <= 
	  2*(Variable_translation.length var_translation))
      
  (****************************************************************************)
  (* Expressions are reduced to an "elementary-form" as a multi-input
     and-inverter graph.
     
     The elementary form is semi-canonical:
     
     - The literal inputs to an mult-input AIG node are sorted.
     
     - When we could represent "NOT x" either as "Nand [x]", or "And
     [~x]", we choose the former.
     
     - When we could represent "x" as wither "And [x]" or "Nand [~x]",
     we choose the former.

     - Occurrences of true_lit, false_lit, or a literal and its
     negation in the inputs to a multi-input AIG node are
     simplified. *)
      
  (* Elementary form multi-input AIG node for and.

     We normalize the inputs by sorting and simplifying. *)
  method private elementary_and lits =  
    let rec sorted_normal acc = function
      | (l0::(l1::_ as tail)) ->
	  assert (Dpll_literal.Literal.compare l0 l1 <= 0);
	  if 
	    Dpll_literal.Literal.equal l0 l1 
	    || Dpll_literal.Literal.equal l0 true_lit
	  then
	    sorted_normal acc tail
	  else if 
	    Dpll_literal.Literal.equal l0 (Dpll_literal.negation l1) 
	    || Dpll_literal.Literal.equal l0 false_lit 
	  then
	    [false_lit]
	  else
	    sorted_normal (l0::acc) tail
      | [l] -> l::acc
      | [] -> assert false
    in
      (* Literal.compare ensures that, if present, a literal and its
	 negation are adjacent when sorted. *)
      Multi_aig.conjunction 
        (sorted_normal [] 
           (List.sort Dpll_literal.Literal.compare (Bunch.to_list lits)))

  (* Elementary form for a mult-input AIG node for nand. *)
  method private elementary_nand lits =
    Multi_aig.negation (self#elementary_and lits)
      
  (* Elementary form for a multi-input AIG node for or. *)
  method private elementary_or lits = 
    Multi_aig.negation 
      (self#elementary_and (Bunch.map Dpll_literal.negation lits))
      
  (* Elementary form for a multi-input AIG node for nor. *)
  method private elementary_nor lits = 
    self#elementary_and (Bunch.map Dpll_literal.negation lits)

  (****************************************************************************)
  (* Input terms are traversed to find recurring subterms.  Information
     is stored in the table "recurring_term".  If a term does not have an
     entry in the table then it has yet to be traversed.  If the table
     maps the term to false, then no other occurrences of the term are
     known.  If the table maps the term to true, then it is a recurring
     subterm. *)

  (* Have we searched this term for subterm recurrences? *)
  method private recurrence_analysed term =
    Term_hashtbl.mem recurring_term term

  (* Search a term for subterm recurrence. *)
  method private find_recurrence term = 
    if Term_hashtbl.mem recurring_term term then
      Term_hashtbl.replace recurring_term term true
    else begin
      Term_hashtbl.add recurring_term term false;
      match term with
	| Application (_, f, x) ->
            self#find_recurrence f;
            self#find_recurrence x;
	| _ -> ()
    end;
    assert (self#recurrence_analysed term)
      
  (* Is this a (previously examined) recurring subterm? *)
  method private is_recurring_term term =
    assert (self#recurrence_analysed term);
    Term_hashtbl.find recurring_term term

  (****************************************************************************)

  (* Make the term variable term_var known to the term-level DPLL
     solver by defining it as equivalent to the DPLL variable var. *)
  method private add_translation dpll_var term_var =
    assert (is_variable term_var);
    assert (not (Term_set.mem term_var variables));
    variables <- Term_set.add term_var variables;
    Variable_translation.add var_translation dpll_var term_var

  (* A term variable (defined) to represent a given DPLL variable. *)
  method private term_var_translation dpll_var = 
    try
      Variable_translation.find var_translation dpll_var 
    with Not_found -> 
      (* No previously recorded translation, define one. *)
      let term_var = variable (Variable.fresh ()) in
	self#add_translation dpll_var term_var;
	term_var
	  
  (* A DPLL variable (defined) to represent a given term variable. *)
  method private dpll_var_translation term_var =
    assert (is_variable term_var);
    try
      Variable_translation.find_inv var_translation term_var
    with Not_found ->
      (* No previously recorded translation, define one. *)
      let dpll_var = Dpll_core.fresh_variable core in
 	self#add_translation dpll_var term_var;
 	dpll_var

  (* Term literal known to correspond to a DPLL literal. *)
  method private term_lit_of dpll_lit =
    let term_var = self#term_var_translation (Dpll_literal.variable dpll_lit) in
    let term_lit = Term_literal.literal term_var in
      if Dpll_literal.is_positive dpll_lit then
	term_lit
      else
	Term_literal.negation term_lit
	  
  (* The DPLL literal known to correspond to a term literal. *)
  method private dpll_lit_of term_lit =
    assert (self#is_interface_literal term_lit);
    let term_var = Term_literal.variable term_lit in
    let dpll_var = Variable_translation.find_inv var_translation term_var in
    let dpll_lit = Dpll_literal.literal dpll_var in
      if Term_literal.is_positive term_lit then
	dpll_lit
      else
	Dpll_literal.negation dpll_lit
	  
  (* The DPLL clause known to correspond to a term clause. *)
  method private dpll_clause_of term_clause =
    List.map self#dpll_lit_of term_clause

  (****************************************************************************)

  (* DPLL literal (defined to be) equisatisfiable to an elementary
     form. *)
  method private literal_for_elementary aig = 
    (* Literal (defined to be) equisatisfiable to a conjunction. *)
    let defined_literal = function
      | [l] -> l
      | ls ->
	  try
	    Dpll_literal.literal (Variable_definition.find var_definition ls)
	  with Not_found ->
	    let var = Dpll_core.fresh_variable core in
	    let lit = Dpll_literal.literal var in
	    let not_lit = Dpll_literal.negation lit in
	      (* . (v = (l0 && l1 && ... ))
		 = (v ==> (l0 && l1 && ...)) && 
		 . ((l0 && l1 && ...) ==> v)
		 = (not v || (l0 && l1 && ...)) && 
		 . (not (l0 && l1 && ...) || v)
		 = (not v || l0) && (not v || l1) && ...
		 . (not l0 || not l1 || ... || v)        *)
	      List.iter (fun l -> Dpll_core.add_clause core [not_lit; l]) ls;
	      Dpll_core.add_clause core 
		(lit::(List.map Dpll_literal.negation ls));
	      Variable_definition.add var_definition ls var;
	      lit
    in
      if Multi_aig.is_positive aig then
	defined_literal (Multi_aig.inputs aig)
      else
	Dpll_literal.negation (defined_literal (Multi_aig.inputs aig))
	  
  (* DPLL literal equisatisfiable to a multi-input and of DPLL literals. *)
  method private literal_and lits = 
    self#literal_for_elementary (self#elementary_and lits)

  (* DPLL literal equisatisfiable to a multi-input nand of DPLL literals. *)
  method private literal_nand lits = 
    self#literal_for_elementary (self#elementary_nand lits)

  (* DPLL literal equisatisfiable to a multi-input or of DPLL literals. *)
  method private literal_or lits =
    self#literal_for_elementary (self#elementary_or lits)

  (* DPLL literal equisatisfiable to a multi-input nor of DPLL literals. *)
  method private literal_nor lits =
    self#literal_for_elementary (self#elementary_nor lits)

  (* DPLL literals whose disjunction is equisatisfiable to a term
     disjunction. *)
  method private or_disjuncts x y =
    (* . (x0 || x1 || ...) || (y0 || y1 || ...)
       = x0 || x1 || ... || y0 || y1 || ...     *)
    let dx = self#disjuncts x in
    let dy = self#disjuncts y in
      Bunch.join dx dy

  (* DPLL literals whose conjunctions is equisatisfiable to a term
     conjunction. *)
  method private and_conjuncts x y =
    (* . (x0 && x1 && ...) && (y0 && y1 && ...)
       = x0 && x1 && ... && y0 && y1 && ...     *)
    let cx = self#conjuncts x in
    let cy = self#conjuncts y in
      Bunch.join cx cy

  (* DPLL literals whose disjunction is equisatisfiable to a term negation. *)
  method private not_disjuncts x =
    Bunch.map Dpll_literal.negation (self#conjuncts x)

  (* DPLL literals whose conjunction is equisatisfiable to a term negation. *)
  method private not_conjuncts x =
    Bunch.map Dpll_literal.negation (self#disjuncts x)

  (* DPLL literals whose disjunction is equisatisfiable to a term
     implication. *)
  method private implication_disjuncts x y =
    (* . ((x0 && x1 && ...) ==> (y0 || y1 || ...))
       = not (x0 && x1 && ...) || (y0 || y1 || ...) 
       = (not x0 || not x1 || ...) || (y0 || y1 || ...)
       = not x0 || not x1 || ... || y0 || y1 || ...     *)
    let cx = self#conjuncts x in
    let dy = self#disjuncts y in
      Bunch.join (Bunch.map Dpll_literal.negation cx) dy

  (* DPLL literals whose disjunction is equisatisfiable to a term
     equivalence. *)
  method private iff_disjuncts x y =
    (* . (x = y) 
       = (x && y) || (not x && not y)
       = (x && y) || not (x || y)     *)
    Bunch.couple
      (self#literal_and (self#and_conjuncts x y))
      (self#literal_nor (self#or_disjuncts x y))

  (* DPLL literals whose conjunction is equisatisfiable to a term
     equivalence. *)
  method private iff_conjuncts x y =
    (* . (x = y) 
       = (x ==> y) && (y ==> x) *)
    Bunch.couple
      (self#literal_or (self#implication_disjuncts x y))
      (self#literal_or (self#implication_disjuncts y x))

  (* DPLL literals whose disjunction is equisatisfiable to a term
     difference. *)
  method private xor_disjuncts x y =
    (* . (x != y)
       = not (x = y)
       = not ((x ==> y) && (y ==> x))
       = not (x ==> y) || not (y ==> x) *)
    Bunch.couple
      (self#literal_nor (self#implication_disjuncts x y))
      (self#literal_nor (self#implication_disjuncts y x))

  (* DPLL literals whose conjunction is equisatisfiable to a term
     difference. *)
  method private xor_conjuncts x y =
    (* . (x != y)
       = (x = not y)
       = (x ==> not y) && (not y ==> x)
       = (not x || not y) && (y || x)
       = not (x && y) && (y || x)       *)
    Bunch.couple
      (self#literal_nand (self#and_conjuncts x y))
      (self#literal_or (self#or_disjuncts y x))

  (* DPLL literals whose disjunction is equisatisfiable to a term
     conditional. *)
  method private conditional_disjuncts x y z =
    (* . (if x then y else z)
       = (x && y) || (not x && z)
       = (x && y) || not (x || not z)
       = (x && y) || not (z ==> x)    *)
    Bunch.couple
      (self#literal_and (self#and_conjuncts x y))
      (self#literal_nor (self#implication_disjuncts z x))

  (* DPLL literals whose conjunction is equisatisfiable to a term
     conditional. *)
  method private conditional_conjuncts x y z =
    (* . (if x then y else z)
       = (x ==> y) && (not x ==> z)
       = (x ==> y) && (x || z)      *) 
    Bunch.couple
      (self#literal_or (self#implication_disjuncts x y))
      (self#literal_or (self#or_disjuncts x z))

  (* DPLL literal equisatisfiable to a term. *)
  method private literal_for_term term =
    assert (self#recurrence_analysed term);
    self#literal_for_elementary (self#elementary_form term)
      
  (* DPLL literals whose conjunction is equisatisfiable to a term. *)
  method private conjuncts term =
    assert (self#recurrence_analysed term);
    match term with
      | _ when self#is_recurring_term term ->
	  Bunch.singleton (self#literal_for_term term)
      | Application (_, f, x)
          when term_equal f Term_constant.negation ->
	  self#not_conjuncts x
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.conjunction ->
	  self#and_conjuncts x y
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.equivalence ->
	  self#iff_conjuncts x y
      | Application (_, Application (_, f, x), y)
	  when term_equal f Term_constant.difference ->
	  self#xor_conjuncts x y
      | Application (_, Application (_, Application (_, f, x), y), z) 
          when term_equal f Term_constant.conditional ->
	  self#conditional_conjuncts x y z 
      | _ -> 
	  Bunch.singleton (self#literal_for_term term)

  (* DPLL literal whose disjunction is equisatisfiable to a term. *)
  method private disjuncts term =
    assert (self#recurrence_analysed term);
    match term with 
      | _ when self#is_recurring_term term ->
	  Bunch.singleton (self#literal_for_term term)
      | Application (_, f, x)
	  when term_equal f Term_constant.negation ->
	  self#not_disjuncts x
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.disjunction ->
	  self#or_disjuncts x y
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.implication ->
	  self#implication_disjuncts x y
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.equivalence ->
	  self#iff_disjuncts x y
      | Application (_, Application (_, f, x), y) 
          when term_equal f Term_constant.difference ->
	  self#xor_disjuncts x y
      | Application (_, Application (_, Application (_, f, x), y), z) 
          when term_equal f Term_constant.conditional ->
	  self#conditional_disjuncts x y z
      | _ -> 
	  Bunch.singleton (self#literal_for_term term)

  method private elementary_form term =
    assert (self#recurrence_analysed term);
    try
      Term_hashtbl.find elementary_form_cache term 
    with Not_found -> 
      let elementary = match term with 
	| _ when term_equal term Term_constant.truth ->
	    Multi_aig.singleton true_lit
	| _ when term_equal term Term_constant.falsity ->
	    Multi_aig.singleton false_lit
	| Variable _ ->
	    Multi_aig.singleton 
	      (Dpll_literal.literal (self#dpll_var_translation term))
	| Application (_, f, negated_term)
            when term_equal f Term_constant.negation ->
	    Multi_aig.negation (self#elementary_form negated_term)
	| Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.conjunction ->
	    self#elementary_and (self#and_conjuncts x y)
	| Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.disjunction ->
	    self#elementary_or (self#or_disjuncts x y)
	| Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.implication ->
	    self#elementary_or (self#implication_disjuncts x y)
	| Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.equivalence ->
	    self#elementary_and (self#iff_conjuncts x y) 
	| Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.difference ->
	    self#elementary_and (self#xor_conjuncts x y) 
	| Application (_, Application (_, Application (_, f, x), y), z) 
            when term_equal f Term_constant.conditional ->
	    self#elementary_and (self#conditional_conjuncts x y z)
	| _ -> assert false
      in
	Term_hashtbl.add elementary_form_cache term elementary;
	elementary

  (* A DPLL variable equisatisfiable to a DPLL literal.
     
     - straight forward if the literal is positive
     
     - if negative we find (or define) a variable equisatisfiable to
     the negation 

     This is needed only to produce proxy term variables for negative
     terms. *)
  method private var_for_literal lit =
    let var = Dpll_literal.variable lit in
      if Dpll_literal.is_positive lit then
	var
      else
	try
	  Variable_negation.find var_negation var
	with Not_found ->
	  let var_lit = Dpll_literal.literal var in
	  let not_var_lit = Dpll_literal.negation var_lit in
	  let dual = Dpll_core.fresh_variable core in
	  let dual_lit = Dpll_literal.literal dual in
	  let not_dual_lit = Dpll_literal.negation dual_lit in
	    (* . (dual != var)
	       = (dual = not var)
	       = (dual ==> not var) && (not var ==> dual)
	       = (not dual || not var) && (dual || var)   *)
	    Dpll_core.add_clause core [not_dual_lit; not_var_lit];
	    Dpll_core.add_clause core [dual_lit; var_lit];
	    Variable_negation.add var_negation dual var;
	    Variable_negation.add var_negation var dual;
	    dual

  method proxy term =
    self#find_recurrence term;
    let dpll_lit = self#literal_for_term term in
    let dpll_var = self#var_for_literal dpll_lit in
    let term_var = self#term_var_translation dpll_var in
      assert self#invariant;
      assert (is_variable term_var);
      term_var

  method add_clause term_clause = 
    let lit_disjuncts = Function.compose Bunch.to_list self#disjuncts in
      List.iter self#find_recurrence term_clause; 
      Dpll_core.add_clause core 
	(List.flatten (List.map lit_disjuncts term_clause)); 
      assert self#invariant
      
  method owns_constant term = 
    assert (Term.is_constant term);
    Term_constant.is_boolean_constant term

  method variables = 
    variables

  method add_interface_variable term_var = 
    assert (is_variable term_var);
    assert (Term_set.mem term_var variables);
    Dpll_core.add_interface_variable core 
      (Variable_translation.find_inv var_translation term_var);
    assert self#invariant

  method decision_level = 
    let level = Dpll_core.decision_level core in
      assert (0 <= level);
      level

  method decision_possible =
    Dpll_core.decision_possible core

  method inference_possible =
    Dpll_core.inference_possible core

  method conflict_detected =
    Dpll_core.conflict_detected core

  method backjump_possible =
    Dpll_core.backjump_possible core

  method satisfied = 
    let result = Dpll_core.satisfied core in
      assert (result = (not (self#conflict_detected 
                             || self#decision_possible 
                             || self#inference_possible)));
      result

  method unsatisfiable = 
    let unsat = Dpll_core.unsatisfiable core in
      assert (not unsat || self#conflict_detected);
      unsat

  method is_interface_literal term_lit =
    assert (Term_literal.is_literal term_lit);
    Variable_translation.mem_inv var_translation 
      (Term_literal.variable term_lit)
	
  method is_true term_lit =
    assert (self#is_interface_literal term_lit);
    Dpll_core.is_true core (self#dpll_lit_of term_lit)

  method private is_false term_lit =
    assert (self#is_interface_literal term_lit);
    Dpll_core.is_true core (Dpll_literal.negation (self#dpll_lit_of term_lit))

  method decide = 
    assert (not self#conflict_detected);
    assert (not self#inference_possible);
    assert self#decision_possible;
    Dpll_core.decide core;
    assert self#invariant;
    assert (self#decision_level > 0);
    assert (not self#conflict_detected)

  method infer = 
    assert (not self#conflict_detected);
    assert (self#inference_possible);
    Dpll_core.infer core;
    assert self#invariant

  method set_literal term_lit =
    assert (not self#conflict_detected);
    assert (self#is_interface_literal term_lit);
    Dpll_core.set_literal core (self#dpll_lit_of term_lit);
    assert self#invariant;
    (* Note: We ignore literals that conflict with DPLL assignments.
       It is left to the external agent to detect the conflict (when
       it learns of the DPLL assignment through get_literal). *)
    assert (not self#conflict_detected)

  method get_literal_possible =
    Dpll_core.get_literal_possible core
      
  method get_literal = 
    assert self#get_literal_possible;
    let term_lit = self#term_lit_of (Dpll_core.get_literal core) in
      assert self#invariant;
      assert (self#is_true term_lit);
      term_lit

  method set_conflict_clause term_clause = 
    assert (not self#conflict_detected);
    assert (List.for_all self#is_false term_clause);
    Dpll_core.set_conflict_clause core (self#dpll_clause_of term_clause);
    assert self#invariant;
    assert self#conflict_detected

  method literal_to_explain = 
    assert self#conflict_detected;
    assert (not self#unsatisfiable);
    assert (not self#backjump_possible);
    let term_lit = self#term_lit_of (Dpll_core.literal_to_explain core) in
      assert (self#is_true term_lit);
      term_lit

  method set_literal_explanation term_clause = 
    assert self#conflict_detected;
    assert (not self#unsatisfiable);
    assert (not self#backjump_possible);
    assert (List.mem self#literal_to_explain term_clause);
    assert
      (List.for_all (fun l -> l = self#literal_to_explain || self#is_false l) 
	 term_clause);
    Dpll_core.set_literal_explanation core (self#dpll_clause_of term_clause);
    assert self#invariant;
    assert self#conflict_detected

  method backjump = 
    assert self#backjump_possible;
    let start_level = Dpll_core.decision_level core in
      Dpll_core.backjump core;    
      assert (Dpll_core.decision_level core < start_level);
      assert self#invariant;
      assert (not self#conflict_detected);
      assert (self#inference_possible)

  method forget_some_learned_clauses = 
    assert (not self#conflict_detected);
    Dpll_core.forget_some_learned_clauses core;
    assert self#invariant;
    assert (not self#conflict_detected)

  method restart = 
    assert (not self#conflict_detected);
    Dpll_core.restart core;
    assert self#invariant;
    assert (self#decision_level = 0);
    assert (not self#conflict_detected)

  method num_vars = 
    Dpll_core.num_vars core

  method num_clauses = 
    Dpll_core.num_clauses core

  method num_learneds = 
    Dpll_core.num_learneds core

  method num_conflicts = 
    Dpll_core.num_conflicts core

  method num_decisions = 
    Dpll_core.num_decisions core

  method num_inferences = 
    Dpll_core.num_inferences core

  method num_externally_set = 
    Dpll_core.num_externally_set core
end
