(* $Id: dpll_literal.mli 1788 2007-09-28 18:23:28Z jgrundy $ *)
(** Literals and variables. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** Variables are represented by unique identifiers.

    Treat this as if Id.t was abstract, but if we make it so then the
    compiler generates slower code. *)
module Variable: Id.Generator with type t = int

(** Maps keyed off variables. *)
module Variable_map: Index.Map with type key = Variable.t

(** Sets of variables. *)
module Variable_set: Index.Set with type elt = Variable.t

(** Priorities queues of variables. *)
module Variable_queue: Index.PriorityQueue with type elt = Variable.t

(** Literals are unique identifiers.

    The compare operation on literals has the property that no literal
    is considered between a literal and the negation of that literal.
    That is, if you sort a list of literals then literals and their
    negations (if present) will be adjacent.

    Treat this as if Id.t was abstract, but if we make it so then the
    compiler generates slower code. *)
module Literal: Id.T with type t = int

(** Maps keyed off literals. *)
module Literal_map: Index.Map with type key = Literal.t

(** Sets of literals. *)
module Literal_set: Index.Set with type elt = Literal.t

val variable: Literal.t -> Variable.t
  (** [variable l] is the variable in literal [l].

      - Require [Literal.is_valid l]
      - Ensure  [Variable.is_valid (variable l)] *)

val is_positive: Literal.t -> bool
  (** [is_positive l] is true when [l] is a literal positive.

      - Require [Literal.is_valid l] *)

val literal: Variable.t -> Literal.t
  (** [literal v] is the positive literal corresponding to variable [v]. 

      - Require [Variable.is_valid v]
      - Ensure  [Literal.is_valid (literal v)] 
      - Ensure  {!is_positive}[ (literal v)]
      - Ensure  {!variable}[ (literal v) = v] *)
  
val negation: Literal.t -> Literal.t
  (** [negation l] is the negation of literal [l].

      - Require [Literal.is_valid l]
      - Ensure  [Literal.is_valid (negation l)] 
      - Ensure  {!is_positive}[ (negation l) = not (]{!is_positive}[ l)]
      - Ensure  {!variable}[ (negation l) = ]{!variable}[ l] *)

