(* $Id: solver_api.mli 1800 2007-09-29 00:17:21Z jgrundy $ *)
(** Term level interface specifications. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** General term level interfaces to DPLL and theory solvers for use
    with combination schemes.  The interfaces are essentially in the
    style of the following paper

    - Sava Krstic, Amit Goel.
    {{:http://dx.doi.org/10.1007/978-3-540-74621-8_1} Architecting
    Solvers for SAT Modulo Theories: Nelson-Oppen with DPLL}.  {i
    Lecture Notes in Computer Science} 4720:1-27, 2007. *)


open Term

(** Term level interface to a DPLL solver. A newly created DPLL solver
    has the following initial properties:

    - {!decision_level}[ = 0]
    - [not ]{!conflict_detected} 
    - [Term_set.is_empty ]{!variables} *)
class type dpll = 
object
  (** {2 Problem Setup and Solution} *)

  method owns_constant: term -> bool
    (** [owns_constant c] holds when [c] is a boolean constant or
	connective.

	- Require {!Term.is_constant}[ c] *)

  method add_clause: term list -> unit
    (** Adds the clause (disjunction of pure Boolean terms) to the
	solver. *)

  method proxy: term -> term
    (** A proxy for a pure boolean term. *)

  method add_interface_variable: term -> unit
    (** [add_interface_variables v] treats the variable [v] and its
	negation as an interface literal.

	- Require {!Term.is_variable}[ v]
	- Require {!Set.S.mem}[ v ]{!variables} *)

  (** {2 Solver Status} *)

  method variables: Term_set.t
    (** Set of all variables known to this solver. 

	- Ensure {!Set.S.for_all}[ ]{!Term.is_variable}[ variables] *)

  method decision_level: int
    (** Decision level of the solver.

	- Ensure [0 <= decision_level] *)

  method inference_possible: bool
    (** Might an inference be made? *)
    
  method get_literal_possible: bool
    (** Is there an interface literal in the current partial model
	that has not been dispatched by {!get_literal} since it was
	assigned? *)

  method conflict_detected: bool
  (** Has the solver deduced that the current (partial) model cannot
      be (extended to) a satisfying variable assignment? *)

  method decision_possible: bool
    (** Might further arbitrary variable assignment decisions be
	attempted? *)

  method backjump_possible: bool
    (** Might a backjump be used to resolve a conflict?

        - Ensure {!conflict_detected}[ || not ]
        {!backjump_possible}
        - Ensure [not (]{!backjump_possible}[ && ]
        {!unsatisfiable}[)] *)
    
  method satisfied: bool
    (** Has the solver deduced that the current model is a satisfying
	assignment for every variable?
	
	- Ensure 
	[satisfied = not (]
	{!conflict_detected}[ || ]{!decision_possible}[ || ]
        {!inference_possible}[)] *)

  method unsatisfiable: bool
    (** Has the solver deduced that the problem is unsatisfiable?

	- Ensure [not unsatisfiable || ]{!conflict_detected} *)

  (** {2 Literal Status} *)

  method is_interface_literal: term -> bool
    (** Is the literal [l] an interface literal for the solver.

	- Require {!Term_literal.is_literal}[ l] 
	- Ensure 
	[is_interface_literal l = ]
	{!Set.S.mem}[ (]{!Term_literal.variable}[ l) ]{!variables} *)

  method is_true: term -> bool
    (** Is the literal [l] assigned true in the solver state? 
	
	- Require {!is_interface_literal}[ l] *)

  (** {2 Search and Inference} *)

  method decide: unit
    (** Makes an arbitrary assignment to an unassigned variable. 

	- Require [not ]{!conflict_detected} 
	- Require [not ]{!inference_possible}
	- Require {!decision_possible}
	- Ensure {!decision_level}[ > 0]
	- Ensure [not ]{!conflict_detected} *)

  method infer: unit
    (** Attempt to infer a unit propagation or a conflict.
	
	- Require [not ]{!conflict_detected} 
	- Require {!inference_possible} *)

  method set_literal: term -> unit
    (** [set_literal l] attempts to make a variable assignment
	inferred by an external agent.  The interface literal is
	ignored if the corresponding variable is already assigned in
	the current partial model.

	- Require [not ]{!conflict_detected} 
	- Require {!is_interface_literal}[ l] *)

  method get_literal: term
    (** Choose an interface literal assigned true in the current
	partial model that has not been dispatched since it was
	assigned.

	- Require {!get_literal_possible}
	- Ensure {!is_true}[ get_literal] *)

  (** {2 Conflict Resolution} *)

  method set_conflict_clause: term list -> unit
    (** [set_conflict_clause ls] initializes the conflict clause with
	the disjoined literals.
	
	- Require [not ]{!conflict_detected}
	- Require 
	[List.for_all (fun l -> ]{!is_true}[ (]{!Term_literal.negation}[ l)) ls]
	- Ensure {!conflict_detected} *)

  method literal_to_explain: term
    (** Returns the next conflict literal to explain for conflict
	resolution.
	
	- Require {!conflict_detected}
	- Require [not ]{!unsatisfiable}
        - Require [not ]{!backjump_possible}
	- Ensure {!is_true}[ literal_to_explain]  *)

  method set_literal_explanation: term list -> unit
    (** [set_literal_explanation ls] removes {!literal_to_explain}
	from the set of conflict literals using the provided
	explanation [ls].

	- Require {!conflict_detected}
	- Require [not ]{!unsatisfiable}
        - Require [not ]{!backjump_possible}
	- Require {!List.mem}[ ]{!literal_to_explain}[ ls]
	- Require 
	{!List.for_all} 
	[(fun l -> l = 
	]{!literal_to_explain}[ 
	|| ]{!is_true}[ (]{!Term_literal.negation}[ l)) ls]
	- Ensure {!conflict_detected}
        - Ensure [not ]{!unsatisfiable} *)

  method backjump: unit
    (** Resolve a conflict by backjumping to undo conflicting
	assignments from the current partial model. A learned clause is
	added to the solver to prevent the same conflict later.

        - Require {!backjump_possible}
	- Ensure [not ]{!conflict_detected} *)

  (** {2 Search Optimization} *)

  method forget_some_learned_clauses: unit
    (** Forget some clauses learned by the solver though conflict
	analysis.
	
	- Require [not ]{!conflict_detected} 
	- Ensure [not ]{!conflict_detected} *)

  method restart: unit
    (** Undo all arbitrary variable assignment decisions.

	- Require [not ]{!conflict_detected}
	- Ensure [not ]{!conflict_detected} 
	- Ensure {!decision_level}[ = 0] *)

  (** {2 Usage Statistics} *)

  method num_vars: int
    (** Number of variables created. *)

  method num_clauses: int
    (** Number of added clauses that are tracked in current state. *)

  method num_learneds: int
    (** Number of clauses learned through conflict analysis in current
	state. *)

  method num_conflicts: int
    (** Number of conflicts encountered. *)

  method num_decisions: int
    (** Number of arbitrary variable assignment ever tried. *)

  method num_inferences: int
    (** Number of variable assignment ever made by unit propagation. *)

  method num_externally_set: int
    (** Number of variable assignments ever made by {!set_literal}. *)

end

(** Term level interface to all theory solvers. A newly created theory
    solver has the following initial properties:

    - {!decision_level}[ = 0]
    - [not ]{!conflict_detected}
    - [Term_set.is_empty ]{!variables} *)
class type theory =
object
  (** {2 Problem Setup and Solution} *)

  method owns_constant: term -> bool
    (** [owns_constant t] holds when the solver responsible for
	reasoning about the constant [t].

	- Require {!Term.is_constant}[ t] *)

  method proxy: term -> term
    (** A proxy for a pure term. *)

  method add_interface_variable: term -> unit
    (** [add_interface_variable v] treats the variable [v] and its
	negation as an interface literal.

	- Require {!Term.is_variable}[ v]
	- Require {!Set.S.mem}[ v ]{!variables} *)


  (** {2 Solver Status} *)

  method variables: Term_set.t
    (** Set of all variables known to this solver. 

	- Ensure {!Set.S.for_all}[ ]{!Term.is_variable}[ variables] *)

  method decision_level: int
    (** Decision level of the solver.

	- Ensure [decision_level >= 0] *)

  method inference_possible: bool
    (** Might an inference be made? *)

  method get_literal_possible: bool
    (** Is there a literal that has been inferred, but not dispatched
	by {!get_literal}? *)

  method conflict_detected: bool
    (** Has a conflict been detected in the current solver state? *)


  (** {2 Literal Status} *)

  method is_interface_literal: term -> bool
    (** Is the literal [l] an interface literal for the solver.

	- Require {!Term_literal.is_literal}[ l] 
	- Ensure 
	[is_interface_literal l = ]
	{!Set.S.mem}[ (]{!Term_literal.variable}[ l) ]{!variables} *)

  method is_true: term -> bool
    (** Is the literal [l] assigned true in the solver state? 

	- Require {!is_interface_literal}[ l] *)


  (** {2 Inference} *)

  method infer: unit
    (** Attempt to make inferences from currently assigned literals.

	- Require [not ]{!conflict_detected} 
	- Require {!inference_possible} *)

  method set_literal: term -> unit
    (** [set_literal l] attempts to make a variable assignment
	inferred by an external agent.

	- Require {!is_interface_literal}[ l]
	- Require [not ]{!conflict_detected} *)

  method get_literal: term
    (** Choose an inferred literal that has not been dispatched
	before.
	
	- Require {!get_literal_possible}
	- Ensure {!is_true}[ get_literal] *)

  method increment_level: unit
    (** Increment decision_level.

	- Require [not ]{!conflict_detected} 
        - Require [not ]{!inference_possible}
        - Require [not ]{!get_literal_possible}
	- Ensure [not ]{!conflict_detected} 
        - Ensure [not ]{!inference_possible}
        - Ensure [not ]{!get_literal_possible}
	- Ensure {!decision_level}[ > 0] *)

  (** {2 Conflict Resolution} *)

  method backjump: int -> unit
    (** Backjumps to the given level.

	For [backjump level]:
	- Require [0 <= level]
	- Require [level < ]{!decision_level}
	- Ensure [level = ]{!decision_level}
	- Ensure [(not ]{!conflict_detected}[)] *)

  method restart: unit
    (** Backjumps to level 0 if not already there.

        - Require [not ]{!conflict_detected}
	- Ensure {!decision_level}[ = 0] 
        - Ensure [not ]{!conflict_detected} *)

  method conflict_clause: term list
    (** An explanation for a conflict. The explanation clause consists
	of interface literals that have been set in the solver.

	- Require {!conflict_detected}
	- Ensure 
	{!List.for_all} 
	[(fun l -> ]{!is_true}[ (]{!Term_literal.negation}[ l))
	conflict_clause] 
	- Ensure [conflict_clause != []] *)

  method literal_explanation: term -> term list
    (** An explanation clause for a literal. The explanation clause
	consists of literals that were given to {!set_literal}.

	[let cls = literal_explanation l]
	- Ensure
	{!List.for_all}[ ]{!is_interface_literal}[ cls]
	- Ensure {!List.mem}[ l cls]
	- Ensure 
	{!List.for_all} 
	[(fun m -> ]{!is_true}[ (]{!Term_literal.negation}[ m) 
	|| ]{!Term.term_equal}[ l m) 
	cls] *)

end

(** The interface for a solver. *)
class type solver =
object
  method add_clause: term list -> unit
    (** Adds the clause (disjunction of pure Boolean terms) to the
	solver. *)

  method solve: bool
    (** Determines the satisfiability of the conjunction of clauses
        added to the solver. *)
end
