(* $Id: term.ml 1793 2007-09-28 19:12:36Z jgrundy $ *)
(* The structure of of terms. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

type key = int
type db_index = int

type term =
  | Bound       of db_index       
  | Variable    of key * Variable.t
  | Constant    of key * Constant.t
  | Application of key * term * term
  | Abstraction of key * Variable.t * term
      
(*****************************************************************************)
(* Term data structures. *)

(* Equality for hash-consed terms. *)
let term_equal = (==)

(* Structural comparison for hash-consed terms. *)
let rec term_compare t1 t2 =
  if term_equal t1 t2 then
    0
  else
    let comparison = match t1, t2 with
      | Bound i1,         Bound i2         -> Pervasives.compare i1 i2
      | Variable (_, s1), Variable (_, s2) -> Variable.compare s1 s2
      | Constant (_, c1), Constant (_, c2) -> Constant.compare c1 c2
      | Application (h1, f1, a1), Application (h2, f2, a2) ->
          let hash_comparison = Pervasives.compare h1 h2 in
            if hash_comparison != 0 then
              hash_comparison
            else
              let fn_comparison = term_compare f1 f2 in
                if fn_comparison != 0 then 
                  fn_comparison
                else 
                  term_compare a1 a2
      | Abstraction (h1, _, b1), Abstraction (h2, _, b2) ->
          let hash_comparison = Pervasives.compare h1 h2 in
            if hash_comparison != 0 then 
              hash_comparison
            else 
              term_compare b1 b2
      | _, _ -> Pervasives.compare t1 t2
    in
      (* Pointer disequal hash-consed terms can't be structurally equal. *)
      assert (comparison != 0);
      comparison

(* Extract the previously computed hash-key for the term. *)
let term_hash = function
  | Bound (h) 
  | Variable (h,_) 
  | Constant (h,_) 
  | Application (h,_,_)
  | Abstraction (h,_,_) -> h

module Term_compared = struct
  type t = term
  let equal = term_equal
  let hash = term_hash
  let compare = term_compare
end

module Term_hashtbl = Hashtbl.Make (Term_compared)
module Term_set     = Set.Make     (Term_compared)
module Term_map     = Map.Make     (Term_compared)

(*****************************************************************************)
(* Hash-consing. *)

(* Note: We would like to use a weak-table for hash-consing, but the
   performance of weak tables is much worse than hashtbls. *)
module Term_pool = Hashtbl.Make 
  (struct
     type t = term

     (* Note: We are assuming here that the subterms are hash-consed. *)
     (* Check alpha-equivalence of terms. *)
     let equal t1 t2 =  
       match t1, t2 with
         | Bound i,         Bound j         -> i=j
         | Variable (_, n), Variable (_, m) -> n = m
	 | Constant (_, c), Constant (_, d) -> Constant.equal c d
         | Application (_, f, x), Application (_, g, y) -> 
             term_equal f g && term_equal x y
         | Abstraction (_, _, x), Abstraction (_, _, y) -> 
             term_equal x y
         | _ -> false
             
     let hash = term_hash
   end)

(* A hash-table of terms for hash-consing. *)
let pool = Term_pool.create 0

let hash_cons tm =
  try Term_pool.find pool tm
  with Not_found -> 
    Term_pool.add pool tm tm;
    tm

(*****************************************************************************)
(* Hash key computation for a term *)

let prime = 19
let variable_hash var      = 1 + prime * Variable.hash var
let constant_hash const    = 2 + prime * Constant.hash const
let abstraction_hash body  = 3 + prime * term_hash body
let application_hash f x   = 4 + prime * (term_hash f + prime * term_hash x)

(*****************************************************************************)
(* Internal constructors -- compute hash key, and hash-cons. *)

let make_bound i =
  hash_cons (Bound i)

let make_variable var  =
  hash_cons (Variable (variable_hash var, var))

let make_constant const = 
  hash_cons (Constant (constant_hash const, const))

let make_abstraction tm var = 
  hash_cons (Abstraction (abstraction_hash tm, var, tm))

let make_application f x =
  hash_cons (Application (application_hash f x, f, x))

(*****************************************************************************)
(* External constructors and accessors *)

(*****************************************************************************)
(* Variables *)

let is_variable = function
  | Variable _ -> true
  | _ -> false

let variable v = make_variable v

(*****************************************************************************)
(* Constants *)

let is_constant = function
  | Constant _ -> true
  | _ -> false

let constant = make_constant

let constant_value = function
  | Constant (_,c) -> c
  | _ -> assert false

let is_symbolic_constant = function
  | Constant (_,c) -> Constant.is_symbolic c
  | _ -> false

let symbolic_constant s = constant (Constant.symbolic s)

let symbolic_value = function
  | Constant (_,c) -> Constant.symbolic_value c
  | _ -> assert false

let is_integer_constant = function
  | Constant (_,c) -> Constant.is_integer c
  | _ -> false

let integer_constant i = constant (Constant.integer i)

let integer_value = function
  | Constant (_,c) -> Constant.integer_value c
  | _ -> assert false

let is_rational_constant = function
  | Constant (_,c) -> Constant.is_rational c
  | _ -> false

let rational_constant r = constant (Constant.rational r)

let rational_value = function
  | Constant (_,c) -> Constant.rational_value c
  | _ -> assert false

(*****************************************************************************)
(* Applications *)

let is_application = function
  | Application _ -> true
  | _ -> false

let application = make_application

let application_function = function
  | Application (_, f, _) -> f
  | _ -> assert false

let application_argument = function
  | Application (_, _, x) -> x
  | _ -> assert false

let binary_application f x y = 
  application (application f x) y

let ternary_application f x y z = 
  application (application (application f x) y) z

let nary_application f args = 
  List.fold_left application f args

let nary_application_decomposition tm =
  let rec un_apply_rec t acc = match t with 
    | Application (_, f, x) -> un_apply_rec f (x::acc) 
    | _ -> (t, acc) 
  in 
    un_apply_rec tm [] 


(*****************************************************************************)
(* Abstractions *)

let is_abstraction = function 
  | Abstraction _ -> true
  | _             -> false

module Var_map = Map.Make (Variable)

(* A map the variables in list to their positions. *)
let index_map vars =
  let rec index_rec map i = function
    | [] -> map
    | (v::vs) -> index_rec (Var_map.add v i map) (i+1) vs
  in
    index_rec Var_map.empty 0 vars

(* Replace variables in indices map with de Bruin style bound variables. *)
let rec abstract_rec indices depth tm = match tm with
  | Variable (_,v) ->
      begin
	try 
	  let i = Var_map.find v indices in
	    make_bound (depth + i)
	with Not_found -> tm
      end
  | Abstraction (_, var, body) ->
      let body' = abstract_rec indices (depth+1) body in
	make_abstraction body' var
  | Application (_, f, x) ->
      let x' = abstract_rec indices depth x in
      let f' = abstract_rec indices depth f in
	application f' x'
  | _ -> tm

let abstract_rev vars tm = 
  if vars = [] then 
    tm
  else 
    let tm' = abstract_rec (index_map vars) 0 tm in
      List.fold_left make_abstraction tm' vars

let abstraction v tm =
  abstract_rev [v] tm

let nary_abstraction vs tm =
  abstract_rev (List.rev vs) tm

(*****************************************************************************)

let rec beta bvars args env tm =
  match tm with
    | Bound i -> 
	beta bvars args env (List.nth env i)
    | Constant _ | Variable _ -> 
	abstract_rev bvars (nary_application tm args)
    | Application (_, f, x) ->
	let x' = beta [] [] env x in
	  beta bvars (x'::args) env f
    | Abstraction (_, _, body) ->
	match args with 
	  | [] ->  
	      let v = Variable.fresh () in
		beta (v::bvars) [] ((variable v)::env) body
	  | (x::xs) ->
	      beta bvars xs (x::env) body

let beta_normal_form tm = beta [] [] [] tm

(*****************************************************************************)
(* Pretty printing. *)

open Format

let indent = 2

let rec print_list formatter print_elt print_sep  = function
  | [] -> ()
  | [u] -> print_elt formatter u
  | (h::t) -> 
      begin
	print_elt formatter h;
	print_sep formatter ();
	print_list formatter print_elt print_sep t
      end
	
let rec pp_print_rec bindings formatter = function
  | Bound (i) -> 
      Variable.pp_print formatter (List.nth bindings i)
  | Variable (_, v) -> 
      Variable.pp_print formatter v
  | Constant (_, c) -> 
      Constant.pp_print formatter c
  | Application _ as appl ->
      let (f,xs) = nary_application_decomposition appl in
      pp_print_string formatter "(";
      begin
        pp_open_hovbox formatter indent;
	print_list formatter (pp_print_rec bindings) pp_print_space (f::xs);
        pp_close_box   formatter ();
      end;
      pp_print_string formatter ")"
  | Abstraction (_, v, tm)  -> 
      let v' = Variable.variant bindings v in
	pp_print_string formatter "(";
	begin
          pp_open_hovbox formatter indent;
	  pp_print_string formatter "\\";
	  Variable.pp_print formatter v';
          pp_print_string formatter ".";
          pp_print_space formatter ();
	  begin
            pp_open_hovbox formatter indent;
            pp_print_rec (v'::bindings) formatter tm;
	    pp_close_box formatter ();
	  end;
	  pp_close_box formatter ();
	end;
	pp_print_string formatter ")"
	  
let pp_print_term = pp_print_rec []

let print_term = pp_print_term std_formatter

let log_term level tm = Log.log pp_print_term level tm

let rec string_of_list string_of_elt sep = function
  | [] -> ""
  | [u] -> string_of_elt u
  | (h::t) -> 
      (string_of_elt h) ^ sep ^ (string_of_list string_of_elt sep t)

let rec string_of bindings = function
  | Bound (i) -> 
      Variable.string_of (List.nth bindings i)
  | Variable (_, v) ->
      Variable.string_of v
  | Constant (_, c) ->
      Constant.string_of c
  | Application _ as appl ->
      let (f, xs) = nary_application_decomposition appl in
        "(" ^ (string_of_list (string_of bindings) " " (f::xs)) ^ ")"
  | Abstraction (_, v, tm) ->
      let v' = Variable.variant bindings v in
        "\\" ^ (Variable.string_of v') ^ "." ^ (string_of (v'::bindings) tm)

let string_of_term = string_of []

(*****************************************************************************)
(* Compact applicative pretty printing. *)

let rec count_sharing tbl tm =
  try 
    let (count,proxy) = Term_hashtbl.find tbl tm in
      Term_hashtbl.replace tbl tm (count+1,proxy)
  with Not_found -> 
    if is_application tm then
      let (_,xs) = nary_application_decomposition tm in
      let proxy = variable (Variable.fresh ()) in 
	Term_hashtbl.add tbl tm (1, proxy);
	List.iter (count_sharing tbl) xs

let rec share tbl tm =
  try
    let (count,p) = Term_hashtbl.find tbl tm in
      if count > 1 then
	p
      else
	begin
	  Term_hashtbl.remove tbl tm;
	  raise Not_found
	end
  with Not_found -> 
    if is_application tm then
      let (f,xs) = nary_application_decomposition tm in
      let xs' = List.map (share tbl) xs in
	nary_application f xs'
    else
      tm

let share_sub tbl tm =
  if is_application tm then
    let (f,xs) = nary_application_decomposition tm in
    let xs' = List.map (share tbl) xs in
      nary_application f xs'
  else
    tm

let print_let formatter tbl tm (count, proxy) =
  if count > 1 then
    begin
      pp_open_hovbox formatter indent;
      pp_print_string formatter "let";
      pp_print_space formatter ();
      begin
	pp_open_hovbox formatter indent;
	pp_print_term formatter proxy;
	pp_print_space formatter ();
	pp_print_string formatter "=";
	pp_print_space formatter ();
	pp_print_term formatter (share_sub tbl tm);
	pp_close_box formatter ();
      end;
      pp_print_space formatter ();
      pp_print_string formatter "in";
      pp_close_box formatter ();
      pp_print_cut formatter ();
    end

let pp_print_term formatter tm =
  let tbl = Term_hashtbl.create 0 in
    count_sharing tbl tm;
    let tm' = share_sub tbl tm in
      begin
	pp_open_vbox formatter 0;
	Term_hashtbl.iter (print_let formatter tbl) tbl;
	pp_print_term formatter tm';
	pp_close_box formatter ()
      end
  
let print_term = pp_print_term std_formatter

let log_term level tm = Log.log pp_print_term level tm
