(* $Id: term_logic.mli 1804 2007-09-30 14:50:25Z agoel3 $ *)
(** Term constructors. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Term

(** Signature for logical constructors. *)
module type Constructor = sig

  val negation:     term -> term
    (** Negation of a term. *)

  val conjunction:  term -> term -> term
    (** Conjunction of two terms. *)

  val disjunction:  term -> term -> term
    (** Disjunction of two terms. *)

  val equivalence:  term -> term -> term
    (** Logical equivalence of two (Boolean) terms. *)

  val difference:   term -> term -> term
    (** Logical dis-equivalence of two (Boolean) terms. *)

  val implication:  term -> term -> term
    (** Implication of two terms. *)

  val conditional:  term -> term -> term -> term 
    (** Conditional expression: Boolean condition, and Boolean alternatives. *)

  val equality:     term -> term -> term
    (** Equality of two non-Boolean terms. *)
    
  val ite:          term -> term -> term -> term
    (** Conditional expression: Boolean condition, non-Boolean alternatives. *)
    
end

(** Logical constructors with no syntactic simplifications. *)
module Standard: Constructor

(** Logical constructors with syntactic simplification based on
    Reduced Boolean Circuits.

    Reduced Boolean Circuits are defined here:

    Per Bjesse, Arne Boralv.
    {{:http://dx.doi.org/10.1109/ICCAD.2004.1382541} DAG-Aware Circuit
    Compression for Formal Verification}. ICCAD: 42-49, 2004. *)
module Rbc: Constructor


(** Signature for Term Rewriters. *)
module type Rewriter = sig

  val rewrite: term -> term
    (** A term transformation. *)

end

(** A term rewriter based on Reduced Boolean Circuits. *)
module Rbc_rewriter : Rewriter
