(* Copyright 2008  The Regents of the University of California.

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License.
*)

(* $Id: abs_focimc.ml,v 1.5 2010/02/22 22:29:34 rupak Exp $
 *
 * This file is part of the BLAST Project.
 *)

(*****************************************************************************) 
(*********** Transition Relation Refinement [CAV 05] for -focisymm ***********) 
(*****************************************************************************)
module O = Options
module M = Message
module S = Bstats
module E = Ast.Expression
module P = Ast.Predicate
module C_SD = BlastCSystemDescr.C_System_Descr
module Cmd = C_SD.Command
module PT = PredTable 
module TP = Abs_config.TP 
module ITP =  Abs_config.ITP

module R = Abs_region.Region
module Op = Abs_op.Operation
module U = Abs_util
module Ar = Abs_refine
open Abs_op.Defs

(* TBD : options *)
let oBool = O.getValueOfBool

(* global *)
let lvm = new Lvmap.lvmap 

(*************************************************************)
(************** Printing Abstract Counterexamples ************)
(*************************************************************)

let print_upd_tr (x,y) (x',y') p =
  let s = P.toString p in
  M.msg_string M.Debug (Printf.sprintf "upd_trel: (%d,%d)--->(%d,%d): %s" x y x' y' s)

let printop i op = 
  M.msg_string M.Debug (Printf.sprintf "%d : %s" i (Op.to_string op))

let print_counterexample opa reva rega ps =
  Misc.fassert (Array.length rega = Array.length opa + 1) "print_ctrex"; 
  M.msg_string M.Debug ("SAT preds: "^(Misc.strList ps));
  let pt = Hashtbl.create 101 in
  List.iter 
    (fun s -> 
      try
        if Misc.is_prefix PT.bp_prefix s then 
        let (pi,t) = Misc.map_pair int_of_string (Misc.last_two (Misc.chop s "_")) in
        Hashtbl.add pt t pi;
      with _ -> ()) ps;
  M.msg_string M.Normal "*********Abstract Counterexample**********";
  let (_,_,cubes) = 
    Array.fold_left
    (fun (i,j,cs) opi -> 
      let ps  = List.map PT.getPred (Hashtbl.find_all pt j) in
      let c   = Misc.strList (List.map P.toString ps) in
      let j'  = if i = reva.(j) then j+1 else j in
      M.msg_string M.Normal (Printf.sprintf "Time: %d : %s \n %s" i c (Op.to_string opi));
      (i+1,j',c::cs)) (0,0,[]) opa in
  let rv = List.rev ("True"::cubes) in
  Abs_html.htmlize_trace (Array.to_list opa) (List.tl (Array.to_list rega)) [] "Abstract Counterexample";
  rv

(*********************************************************************************)
(********************************** helpers **************************************)
(*********************************************************************************)

let make_op_id op l' = 
  Intra (fst (C_SD.get_edge_coords op),C_SD.location_coords l')

let make_boolean_state_var i = 
  let v = PT.bp_prefix^(string_of_int i) in
  let v' = PT.bp_next_prefix^(string_of_int i) in
  (P.Atom (E.Lval (E.Symbol v)),P.Atom (E.Lval (E.Symbol v')))

let copy_update_constraints is =
  P.conjoinL (List.map (fun i -> let (x,y) = make_boolean_state_var i in P.Iff (x,y)) is)
 
let is_pure_skip op e_id (is,is') =
  if is != is' then false else
  match (C_SD.get_command op).Cmd.code with Cmd.Pred e -> 
    let lvs = P.lvals_of_predicate (P.And [e; (U.pred_to_mccarthy e)]) in 
    not (List.exists PT.is_relevant_lval lvs)
  | _ -> false (* not (List.exists (PT.is_pred_affected e_id) is) *)

let process_inconsistent_ops =
  let tick = ref 0 in
  fun opa maska opids ->
    M.msg_string M.Normal (Printf.sprintf "[INF%d] \n\n\n" !tick);
    tick := !tick + 1;
    List.fold_left 
    (fun () i ->
      let opi = opa.(i) in
      M.msg_string M.Normal (Printf.sprintf "[INF%d] %s \n\n" !tick (Op.to_string opi));
      if C_SD.isPredicate opi then (Array.set maska i true)) () opids

(********************************************************************************)
(************** block constraints (with "mccarthy" / splitprover) ***************)
(********************************************************************************)
 
let block_cons_op =
  fun op l' op_id ->  
    let _ = lvm#clear () in
    let rv1 = Ar.block_wp lvm op l' in
    let rv2 = Oo.copy lvm in
    let _ = lvm#clear () in 
    (rv1,rv2)

let block_cons_in lvm op_id i =
  let (v,_) = make_boolean_state_var i in
  P.iff v (lvm#map_pred (U.pred_to_mccarthy (PT.getPred i)))

let block_cons_out asgns op_id i = 
  let p = U.pred_to_mccarthy (PT.getPred i) in
  let _ = if p <> Abs_image.wp p asgns then PT.update_pred_affected op_id i in
  P.iff (lvm#map_pred p) (snd (make_boolean_state_var i))

let block_p_update_cons op l' (is,is') =
  M.msg_string M.Debug ("block_p_update" );
  let op_id = make_op_id op l' in
  let (c,lvm') = block_cons_op op l' op_id in
  let chain = List.map (block_cons_in lvm' op_id) is in
  let (_,asgns) = U.make_guarded_asgn op l' in
  let wpc = List.map (block_cons_out asgns op_id) is' in
  (chain,wpc,[c])

(**********************************************************************************)
(****************************** regular constraints *******************************)
(**********************************************************************************)
 
let cons_in =
  let memo = Hashtbl.create 1009 in
  fun i ->
    Misc.memoize memo i
    (fun () -> P.iff (fst (make_boolean_state_var i)) (U.pred_to_mccarthy (PT.getPred i)))
    () 

let cons_out = 
  let memo = Hashtbl.create 1009 in
  fun asm asgns op_id i -> 
    Misc.memoize memo (op_id, i)
    (fun () ->
      let (v,v') = make_boolean_state_var i in
      let p = U.pred_to_mccarthy (PT.getPred i) in
      let wpa = Abs_image.wp p asgns in 
      let _ = if p <> wpa then PT.update_pred_affected op_id i in
      P.iff (P.Implies (U.pred_to_mccarthy asm , wpa)) v')
    () 

let p_update_cons op l' (is,is') =  
  M.msg_string M.Debug ("p_update_cons" );
  let op_id = make_op_id op l' in
  let chain = List.map cons_in is in
  let (asm,asgns) = U.make_guarded_asgn op l' in
  let wpc = List.map (cons_out asm asgns op_id) is' in
  (chain,wpc,[U.pred_to_mccarthy asm])

(*************************************************************)

let predicate_update_constraints op l' isp =
  let rv = 
    if is_pure_skip op (make_op_id op l') isp then P.True else   
    let (a,b,c) = 
      if O.getValueOfInt "focimax" = -1 
      then p_update_cons op l' isp 
      else block_p_update_cons op l' isp in
    P.conjoinL (a@b@c) in
  let _ = M.msg_string M.Debug 
  (Printf.sprintf "predicate_update_constraints: op = %s, rv = %s"
  (Op.to_string op) (P.toString rv)) in
  rv

(*************************************************************)
(***************** Conversion Functions **********************)
(*************************************************************)

let interpreted_rename_suffix s e = 
  match e with 
    E.FunctionCall (E.Lval (E.Symbol f),es) 
    when E.is_interpreted f && E.is_interpreted_renamable f ->
      (match Misc.flap E.lvals_of_expression es with _::_ -> (e,false) | _ -> 
        (E.FunctionCall (E.Lval(E.Symbol(f^s)),es),true))
  | E.Lval (E.Symbol x) when E.is_interpreted x && E.is_interpreted_renamable x -> 
        (E.Lval (E.Symbol (x^s)),true) 
  | _ -> (e,true)

let retime_pred i p =
  let new_name su su' l = 
    match l with 
      E.Symbol s when E.is_interpreted s -> E.Symbol s
    | E.Symbol s when Misc.is_prefix PT.bp_next_prefix s -> 
        E.Symbol (PT.bp_prefix^(Misc.chop_after_prefix PT.bp_next_prefix s)^su')
    | E.Symbol s -> E.Symbol (s^su) in 
  let i_s = "_"^(string_of_int i) in
  let p' = P.symbol_alpha_convert (new_name i_s ("_"^(string_of_int (i+1)))) p in
  P.atom_transform (E.deep_transform_top_cond (interpreted_rename_suffix i_s)) p'
  
let pred_to_bpred p =
  P.deep_transform 
  (fun p -> 
    match p with P.Atom e -> 
      let (i,b) = PT.getAtomIndex p in
      let p' = P.Atom (E.Lval (E.Symbol (PT.bp_prefix^(string_of_int i)))) in
      if b then p' else (P.negate p')
    | _ -> p) p

let bpred_to_pred j bp =
  P.deep_transform
  (fun p -> 
    match p with P.Atom (E.Lval (E.Symbol s))->
        assert (Misc.is_prefix PT.bp_prefix s);
        let n,i = 
          match (List.rev (Misc.chop s "_")) with
          | a::b:: _ -> (int_of_string a, int_of_string b)
          | _ -> assert false 
        in
        assert (n = j || n = j+1);
        if (n = j) then PT.getPred i else (P.Next (PT.getPred i))
    | P.Atom _ -> failwith "error:bpred_to_pred" (* QUARC *)
    | _ -> p) bp 

let bpred_to_bdd j p = 
  Misc.do_catch ("bpred_to_bdd:"^(P.toString p))
  (PT.convertPredToBdd_map 
  (fun p ->
    match p with P.Atom (E.Lval (E.Symbol s)) -> 
      assert (Misc.is_prefix PT.bp_prefix s);
      print_string "PRINTING S" ; print_string s ; print_newline();
      let n,i = 
        match (List.rev (Misc.chop s "_")) with
        | a::b:: _ -> (int_of_string a, int_of_string b)
        | _ -> assert false 
      in
      assert (n= j || n= j+1);
      if n = j then PT.getPredIndexBdd i else PT.getPredIndexNextBdd i)) p

      
(* WARNING: DNF *)  
let bdd_to_bpred b =
  PT.ac_convertBddToPred 
  (fun i -> 
    let s = if PT.isCurrentIndex i then PT.bp_prefix else PT.bp_next_prefix in
    P.Atom (E.Lval (E.Symbol (s^(string_of_int i))))) b 

(*************************************************************************)
(*****************  Building Constraints from Traces *********************)
(*************************************************************************)
  
let build_cons aps cpc rega cba ccba reva tps_o (i,j) opi = 
  let l = rega.(i).R.location in
  let l' = rega.(i+1).R.location in
  let (ps,ps') = 
    match tps_o with 
     Some a -> (Misc.sort_and_compact a.(i), Misc.sort_and_compact a.(i+1)) 
   | None when O.getValueOfInt "craig" = 2 -> 
       (PT.lookup_location_predicates (C_SD.location_coords l),
        PT.lookup_location_predicates (C_SD.location_coords l'))
   | _ -> (aps,aps) in
  let wpc = predicate_update_constraints opi l' (ps,ps') in
  if wpc = P.True then (i+1,j) else 
    let consi = retime_pred j wpc in
    let copyi = retime_pred j cpc in
    let _ = Array.set cba j consi in
    let _ = Array.set ccba j copyi in
    let _ = Array.set reva j i in
    (i+1,j+1)

let build_fmc_constraint_arrays rega opa tps_o =
  let n_ops = Array.length opa in 
  M.msg_string M.Debug ("build_fmc_cons: n_ops= "^(string_of_int n_ops));
  assert (Array.length rega = n_ops + 1);
  let cba = Array.create (n_ops+1) P.True in
  let ccba = Array.create (n_ops+1) P.True in
  let aps = PT.get_all_pred_indexes () in
  let cpc = copy_update_constraints aps in
  let reva = Array.create (n_ops+1) (-1) in  
  let (_,n) = Array.fold_left (build_cons aps cpc rega cba ccba reva tps_o) (0,0) opa in
  Array.set cba n (retime_pred n (pred_to_bpred(R.to_predicate(rega.(n_ops))))); 
  M.msg_string M.Debug "Non-skip constraints start";
  for i = 0 to n - 1 do 
    M.msg_string M.Debug (Printf.sprintf "%d: :%s" reva.(i) (Op.to_string opa.(reva.(i))))
  done;
  M.msg_string M.Debug "Non-skip constraints end";
  (cba,ccba,reva,n)

(*************************************************************************)
(*****************  Strengthen TR using interpolants *********************)
(*************************************************************************)

let get_cons maska ccba cba n_ops =
  let rec g i l =
    if i > n_ops then (List.rev l) else 
      let h = if maska.(i) then ccba.(i) else cba.(i) in
      g (i+1) (h::l) in
  g 0 []

let check_symm_itp c itp =               
  if oBool "checktree" then
    if ITP.interpolant_sat c (P.Not itp) = ITP.Sat then 
      failwith (Printf.sprintf "Invalid itp: %s :: %s \n" 
      (P.toString itp) (P.toString c))

let update_skip_ops rega start stop pis =
  let b = PT.make_eq_bdd pis in
  let p = P.And (List.map (fun i -> let p = PT.getPred i in P.Iff (p,P.Next p)) pis) in
  for i = start to stop do
    let loc = C_SD.location_coords rega.(i).R.location in
    let loc' = C_SD.location_coords rega.(i+1).R.location in
    print_upd_tr loc loc' p;              
    PT.update_transrel_bdd (Intra (loc,loc')) b 
  done

let strengthen_trel_symm rega reva cs itps =
  assert (List.length cs = List.length itps); 
  let (min_i,_,inc_ops) = 
    List.fold_left2 
    (fun (min_i,clock,inc_ops) c itp -> 
      if itp = P.True then (min_i, clock + 1, inc_ops) else
        let _ = check_symm_itp c itp in
        let i = reva.(clock) in
        let l = C_SD.location_coords rega.(i).R.location in
        let l' = C_SD.location_coords rega.(i+1).R.location in
        let bi = bpred_to_bdd clock itp in
        let (ps,ps') = 
          let aps = PT.bdd_support bi in
          (List.filter PT.isCurrentIndex aps, 
           Misc.map_partial PT.getCIndexFromNIndex aps) in
        let _ = print_upd_tr l l' (bpred_to_pred clock itp) in             
        let _ = PT.update_transrel_bdd (Intra (l,l')) bi in
        let _ = PT.add_local_preds l' ps' in
        let _ = if clock > 0 then update_skip_ops rega (reva.(clock-1)+1) (i-1) ps in
        ((if 0 <= min_i then min_i else i), clock+1, i::inc_ops)) (-1,0,[]) cs itps in
  ((if 0 <= min_i then Some (min_i + 1) else None), inc_ops)

let get_lazysymm_itps cs reva itpa = 
  M.msg_string M.Normal "in get_lazysymm_itps";
  assert (List.length cs > 0);
  let ren (i,p) = retime_pred i (pred_to_bpred (P.eq_canonicize (P.unprime p))) in
  snd(Misc.do_catch "get_lazysymm_itps" 
     (List.fold_left
       (fun (clock,rs) c ->
          let i = reva.(clock) in 
          let ri = ren (clock,itpa.(i)) in
          let ri' = ren (clock+1,itpa.(i+1)) in
          let (rv::_) = ITP.extract_interpolantN_symm ([c;P.And [ri;P.Not ri']]) in
          (clock + 1,rv::rs)) (0,[])) cs)

(* may throw SAT exn *)
let refine_trel_symm rega cba ccba maska reva n acc_itp_o =
  let cs = get_cons maska ccba cba n in
  let itps = 
    match (oBool "slicesymm", acc_itp_o)  with 
      (true, Some a) -> Stats.time "get lazysymm itp" (get_lazysymm_itps cs reva) a 
    | _ -> (Stats.time "get symm itp" ITP.extract_interpolantN_symm cs) in 
  Stats.time "strengthen_trel_symm" (strengthen_trel_symm rega reva cs) itps 

let omin xo yo = 
  match xo, yo with
  | None, _ -> yo
  | _, None -> xo
  | Some x, Some y -> Some (min x y)


let absref regs ops = 
  let _ = Misc.fassert (oBool "focisymm") "absref 1";
          Misc.fassert (List.length regs = List.length ops + 1) "absref2" in
  let rega = Array.of_list regs in
  let opa = Array.of_list ops in
  let n_rega = Array.length rega in 
  let maska = Array.create n_rega false in 
  let (itp_o,tps_o) =
    if not (oBool "lazysymm") then (None,None) else
      let (rsp,x) = Ar.block_analyze_trace (Array.to_list rega) (Array.to_list opa) in
      match rsp with None -> raise ITP.FociSatException | _ -> x in
  let (cba,ccba,reva,n) = build_fmc_constraint_arrays rega opa tps_o in
  let _ = M.msg_string M.Debug "HERE 1" in
  let rec _do ((rsp,ibs,regss) as args) =
    let _ = M.msg_string M.Debug "HERE 2" in
    if ((oBool "lazysymm") || not (oBool "multistr")) && rsp <> None then args else
      try 
        let (rsp',ops') = refine_trel_symm rega cba ccba maska reva n itp_o in
        process_inconsistent_ops opa maska (List.rev ops');
        _do (omin rsp rsp', ops'::ibs, regss)
      with ITP.FociSatExceptionAsgn ps -> 
        let _ = M.msg_string M.Debug ("HERE 3: rsp ="^(Misc.ostring string_of_int rsp)) in
        (rsp,ibs,if rsp <> None then regss else print_counterexample opa reva rega ps) 
      | ex -> (M.msg_string M.Debug ("HERE 4: exn "^(Printexc.to_string ex)); raise ex) in
  _do (None, [], Misc.clone "" n_rega)

(*********************************************************************)
(******************************* API *********************************)
(*********************************************************************)

(* API *)
let fociModelChecker_refine_trace_abstract regs ops =
  Misc.fassert (List.length regs = List.length ops + 1) "fmc_refine_trace_abstract 1";
  try
     if (oBool "fmcpa") then (fst (Ar.block_analyze_trace regs ops),regs) else
       let (rsp,ibs,regss) = absref regs ops in
       let _ = M.msg_string M.Debug "back from absref" in
       match () with
       | _ when rsp <> None && !PT.refined_flag ->
           let _ = M.msg_string M.Debug "CASE 1" in
           let _ = Abs_html.htmlize_trace ops regs ibs "Weak Transition Relation" in
           (rsp,regs)
       | _ when rsp <> None && oBool "localrestart" ->
           let _ = M.msg_string M.Debug "CASE 2" in
           (U.dump_abs (); raise NoNewPredicatesException)
       | _ when rsp <> None && oBool "lazysymm" ->
           let _ = M.msg_string M.Debug "CASE 3" in
           let _ = O.setValueOfBool "localrestart" true; U.dump_abs () in
           (rsp,regs)
       | _ when rsp = None && oBool "lazysymm" -> 
           let _ = M.msg_string M.Debug "CASE 4" in
           (rsp, regs)
       | _ ->
           let _ = M.msg_string M.Debug "CASE 5: lets get some predicates" in
           let (rsp,_) = Ar.block_analyze_trace regs ops in
           (if rsp <> None && not !PT.refined_flag then raise NoNewPredicatesException);
           if (rsp <> None && not (oBool "hybrid")) then let (rsp,_,_) = absref regs ops in (rsp,regs)
           else (rsp,regs)  
    with ITP.FociSatException -> (None,[])

(* API *)
let check_abstract_transition reg op reg' = 
  M.msg_string M.Error "WARNING: in check_abstract_transition";
  let p = retime_pred 0 (bdd_to_bpred reg.R.data.R.bdd) in
  let p' = retime_pred 1 (bdd_to_bpred reg'.R.data.R.bdd) in
  let l' = reg'.R.location in
  let is = PT.get_all_pred_indexes () in
  let c = retime_pred 0 (predicate_update_constraints op l' (is,is)) in
  try ignore (ITP.extract_interpolantN [P.And [p;c];P.Not p']); true with e -> 
    let [s;s'] = List.map P.toString [p;p'] in
    let _ = M.msg_string M.Normal ("[FAIL] Bad transition! [IN] "^s^" [OUT] "^s') in 
    false
