
(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

module O = Options
module Op = Abs_op.Operation
module C_SD = BlastCSystemDescr.C_System_Descr

open Misc
open Printf

module P = Ast.Predicate
module R = Abs_region.Region
module M = Message

(************* Code for generating HTML versions of traces ***************)
(* Files:
 * static: trace.js,trace.css
 * gen (fixed): code.html, index.html, result.html
 * gen (for each trace i): index.i.html,header.i.html,trace.i.html
 *)

let traceCount = ref 0
let sourceName = ref ""
let htmlDir () = "trace_"^(!sourceName)

let rootIndexFile () = sprintf "%s/index.html" (htmlDir ())
let headerFile i = sprintf "%s/header.%d.html" (htmlDir ()) i
let indexFile  i = sprintf "%s/index.%d.html" (htmlDir ())  i
let traceFile  i = sprintf "%s/trace.%d.html" (htmlDir ())  i
let resultFile () = sprintf "%s/result.html" (htmlDir ())

let sanitize = substitute_substrings [("<"," &lt; "); (">"," &gt; ")]

let lookupPredIndex pred =
    try PredTable.getPredIndex pred
    with Not_found ->
        try PredTable.getPredIndex <| P.negate pred
        with Not_found -> -1

(* XXX: this sucks... the use r.data.preds instead of lookup *) 
let rec printPred = 
    let group_count = ref 0 in 
    fun pred -> 

    let next_group () = 
        incr group_count;
        !group_count - 1 in

    match pred with
    | P.True -> "True"
    | P.False -> "False"
    | P.And ps -> 
        let gi = next_group () in
        sprintf "And { <br> <div id='pred_group_%d'
                          onmouseover='hl_pred_group(%d);'
                           onmouseout='unhl_pred_group(%d);'>
                         %s
                     </div> } <br>"
                   gi gi gi (String.concat "" <| List.map printPred ps)
    | P.Or ps -> 
        let gi = next_group () in
        sprintf "Or { <br> <div id='pred_group_%d'
                           onmouseover='hl_pred_group(%d);'
                           onmouseout='unhl_pred_group(%d);'>
                         %s
                     </div> } <br>"
                   gi gi gi (String.concat "" <| List.map printPred ps)
    | pred ->
        let pi = lookupPredIndex pred in
        sprintf 
           "<span class='p%d'>
                <span class='p%d_p' onclick='hl_pred(%d);'>%s</span>
                <span class='pred_x_button' onclick='hide_pred(%d);'>[x]</span>
                <br>
            </span>"
            pi pi pi (sanitize << P.toString <| PredTable.getPred pi) pi



let headString = 
  "<html><head><title>BLAST Counterexample Viewer</title>
   <style type=\"text/css\" media=\"screen\">@import \"trace.css\";</style>
   <script src=\"trace.js\" type=\"text/javascript\"></script></head>"
   
let rootIndexString = 
    "<html><head><meta http-equiv=\"Refresh\" content=\"0; URL=index.1.html\"></head></html>"

let resultIndexString = 
    "<html><head><meta http-equiv=\"Refresh\" content=\"0; URL=result.html\"></head></html>"

let predSelString pi =
    sprintf
       "<span class='ps%d' style='border-stle: solid; line-height: 1.5em; font-size: small'>
            <span class='p%d_p' onclick='hl_pred(%d);'> %s </span>
            <span class='p%d_o_button' onclick='unhide_pred(%d);'>
                <span class='pred_x_button'>[o]</span>
            </span>
        </span>"
        pi pi pi (sanitize << P.toString <| PredTable.getPred pi) pi pi

let indexString i = 
    headString ^ "\n" ^ sprintf 
       "<frameset rows=\"170,*\">
        <frame name=\"header\" scrolling=\"no\" src=\"header.%d.html\">
        <frameset cols=\"267,*\">
        <frame name=\"trace\" src=\"trace.%d.html\" scrolling=\"auto\">
        <frame name=\"code\" src=\"code.html\" scrolling=\"auto\">
        </frameset>
        <noframes>
        <body>
        <p>This page uses frames, but your browser doesn't support them.</p>
        </body>
        </noframes>
        </frameset>
        </html>" i i

let headerString i srcfile comment_s infblock_s ps = 
    headString ^ "\n" ^ sprintf 
       "<body align=center>
        <div align=center> BLAST Counterexample %d for %s </div>
        <table align=center>
        <tr><td>%s</td></tr>
        <tr> 
        <td>Infeasible Blocks:</td>
        <td><a href=\"javascript://\" onclick=\"top.trace.colorRed()\";>-</a></td>
        %s 
        </tr></table>
        <table border align=center>
        <tr>
          <td><a href=\"javascript://\" onclick=\"top.trace.showfootnote()\";>Toggle State</a></td>
          <td><a href=\"index.%d.html\" target=\"_top\">Previous</a></td>
          <td><a href=\"index.%d.html\" target=\"_top\">Next</a></td>
          <td><a href=\"result.html\" target=\"_top\">Result</a></td>
        </tr>
        </table>
        <br>
        <div align=center>
            %s
        </div>
        </body></html>" i srcfile comment_s infblock_s (if i > 1 then i-1 else
            1) (i+1) (String.concat "&nbsp;" <| List.map predSelString ps)

let resultString res_s = 
    headString ^ "\n"
    ^ sprintf "<body> <table align=center> 
        <tr><td align=center> <img src=\"%s.gif\" height=30> </td></tr> 
        <tr><td> %s <td></tr> 
        </table></body> </html>" 
        (if is_substring res_s ":-)" then "safe" 
         else if is_substring res_s ":-(" then "unsafe" 
         else "dontknow") 
        res_s

let traceString s = 
    headString ^ "\n" ^ sprintf "<body><ol> %s </ol></body></html>" s

let opString i op reg = 
    let clean op_s = 
        match chop op_s "\t" with 
        | _::x::xs -> String.concat "" (x::xs)
        | _ -> op_s in

    let fix_locs l l' =
        match l, l' with
        | -1, -1 -> -1, -1
        | -1, j  -> j-1, j
        | i, j when j <= i -> i, i+1
        | i, j -> i, j in
            
    let j, k = fix_locs
        (Op.op_to_line_number op) 
        (Op.op_to_line_number op) in

    sprintf 
       "<li onclick=\"colorme(%d)\";> 
        <div onMouseOver=\"this.className='highlight'\"
             onMouseOut=\"this.className='normal'\"
             onclick=\"top.code.colorCode('td',%d,%d)\";> %s </a><br><br>
        <div class=\"Desc\"> %s </div></div></li><br />"
        i j k 
        (clean <| Op.to_string op) 
        (try printPred <| R.to_predicate reg 
         with e -> String.concat "," << List.map sanitize << fst << R.to_strings <| reg)

let feasibleTraceString = 
    "<td><font color=\"#ff0000\"> Feasible Trace </font></td>"

(**********************************************************************************)
(****************************** API functions *************************************)
(**********************************************************************************)

let sys_command s = 
  M.msg_string M.Debug (Printf.sprintf "sys: %s" s);
  Sys.command s

let htmlize_trace ops regs infbs comment =
  M.msg_string M.Debug 
  <| Printf.sprintf "htmlize_trace: [ops] = %d , [regs] = %d" (List.length ops) (List.length regs);
  
  let ps = List.fold_left (fun ps reg -> union ps <| reg.R.data.R.preds) [] regs in
  
  let init () =  (* TBD:HTMLCOPY *)
      sourceName := O.getValueOfString "mainsourcename";
      let src = try List.hd <| O.getSourceFiles () 
                with _ -> failwith "Wrong number of source files" in
      let dir = htmlDir () in
      ignore << sys_command <| sprintf "mkdir %s" dir; 
      ignore << sys_command <| sprintf "rm %s/*.html" dir; 
      ignore << sys_command <| sprintf "cp $B2HOME/doc/html/* %s/" dir;
      ignore << Sys.command <| sprintf "$B2HOME/bin/htmlize %s > %s/code.html" src dir;
      write_to_file (rootIndexFile ()) rootIndexString;
      traceCount := 1

  and write_index i = 
      write_to_file (indexFile i) (indexString i)

  and write_header i = 

      let htmlize_make_infbs infb_l = 
          match infb_l with
          | [] -> feasibleTraceString
          | xs -> list_fold_lefti 
              (fun s i x -> 
                  s ^ sprintf 
                      "<td><a href=\"javascript://\" onclick=\"top.trace.colorRed(%s)\";>%d</a></td>\n"
                      (string_of_int_list x)
                      (i+1)) 
              "" xs in

      write_to_file (headerFile i) 
      <| headerString i !sourceName comment (htmlize_make_infbs infbs) ps

  and write_trace i =
      write_to_file (traceFile i) 
      << traceString
      << list_fold_lefti (fun s i (op,reg) -> s^(opString i op reg)^"\n\n") ""
      <| try List.combine ops regs with ex -> failwith "OHO" in
  
  if Options.getValueOfBool "traces" then begin
      if !traceCount = 0 then init ();
      write_index !traceCount;
      write_header !traceCount;
      write_trace !traceCount;
      incr traceCount
  end

let htmlize_trace_final res_s =
    if O.getValueOfBool "traces" then begin
        write_to_file (resultFile ()) <| resultString res_s;
        write_to_file (indexFile !traceCount) resultIndexString
    end
