(* Copyright 2008  The Regents of the University of California.

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License.
*)

module E = Ast.Expression
module P = Ast.Predicate
module C = Ast.Constant

module M = Message
open BlastArch
module C_SD = BlastCSystemDescr.C_System_Descr
module Command = C_SD.Command
module Op = Abs_op.Operation

module EnumerativeLattice = 
  struct
    
    module DataMap = Map.Make(struct
      type t = E.lval 
      let compare = compare
    end)
    
    type lattice = Cmap of C.constant DataMap.t | Top | Bottom
    type edge = C_SD.edge

    let bottom = Bottom 
    let top = Top 
    let init = Top
   
    let join l l' = l 
    let meet l l' =
	match l' with 
          Top -> l  
        | Bottom -> Bottom
        | _ -> (match l with Top -> l' | Bottom -> Bottom | _ -> failwith "Enum::Meet unimplemented")
 
    let leq l l' = (* CHECK *)
      match (l, l') with
      | (_, Top) -> true
      | (Bottom , _) -> true
      | (Top, _) -> false
      | (_, Bottom) -> false
      | (Cmap d, Cmap d') -> 
         try 
           DataMap.iter (fun k data -> let data' = DataMap.find k d' in
             if data = data' then () else raise Not_found) d ;
           true
         with Not_found -> false
          
	
 
    let eq l l' =
      match (l, l') with
      | (Top, Top) -> true
      | (Bottom, Bottom) -> true
      | (Cmap d, Cmap d') -> 
        DataMap.equal (fun a b -> Pervasives.compare a b = 0) d d' 
      | _ -> false

    let focus () = () 

    let is_consistent l bdd = if l = Bottom then false else true 

    let print fmt l =
      match l with
	  Top -> Format.fprintf fmt "Enum::Top"
	| Bottom -> Format.fprintf fmt "Enum::Bottom"
        | Cmap e -> 
            let f k d s = s^(Printf.sprintf " : %s -> %s " (E.lvalToString k) (C.toString d)) in
            let s = DataMap.fold f e "" in
            Format.fprintf fmt "@ Enummap = %s" s

    let toString l = 
      match l with
	  Top -> "Enum::Top"
	| Bottom -> "Enum::Bottom"
        | Cmap e -> 
            let f k d s = s^(Printf.sprintf " : %s -> %s " (E.lvalToString k) (C.toString d)) in
            let s = DataMap.fold f e "" in
  	    s
 
    
    
    (* type _query_fn = Ast.Predicate.predicate -> Ast.predicateVal*)
    let query_fn (lemnt:lattice) (pred:Ast.Predicate.predicate) = Ast.Dontknow

  let constantfold_lval l maptable =
    match l with
      | E.Symbol _ -> l
      | E.This -> l
      | E.Access (op, e, label) -> E.Access (op, E.constantfold e maptable, label)
      | E.Dereference e -> E.Dereference (E.constantfold  e maptable)
      | E.Indexed (e1, e2) -> (E.Indexed (E.constantfold e1 maptable, E.constantfold  e2 maptable))

  let post_asgn emap (lval, e) =
    M.msg_string M.Debug (Printf.sprintf "Enum::post_asgn : %s = %s" (E.lvalToString lval) (E.toString e));
          let (enum, range) = C_SD.is_enum lval in
M.msg_string M.Debug ("enum = "^(if enum then "true" else "false"));
    if enum then 
            let emht = Hashtbl.create 13 in
            DataMap.iter (fun k d -> Hashtbl.add emht k (E.Constant d) ) emap;
            let l = constantfold_lval lval emht in
            let ce = E.constantfold e emht in
    M.msg_string M.Debug (Printf.sprintf "%s = %s::post_asgn" (E.lvalToString l) (E.toString ce));
            match ce with 
            | E.Constant c -> DataMap.add l c emap
            | _ -> DataMap.remove l emap
   else emap 

  let post_block emap stmts =
    M.msg_string M.Debug "Enum::post_block" ;
    M.msg_string M.Debug (toString (Cmap emap));
    List.fold_left 
      (fun sofar thiscmd ->
       match thiscmd with   
         Command.Expr (E.Assignment (E.Assign, lval, e)) ->
           post_asgn sofar (lval,e)
       | Command.Expr _ -> sofar
       | Command.Return e -> sofar
      )    emap stmts

  let post_cmd emap op =
        begin
          match (C_SD.get_command op).Command.code with
          | Command.Skip -> Cmap emap
          | Command.Pred p ->
            begin
              let maptable = Hashtbl.create 13 in
              DataMap.iter (fun key data -> Hashtbl.add maptable key (E.Constant data) ) emap ;
              let p' = P.constantfold p maptable 
              in
              match p' with 
              | P.False -> Bottom
              | _ -> 
                begin
                  match p with
                  | P.Atom (E.Binary(E.Eq, E.Lval lv, E.Constant c))
                  | P.Atom (E.Binary(E.Eq, E.Constant c, E.Lval lv)) ->
                    if fst (C_SD.is_enum lv) then
                      Cmap (DataMap.add (constantfold_lval lv maptable) c emap)
                    else Cmap emap
                  | _ -> Cmap emap 
                end
            end
         | Command.Block stmts ->
		Cmap (post_block emap stmts)
         | Command.Havoc l -> let emap' = DataMap.remove l emap in Cmap emap'
         | Command.FunctionCall fce ->
           begin
             let (fn,subs,t) = C_SD.deconstructFunCall fce in
             M.msg_string M.Debug ("Function call "^fn);
	     List.iter (fun (x,e) -> M.msg_string M.Debug (Printf.sprintf "%s = %s" x (E.toString e))) subs;
	     M.msg_string M.Debug ("t = "^(E.lvalToString t));
             let asgns =
               if (C_SD.is_defined fn || t = C_SD.junkRet) then
                 (List.map (fun (x,e) -> (E.Symbol x,e)) subs) @
                 (if C_SD.is_noexpand fn then [(t,fce)] else [])
               else [(t, E.Lval C_SD.junkRet)] 
	     in
             let m = (List.fold_left (fun sofar thisasgn -> post_asgn sofar thisasgn) emap asgns) in
             Cmap (DataMap.fold (fun key _ sofar -> if C_SD.is_global key || C_SD.is_in_scope fn key then sofar else DataMap.remove key sofar) m m)
           end 
         | Command.GuardedFunctionCall expr -> failwith "GFcall: TODO" 
         | Command.SymbolicHook s -> Cmap emap
        end
 
  let post (a : lattice) op _ =
    M.msg_string M.Debug ("Enum.post: "^(C_SD.get_edge_name op));
    M.msg_printer M.Debug print a;
    let a' = 
      match a with 
      | Bottom -> Bottom
      | Top -> 
        (let ec = DataMap.empty in 
        let lec' = post_cmd ec op in
        match lec' with 
          Cmap ec' -> if DataMap.is_empty ec' then Top else Cmap ec'
        | _ -> lec')
      | Cmap ec -> 
        (let lec' = post_cmd ec op in
        match lec' with 
          Cmap ec' -> if DataMap.is_empty ec' then Top else Cmap ec'
        | _ -> lec')
    in
    ([],a')

    let summary_post caller exit callop inbdd =
     M.msg_string M.Debug (Printf.sprintf "Enum.summary_post: caller = %s \n exit = %s \n callop = %s\n" (toString caller) (toString exit) (Op.to_string callop));
     let scope = C_SD.get_location_fname (C_SD.get_source callop) in
     let rv = match exit with
     | Top -> Top
     | Bottom -> Bottom
     | Cmap emap ->
        begin
          match caller with 
          | Top -> exit
          | Bottom -> assert false
          | Cmap emapcaller -> 
            Cmap (DataMap.fold (fun k d sofar -> if C_SD.is_global k || C_SD.is_in_scope scope k then DataMap.add k d sofar else sofar) emap emapcaller)
        end
     in
     M.msg_string M.Debug "Enum.summary_post: return = \n";
     M.msg_printer M.Debug print rv ;
     rv

    let pre a _ = a

    let initialize () = ()
    
    let enabled_ops l = []

 end


