(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

(* $Id: blastArch.ml,v 1.5 2010/02/22 22:29:34 rupak Exp $
 * This file is part of the BLAST Project.
 *)


(**
 * This module defines the main signatures for Blast.
 *)
module P = Ast.Predicate
module E = Ast.Expression

(* DUMMY -- "unit" for functors *)
module type MYUNIT = sig val myunit : unit end

(**
 * The PRINTABLE signature specifies a type equipped with a print function
 * that allows to pretty print values of this type.
 *)
module type PRINTABLE =
sig
  type t
  val print : Format.formatter -> t -> unit
  val to_string : t -> string
end

(**
 * The COMMAND signature specifies a set of commands.  These commands
 * typically label the edges of a System description.
 *)
module type COMMAND =
sig
  include PRINTABLE
end


(**
 * The SYSTEM_DESCRIPTION signature specifies how an input system shall be
 * described.  This description is an automaton-like representation suitable
 * for model checking.
 * 
 * A system description consists of a list of functions, where each function
 * is represented as a control flow automaton (CFA). A CFA is a directed graph,
 * where each edge is labeled with a command.
 * Each function has a unique integer id. Each node in the CFA for a function
 * has a unique identifier.
 *)
module type SYSTEM_DESCRIPTION =
  sig
    module Command : COMMAND
	(** a command that does nothing *)
    val skip : Command.t
	
	(** 
	 * Data structure and access functions for a control flow automaton 
	 *)
    type location
    type edge
	  
	  (** used to hash edges, its (int*int)*(int*int) *)
    type edge_id_t = ((int * int) * (int * int))
	  
	  (** a location, used to build regions where only the data part is relevant *)
    val fake_location : location
	(** a location, used to build regions where only the data part is relevant *)
    val fake_edge : edge (* fake_location, skip, fake_location *)
	
	(** get the source position for a location *)
    val get_source_position : location -> (string * int * int) option
	
	(** get the locations for a source position *)
    val get_position_sources : string * int -> location list
	
	(** the pretty-printer for locations *)
    val print_location : Format.formatter -> location -> unit
	
	(** the pretty-printer for edges *)
    val print_edge     : Format.formatter -> edge     -> unit
	
	(** the string converter for locations *)
    val location_to_string : location -> string
	
	(** the string converter for edges *)
    val edge_to_string : edge -> string
	
	(** map a location to unique co-ordinates: the first int in the
	   pair is the same for all locations in the same function *)
    val location_coords : location -> int * int

	(** return the location, given the location coordinates *)
    val lookup_location_from_loc_coords : int * int -> location

	(** get the co-ordinates for the source and target nodes *)
    val get_edge_coords : edge -> edge_id_t
	
    val make_edge_coords : location -> location -> edge_id_t
	
	(** Extracting an edge's associated command *)
    val edge_to_command : edge -> Command.t
	
	(** gives the list of edges originating from a given location *)
    val get_outgoing_edges : location -> edge list
	
	(** gives the list of edges pointing to a given location *)
    val get_ingoing_edges : location -> edge list
	
	(** gives the source of an edge *)
    val get_source  : edge -> location
	
	(** gives the target of an edge *)
    val get_target  : edge -> location
	
	(** gives the command labeling an edge *)
    val get_command : edge -> Command.t
	
	(** Create a unique name for the edge *)
    val get_edge_name : edge -> string
	
	(** lvalues read and written by the command on an edge *)    
    val reads_and_writes_at_edge : edge -> E.lval list * E.lval list
	
	(** is the command on this edge a function call? *)
    val is_call_on_edge : edge -> bool
	(** return the function called on this edge, if it is a call.
         * Note: We do not handle function pointers, as the called
         * function is assumed to be a Symbol.symbol
         *)
    val call_on_edge : edge -> (Ast.Symbol.symbol * E.expression) option	
	
	(** make_edge src target cmd : make a new edge with command cmd between src and target,
	   but don't add this to the CFA. *)
    val make_edge : location -> location -> Command.t -> edge

  (** the locations corresponding to a label name in the C source *)
  val get_locations_at_label : string -> location list
	
	(**
	 * Interface functions to access the list of functions and the CFA
	 * associated with functions.
	 *)
	(** return list of functions *)
    val list_of_functions : unit -> string list

    exception NoSuchFunctionException of string
	
	(** returns true if the function is defined, false otherwise *)
    val is_defined : string -> bool
        
        (** find the entry location for a function.
	   Can raise NoSuchFunctionException
	 *)
    val lookup_entry_location : string -> location
	
	(** Given some location, return what function it belongs to *)
    val get_location_fname : location -> string

        (** get source line information from a location, for trace output *)
    val loc_to_line_number : location -> int
	
	(** Given some fname, returns all the locations for it 
	   Can raise NoSuchFunctionException
	 *)
    val lookup_all_locations : string -> location list
	
	(** Get the unique identifier of a function *)
    val get_fname_id : string -> int

	(** return true if the function name corresponds to an allocation function *)
    val is_alloc : string -> bool
	(** return true if the function name corresponds to a deallocation function *)
    val is_dealloc : string -> bool

        (** return true and the range if the lvalue is to be enumerated in the model checking *)
    val is_enum : Ast.Expression.lval -> (bool * int list)
	
	(** return true if the function should not be expanded by Blast *) 
    val is_noexpand : string  -> bool 
	
	(** return true if the function does not return -- hence no summaries created *) 
    val is_noreturn : string  -> bool 	

  (** parse a function expression, returning a triple of
      (function name, (formal name, formal value expr) list, result target lval) *)
  val deconstructFunCall : E.expression ->
       Ast.Symbol.symbol * (string * E.expression) list *
         E.lval

  (** Type of formal arguments to a function. *)
  type formals_kind =
      Fixed of string list
    | Variable of string list (** variable arguments *)

  (** gives the locals of a function *)
  val lookup_locals : string -> string list

  (** returns ALL the lvals that may be smashed by fname -- and are in scope of fname *)
  val lookup_local_lvals : string -> E.lval list
	
  (** gives the formals of a function *)
  val lookup_formals : string -> formals_kind


	
	(****************************************************************************)
    val stats_nb_cached_exp_closure : int ref
    val stats_nb_recomp_exp_closure : int ref

	(** the first argument is true if its to ALL, false for IMAGE *) 
    val add_field_to_unroll : bool -> string -> bool 
	(* was it added (true) or already there (false) *)
	
    val expression_closure : E.expression -> E.expression list
	(** again -- first arg is true if for ALL e.g. trace-project-refine etc. Second for image *)   
    val expression_closure_stamp :  bool -> E.expression -> E.expression list
	
    val lvalue_closure : E.lval ->  E.lval list
    val lvalue_closure_stamp : bool -> E.lval ->  E.lval list
    val iter_global_variables : (E.lval -> unit) -> unit
	
    val iterate_all_lvals : (E.lval -> unit) -> unit
	
    val iterate_all_scope_lvals : string -> (E.lval -> unit) -> unit
	
	(** Functions that return the Cil type of an lvalue. *)
	
	(** return the type of an lvalue *)
    val get_type_of_lval : E.lval -> Cil.typ
	
	(** return the type of an expression. Right now this only deals with Lvals and Address-es*)
    val get_type_of_exp : E.expression -> Cil.typ
	
    val is_struct : E.lval -> bool
    val is_array : E.lval -> bool
    val is_ptr : E.lval -> bool
    val is_skolem : E.lval -> bool
	

	
	(** returns true if the symbol is in the table of global variables *)
    val is_global : E.lval -> bool
	
	(** returns true if the symbol is a local for the specified function *)
    val is_in_scope : string -> E.lval -> bool
	
    val globals_of_type : Cil.typ -> E.lval list
	
	(** returns the possible "scopes" of an lval -- should be a single string.
	   "" indicates global scope *)
    val scope_of_lval : E.lval -> string list
	
	

	
	(** always_reach loc1 loc2 returns true iff loc1,loc2 are two locations in the same CFA,
	   s.t. all acyclic paths from loc1 (eventually) go through loc2 *)
    val always_reach : location -> location -> bool
	
	(** mods_on_path loc1 loc2 returns the set of lvals written to along all paths from loc1 to loc2.
	   again, the two locations need to be in the same CFA (or you get the empty set) *)
    val mods_on_path : location -> location -> E.lval list
	
    val may_be_modified : E.lval list -> location -> location -> bool
    val get_fname : E.expression -> string
    val local_mod_on_edge : edge -> E.lval -> bool
    val global_mod_on_call : string -> E.lval -> bool
	
	(** returns the list of global lvals that can be modified by a function *)
    val global_lvals_modified : string -> E.lval list
	
	(** returns the list of formal lvals that can be modified by a function *)
    val formal_lvals_modified : string -> E.lval list
      
    val global_lvals_mod_by : E.lval -> string list
	
	(* the list of funs that may mod an lv, via copyback effects *)
    val may_mod_by : E.lval -> string list
    
      
      
  val get_next_location : location -> location
      
      
  exception FunctionPointerException 
      
      (** Return target used by deconstructFunCall when function is
	 void *)
  val junkRet : E.lval
      
      (** Name for irrelevant functions *)
  val __SKIPFunctionName : string
      
  val __NotImplementedFunctionName : string
      
  val __BLAST_DummyFunctionName : string

  val __BLAST_DispatchFunctionName : string




  (** is a given location a join location ? actually are there 
  multiple edges _into_ that location *)
  val is_join : location -> bool

  (** is a given location l on a loop -- i.e. is l on a cycle in its CFA ?*)
  val is_loopback : location -> bool
 
  val isEntry : int * int -> bool

  (* various boolean utilities *)
  val isReturn : edge -> bool
  
  val returnExp : edge -> E.expression
  
  val isBlock : edge -> bool
  
  val isPredicate : edge -> bool

  val isFunCall : edge -> bool
  
  val isDefinedFunCall : edge -> bool
  (* utilities for async/events *)
 
  val isEventHandler_fname : string -> bool

  val is_dispatch_loc : location -> bool
    
  val isAsyncCall : edge -> bool
  
  val isAsyncExec : edge -> bool
  
  val isDispatchCall : edge -> bool
    
  val acAsyncCall : edge -> edge 
  
  val enter_call : edge -> bool

  val replace_code_in_cfa : edge -> E.t -> unit



  (** gives the list of global variables *)
  val globals : unit -> string list

	
  val is_event : location -> bool



  (** map_edges f applies f to all the edges in the program *)
  val map_edges : (edge -> 'a) -> 'a list

  (** map_edges f fname applies f to all the edges in the function fname *)
  val map_edges_fname : (edge -> 'a) -> string -> 'a list

  (** call graph related functions *)
  module type CALLGRAPH =
    sig
      val compute_callgraph : unit -> unit
      val output_callgraph : string -> unit
      val calls_made_by : string -> (E.expression * string) list
      val calls_made_into :
          string -> ((edge_id_t * E.expression) * string) list
      val get_reachable_functions : unit -> string list
      val caller_closure : string list -> string list
      val backwards_caller_closure : string list -> string list
      val bounded_backwards_caller_closure :
          int -> string list -> string list
      val shortest_path : string -> string -> string list
      val output_call_paths : unit -> unit
	  
      val call_depth : location -> int
	  
	  (* SCC interface *)
      type call_graph_scc_info = { 
	scc_fname_map : int -> string list;
          fname_scc_map : string -> int;
          scc_pre : int -> int list;
          scc_sinks : int list ;
          scc_size : int;
          scc_post : int -> int list;
	  topo_sort : int list
	}
      val this_fname_scc : string -> int
      val this_scc_fnames : int -> string list
      val this_scc_info : call_graph_scc_info option ref
      val scc_caller_closure : string -> string list
    end

  module CallGraph : CALLGRAPH

  
  (**** Reaching definitions *****)

  (** Return a list of lvals which have multiple paths to the specified location *)
  val get_all_multiply_defined_lvals : location -> E.lval list
  (** Do the analysis for reaching definitions *)
  val compute_reaching_definitions : unit -> unit

(******* Enable the passing of the "error"/"spec" predicate between main and the
   modelchecker -- this is a single pred now, but really one should know what all
   the spec lvals are and seed skipfun with those *)
  val set_error_lvals : E.lval list -> unit
  val get_error_lvals : unit -> E.lval list
end

module type OPERATION =
sig

  include PRINTABLE
  type info =
    | Normal
    | Call of P.predicate * 
              E.lval option * 
              string * E.expression list
    | Ret of E.expression 

  val op_to_line_number : t -> int     
  val get_info : t -> info
  val paren_fun : t -> string
  val opFunName : t -> string
  val fcexp_to_string : E.expression -> string
end

module type REGION =
sig
  include PRINTABLE
  val log_region : Message.logmsg_level -> Message.logmsg_type -> t -> unit
  
  val eq : t -> t -> bool
  val leq : t -> t -> bool
  val cup : t -> t -> t
  val cap : t -> t -> t
  val negate : t -> t
  val is_empty : t -> bool
  val to_strings : t -> (string list * string)
  val to_predicate: t -> P.predicate
  val get_call_depth : t -> int
  val get_function_name : t -> string
  val get_location_id : t -> (int * int)
  val is_noreturn : t -> bool
  val is_coverable : t -> bool
  val erase_data : t -> t
  val cubes : t -> int 
  (* coverers store data structures for quick lookup of a region which covers another *)
  type 'a coverers
  val emptyCoverers : unit -> 'a coverers
  val addCoverer : 'a coverers -> t -> 'a -> unit
  val deleteCoverer : 'a coverers -> t -> 'a -> unit
  val deleteCoverers : 'a coverers -> (t * 'a) list -> unit
  val findCoverer : 'a coverers -> t -> 'a option
  val findCoverUnion : 'a coverers -> t -> t
  val findExactCoverer : 'a coverers -> t -> 'a option
  val sizeCoverer : 'a coverers -> int
  val printCoverer : 'a coverers -> unit


end


(**
   Module that implements a complete lattice (with top and bottom elements)
   *)

(**
  Lattice Query Interface
  =======================
  
  The query interface enables lattices to exchange predicates with each other
  and with the abstraction layer (which maintains the predicate abstraction
  state). Here is how the layers fit together:

                       ----------------------------
                       |                          |
                       |    Abstraction Layer     |
                       |                          |
                       ----------------------------
                           |      /|\        | 
                          1|       |        7|
                           |      4|         |
                          \|/      |        \|/
                       ----------------------------
                       |                          |
                       |      Union Lattice       |
                       |                          |
                       ----------------------------
                        |  /|\  |     |    |   |
                       2|  3|  8|    5|   6|  9|
                       \|/  |  \|/   \|/  \|/ \|/
                       -----------   --------------         ---------
                       |         |   |            |         |       |
                       |  Set    |   |  SymExe    |  . . .  |       |
                       | Lattice |   |  Lattice   |         |       |
                       -----------   --------------         ---------

  The query interface is called during a post operation. Here's how it works:
  Step(s)     What Happens
  =======     ==================================================
  1           The abstraction layer calls post on the union lattice. It passes
              in a current lattice element, a command, and a query function.
              The query function, when called, executes the against the
              predicate abstraction state.

  2, 6        The union lattice then executes post on each of the sub-lattices.
              For each sub-lattice, it passes in a different query function -
              each closure includes the identity of the lattice it is passed
              to. This is used by the query function to avoid calling back to
              the same lattice.

  3, 4, 5     During the post function, a lattice can make a call to the
              query function that was passed in. This query will then execute
              queries against the abstraction layer and each of the other
              lattices. A query can return 0, 1, or Top. A contradiction is
              an error. Otherwise, the query returns the most precise value
              returned. Note that queries are executed against the OLD version
              of each lattice, not the new one being computed.

  2, 6, 1     Each lattice's post function returns a new lattice element and
              a predicate. When the union lattice returns from post, it returns
              an array of lattice elements and the conjunction of the
              sub-lattice predictes. Any predicates which are just TRUE may be
              dropped from the conjunction.

  7, 8, 9     The abstraction layer creates an assume command from the returned
              predicate using C_SD.command.pred. This command is first executed
              against the predicate state. The abstraction layer then calls
              the post of the union lattice with the assume command. The union
              lattice, in turn, calls the post of each sublattice.

  The lattice elements and predicate states from the second post are the
  final result used to create the new region.


  Interface Changes
  -----------------
  blastArch.ml will include the following new definitions:
   type lattice_qry_result = Top | True | False
   type lattice_qry_fn =  Ast.predicate -> lattice_qry_result

  The post functions of the set, symbolic execution, and union lattices will
  be changed to have the following signature:
   post: lattice -> edge -> lattice_qry_fun -> (lattice, Ast.prediate)

  Each lattice must also implement a function called "query" which is of
   type lattice_qry_fn.
  

  Design Decisions
  ----------------
  1. When the post function of a lattice is called with an assume command, it
     must not return any new predicates. This ensures termination of our overall
     post.

  2. A query is executed against the old version of each lattice element. This
     is consistent with the semantics of the union lattice -- each sub-lattice
     is an independent representation of the program state. A post operation
     is executed "in parallel" against all sub-lattices.
**)
module type LATTICE =
  sig
    module C_SD : SYSTEM_DESCRIPTION
    type lattice
    type edge
    val top : lattice (** No information known *)
    val bottom : lattice (** Empty region -- generally means that state is not reachable *)
    val init : lattice (** initial lattice element *)
    val join : lattice -> lattice -> lattice (** Join is used to take the union of two regions *)
    val meet : lattice -> lattice -> lattice (** Meet is used to take the interection of two regions *)
    val leq : lattice -> lattice -> bool (** Check for containment of regions *)
    val eq : lattice -> lattice -> bool (** Check for equality of regions *)
    val focus : unit -> unit (* for the moment it is unit -> unit *)
    val is_consistent : lattice -> CaddieBdd.bdd -> bool (** Is lattice consistent with predicates? *)
    val post : lattice -> edge -> (P.predicate -> P.predicate) ->
                                                                          (P.predicate list * lattice)
                    (** Compute the lattice which is result after applying computation in edge to initial lattice
		        Also return the predicates that you might have discovered to the mother code *)
    val summary_post : lattice -> lattice -> edge -> edge -> lattice
    val pre : lattice -> edge -> lattice  (** Not currently used *)
    (** Called after source code has been initialized *)
    val initialize : unit -> unit
    val query_fn : lattice -> P.predicate -> Ast.predicateVal
    val enabled_ops : lattice -> C_SD.Command.t list

    val print : Format.formatter -> lattice -> unit (** Pretty printer for this lattice *)
    val toString : lattice -> string 
  end


module type EVENTS =
  sig
    module C_SD : SYSTEM_DESCRIPTION
    type eventHandler 
    val eventHandler_to_string : eventHandler -> string
    val eventHandler_to_command : eventHandler -> C_SD.Command.t
    val deconstructAsyncCall : C_SD.edge -> (eventHandler * CaddieBdd.bdd)
    val constructAsyncCall : C_SD.edge * CaddieBdd.bdd -> C_SD.edge
  end

(**
 * The ABSTRACTION signature groups an operation module and a region module
 * and specifies abstract post and pre functions.  These two functions are
 * supposed to be conservative with respect to the precise semantics of the
 * operations.  It also specifies the precise and focus functions used for
 * refinement.
 *)
module type ABSTRACTION =
sig
  module Operation : OPERATION
  module Region : REGION

  val create_region : bool -> P.predicate -> P.predicate -> Region.t

  val initialize_abstraction : unit -> unit 

  val post : Region.t -> Operation.t -> Region.t

  (** [summary_post caller exit retop ] creates a region to be used 
   * after the function call whose post-condition caller represents, 
   * where exit is the region of the used summary node's exit *)
  val summary_post : Region.t -> Region.t -> Operation.t -> Operation.t -> Region.t

  (** check if a tree edge is sound check_abstract_transition reg op reg' checks 
  if post.(reg,op) \subseteq reg' *)
  val check_abstract_transition : Region.t -> Operation.t -> Region.t -> bool

  (** notify abstraction module of final outcome, for GUI *)
  val notify_result : string -> unit
  
  val dump_abs : unit -> unit 

  (** Raised when no new predicates are found after counterexample analysis. (Blast fails) *)
  exception NoNewPredicatesException
  exception RecursionException  
  exception RefineFunctionException of string (* refine from all entries of fname *)

  (** was a location belonging to fn refined in the last refine ? *)
  val is_fun_refined : string -> bool

  (** analyze a trace as a bunch of blocks foci *)
  val block_analyze_trace : Region.t list -> Operation.t list -> int option

  val print_stats : unit -> unit

  (** gives the possibly enabled operations in a given region *)
  val enabled_ops  : Region.t -> Operation.t list

  val reset : unit -> unit

  val fociModelChecker_refine_trace_abstract : 
    Region.t list -> Operation.t list -> (int option *  Region.t list)

end

module ModelCheckOutcome =
struct
  (* Outcome type polymorphic in error_path type *)
  type 'a outcome =
    | No_error
    | Error_reached of 'a
    | Race_condition of 'a * 'a * E.lval * E.lval
end

module type MAKE_LMC = 
  functor(A : ABSTRACTION) ->
  sig
    val model_check : A.Region.t  -> A.Region.t -> int
  end 

module type THEOREMPROVER_QUERY = 
sig
  val block_reset: unit -> unit
  val block_assert: P.predicate -> unit
  val block_check_unsat: P.predicate -> bool
  val is_contra: unit -> bool
  val block_pop: unit -> unit
  val reset: unit -> unit
  val queryExp: P.predicate -> bool
  val queryExpContext: P.predicate -> P.predicate -> bool
  val print_stats: unit -> unit

(*
  val get_model : unit -> (E.lval, Ast.Constant.constant) Hashtbl.t
*)
  val get_allsat : P.predicate -> P.predicate list -> P.predicate list -> P.predicate
 

  (* identifying name for decision procedure *)
  val whoami : unit -> string
end

module type THEOREMPROVER_INTERPOLATION =
sig
  exception FociSatException
  exception FociSatExceptionAsgn
  type t = Sat | Unsat of P.predicate
  val interpolant_sat: (P.predicate * P.predicate) -> t 
  val extract_interpolantN_symm: P.predicate list -> P.predicate list
  val extract_interpolantN: P.predicate list -> P.predicate list
  val interpolant_unsat: P.predicate -> bool
  val reset: unit -> unit
  val print_stats: unit -> unit
  val extract_preds: P.predicate list * P.predicate list -> P.predicate list

  (* identifying name for interpolation procedure *)
  val whoami : unit -> string
end

(*
module type PREDTABLE =
  sig
    module C_SD : SYSTEM_DESCRIPTION
    val refined_flag : bool ref
    val bddZero : CaddieBdd.bdd
    val bddOne : CaddieBdd.bdd
    val bdd_support : CaddieBdd.bdd -> int list
    val bdd_cube_cover : CaddieBdd.bdd -> CaddieBdd.bdd
    val lits_of_cube : CaddieBdd.bdd -> (int * bool) list
      
    val tableOfPreds : (int * P.predicate * CaddieBdd.bdd) list ref
    val hashTableOfPreds : (P.predicate, int) Hashtbl.t
      
    val getCurrentIndex : unit -> int
    val printTableOfPreds : unit -> unit
      
    val getPred : int -> P.predicate
    val getPredIndex :P.predicate -> int
    val getPredBdd : P.predicate -> CaddieBdd.bdd
    val getPredIndexBdd : int -> CaddieBdd.bdd
     
    val getNIndexFromCIndex : int -> int option
    val getCIndexFromNIndex : int -> int option
      
    val isCurrentIndex : int -> bool
      
    val getPredIndexNextBdd : int -> CaddieBdd.bdd
    val transrel_bdd_table : ((int * int) * (int * int), CaddieBdd.bdd) Hashtbl.t
    val get_transrel_bdd : (int * int) * (int * int) -> CaddieBdd.bdd
    val bddVarMap_list : (int * int) list ref
    val curr_var_map : CaddieBdd.varMap ref
    val curr_var_cube : CaddieBdd.bdd ref
    val next_var_map : CaddieBdd.varMap ref
    val next_var_cube : CaddieBdd.bdd ref
    val update_bdd_info : int * int -> unit
    val make_unprimed_bdd : CaddieBdd.bdd -> CaddieBdd.bdd
    val make_primed_bdd : CaddieBdd.bdd -> CaddieBdd.bdd
    val bdd_postimage : CaddieBdd.bdd -> CaddieBdd.bdd -> CaddieBdd.bdd
    val bdd_preimage : CaddieBdd.bdd -> CaddieBdd.bdd -> CaddieBdd.bdd
    val bdd_exist_quantify : CaddieBdd.bdd -> int list -> CaddieBdd.bdd
    val make_eq_bdd : int list -> CaddieBdd.bdd
    val make_impl_bdd : int list -> CaddieBdd.bdd
    val add_new_pred_to_table : P.predicate -> int
    val get_all_pred_indexes : unit -> int list
    val extract_full_cube : CaddieBdd.bdd -> CaddieBdd.bdd
    val convertPredToBdd_map :
      (P.predicate -> CaddieBdd.bdd) ->
        P.predicate -> CaddieBdd.bdd
    val convertPredToBdd : P.predicate -> CaddieBdd.bdd
    val ac_convertCubeToPred :
        (int -> P.predicate) -> int array -> P.predicate
    val convertCubeToPred : int array -> P.predicate
    val ac_convertBddToPred :
        (int -> P.predicate) -> CaddieBdd.bdd -> P.predicate
      val convertBddToPred : CaddieBdd.bdd -> P.predicate
      val bdd_to_string : CaddieBdd.bdd -> string
      val bdd_to_string_pretty : CaddieBdd.bdd -> string
      val update_transrel_bdd :
        (int * int) * (int * int) -> CaddieBdd.bdd -> unit
      val update_global_inv_bdd : CaddieBdd.bdd -> unit
      val bdd_node_count : CaddieBdd.bdd -> int
    (*  val bdd_top_var : CaddieBdd.bdd -> int *)
      val bdd_hash : CaddieBdd.bdd -> int * int
      val string_to_atom : Ast.Symbol.symbol -> P.predicate
      class ['a] bdd_hash :
        object
          val bdd_table : (int * int, (CaddieBdd.bdd * 'a) list) Hashtbl.t
          method add_new : CaddieBdd.bdd -> 'a -> unit
          method find : CaddieBdd.bdd -> 'a
        end
      val bdd_fold :
        (int -> 'a -> 'a -> 'a) -> (bool -> 'a) -> CaddieBdd.bdd -> 'a
      val bdd_tmp_var_idx : int ref
      val next_bdd_tmp_atom : unit -> P.predicate
      val bdd_to_pred_map :
        (int -> P.predicate) -> CaddieBdd.bdd -> P.predicate
      val bdd_to_predicate : CaddieBdd.bdd -> P.predicate
    val refine_counter : int ref
val refined : unit -> unit
val last_refine_counter : int ref
val refine_check : unit -> bool
val globally_useful_preds : int list ref
val blast_error_lval : E.lval
val globally_important_lvals : E.lval list ref
val globally_important_lvals_table : (E.lval, bool) Hashtbl.t
val all_important_lvals_table : (E.lval, bool) Hashtbl.t
val __SKIPFunctionName : string
val __NotImplementedFunctionName : string
val always_take_fun_table : (string, bool) Hashtbl.t
val always_skip_fun_table : (string, bool) Hashtbl.t
val add_skip_fun : string -> unit
val delete_skip_fun : string -> unit
val stats_nb_take_funs : int ref
val add_take_fun : string -> unit
val always_take : string -> bool
val always_skip : string -> bool
val relevant_globals_mod_table : (string, E.lval list) Hashtbl.t
val relevant_fields_table : (string, bool) Hashtbl.t
val update_global_lvals : E.lval -> unit
val ignorable_globals : E.lval list
val add_relevant_fields : E.lval -> unit
val is_relevant_field_image : string -> bool
val update_imp_lv_table : P.predicate -> unit
val ac_is_relevant_lval : E.lval -> bool
val is_relevant_lval : E.lval -> bool
val relevant_globals_mod : string -> E.lval list
val pred_affected_table : (C_SD.edge_id_t * int, bool) Hashtbl.t
val update_pred_affected_table : C_SD.edge_id_t -> int -> unit
val is_pred_affected : C_SD.edge_id_t -> int -> bool
val is_pred_global : int -> bool
val isConstant : E.expression -> bool
val lval_constant_table :
  (E.expression, E.expression list) Hashtbl.t
val make_const_table : unit -> unit
val getConstants : P.predicate -> E.expression list
val constant_preds : P.predicate -> P.predicate list
val actual_addPred : P.predicate -> int
val addPred : P.predicate -> int
val just_addPred : P.predicate -> int
val add_pred_to_scope : int -> int option
val getPredsContainingVar : Ast.Symbol.symbol -> P.predicate list
val theoremProverCache : (string , bool) Hashtbl.t
val statTotalNumberOfQueries : int ref
val statCoveredQueries : int ref
val statTotalCached : int ref
val statPostBddCalls : int ref
val statWorstCaseQueriesForPost : int ref
val statActualQueriesForPost : int ref
val statNonDCQueriesForPost : int ref
val statAssumeQueriesForPost : int ref
val statNumPost : int ref
val statNumAssumePost : int ref
val statReachedCubes : int ref
val statVisitedFunctionTable : (string, bool) Hashtbl.t
val reset_abs_stats : unit -> unit
exception SatisfiableException
val isSatisfiable : CaddieBdd.bdd -> P.predicate list -> bool
val askTheoremProver : P.predicate -> bool
val askTheoremProverContext :
  P.predicate -> P.predicate -> bool
val theoremProverCache_proof :
  (P.predicate, Proof.proof * int) Hashtbl.t
val trivialProof : Proof.proof
val statTotalNumberOfQueries_proof : int ref
val statTotalCached_proof : int ref
val statTotalMismatches : int ref
val vampyre_lemma_counter : int ref
val generateLfOutput : P.predicate -> Proof.proof -> unit
val generateLfOutput_noproof : P.predicate -> unit
val askTheoremProver_proof : P.predicate -> int option
val askTheoremProver_noproof :
  P.predicate -> P.predicate -> int option
val bat_counter : int ref
val refined_fun_table : (string, bool) Hashtbl.t
val cfb_local_refine_table : (string, bool) Hashtbl.t
val block_a_t_reset : unit -> unit
val is_fun_refined : string -> bool
val add_local_preds : int * int -> int list -> unit
val lookup_location_predicates : int * int -> int list
val print_pred_vector : unit -> unit
val ctrbddManager : CaddieBdd.ddMgr
val ctrbddZero : CaddieBdd.bdd
val ctrbddOne : CaddieBdd.bdd
val freeBddVars : CaddieBdd.bdd list ref
val getBddVar : unit -> CaddieBdd.bdd
val bddMap : (int * CaddieBdd.bdd list) list ref
val allocateCtrBdds : int list -> int -> unit
val destroyCtrBdds : unit -> unit
val getBdd : int -> Ast.Counter.counter -> CaddieBdd.bdd
val ctrmap_to_bdd : Ast.Counter.counter array -> CaddieBdd.bdd
val ctrmap_check_covered : Ast.Counter.counter array -> CaddieBdd.bdd -> bool
val ctrmap_exists_stuck : CaddieBdd.bdd -> bool
type abstraction_info =
    AbsPred of P.predicate
  | LocPredTuple of ((int * int) * int list)
  | SymLvalTuple of (string * (E.lval * E.lval) list)
val get_AbsPred : abstraction_info -> P.predicate option
val get_LocPredTuple : abstraction_info -> ((int * int) * int list) option
val get_SymLvalTuple :
  abstraction_info ->
  (string * (E.lval * E.lval) list) option
val relevant_symvar_table : (string, E.lval list) Hashtbl.t
val add_relevant_symvar : string -> E.lval -> unit
val get_relevant_symvar : string -> E.lval list
val initialize_skipfun : unit -> unit
val dump_abs : unit -> unit
val load_abs : unit -> unit
  end
*)

