
(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

(*************************************************************************************)
(*************************************************************************************)
(*                               Interface to CSIISAT                                *)
(*************************************************************************************)
(*************************************************************************************)
open Ast
module P = Predicate
module E = Expression
module M = Message
module O = Options

open Printf
open Misc

exception AtomInExpressionException
module CSAT = Csisat_ast
module Interpolate = Csisat_interpolate
module FociPrinter = Csisat_fociPrinter

(* API *)
exception FociSatException
(* API *)
exception FociSatExceptionAsgn of string list

(* API *)
type t = Sat | Unsat of P.predicate

let stats_nb_foci_queries = ref 0

(* API *)
let print_stats () = 
  M.msg_string M.Normal (Printf.sprintf "Number of interpolation queries = %d" !stats_nb_foci_queries)

(* API *)
let reset () = ()

(* API *)
let whoami () = 
  M.msg_string M.Normal "Csisat 1.1: By Dirk Beyer and Damien Zufferey";
  "CSISAT"

let csi_interpolate a b : CSAT.predicate =  
  incr stats_nb_foci_queries;
  try
    let a' = Csisat_astUtil.integer_heuristic a and
        b' = Csisat_astUtil.integer_heuristic b in
    M.msg_string M.Normal (FociPrinter.print_foci [a'; b']);
    let it = Csisat_LIUtils.round_coeff (Interpolate.interpolate_with_proof a' b') in
    M.msg_string M.Normal (FociPrinter.print_foci [it]);
    it
  with CSAT.SAT_FORMULA _ -> raise FociSatException




let csisat_expression_toString = Csisat_astUtil.print_expr
let csisat_predicate_toString = Csisat_astUtil.print_pred


let isSat a = match a with
  | Sat     -> true
  | Unsat _ -> false

(* For the benefit of foci_symm -- we need to rename the UFs used with this
 * suffix so that they are disjoint *)

let csisat_make_disjoint_flag = ref false
let set_disjoint_flag () = 
  M.msg_string M.Debug "Set Disjoint flag";
  csisat_make_disjoint_flag := true

let unset_disjoint_flag () = 
  M.msg_string M.Debug "Unset Disjoint flag";
  csisat_make_disjoint_flag := false

let csisat_UF_idx = ref 0
let csisat_UF_idx_tick () = incr csisat_UF_idx
let csisat_UF_idx_reset () = csisat_UF_idx := 0

let csisat_UF_string s =
    if !csisat_make_disjoint_flag && not (E.is_interpreted s) 
    then s^"_"^(string_of_int !csisat_UF_idx) 
    else s

let clean_csisat_UF_string s = 
  if not !csisat_make_disjoint_flag then s
  else let pieces = Misc.chop s "_" in
  if List.length pieces > 1 then String.concat "_" (Misc.list_init pieces)
  else s
      
let csisat_var_counter = ref 0 
let new_csisat_var () = 
    incr csisat_var_counter;
    "csisat"^(string_of_int !csisat_var_counter)

let listOfAddrVarsCsisat = ref []
let chlval_table = Hashtbl.create 109
let chlval_map_table  = Hashtbl.create 109
let int_table_to = Hashtbl.create 23
let int_table_from = Hashtbl.create 23
let int_table_seen = Hashtbl.create 23

let max_int = 999999 (* some RANDOM number I made up at 1am -- RJ *)
let make_new_num i = int_of_string (Misc.suffix 2 (string_of_int i))

let sign i = if i > 0 then 1 else if i = 0 then 0 else -1 
 
let int_map_to i =
  (* add i to the list of constants we've passed to csisat *)
  Hashtbl.replace int_table_seen i true;
  let i' = abs i in
  let j = (sign i) 
      * (if i' < max_int then i' else begin
        if Hashtbl.mem int_table_to i' 
        then Hashtbl.find int_table_to i' 
        else let key = make_new_num i' in 
        Hashtbl.add int_table_to i' key; 
        Hashtbl.add int_table_from key i';
        key
      end) in
  if (j <> i) then M.msg_string M.Normal "WARNING :int_maps_to overflow!";
  j
    
let int_map_from i = 
  let i' = abs i in
  let seen = 0::(Misc.hashtbl_keys int_table_seen) in
  let csisatmax = O.getValueOfInt "focimax" in
  if false (* XXX: what's going on here? -- mje *)
      && (csisatmax <> -1) 
      && (not (Hashtbl.mem int_table_seen i)) 
      && (i' > csisatmax) 
      && List.for_all (fun k -> abs ((abs k) - i') > csisatmax) seen (* TBD:SPEED *)
  then begin
    M.msg_string M.Error ("Invalid Interpolant -- large constant:"^(string_of_int i));
    raise FociSatException
      (* this will trickle up all the way and be caught by whoever *)
  end;
  (try Hashtbl.find int_table_from i' with Not_found -> i') * (sign i)
    
let getConstant e = 
  match e with
  | E.Constant (Constant.Int i) -> Some i
  | E.Unary (E.UnaryMinus,E.Constant (Constant.Int i)) -> Some (-i)
  | E.Unary (E.UnaryPlus,E.Constant (Constant.Int i)) -> Some i
  | _ -> None



let get_string_constant,
    get_const_string =

    let string_const_ctr_ref = ref 0
    and string_const_table : (string,string) Hashtbl.t = Hashtbl.create 101
    and const_string_table : (string,string) Hashtbl.t = Hashtbl.create 101 in

    (fun s ->
    try Hashtbl.find string_const_table s 
    with Not_found -> 
        incr string_const_ctr_ref; 
        let c = string_of_int !string_const_ctr_ref in 
        Hashtbl.replace string_const_table s c; 
        Hashtbl.replace const_string_table c s; 
        c),

    fun c ->
    try Hashtbl.find const_string_table c 
    with Not_found -> failwith "Unknown string constant c"

let symbol_to_int, int_to_symbol, clear_symbol_int_table =
  let sctr = ref (-1) in
  let sitbl : (string, int) Hashtbl.t = Hashtbl.create 101 and
      istbl : (int, string) Hashtbl.t = Hashtbl.create 101 in
  (fun s -> 
     try Hashtbl.find sitbl s 
     with Not_found ->
       incr sctr; 
       Hashtbl.add sitbl s !sctr ; 
       Hashtbl.add istbl !sctr s ; 
       !sctr) ,
  (fun i -> 
     try Hashtbl.find istbl i 
     with Not_found -> failwith "Unknown atomic literal i") ,
  (fun () -> Hashtbl.clear sitbl; Hashtbl.clear istbl)

let rec convertExpCsisat e =
  let convert_times e1 e2 = 
    match (getConstant e1, getConstant e2) with
    | (Some i1, Some i2) -> CSAT.Constant (float_of_int (i1 * i2))
    | (Some i1, None) -> CSAT.Coeff (float_of_int i1, (convertExpCsisat e2))
    | (None, Some i2) -> CSAT.Coeff (float_of_int i2, (convertExpCsisat e1))
    | (None, None) -> begin
	M.msg_string M.Debug (Printf.sprintf "WARNING! convertExpCsisat: times of %s * %s  " 
				(E.toString e1) (E.toString e2));
        CSAT.Application 
          (csisat_UF_string "times", List.map convertExpCsisat [e1;e2])
    end in
  
  let csisat_minus e' =
    match getConstant e' with
    | Some i -> CSAT.Constant (float_of_int (int_map_to (-i)))
    | None -> CSAT.Coeff (-1.0, (convertExpCsisat e')) in
  
  let chlval_convert_table e = 
    try CSAT.Variable (Hashtbl.find chlval_map_table e) 
    with Not_found ->
      let s_e = new_csisat_var () in
      Hashtbl.add chlval_table s_e e;
      Hashtbl.add chlval_map_table e s_e;
      CSAT.Variable s_e in
  
  let chlval_convert_recursive (l,s) = 
    let ufs = csisat_UF_string "chlval" in
    CSAT.Application (ufs, [ convertLvalCsisat l; CSAT.Variable ("_"^s) ]) in
  
  match e with
  | E.Lval lv -> convertLvalCsisat lv
  | E.Chlval (l, s) -> 
      if O.getValueOfBool "fociUF"
      then chlval_convert_recursive (l,s) 
      else chlval_convert_table e
	  
	  (* So the chlval thing didnt work out so well. look at test/foci/ex3.foci 
	     so now we just convert it to string and for reconversion, stuff the result in a hashtbl!
	     CSAT.Application ("chlval", [ convertLvalCsisat l; CSAT.Variable ("_"^s) ])
	   *)
	  
  | E.Binary (op, e1, e2) -> begin
      match op with 
      | E.Minus -> 
	  if (O.getValueOfBool "uif") then
            let ufs = csisat_UF_string "UIminus" in
	    CSAT.Application (ufs, [convertExpCsisat e1; convertExpCsisat e2])
	  else
	    CSAT.Sum [convertExpCsisat e1; csisat_minus e2]
      | E.Plus  -> 
	  if (O.getValueOfBool "uif") then
            let ufs = csisat_UF_string "UIplus" in
	    CSAT.Application (ufs, [convertExpCsisat e1; convertExpCsisat e2])
	  else
	    CSAT.Sum[convertExpCsisat e1; convertExpCsisat e2]
      | E.Mul   -> convert_times e1 e2
	    
      | E.Offset | E.BitAnd | E.BitOr | E.Div 
      | E.Rem | E.RShift | E.Xor | E.Comma -> 
          let ufs = csisat_UF_string (E.binaryOpToVanillaString op ) in
          CSAT.Application (ufs, List.map convertExpCsisat [e1;e2])
      | E.LShift -> begin
	  match e2 with
          | E.Constant (Constant.Int k) -> 
              convert_times e1 (E.Constant (Constant.Int (Misc.power 2 k)))
	  | _ -> 
              let ufs = csisat_UF_string (E.binaryOpToVanillaString op ) in 
              CSAT.Application (ufs, List.map convertExpCsisat [e1;e2])
      end
      | E.FieldOffset ->
	  (* the second argument to field offset is a field id *)
	  let v = convertExpCsisat e1 in
	  let s = match e2 with
          | E.Lval (E.Symbol s) -> s
	  | _ -> failwith ("convertExp : (FOCI) Second operator of field offset not a field id "^
			   (E.toString e)) in
	  let ufs_fof = csisat_UF_string "foffset" in
          let ufs_fld = csisat_UF_string ("@field_"^s) in
          CSAT.Application(ufs_fof, [v; CSAT.Variable ufs_fld])
	    
      | _ -> 
          if E.isRelOp op then raise AtomInExpressionException
          else failwith ("convertExp (Csisat): This operator should not be handled here "
                         ^(E.toString e))
  end
  | E.Constant (c) -> begin
      match c with
      | Constant.Int i -> CSAT.Constant (float_of_int (int_map_to i))
      | Constant.String s -> begin
	      M.msg_string M.Debug ("convertExpCsisat: String Constant! in"^(E.toString e));
	  CSAT.Variable ("_STRINGCONSTANT"^(get_string_constant s))
      end
      | Constant.Float f -> CSAT.Constant f
      | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
  end
    | E.Unary (uop, e) -> begin
	match uop with 
	| E.UnaryMinus -> csisat_minus e
	| E.UnaryPlus -> convertExpCsisat e
	| E.Address -> 
            let v = convertExpCsisat e in
            if (List.mem v (!listOfAddrVarsCsisat)) 
            then () 
            else listOfAddrVarsCsisat := v::(!listOfAddrVarsCsisat);
            CSAT.Application (csisat_UF_string "addrOf", [v])
	| _ -> CSAT.Application (csisat_UF_string (E.unaryOpToVanillaString uop), 
				 [convertExpCsisat e])
    end
    | E.FunctionCall(E.Lval (E.Symbol f), args) ->
        CSAT.Application(csisat_UF_string f, List.map convertExpCsisat args)
    | E.Select (e1,e2) -> CSAT.Application("select",List.map convertExpCsisat [e1;e2])
    | E.Store  (e1,e2,e3) -> CSAT.Application("store",List.map convertExpCsisat [e1;e2;e3])
    | _ -> failwith ("convertExpCsisat: " ^ (E.toString e) ^ "not handled")
	  
and convertLvalCsisat l =
    match l with
    | E.This -> failwith "This not handled in convertLval"
    | E.Symbol(str) -> CSAT.Variable (str) 
    | E.Access(E.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExpCsisat e in (* hack to get around foci bug! *)
        let ufs = csisat_UF_string ("sel@field_"^s) in
	CSAT.Application( ufs, [v(*; CSAT.Variable ("@field_"^s)*)])
    | E.Access(E.Arrow, e, s) ->
	failwith "csisat doesnt like Arrow"
    | E.Dereference e ->
        let ufs = csisat_UF_string "sel" in
	CSAT.Application(ufs, [convertExpCsisat e; CSAT.Constant 0.0])
    | E.Indexed (e1, e2) -> 
        let ufs = csisat_UF_string "sel" in
	CSAT.Application(ufs, List.map convertExpCsisat [e1;e2])

and convertAtomCsisat e = (* this returns a csisatpredicate *)
  let make_eq e1' e2' = 
    try CSAT.Eq (convertExpCsisat e1', convertExpCsisat e2') 
    with AtomInExpressionException -> begin
      match e1' with
      | E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) -> begin
          let s1 = convertExpCsisat e1' in
          let s2 = convertAtomCsisat e2' in
          CSAT.Or [ 
          (CSAT.And [s2 ; 
		     CSAT.Eq (s1, CSAT.Constant 1.0)]); 
          (CSAT.And [CSAT.Not s2 ; 
		     CSAT.Eq (s1, CSAT.Constant 0.0)]) ]
      end
      | _ -> begin
          match e2' with
          | E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) -> begin
              let s1 = convertExpCsisat e2' in
              let s2 = convertAtomCsisat e1' in
              CSAT.Or [ 
              (CSAT.And [s2 ; 
			 CSAT.Eq (s1, CSAT.Constant 1.0)]);
              (CSAT.And [CSAT.Not s2 ; 
			 CSAT.Eq (s1, CSAT.Constant 0.0)]) ]
          end
          | _ -> failwith ("convertAtomCsisat: Cannot handle this expression: "^(E.toString e))
      end 
    end in
  let make_leq e1' e2' = CSAT.Leq (convertExpCsisat e1', convertExpCsisat e2') in
  match e with
  | E.Binary (op, e1, e2) -> begin
      match op with 
      | E.Eq  -> make_eq e1 e2 
      | E.Ne  -> CSAT.Not (make_eq e1 e2)
      | E.Le  -> make_leq e1 e2	      
      | E.Lt  -> CSAT.Not (make_leq e2 e1)
      | E.Gt  -> CSAT.Not (make_leq e1 e2)
      | E.Ge  -> make_leq e2 e1 
      | _ -> failwith ("convertAtomCsisat: Cannot handle this expression: "^(E.toString e))
  end	
  | E.FunctionCall(E.Lval (E.Symbol f), args) -> make_eq e (E.Constant (Constant.Int 0))
  | E.Lval (E.Symbol s) -> CSAT.Atom (symbol_to_int s)
  | _ -> failwith ("convertAtomCsisat: "^(E.toString e)^ " Not a binary relation operator")
	
and convertPredCsisat e =
  match e with
  | P.True -> CSAT.True
  | P.False -> CSAT.False
  | P.And plist -> 
      CSAT.And (List.map convertPredCsisat plist)
  | P.Or plist -> 
      CSAT.Or (List.map convertPredCsisat plist)
  | P.Implies (e1, e2) -> 
      CSAT.Or [CSAT.Not (convertPredCsisat e1); (convertPredCsisat e2)]
  | P.Not e1 -> CSAT.Not (convertPredCsisat e1)
  | P.Atom e ->  (convertAtomCsisat e)
  | P.Iff (e1,e2) ->  convertPredCsisat 
        (P.Or [(P.And [e1;e2]);(P.And [P.Not e1 ; P.Not e2])])
(*  
  | P.All (s,e1) -> CSAT.Forall ([s],convertPredCsisat e1)  
*)
  | _ -> failwith ("Csisat Cannot handle this predicate "^(P.toString e))
	
and convertToCsisatSyntax (e: P.predicate) =
  listOfAddrVarsCsisat := [];
  let v_exp = convertPredCsisat (P.constantfold e (Hashtbl.create 1)) in
  let ufs_sel = csisat_UF_string "sel" in
  let ufs_aof = csisat_UF_string "addrOf" in
  let genAddrOfEq b =
      CSAT.Eq(b, CSAT.Application(ufs_sel, 
          [CSAT.Application(ufs_aof, [b]); CSAT.Constant 0.0])) in
  let genAddrPositive b = CSAT.Not  (CSAT.Leq 
      (CSAT.Application(ufs_aof, [b]), (CSAT.Constant 0.0))) in
  begin
    let vv = CSAT.And 
        ((v_exp :: (List.map (fun b -> CSAT.Not (CSAT.Eq
            ((CSAT.Application(ufs_aof, [b])), CSAT.Constant 0.0))) !listOfAddrVarsCsisat))
            @ (List.map genAddrOfEq !listOfAddrVarsCsisat)
            @ (List.map genAddrPositive !listOfAddrVarsCsisat)) in
    vv
  end

(* convert from Csisat to BLAST. *)
let rec convertCsisatLval l = 
  match convertCsisatExp l with
    | E.Lval l' -> l'
    | _ -> failwith ("convertCsisatLval failed on argument"^(Csisat_astUtil.print_expr l))
    
and make_bin op e1 e2 = E.Binary(op,convertCsisatExp e1, convertCsisatExp e2)
and make_eq e1 e2 = make_bin (E.Eq) e1 e2
and make_leq e1 e2 = make_bin (E.Le) e1 e2
and make_lt e1 e2 = make_bin (E.Lt) e1 e2

and gatherSum el = 
  match el with
    | [] -> E.Constant (Constant.Int 0)
    | h::tl -> 
       List.fold_left (fun a b -> E.Binary (E.Plus,a,convertCsisatExp b)) 
                      (convertCsisatExp h) tl
	
and convertCsisatExp (e : CSAT.expression) : E.expression = 
  match e with
    | CSAT.Constant f -> 
	let i = int_of_float f in
	assert ( f -. (ceil f) = 0.0 );
	E.Constant (Constant.Int i)
    | CSAT.Variable v -> 
	begin
          match (Hashtbl.mem chlval_table v) with
          | true -> Hashtbl.find chlval_table v 
          | false -> 
              if Misc.is_prefix "_STRINGCONSTANT" v then
                let c = Misc.chop_after_prefix "_STRINGCONSTANT" v in 
                let s = get_const_string c in
                E.Constant (Constant.String s)
              else E.Lval (E.Symbol v)
	end
    | CSAT.Application (f',el) -> begin
        let f = clean_csisat_UF_string f' in
        match f with
          | "select" -> begin
              try assert (List.length el = 2);
                  let (e1,e2) = (List.nth el 0, List.nth el 1) in
                  E.Select(convertCsisatExp e1, convertCsisatExp e2)
              with _ -> failwith "exn in convertCsisatExp @ select"
          end
          | "store" -> begin
              try assert (List.length el = 3);
                  let (e1',e2',e3') = 
                      let l' = (List.map convertCsisatExp el) in 
                      List.nth l' 0, List.nth l' 1, List.nth l' 2 in
                  E.Store(e1',e2',e3')
              with _ -> failwith "exn in convertCsisatExp @ store"
          end
          | "chlval" -> begin
              match el with
                | [l;(CSAT.Variable s)] -> 
                        E.Chlval (convertCsisatLval l, Misc.chop_after_prefix "_" s)
                | _ -> failwith ("unknown chlval args in convertCsisatExp! "^(csisat_expression_toString e))
          end
          | "*" -> begin
              try make_bin E.Mul (List.hd el) (List.nth el 2)
              with _ ->
                  failwith ("unknown * args in convertCsisatExp! "^(csisat_expression_toString e))
          end

          (* Hopefully we wont have to deal with wacko spl cases here! *)
	  
          | "addrOf" -> begin
              try let e1 = Misc.list_only el in E.Unary(E.Address, convertCsisatExp e1) 
              with _ -> failwith ("unknown & args in convertCsisatExp! "^(csisat_expression_toString e))
          end
	  
          | "sel" -> begin
              try let e1, e2 = Misc.list_only2 el in 
                  begin
                      match e2 with
                        | CSAT.Variable s -> begin
                              if (Misc.is_prefix "@field_" s)
                                  then E.Lval (E.Access(E.Dot, convertCsisatExp e1, 
                                      (Misc.chop_after_prefix "@field_" s)))
                                  else E.Lval (E.Indexed(convertCsisatExp e1,convertCsisatExp e2))
                        end
                        | _ -> (* JHALA: flaky! TBD:ARRAYCANON? *(x+i), x[i] *)
                            if (e2 = CSAT.Constant 0.0) 
                                then E.Lval (E.Dereference(convertCsisatExp e1))
                                else E.Lval (E.Indexed (convertCsisatExp e1, convertCsisatExp e2))
                  end
              with _ -> failwith ("unknown sel args in convertCsisatExp! "^(csisat_expression_toString e))
          end
	  
          | "+" -> begin
              M.msg_string M.Normal "csisat uses + in app";
              gatherSum el
          end
	      
          | "offset" -> 
              let e1, e2 = Misc.list_only2 (List.map convertCsisatExp el) in
              E.Binary(E.Offset,e1,e2)
          | "UIplus" -> begin
              M.msg_string M.Normal "csisat uses UIplus in app";
              gatherSum el
          end
	  | "UIminus" ->
              if (List.length el <> 2) then failwith "Bad use of UIminus";
              let e1, e2 = Misc.list_only2 (List.map convertCsisatExp el) in E.Binary(E.Minus,e1,e2)
          | "foffset" -> begin
              match el with
                | [e1;(CSAT.Variable s)] -> 
                    let fld = 
                        if Misc.is_prefix "@field_" s 
                            then Misc.chop_after_prefix "@field_" s 
                            else failwith "bad foffset!" in
                    E.Binary (E.FieldOffset, (convertCsisatExp e1), (E.Lval (E.Symbol fld)))
                | _ -> failwith "bad number of args to foffset!"
          end
          | _ -> begin
              if Misc.is_prefix "sel@field_" f then
                  (* a field "access" thing *)
                  let field_name = Misc.chop_after_prefix "sel@field_" f in
                  try let e = Misc.list_only el in
                      E.Lval (E.Access(E.Dot, convertCsisatExp e, field_name))
                  with _ -> failwith "multiple args to sel@field_ ! ?"
              else
              try let op = E.binaryOpFromVanillaString f in
                  let e1, e2 = Misc.list_only2 el in
                   E.Binary (op, convertCsisatExp e1,convertCsisatExp e2)
              with _ -> E.FunctionCall(E.Lval (E.Symbol f), List.map convertCsisatExp el)
          end
    end
    | CSAT.Sum el -> gatherSum el
    | CSAT.Coeff (f,e') ->
	assert (f = float_of_int (int_of_float f)) ;
	let i = int_of_float f in
        if i = -1 then 
            E.Unary (E.UnaryMinus, (convertCsisatExp e'))
        else if i < 0 then 
            E.Binary(E.Mul, E.Constant (Constant.Int (int_map_from i)), convertCsisatExp e')
        else if i = 1 then 
            convertCsisatExp e'
        else 
            E.Binary(E.Mul, E.Constant (Constant.Int (int_map_from i)), convertCsisatExp e')
	      

and convertCsisatPred p =
  match p with
    | CSAT.True -> P.True
    | CSAT.False -> P.False
    | CSAT.And plist -> P.And (List.map convertCsisatPred plist)
    | CSAT.Or plist -> P.Or (List.map convertCsisatPred plist)
    | CSAT.Not p' -> P.Not (convertCsisatPred p')
    | CSAT.Eq (e1,e2) -> P.Atom (make_eq e1 e2)
    | CSAT.Leq (e1,e2) -> P.Atom (make_leq e1 e2)
    | CSAT.Lt (e1,e2) -> P.Atom (make_lt e1 e2)
    | CSAT.Atom i -> P.Atom (E.Lval (E.Symbol (int_to_symbol i)))


let csisat_plist_toString = 
    String.concat "\n ; \n" 
    << List.map csisat_predicate_toString

let fq_dump num_cons =
    let big_fq_count_ref = ref 0 in
    fun cons_l_string ->
    let file = 
        sprintf "%s.%s" (O.getValueOfString "mainsourcename") <| 
        if num_cons > 2 then begin 
            incr big_fq_count_ref; 
            sprintf "%N.fq" !big_fq_count_ref end 
        else "csisatqs" in

    M.msg_string M.Normal <| sprintf "writing to: %s" file;
    Misc.write_to_file file <| sprintf "\n\n\n %s\n\n\n" cons_l_string


(* DEBUG *)
(* check that every symbol in itp is in both phi1 and phi2 *)
let check_interpolant_vars itp phi1 phi2 =
    match List.filter (fun x -> not ((List.mem x <| P.allVarExps phi1) 
            && (List.mem x <| P.allVarExps phi2))) 
        <| P.allVarExps itp with
    | [] -> ()
    | exps -> begin
        ignore 
        << map_pair2 (M.msg_string M.Normal) (M.msg_string M.Error) 
        << double <| sprintf 
            "Interpolant issue: %s\n phi1: %s\n phi2: %s\n" 
            (P.toString itp) (P.toString phi1) (P.toString phi1);
        List.iter (M.msg_string M.Normal << E.toString) exps
    end

(* DEBUG *)
let check_interpolant it a b = 
  if (Csisat_satPL.is_sat 
	(Csisat_astUtil.simplify 
	  (CSAT.And[ a ; CSAT.Not it]))) then 
	M.msg_string M.Error ("FAILURE: A |= I");
  if (Csisat_satPL.is_sat (Csisat_astUtil.simplify (CSAT.And[ it ; b]))) then 
	M.msg_string M.Error ("FAILURE: I /\\ B")
    
      
let extract_and_process_interpolant itp_process_function (cons_pre,cons_post) =
    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    clear_symbol_int_table ();
    M.msg_string M.Debug "In extract_and_process_interpolant";
    let csisat_c_pre, csisat_c_post = 
        convertToCsisatSyntax (P.And cons_pre), convertToCsisatSyntax (P.And cons_post) in 

    try M.msg_string M.Debug ">>>>>>>>>>>> calling csisat";
        M.msg_string M.Debug <| csisat_predicate_toString csisat_c_pre;
        M.msg_string M.Debug ";";
        M.msg_string M.Debug <| csisat_predicate_toString csisat_c_post;
        M.msg_string M.Debug ">>>>>>>>>>>> csisat done";

        if O.getValueOfBool "fqdump" then 
            fq_dump 2 <| csisat_plist_toString [csisat_c_pre;csisat_c_post];

        let csisat_file_string = 
            sprintf "\n\n\n %s \n\n\n" 
                (FociPrinter.print_foci [ csisat_c_pre;  csisat_c_post]) in

        let fwd_ip = 
            try csi_interpolate csisat_c_pre csisat_c_post
            with e -> begin
                M.msg_string M.Error <| sprintf
                    "Csisat FAILS!(eap_itp) raises exception!! %s"
                    (Printexc.to_string e);
                M.msg_string M.Normal <| sprintf
                    "Csisat FAILS!(eap_itp) raises exception!! %s"
                    (Printexc.to_string e);

                Misc.append_to_file "csisat_failed_q" csisat_file_string;
                CSAT.True end in

        M.msg_string M.Debug 
            << sprintf "Csisat yields Interpolant: %s" 
            <| csisat_predicate_toString fwd_ip;

        let fwd_ip_b = convertCsisatPred fwd_ip in

        M.msg_string M.Debug 
            << sprintf "Csisat yields b_Interpolant: %s" 
            <| P.toString fwd_ip_b;

        check_interpolant_vars fwd_ip_b (P.And cons_pre) (P.And cons_post);

	itp_process_function fwd_ip_b

    with FociSatException -> failwith "Foci raised SAT!"


(* API *)
let extract_preds = 
    extract_and_process_interpolant P.getAtoms

let extract_preds_parity = 
    extract_and_process_interpolant <| P.getAtoms_parity true

let extract_interpolant = 
    extract_and_process_interpolant id


(* API *)
let interpolant_sat cons_pre cons_post =
    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    M.msg_string M.Debug "CSISAT: In interpolant_sat:";
    M.msg_string M.Debug <| P.toString cons_pre;
    M.msg_string M.Debug <| P.toString cons_post;
    let csisat_c_pre = convertToCsisatSyntax cons_pre
    and csisat_c_post = convertToCsisatSyntax cons_post in
    M.msg_string M.Debug "CSISAT converted formulas: ";
    M.msg_string M.Debug <| FociPrinter.print_foci [csisat_c_pre; csisat_c_post];
    try 
      let fwd_ip = csi_interpolate csisat_c_pre csisat_c_post in
      M.msg_string M.Debug <| sprintf 
        "CSISAT yields Interpolant: %s"
        (FociPrinter.print_foci [fwd_ip]);
      let fwd_ip_b = convertCsisatPred fwd_ip in
      M.msg_string M.Debug <| sprintf 
        "Csisat yields b_Interpolant: %s"
        (P.toString fwd_ip_b);
      Unsat fwd_ip_b
    with 
    | FociSatException -> 
        M.msg_string M.Debug "CSISAT says SAT!";  Sat
    | FociSatExceptionAsgn _ -> 
        M.msg_string M.Debug "CSISAT says SATa!"; Sat
    | exc -> begin
        M.msg_string M.Error <| sprintf 
          "CSISAT FAILS!(isat) raises exception!! %s"
          (Printexc.to_string exc);
        Misc.append_to_file "foci_failed_q" <| sprintf 
          "\n\n\n %s\n ;\n  %s\n\n\n" 
          (FociPrinter.print_foci [csisat_c_pre]) 
            (FociPrinter.print_foci [csisat_c_post]);
        raise exc
    end

(* can raise FociSatException *)
let csi_interpolateN plist = 
    let ilistref = ref [] in
    let len = List.length plist in
    assert (len > 0);
    for i = 1 to (len-1) do
      let left, right = Misc.splitAt i plist in
      let itp = csi_interpolate (CSAT.And left) (CSAT.And right) in 
      ilistref := itp :: !ilistref  ;
      () 
    done;
    List.rev !ilistref

let csi_interpolateN_symm plist =
  csi_interpolateN plist (* TODO *)

(* API *)
let interpolateN plist =
  M.msg_string M.Debug "CSISAT: interpolatN";
  List.iter (fun p -> M.msg_string M.Debug <| P.toString p) plist; 
  Hashtbl.clear chlval_table;
  Hashtbl.clear chlval_map_table;
  let csiplist = List.map convertToCsisatSyntax plist in
  let l = csi_interpolateN csiplist in
  List.map convertCsisatPred l

(* API *)
let interpolateN_symm plist =
  interpolateN (* TODO *)
	 
(* API *)     	  
let interpolant_implies a b = 
  not (isSat (interpolant_sat a (P.negate b)))

(* API *)  
let interpolant_unsat a =
  not (isSat (interpolant_sat a P.True))


let extract_and_process_interpolantN itp_fun csisat_syn_conv itp_process_function cons_l 
        : P.predicate list = 

    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    M.msg_string M.Debug "In extract_and_process_interpolantN:";
    List.iter (M.msg_printer M.Debug P.print) cons_l;

    let csisat_cons_l = List.map csisat_syn_conv cons_l in

    M.msg_string M.Debug ">>>>>>>>>>>> calling csisat";
    let cons_l_string = csisat_plist_toString csisat_cons_l in
    M.msg_string M.Debug cons_l_string ;
    M.msg_string M.Debug ">>>>>>>>>>>> csisat done";
    if O.getValueOfBool "fqdump" then fq_dump (List.length cons_l) cons_l_string;

    let itp_l = 
        try let rv = itp_fun csisat_cons_l in 
            M.msg_string M.Debug "Csisat returns"; 
            rv
        with
          | FociSatException -> begin
              M.msg_string M.Normal "Csisat says SAT!"; 
              raise FociSatException end

          | FociSatExceptionAsgn a -> begin
              M.msg_string M.Normal "Csisat says SATa!";
              raise (FociSatExceptionAsgn a) end

          | e -> begin
              M.msg_string M.Normal <| sprintf
                  "Csisat FAILS!(eap_itpN) raises exception!! %s"
                  (Printexc.to_string e);
              M.msg_string M.Error <| sprintf
                  "Csisat FAILS!(eap_itpN) raises exception!! %s"
                  (Printexc.to_string e);
              Misc.append_to_file "foci_failed_q" <| sprintf 
                  "\n\n\n %s\n\n\n"
                  cons_l_string;
              raise FociSatException 
          end in
    M.msg_string M.Debug <| sprintf 
        "Csisat yields Interpolants: %s\n"
        (csisat_plist_toString itp_l);
    let itp_b_l = List.map convertCsisatPred itp_l in
    M.msg_string M.Debug 
        << sprintf "Csisat yields b_Interpolant: %s"
        << String.concat "\n ; \n" <| List.map P.toString itp_b_l;
    List.map itp_process_function itp_b_l

let conv_fun p = 
    if O.getValueOfBool "fmc" then 
        csisat_UF_idx_tick (); 
    convertToCsisatSyntax p
  
let extract_interpolantN = 
    extract_and_process_interpolantN 
	csi_interpolateN 
	conv_fun 
	id

let extract_interpolantN_noconv =
    extract_and_process_interpolantN csi_interpolateN convertToCsisatSyntax id
  
let extract_interpolantN_symm arg =
    set_disjoint_flag ();
    try let ps = 
            extract_and_process_interpolantN 
                csi_interpolateN_symm 
                conv_fun 
                id 
                arg in
        unset_disjoint_flag ();
        ps
    with e -> begin
        unset_disjoint_flag (); 
        raise e
    end
  
(*
let extract_minp_interpolantN cs =
    M.msg_string M.Debug "extract_minp_itpN";
    ignore <| extract_interpolantN cs;
    let n = List.length cs in
    let k = binary_search (fun i ->
        M.msg_string M.Debug "extract_minp_itpN:tick";
        try ignore << extract_interpolantN <| take (i+1) cs; 
            true
        with FociSatException | FociSatExceptionAsgn _ -> false)
        0 (n-1) in
    M.msg_string M.Debug <| sprintf "extract_minp_itpN: k/n = %d/%d" k n;
    extract_interpolantN <| take (k+1) cs @ clone P.True (n-k-1)
*)

(*
let extract_interpolantN_dasdill' trel_cons_l cube_cons_l =
    assert (List.length trel_cons_l + 1 = List.length cube_cons_l);

    M.msg_string M.Debug "Calling extract_and_process_interpolantN *for dasdill*:";
    List.iter (M.msg_printer M.Debug P.print) trel_cons_l;

    let csisat_trel_cons_l = compile_sets <| List.map conv_fun trel_cons_l in
    M.msg_string M.Debug ">>>>>>>>>>>> calling csisat:trel";
    M.msg_string M.Debug <| csisat_plist_toString csisat_trel_cons_l;

    extract_and_process_interpolantN 
        (interpolateN_dasdill csisat_trel_cons_l)
        conv_fun 
        id 
        cube_cons_l
*)


