(* $Id: cnf_sat.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** SAT solver executable. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

IFDEF FAST THEN
let fast_flag = true
ELSE
let fast_flag = false
ENDIF

let read_cnf f =
  let solver = new Solver_dpll.make (new Dpll.make) in
  let a_var = Term.variable (Variable.fresh ()) in
  let var_map = Index.Nat_map.create a_var in
  let lexer = Genlex.make_lexer ["c"; "p"; "cnf"] in

  let add_vars t = 
    let _ = Stream.next t in 
    let k = Stream.next t in 
      match k with
        | Genlex.Int n -> 
            for i = 0 to (n-1) do
              let v = Term.variable (Variable.fresh ()) in
	        Index.Nat_map.add var_map i v
            done 
        | _ -> failwith "read_cnf: parse error"
  in
    
  let to_term_lit cnf_lit = 
    let cnf_var  = (abs cnf_lit)-1 in
    let term_var = Index.Nat_map.find var_map cnf_var in
    let term_lit = Term_literal.literal term_var in
      if cnf_lit >= 0 then term_lit else Term_literal.negation term_lit
  in

  let rec read_clause cnf_clause t =
    match Stream.next t with
      | Genlex.Int 0 -> List.map to_term_lit cnf_clause
      | Genlex.Int x -> read_clause (x::cnf_clause) t
      | _ -> failwith "read_cnf: parse error"
  in

  let read_line () = 
    try
      let l = input_line f in
      let t = lexer (Stream.of_string l) in
      let c = Stream.next t in
      let _ = match c with
        | Genlex.Kwd "c" -> ()
        | Genlex.Kwd "p" -> add_vars t
        | Genlex.Int x -> solver#add_clause (read_clause [x] t) 
        | _ -> failwith "read_cnf: parse error"
      in
        true
    with End_of_file -> false 
  in
      
  while read_line () do () done;
  close_in f;
  solver

let log_level = ref 1

let file = ref ""

let arg_spec = Arg.align
  [("-v", Arg.Set_int log_level,"n Verbosity, default is 1")]

let usage_msg = String.concat "\n"
  ["DPLL-based SAT solver for DIMACS cnf files.";
   "";
   "Usage: "^(Filename.basename Sys.executable_name)^" [OPTION]... FILE";
   "Exit status 0 if satisfiable, 1 if unsatisfiable";
   "";
   "OPTIONS:"]

let set_file x = 
  file := x

let _ = 
  Arg.parse arg_spec set_file usage_msg;
  if !file = "" then begin
    Arg.usage arg_spec usage_msg;
    raise (Arg.Bad "missing file")
  end;
  Log.set_log_level !log_level;
  let f = open_in !file in 
  let solver = read_cnf f in
    Log.set_log_level !log_level;
    Log.log_string 2 "EXECUTABLE = ";
    Log.log_string 2 Sys.executable_name;
    Log.log_newline 2 ();
    Log.log_string 2 "BENCHMARK = ";
    Log.log_string 2 !file;
    Log.log_newline 2 ();
    Log.log_string 2 "FAST = ";
    Log.log_bool 2 fast_flag;
    Log.log_newline 2 ();
    if solver#solve then begin
      Log.log_string 1 "Satisfiable";
      Log.log_newline 1 ();
      exit 1
    end
    else begin
      Log.log_string 1 "Unsatisfiable";
      Log.log_newline 1 ();
      exit 0
    end
