(* $Id: bihashtbl.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Partial bijective hash tables. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

module type S =
sig
  type key
  type elt
  type t

  val length: t -> int
  val create: int -> t
  val clear: t -> unit
  val copy: t -> t
  val mem: t -> key -> bool
  val mem_inv: t -> elt -> bool
  val add: t -> key -> elt -> unit
  val find: t -> key -> elt
  val find_inv: t -> elt -> key
  val remove_key: t -> key -> unit
  val remove_elt: t -> elt -> unit
  val replace: t -> key -> elt -> unit
  val iter: (key -> elt -> unit) -> t -> unit
  val fold: (key -> elt -> 'a -> 'a) -> t -> 'a -> 'a
end

module Make (K: Hashtbl.HashedType) (E: Hashtbl.HashedType) =
struct
  type key = K.t
  type elt = E.t

  module T = Hashtbl.Make (K)
  module I = Hashtbl.Make (E)

  type t = {
      tbl: elt T.t;
      inv: key I.t
    }

  let length b = T.length b.tbl

  let create i = 
    {
      tbl = T.create i;
      inv = I.create i
    }

  let clear b = 
    T.clear b.tbl;
    I.clear b.inv

  let copy b = 
    {
      tbl = T.copy b.tbl;
      inv = I.copy b.inv
    }

  let mem b k = T.mem b.tbl k

  let mem_inv b e = I.mem b.inv e

  let find b k = T.find b.tbl k

  let find_inv b e = I.find b.inv e

  let add b k e =
    assert (not (T.mem b.tbl k));
    assert (not (I.mem b.inv e));
    T.add b.tbl k e;
    I.add b.inv e k;
    assert (mem b k);
    assert (mem_inv b e);
    assert (E.equal (find b k) e);
    assert (K.equal (find_inv b e) k)

  let remove_key b k = 
    begin
      try
        let e = find b k in
          T.remove b.tbl k;
          I.remove b.inv e
      with 
          Not_found -> ()
    end;
    assert (not (mem b k))

  let remove_elt b e = 
    begin
      try
        let k = find_inv b e in
          T.remove b.tbl k;
          I.remove b.inv e
      with 
          Not_found -> ()
    end;
    assert (not (mem_inv b e))

  let replace b k e =
    remove_key b k;
    add b k e

  let iter f b = T.iter f b.tbl

  let fold f b x = T.fold f b.tbl x
    
end
