(* $Id: bihashtbl.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** Partial bijective hash tables. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** Output signature of the functor {!Bihashtbl.Make} *)
module type S =
sig

  (** Keys in the domain. *)
  type key

  (** Elements in the range. *)
  type elt

  (** Partial bijective hash tables from {!Bihashtbl.S.key} to
      {!Bihashtbl.S.elt}. *)
  type t

  val length: t -> int
    (** Number of entries in a bijective hash table. *)
    
  val create: int -> t
    (** [create n] is a new, empty bijective hash table with capacity
	for [n] entries. 

	- Ensure {!Bihashtbl.S.length}[ (create n) = 0] *)

  val clear: t -> unit
    (** Remove all entries from a hash table. *)

  val copy: t -> t
    (** A copy of the given hash table. *)

  val mem: t -> key -> bool
    (** [mem b k] holds when there is an entry for [k] in [b]. *)

  val mem_inv: t -> elt -> bool
    (** [mem_inv b e] holds when an entry in [b] maps some key to [e]. *)

  val add: t -> key -> elt -> unit
    (** [add b k e] adds an entry mapping [k] to [e] to [b].

        - Require [not (]{!Bihashtbl.S.mem}[ b k)]
        - Require [not (]{!Bihashtbl.S.mem_inv}[ b e)]
        - Ensure {!Bihashtbl.S.mem}[ b k]
        - Ensure {!Bihashtbl.S.mem_inv}[ b e]
        - Ensure {!Bihashtbl.S.find}[ b k = e]
        - Ensure {!Bihashtbl.S.find_inv}[ b e = k] *)

  val find: t -> key -> elt
    (** [find b k] is the element to which [k] is mapped by [b].

	@raise Not_found if [not (]{!Bihashtbl.S.mem}[ b k)] *)

  val find_inv: t -> elt -> key
    (** [find_inv b e] is the key that [b] maps to [e].

	@raise Not_found if [not (]{!Bihashtbl.S.mem_inv}[ b e)] *)

  val remove_key: t -> key -> unit
    (** [remove_key b k] removes any entry that maps the key [k] from
	[b].

        - Ensure [not (]{!Bihashtbl.S.mem}[ b k)] *)

  val remove_elt: t -> elt -> unit
    (** [remove_key b y] removes any entry that maps to the element
	[e] from [b].

        - Ensure [not (]{!Bihashtbl.S.mem_inv}[ b e)] *)

  val replace: t -> key -> elt -> unit
    (** [replace b k e] updates the entries in [b] so that [k] is
	mapped to [e]. This is functionally equivalent to
	{!Bihashtbl.S.remove_key}[ b k] followed by
	{!Bihashtbl.S.add}[ b k e].

        - Ensure {!Bihashtbl.S.mem}[ b k]
        - Ensure {!Bihashtbl.S.mem_inv}[ b e]
        - Ensure {!Bihashtbl.S.find}[ b k = e]
        - Ensure {!Bihashtbl.S.find_inv}[ b e = k] *)

  val iter: (key -> elt -> unit) -> t -> unit
    (** [iter c t] applies [c] to every entry in [b].  Applications
	are made in increasing order with respect
	{!Bihashtbl.S.key}. *)

  val fold: (key -> elt -> 'a -> 'a) -> t -> 'a -> 'a
    (** [fold f b a] is [(f kN eN ... (f k1 e1 a) ... )], where [b] is
	a collection of entries [{(k1,e1) ... (kN,eN)}] ordered with
	respect to {!Bihashtbl.S.key}. *)
end

(** Functor building an implementation of the partial bijective hash
    table structure. *)
module Make (K: Hashtbl.HashedType) (E: Hashtbl.HashedType) :
  S with type key = K.t and type elt = E.t
