(* $Id: bunch.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Collection of elements with fast union. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

type 'a t = 
  | Empty                     (* An empty bunch *)
  | Cons of int * 'a * 'a t   (* Add a single element. *)
  | Node of int * 'a t * 'a t (* Union of two bunches. *)

let size = function
  | Empty -> 0
  | Cons (i,_,_) | Node (i,_,_) -> i

let empty = Empty 

let is_empty b = b = Empty

let singleton x = Cons (1,x,Empty)

let couple x y = Cons (2, x, Cons (1, y, Empty))

let join b1 b2 = 
  if b1 = Empty then b2
  else if b2 = Empty then b1
  else Node (size b1 + size b2, b1, b2)

let add x b = Cons (size b + 1, x, b)

let max3 x y z = if x > y then max x z else max y z

let rec worst = function
  | Empty -> 0
  | Cons (_,_,b) -> max (worst b) (abs (size b) - 1)
  | Node (_,b,c) -> max3 (worst b) (worst c) (abs (size b) - (size c))
      
let imbalance b = max (worst b - 1) 0
  
let rec add_balanced x b =
  let b_size = size b in
    if b_size < 3 then 
      Cons (b_size + 1, x, b)
    else 
      match b with
	| Empty -> assert (false)
	| Cons (i, y, t) -> Node (i+1, Cons (2, x, Cons (1, y, Empty)), t)
	| Node (i, t, u) ->
	    if size u < size t then
	      Node (i+1, t, add_balanced x u)
	    else
	      Node (i+1, add_balanced x t, u)

let rec iter_tail_rec f = function
  | [] -> ()
  | Empty :: t -> iter_tail_rec f t
  | Cons (_, x, b1) :: t -> f x; iter_tail_rec f (b1 :: t) 
  | Node (_, b1, b2) :: t -> iter_tail_rec f (b1 :: b2 :: t)

let iter f b = iter_tail_rec f [b]

let rec iter_const_tail_rec f c = function
  | [] -> ()
  | Empty :: t -> iter_const_tail_rec f c t
  | Cons (_, x, b1) :: t -> f c x; iter_const_tail_rec f c (b1 :: t) 
  | Node (_, b1, b2) :: t -> iter_const_tail_rec f c (b1 :: b2 :: t)

let iter_const f c b = iter_const_tail_rec f c [b] 

let rec filter p b =
  match b with
    | Empty -> Empty
    | Cons (_, x, b1) ->
	let b1' = filter p b1 in
	  if p x then
	    if b1' == b1 then b            (* preserve sharing *)
	    else add x b1'
	  else b1'
    | Node (_, b1, b2) ->
	let b1' = filter p b1 in
	let b2' = filter p b2 in
	  if b1' == b1 && b2' == b2 then b (* preserve sharing *)
	  else join b1' b2'

let rec filter_const p c b =
  match b with
    | Empty -> Empty
    | Cons (_, x, b1) ->
	let b1' = filter_const p c b1 in
	  if p c x then
	    if b1' == b1 then b            (* preserve sharing *)
	    else add x b1'
	  else b1'
    | Node (_, b1, b2) ->
	let b1' = filter_const p c b1 in
	let b2' = filter_const p c b2 in
	  if b1' == b1 && b2' == b2 then b (* preserve sharing *)
	  else join b1' b2'

let rec partition p b =
  match b with
    | Empty -> (Empty, Empty)
    | Cons (_, x, b1) ->
        let (t1,f1) = partition p b1 in
          if p x then
            if f1==Empty then (b, Empty)
            else (add x t1, f1)
          else
            if t1==Empty then (Empty, b)
            else (t1, add x f1)
    | Node (_, b1, b2) ->
        let (t1,f1) = partition p b1 in
        let (t2,f2) = partition p b2 in
          if f1==Empty then
            if f2==Empty then (b, Empty)
            else if t2==Empty then (t1,f2)
            else (join t1 t2, f2)
          else if f2==Empty then
            if t1==Empty then (t2,f1)
            else (join t1 t2, f1)
          else if t1==Empty then
            if t2==Empty then (Empty, b)
            else (t2, join f1 f2)
          else if t2==Empty then
            (t1, join f1 f2)
          else (join t1 t2, join f1 f2)

let rec partition_const p c b =
  match b with
    | Empty -> (Empty, Empty)
    | Cons (_, x, b1) ->
        let (t1,f1) = partition_const p c b1 in
          if p c x then
            if f1==Empty then (b, Empty)
            else (add x t1, f1)
          else
            if t1==Empty then (Empty, b)
            else (t1, add x f1)
    | Node (_, b1, b2) ->
        let (t1,f1) = partition_const p c b1 in
        let (t2,f2) = partition_const p c b2 in
          if f1==Empty then
            if f2==Empty then (b, Empty)
            else if t2==Empty then (t1,f2)
            else (join t1 t2, f2)
          else if f2==Empty then
            if t1==Empty then (t2,f1)
            else (join t1 t2, f1)
          else if t1==Empty then
            if t2==Empty then (Empty, b)
            else (t2, join f1 f2)
          else if t2==Empty then
            (t1, join f1 f2)
          else (join t1 t2, join f1 f2)

let rec fold_tail_rec f a = function
  | [] -> a
  | Empty :: t -> fold_tail_rec f a t
  | Cons (_, x, b1) :: t -> fold_tail_rec f (f a x) (b1 :: t)
  | Node (_, b1, b2) :: t -> fold_tail_rec f a (b1 :: b2 :: t)

let fold f a b = fold_tail_rec f a [b]

let rec map f = function
  | Empty -> Empty
  | Cons (s, x, b1) -> Cons (s, f x, map f b1)
  | Node (s, b1, b2) -> Node (s, map f b1, map f b2)

let rec exists_tail_rec p = function
  | [] -> false
  | Empty :: t -> exists_tail_rec p t
  | Cons (_, x, b1) :: t -> p x || exists_tail_rec p (b1 :: t)
  | Node (_, b1, b2) :: t -> exists_tail_rec p (b1:: b2 :: t)

let exists p b = exists_tail_rec p [b]

let rec for_all_tail_rec p = function
  | [] -> true
  | Empty :: t -> for_all_tail_rec p t
  | Cons (_, x, b1) :: t -> p x && for_all_tail_rec p (b1 :: t)
  | Node (_, b1, b2) :: t -> for_all_tail_rec p (b1 :: b2 :: t)

let for_all p b = for_all_tail_rec p [b]

let rec to_list_rec acc = function
  | Empty -> acc
  | Cons (_, x, b1) -> x :: (to_list_rec acc b1)
  | Node (_, b1, b2) -> to_list_rec (to_list_rec acc b2) b1

let to_list b = to_list_rec [] b

let of_list l = List.fold_right add l Empty
