(* $Id: id.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Abstract unique identifiers and an implementation with natural numbers. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

module type T = sig
  type t
  val invalid: t
  val is_valid: t -> bool
  val equal: t -> t -> bool
  val hash: t -> int
  val compare: t -> t -> int
end

module type Generator = sig
  include T
  type generator
  val new_generator: unit -> generator
  val fresh: generator -> t
  val count: generator -> int
  val iter: (t -> unit) -> generator -> unit
  val fold: ('a -> t -> 'a) -> 'a -> generator -> 'a
end

module Nat = struct

  type t = int

  let invalid = -1

  let possible x =
    x >= -1

  let is_valid x = 
    assert (possible x);
    x >= 0

  let equal x y = 
    assert (possible x && possible y);
    x = y

  let hash x = 
    assert (is_valid x);
    x

  let compare x y = 
    assert (is_valid x && is_valid y);
    (* Note: Subtraction is not safe for int comparison, because of
       overflow.  However, since x and y are both non-negative, this
       is OK. *)
    x - y

end

module Nat_generator = struct
  include Nat

  type generator = int ref
      
  let invariant gen = 
    !gen >= 0
      
  let new_generator () = 
    let gen = ref 0 in
      assert (invariant gen);
      gen

  let fresh gen =
    let result = !gen in
      gen := result + 1;
      assert (invariant gen);
      assert (is_valid result);
      result
	
  let count gen = 
    assert (!gen >= 0);
    !gen

  let iter f gen =
    for i = 0 to (!gen - 1) do
      f i
    done

  let fold f a gen =
    let result = ref a in
      for i = 0 to (!gen - 1) do
        result := f !result i;
      done;
      !result

end
