(* $Id: id.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** Abstract unique identifiers and an implementation with natural numbers. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0x
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** Unique identifiers. *)
module type T = sig

  (** An abstract type of identifiers. *)
  type t
      
  val invalid: t
    (** A distinguished invalid identifier, useful for dummy values.*)
	
  val is_valid: t -> bool
    (** Is an identifier valid?

	- Ensure [is_valid x = (x != ]{!Id.T.invalid}[)] *)

  val equal: t -> t -> bool
    (** Are two identifiers equal?  

	- Ensure [equal x y = (x = y)] *)

  val hash: t -> int
    (** [hash x] is the hash code for [x]

	- Require {!Id.T.is_valid}[ x] *)

  val compare: t -> t -> int
    (** [compare x y] is an arbitrary order for [x] and [y]

	- Require {!Id.T.is_valid}[ x]
	- Require {!Id.T.is_valid}[ y]
	- Ensure [(compare x y = 0) = (x = y)] *)

end

(** Generator of identifiers. *)
module type Generator = sig

  (** The values generated are unique identifiers. *)
  include T

  (** A source of fresh identifiers. *) 
  type generator

  val new_generator: unit -> generator
    (** A new fresh identifier generator. *)

  val fresh: generator -> t
    (** An identifier not previously produced by the generator. *)

  val count: generator -> int
    (** The number of identifiers produced by this generator. 

	- Ensure: [count >= 0] *)

  val iter: (t -> unit) -> generator -> unit
    (** Iterate over all identifiers produced by the generator. *)

  val fold: ('a -> t -> 'a) -> 'a -> generator -> 'a
    (** Fold over all identifiers produced by the generator. *)

end

(** Natural numbers as identifiers. *)
module Nat: T with type t = int

(** Generators for identifiers in an initial region of the naturals. 

    A generator [g] has identifiers in the range [0
    .. Nat_generator.count g - 1]. *)
module Nat_generator: Generator with type t = int
