(* $Id: option.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** Functions for the Option type. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

val is_some: 'a option -> bool
  (** Does an option contain a value? 

      - Ensure [is_some (Some x)]
      - Ensure [not (is_some None)] *)

val is_none: 'a option -> bool
  (** Does an option contain no value?

      - Ensure [not (is_none (Some x))]
      - Ensure [is_none None] *)

val value: 'a option -> 'a
  (** The value contained in an option.

      @raise Invalid_argument when applied to [None]. 

      - Ensure [value (Some x) = x] *)

val default: 'a -> 'a option -> 'a
  (** The value contained in an option, or a default if there is none. 

      - Ensure [default d (Some x) = x]
      - Ensure [default d None = d] *)

val classify: ('a -> bool) -> 'a -> 'a option
  (** Classify a value by placing it in an option. 

      - Ensure [p x  = (classify p x = Some x)]
      - Ensure [not (p x) = (classify p x = None)] *)

val perform: ('a -> unit) -> 'a option -> unit
  (** Perform a command on the value in an option.

      [apply c (Some x)] does [c x], while [apply f None] does
      nothing. *)

val map: ('a -> 'b) -> 'a option -> 'b option
  (** Map a function over the value in an option. 

      - Ensure [map f (Some x) = Some (f x)]
      - Ensure [map f None = None] *)

val filter: ('a -> bool) -> 'a option -> 'a option
  (** Filter the value in an option. 

      [filter p (Some x)] is [Some x] when [p x] holds and [None]
      otherwise, while [filter p None] is [None]. *)

val flatten: 'a option option -> 'a option
  (** Flatten a doubly optional value.

      - Ensure [flatten (Some x) = x]
      - Ensure [Flatten None = None] *)
