(* $Id: vec.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Dynamically sized arrays. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(* Is a value a floating point?
     
   This is a grotesque hack. *)
let is_float x = Obj.tag (Obj.repr x) = Obj.double_tag
  
(* How many elements like x can be stored in an array? *)
let limit_for x =
  if is_float x then
    Sys.max_array_length / 2
  else
    Sys.max_array_length

(* A trivial, in the sense of requiring trivial storage, value to
   store in the unused elements of the representing array.  We store this
   value there because if we simply left the old value there and ignored
   it that old value wouldn't get garbage collected. *)
let trivial x = 
  if is_float x then
    Obj.magic 0.0
  else 
    Obj.magic 0

(* How much capacity should we allocate to store n elements? 

   You could have all sorts of fancy schemes here, the only
   requirements are these:

   * n <= limit_for x
   * capacity_for n x >= n
   * capacity_for n x > 0 *)
let capacity_for n x =
  if (n > limit_for x) then raise (Invalid_argument "exceeds limit");
  max 1 n

(* Suppose we already have capacity for n elements like x, and now we
   want more, how much more should we get?

   You could have all sorts of fancy schemes here, the only
   requirements are these:

   * n > 0
   * capacity_above n x <= limit_for x
   * capacity_above n x > n *)
let capacity_above n x =
  assert (0 < n);
  let result = min (limit_for x) (2*n) in
    if (n >= result) then raise (Invalid_argument "exceeds limit");
    result

(* We want a capacity above n and at-least m. How much should we get?
   
   Again, we could have all sorts of schemes here, the only
   requirements are these:

   * n > 0
   * m > n
   * capacity_atleast m n x <= limit_for x
   * capacity_atleast m n x >= m *)
let capacity_atleast m n x = 
  assert (0 < n && n < m);
  let result = min (limit_for x) (max (2*n) m) in
    if (m > result) then raise (Invalid_argument "exceeds limit");
    result

(* A collection of primitive functions for manipulating sized arrays *)

let array_make size x cap triv =
  assert (size <= cap);
  let a = Array.make cap triv in
    Array.fill a 0 size x;
    a

let array_init size f cap triv =
  assert (size <= cap);
  let a = Array.make cap triv in
    for i = 0 to size - 1 do
      a.(i) <- f i
    done;
    a

let array_extend a size cap triv =
  assert (size <= cap);
  let a' = Array.make cap triv in
    Array.blit a 0 a' 0 size;
    a'

let array_top a size = a.(size - 1)

let array_pop a size triv =
  let last = size - 1 in
  let result = a.(last) in
    a.(last) <- triv;
    result

let array_push a size x =
  assert (size < Array.length a);
  a.(size) <- x

let array_delete a size i triv = 
  let last = size - 1 in
    Array.blit a (i + 1) a i (last - i);  
    a.(last) <- triv;
    last

let array_swap_out a size i triv =
  let last = size - 1 in 
    a.(i) <- a.(last);
    a.(last) <- triv;
    last

let array_insert a size i x =
  assert (i <= size);
  Array.blit a i a (i + 1) (size - i);
  a.(i) <- x

let rec array_fast_filter p a i size triv =
  assert (i <= size);
  if i = size then 
    size
  else if p a.(i) then
    array_fast_filter p a (i+1) size triv
  else begin
    a.(i) <- a.(size - 1);
    a.(size - 1) <- triv;
    array_fast_filter p a i (size - 1) triv
  end

let rec array_fast_filter_const p x a i size triv =
  assert (i <= size);
  if i = size then 
    size
  else if p x a.(i) then
    array_fast_filter_const p x a (i+1) size triv
  else begin
    a.(i) <- a.(size - 1);
    a.(size - 1) <- triv;
    array_fast_filter_const p x a i (size - 1) triv
  end

let rec array_first_not p a i size =
  assert (i <= size);
  if i = size then size 
  else if not (p a.(i)) then i
  else array_first_not p a (i+1) size

let rec array_first_not2 p x a i size =
  assert (i <= size);
  if i = size then size 
  else if not (p x a.(i)) then i 
  else array_first_not2 p x a (i+1) size

let rec array_first_noti p a i size =
  assert (i <= size);
  if i = size then size
  else if not (p i a.(i)) then i
  else array_first_noti p a (i+1) size

let rec array_first_noti2 p x a i size =
  assert (i <= size);
  if i = size then size
  else if not (p x i a.(i)) then i
  else array_first_noti2 p x a (i+1) size

let array_filter p a i size triv =
  assert (i <= size);
  let j = ref (array_first_not p a i size) in
    for k = !j + 1 to size - 1 do
      let x = a.(k) in
	if p x then begin
	  a.(!j) <- x;
	  j := !j + 1
	end
    done;
    Array.fill a !j (size - !j) triv;
    !j

let array_filter_const p x a i size triv =
  assert (i <= size);
  let j = ref (array_first_not2 p x a i size) in
    for k = !j + 1 to size - 1 do
      let y = a.(k) in
	if p x y then begin
	  a.(!j) <- y;
	  j := !j + 1
	end
    done;
    Array.fill a !j (size - !j) triv;
    !j

let array_filteri p a i size triv =
  assert (i <= size);
  let j = ref (array_first_noti p a i size) in
    for k = !j + 1 to size - 1 do
      let x = a.(k) in
	if p k x then begin
	  a.(!j) <- x;
	  j := !j + 1
	end
    done;
    Array.fill a !j (size - !j) triv;
    !j

let array_filteri_const p x a i size triv =
  assert (i <= size);
  let j = ref (array_first_noti2 p x a i size) in
    for k = !j + 1 to size - 1 do
      let y = a.(k) in
	if p x k y then begin
	  a.(!j) <- y;
	  j := !j + 1
	end
    done;
    Array.fill a !j (size - !j) triv;
    !j

let array_first p a i j = 
  assert (0 <= i && i <= j && j <= Array.length a);
  let result = Subarray.first p a i j in
    if result = j then raise Not_found
    else result

let array_first_const p x a i j =
  assert (0 <= i && i <= j && j <= Array.length a);
  let result = Subarray.first_const p x a i j in
    if result = j then raise Not_found
    else result

(* Dynamically resizable arrays and associated functions. *)

type 'a t = {
  mutable contents: 'a array;
  mutable size: int; 
  dummy: 'a
}

let make n x = 
  assert (n >= 0);
  let triv = trivial x in
    {
      contents = array_make n x (capacity_for n triv) triv; 
      size = n;
      dummy = triv
    }

let create = make

let init n f = 
  assert (n >= 0);
  let triv = trivial (f 0) in
    {
      contents = array_init n f (capacity_for n triv) triv;
      size = n;
      dummy = triv
    }
      
let copy v = 
  {
    contents = Array.copy v.contents;
    size = v.size;
    dummy = v.dummy
  }

let ensure_cap v =
  let cap = Array.length v.contents in
    if v.size = cap then
      v.contents <- 
	array_extend v.contents cap (capacity_above cap v.dummy) v.dummy

let of_array a =
  let n = Array.length a in
    assert (n > 0);
    let triv = trivial a.(0) in
      {
	contents = array_extend a n (capacity_for n triv) triv;
	size = n;
	dummy = triv
      }

let of_list l =
  assert (l != []);
  let n = List.length l in
    let triv = trivial (List.hd l) in
    let a = Array.make (capacity_for n triv) triv in
    let rec init i = function
    | [] -> ()
    | (h::t) -> a.(i) <- h; init (i + 1) t
    in
      init 0 l;
      {
	contents = a;
	size = n;
	dummy = triv
      }

let to_array v = 
  let a = Array.make v.size v.dummy in
    Array.blit v.contents 0 a 0 v.size;
    a

let to_list v = Subarray.fold_right (fun h t -> h::t) v.contents 0 v.size []
  
let length v = v.size

let limit v = limit_for v.dummy

let is_empty v = v.size = 0
  
let grow v n x = 
  assert (n > v.size);
  let cap = Array.length v.contents in
    if n > cap then 
      v.contents <- 
        array_extend v.contents v.size 
	(capacity_atleast n cap v.dummy) v.dummy;
    Array.fill v.contents v.size (n - v.size) x;
    v.size <- n

let grow_init v n f =
  assert (n > v.size);
  let cap = Array.length v.contents in
    if n > cap then 
      v.contents <- 
        array_extend v.contents v.size 
	(capacity_atleast n cap v.dummy) v.dummy;
    Subarray.set_fun v.contents v.size n f;
    v.size <- n


let shrink v n =
  assert (0<=n && n<=v.size);
  Array.fill v.contents n (v.size - n) v.dummy;
  v.size <- n

let clear v = shrink v 0

let get v i =
  assert (0<=i && i<v.size);
  v.contents.(i)
    
let set v i x =
  assert (0<=i && i<v.size);
  v.contents.(i) <- x

let set_all v i j x =
  assert (0<=i && i<=j && j<=v.size);
  Subarray.set_all v.contents i j x

let top v = 
  assert (v.size != 0);
  array_top v.contents v.size

let pop v =
  assert (v.size != 0);
  let result = array_pop v.contents v.size v.dummy in
    v.size <- v.size - 1;
    result
  
let push v x = 
  ensure_cap v;
  array_push v.contents v.size x;
  v.size <- v.size + 1
      
let insert v i x =
  assert (0 <= i && i <= v.size);
  ensure_cap v;
  array_insert v.contents v.size i x;
  v.size <- v.size + 1

let delete v i =
  assert (0 <= i && i < v.size);
  v.size <- array_delete v.contents v.size i v.dummy

let swap v i j =
  assert (0 <= i && i < v.size);
  assert (0 <= j && j < v.size);
  Subarray.swap v.contents i j

let swap_out v i =
  assert (0 <= i && i < v.size);
  v.size <- array_swap_out v.contents v.size i v.dummy  

let rev v = Subarray.rev v.contents 0 v.size
      
let iter c v = Subarray.iter c v.contents 0 v.size

let iter_const c x v = Subarray.iter_const c x v.contents 0 v.size

let iteri c v = Subarray.iteri c v.contents 0 v.size

let iteri_const c x v = Subarray.iteri_const c x v.contents 0 v.size

let modify f v = Subarray.modify f v.contents 0 v.size

let modify_const f x v = Subarray.modify_const f x v.contents 0 v.size 

let fold_left f acc v = Subarray.fold_left f acc v.contents 0 v.size
      
let fold_right f v acc = Subarray.fold_right f v.contents 0 v.size acc

let foldi_left f acc v = Subarray.foldi_left f acc v.contents 0 v.size
      
let foldi_right f v acc = Subarray.foldi_right f v.contents 0 v.size acc

let fold_left_const f c acc v = 
  Subarray.fold_left_const f c acc v.contents 0 v.size
      
let fold_right_const f c v acc = 
  Subarray.fold_right_const f c v.contents 0 v.size acc

let foldi_left_const f c acc v = 
  Subarray.foldi_left_const f c acc v.contents 0 v.size
      
let foldi_right_const f c v acc = 
  Subarray.foldi_right_const f c v.contents 0 v.size acc

let first p v = array_first p v.contents 0 v.size

let first_const p x v = array_first_const p x v.contents 0 v.size

let find p v = v.contents.(array_first p v.contents 0 v.size)

let find_const p x v = v.contents.(array_first_const p x v.contents 0 v.size) 

let mem x v = Subarray.mem x v.contents 0 v.size

let exists p v = Subarray.exists p v.contents 0 v.size

let exists_const p x v = Subarray.exists_const p x v.contents 0 v.size

let for_all p v = Subarray.for_all p v.contents 0 v.size

let for_all_const p x v = Subarray.for_all_const p x v.contents 0 v.size

let fast_filter p v = 
  v.size <- array_filter p v.contents 0 v.size v.dummy

let fast_filter_const p x v = 
  v.size <- array_filter_const p x v.contents 0 v.size v.dummy

let filter p v = 
  v.size <- array_filter p v.contents 0 v.size v.dummy
  
let filter_const p x v = 
  v.size <- array_filter_const p x v.contents 0 v.size v.dummy

let filteri p v = 
  v.size <- array_filteri p v.contents 0 v.size v.dummy

let filteri_const p x v = 
  v.size <- array_filteri_const p x v.contents 0 v.size v.dummy

let sort cmp v = Subarray.sort cmp v.contents 0 v.size

let stable_sort cmp v = Subarray.stable_sort cmp v.contents 0 v.size

