(* $Id: smtlib_lexer.mll,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Lexical analysis for SMT-LIB theory, logic, and benchmark files. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

{
  open Lexing
  open Smtlib_attribute
  open Smtlib_grammar

  let parse_identifier value =
    match value with
      | "flet"         -> FLET                 
      | "ite"          -> ITE                  
      | "let"          -> LET                  
      | "logic"        -> LOGIC                
      | "theory"       -> THEORY               
      | "benchmark"    -> BENCHMARK            
	  
      | "true"         -> TRUE
      | "false"        -> FALSE
          
      | "not"          -> CONNECTIVE value
      | "implies"      -> CONNECTIVE value
      | "if_then_else" -> CONNECTIVE value
      | "and"          -> CONNECTIVE value
      | "or"           -> CONNECTIVE value
      | "xor"          -> CONNECTIVE value
      | "iff"          -> CONNECTIVE value
	  
      | "exists"       -> QUANT_SYMB value
      | "forall"       -> QUANT_SYMB value
	  
      | "sat"          -> STATUS Sat
      | "unsat"        -> STATUS Unsat
      | "unknown"      -> STATUS Unknown
	  
      | _              -> IDENTIFIER           value
	  
  let parse_attribute value = 
    match value with  
      | "assumption"   -> ATTRIBUTE_ASSUMPTION
      | "axioms"       -> ATTRIBUTE_AXIOMS    
      | "definition"   -> ATTRIBUTE_DEFINITION
      | "extensions"   -> ATTRIBUTE_EXTENSIONS
      | "formula"      -> ATTRIBUTE_FORMULA   
      | "funs"         -> ATTRIBUTE_FUNS      
      | "extrafuns"    -> ATTRIBUTE_EXTRAFUNS 
      | "extrasorts"   -> ATTRIBUTE_EXTRASORTS
      | "extrapreds"   -> ATTRIBUTE_EXTRAPREDS
      | "language"     -> ATTRIBUTE_LANGUAGE  
      | "logic"        -> ATTRIBUTE_LOGIC     
      | "notes"        -> ATTRIBUTE_NOTES     
      | "preds"        -> ATTRIBUTE_PREDS     
      | "sorts"        -> ATTRIBUTE_SORTS     
      | "status"       -> ATTRIBUTE_STATUS    
      | "theory"       -> ATTRIBUTE_THEORY    
      | _              -> ATTRIBUTE            value
}

let letter             = ['a'-'z' 'A'-'Z']
let digit              = ['0'-'9']

let simple_identifier  = letter (letter | digit | '.' | '\'' | '_')*

let user_value_content = (([^'{' '}' '\\'] | ('\\'_))*)
let string_content     = (([^'"' '\\'] | ('\\'_))*)

let numeral            = '0' | (digit # '0') digit*
let rational           = numeral '.' ('0'*) numeral

let indexed_identifier = simple_identifier '[' numeral (':' numeral)* ']'
let identifier         = simple_identifier | indexed_identifier

let arith_symb         = ['=' '<' '>' '&' '@' '#' '+' '-' '*' '/' '%' '|' '~']+

let comment            = ';'[^ '\n' '\r']*
let blank              = [' ' '\t' '\n' '\r']

rule token = parse
  | '('                                 { LEFT_PAREN }
  | ')'                                 { RIGHT_PAREN }
  | '{'(user_value_content as value)'}' { USER_VALUE value }
  | '"'(string_content as value)'"'     { STRING value } 
  | numeral                             { NUMERAL (lexeme lexbuf) }
  | rational                            { RATIONAL (lexeme lexbuf) }
  | identifier                          { parse_identifier (lexeme lexbuf) }
  | '?'(simple_identifier as name)      { VAR name }
  | '$'(simple_identifier as name)      { FVAR name } 
  | ':'(identifier as value)            { parse_attribute value }
  | arith_symb                          { ARITH_SYMB (lexeme lexbuf) }
  | blank | comment                     { token lexbuf }


