(* $Id: dpll_core.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* DPLL solver state and actions. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Ternary
open Dpll_literal

(* Type of clauses. *)
type clause = {
  literals: Literal.t array;  (* disjunction of literals *)
  mutable activity: float (* relative usefulness of the clause *)
}

(* Type of values used to describe the current state of a literal:
   either that it is unassigned, or why it is assigned. *)
type reason = 
  | Unassigned                 (* an unassigned literal *)
  | Decision                   (* an arbitrarily decided literal *)
  | Propagation of clause      (* a literal inferred from the clause *)
  | External                   (* a literal inferred by an external agent *)

(* State of the solver. *)
type solver = {
  vars: Variable.generator; (* source of fresh dpll variables *)

  (* Clause database. *)

  clause: clause Vec.t;  (* user specified clauses *)
  learned: clause Vec.t; (* learned clauses *)

  (* Current partial model. *)

  trace: Literal.t Vec.t;
  (* The current (partial) model is stored in trace as a sequence of
     literals. *)

  assignment: ternary Variable_map.t;
  (* A variable is assigned 'truth' when it occurs positively in the
     trace, 'falsity' when it occurs negatively, and 'unknown'
     otherwise. *)

  reason: reason Variable_map.t;
  (* The reason map records why a variable was assigned. The reason is :
     - Unassigned            - the variable does not occur in the trace
     - Decision              - the variable was arbitrarily assigned
     - Propagation(cls)      - the variable assignment was inferred (from cls)
     - External              - the variable was assigned by an external agent
  *)

  level: int Variable_map.t;
  (* The level of a variable is the number of variables assigned
     earlier in the trace by Decision. The level is meaningful only
     for assigned variables. *)

  mutable decision_level: int;
  (* The current decision level. This is the total number of variables
     in trace assigned by Decision. *)

  decision_queue: Variable_queue.t;
  (* The decision queue must contain all Unassigned variables. It may
     also contain assigned variables, if we choose to be lazy in
     removing them. *)

  bias: bool Variable_map.t;
  (* The bias for a variable is a hint for which literal to pick when
     making an arbitrary decision for a variable. *)

  (* Conflict information. *)  

  mutable conflict_detected: bool;
  (* Has a conflict been detected in the solver? The solver is
     conflicting when the current partial model cannot be extended to
     a satisfying assignment.  A conflict is detected when all
     literals in a clause are assigned false.  *)

  mutable conflict_level: int;
  (* The level at which the conflict occurs. The conflict_level is
     valid only when conflict_detected, and is the maximum of the
     assignment levels of the conflict_vars -- it can be no larger
     than decision_level. In most cases, conflict_level will be
     decision_level. However, if a clause is added during search, then
     it might be in falsified by assignments prior to the current
     decision_level. *)

  conflict_vars: Variable_set.t;
  (* The set of conflict variables, i.e., a set of variables whose
     current assignment cannot be extended to a satisfying assignment
     (irrespective of the assignments to the non conflict
     variables). The corresponding literals are called conflict
     literals. When conflict_detected is false, conflict_vars should
     be empty. *)

  mutable num_implication_points: int;
  (* The number of implication_points. An implication point is a
     conflict literal assigned at the current conflict level. When
     conflict_detected is false, implication_points should be 0.*)

  mutable lower_conflict_lits: Literal.t list;
  (* A list of conflict literals that are not implication points, and
     therefore assigned below the current conflict level. *)

  (* Clause watching information. *)

  (* For unit propagation and clause detection, every clause watches
     two of its literals which are kept at positions 0 and 1. A
     necessary condition for unit propagation is that one of the two
     watched literals is assigned false. A necessary condition for
     conflict detection is that both watched literals are assigned
     false. *)

  watchers: (clause Vec.t) Literal_map.t;
  (* Watcher for a literal is a list of all clauses watching the
     literal. Every clause in the list will have the literal at
     position 0 or 1. *)

  mutable next_inference_lit: int;
  (* Index into trace for the (negation of the) next literal to use
     for unit propagation and conflict detection. If lit occurs in the
     trace prior to next_inference_lit then all the watcher clauses of
     (negation lit) are satisfied; the other watched literal is
     true. *)

  mutable next_inference_cls: int;
  (* Index into the watcher list of next inference literal for the
     next clause to check for unit propagation or conflict. If lit
     occurs at next_inference_lit in trace and cls occurs in the
     watcher list of (negation lit) prior to next_inference_cls, then
     cls is satisfied. *)

  (* Dispatch information. *)

  mutable interface_variables: Variable_set.t;
  (* Variables about which literals are dispatched. *)

  mutable dispatch_lit: int;
  (* Index into the trace of the literal to dispatch to external
     observers. *)
  
  (* Late clause addition related book-keeping. *)

  pending_clauses: (Literal.t list) Queue.t; 
  (* Clauses pending initial processing. These clauses were added when
     the solver was in a conflict state; they must be processed once
     the conflict is resolved. *)
  
  early_reason: clause Variable_map.t;
  (* The early_reason map records a clause that would have implied a
     variable assignment earlier, had the clause been added
     earlier. *)

  early_level: int Variable_map.t;
  (* The level at which a variable could have been assigned, had the
     early_reason clause for the variable been added earlier. *)

  pending_literals: Literal.t Queue.t;
  (* Literals pending addition to the trace. These are literals that
     have an early_reason; i.e. they could have been assigned earlier
     than they actually were. When the assignment is undone, these
     literals should be reassigned if the level is greater than or
     equal to early_level. *)

  (* Search tuning parameters. *)

  mutable var_increment: float;         (* increment of active variables *)
  mutable cls_increment: float;         (* increment of clauses used *)

  (* Solver statistics. *)

  mutable num_decisions: int;           (* total number of decisions made *)
  mutable num_conflicts: int;           (* total number of conflicts *)
  mutable num_inferences: int;          (* total number of dpll inferences *)
  mutable num_externally_set: int;      (* total number of external sets *)
}

(*****************************************************************************)
(* Usage Statistics *)

let num_vars s              = Variable.count s.vars
let num_clauses s           = Vec.length s.clause
let num_learneds s          = Vec.length s.learned
let num_decisions s         = s.num_decisions
let num_conflicts s         = s.num_conflicts
let num_inferences s        = s.num_inferences
let num_externally_set s    = s.num_externally_set


(*****************************************************************************)
(* Solver Status *)

let unit_propagation_possible s =
  s.next_inference_lit != Vec.length s.trace
  
let inference_possible s = 
  unit_propagation_possible s
  || not (Queue.is_empty s.pending_literals) 
  || not (Queue.is_empty s.pending_clauses)

let decision_level s =
  s.decision_level

let decision_possible s =
  Vec.length s.trace < Variable.count s.vars

let conflict_detected s = 
  s.conflict_detected

let conflict_level s =
  s.conflict_level

let backjump_possible s =
  s.conflict_detected && s.conflict_level > 0 && s.num_implication_points = 1

let unsatisfiable s = 
  s.conflict_detected && s.conflict_level = 0

let satisfied s = 
  not (conflict_detected s || decision_possible s || inference_possible s)


(*****************************************************************************)
(* Literal Inspection *)

let level s lit = Variable_map.find s.level (variable lit)
let reason s lit = Variable_map.find s.reason (variable lit)
let early_level s lit = Variable_map.find s.early_level (variable lit)
let early_reason s lit = Variable_map.find s.early_reason (variable lit)
let is_conflict_lit s lit = Variable_set.mem s.conflict_vars (variable lit)
let is_implication_point s lit = 
  is_conflict_lit s lit && level s lit = s.conflict_level

let is_assigned_var s var = Variable_map.find s.assignment var != unknown

let assignment s lit = 
  let var = variable lit in
  let var_val = Variable_map.find s.assignment var in
    if is_positive lit then var_val else Ternary.negation var_val

let is_true s lit = 
  assignment s lit  = truth

let is_not_true s lit = 
  assignment s lit != truth

let is_false s lit  = 
  let result = (assignment s lit = falsity) in
    assert (result = is_true s (negation lit));
    result

let is_not_false s lit = 
  let result = (assignment s lit != falsity) in
    assert (result = is_not_true s (negation lit));
    result

let is_assigned s lit =
  let result = reason s lit != Unassigned in
    assert (result = is_true s lit || is_false s lit);
    result

let is_decided s lit = 
  assert (is_true s lit);
  reason s lit = Decision

let is_inferred s lit = 
  assert (is_true s lit);
  match reason s lit with
    | Propagation _ -> true
    | _ -> false

let invalid_level = -1

(* Only used in assertions. If a literal is assigned, then its level
   must be valid and no greater than the current decision_level. *)
let level_valid s lit =
  not (is_assigned s lit) 
  || 0 <= level s lit && level s lit <= s.decision_level

(*****************************************************************************)
(* Performance Tuning Constants *) 

let scale = 1.05 (* scale factor for cls_increment and var_increment *)
let shy_of_max = max_float /. scale (* within one scale of overflow *)

(*****************************************************************************)
(* Solver state invariants *)

(* Only used in assertions. *)
let invariant s =
  (* There is a non-negative number of variables. *)
  (Variable.count s.vars >= 0)
  && (* The decision level is non-negative. *)
    (s.decision_level >= 0)
  && (* At least as many literals have been assigned as have been decided. *)
    (Vec.length s.trace >= s.decision_level)
  && (* All variables are assigned or in the still decision queue, or both. *)
    (Vec.length s.trace + Variable_queue.length s.decision_queue >= 
       Variable.count s.vars)
  && (* All literals, including the top, in the trace are assigned true. *)
    (Vec.length s.trace = 0 || is_true s (Vec.top s.trace))
  && (* next_inference_lit is valid trace index except during a conflict *)
    (s.conflict_detected || Vec.length s.trace >= s.next_inference_lit) 
  && (* No decision made without first exhausting inference, which implies: *) 
    (s.next_inference_lit + 1 >= s.decision_level)
  && (* Next literal to infer from always comes from current decision level. *)
    (s.conflict_detected || not (unit_propagation_possible s) ||
       level s (Vec.get s.trace s.next_inference_lit) = s.decision_level)
  && (* The number of implication points is non-negative *)
    (s.num_implication_points >= 0)
  && (* There are zero implication points unless a conflict is detected. *)
    (s.conflict_detected || s.num_implication_points = 0)
  && (* No conflict literals are below conflict_level when not in conflict. *)
    (s.conflict_detected || s.lower_conflict_lits = [])
  && (* Conflict level may not exceed decision level during a conflict. *)
    (not s.conflict_detected || s.conflict_level <= s.decision_level)
  && (* Must be at least one implication point in a resolvable conflict. *)
    (not s.conflict_detected || s.conflict_level = 0 
	|| s.num_implication_points > 0)
  && (* Conflict and decision levels are made equal in resolvable conflicts. *)
    (not s.conflict_detected || s.conflict_level = 0
	|| s.decision_level = s.conflict_level)
  && (* All conflict literals are removed except the last and exterannly set. *)
    (not s.conflict_detected || s.conflict_level = 0 
      || s.num_implication_points = 1 || reason s (Vec.top s.trace) = External)
  && (* clause increment is non-negative and not too big. *)
    (0.0 < s.cls_increment && s.cls_increment <= shy_of_max)
  && (* variable increment is non-negative and not too big *)
    (0.0 < s.var_increment && s.var_increment <= shy_of_max)

(*****************************************************************************)
(* Model communication *)

let model s =
  assert (satisfied s);
  let m = Vec.to_list s.trace in
    assert (satisfied s);
    assert (List.length m = num_vars s);
    m

(* Is there a valid, undispatched, interface literal? *)
let get_literal_possible s =
  while 
    (s.dispatch_lit != Vec.length s.trace)
    && not (Variable_set.mem s.interface_variables 
	      (variable (Vec.get s.trace s.dispatch_lit)))
  do
    s.dispatch_lit <- s.dispatch_lit + 1
  done;
  assert (invariant s);
  s.dispatch_lit != Vec.length s.trace

(* Dispatch a literal to an external observer. *)
let get_literal s = 
  assert (get_literal_possible s);
  let lit = Vec.get s.trace s.dispatch_lit in
    s.dispatch_lit <- s.dispatch_lit + 1;
    assert (invariant s);
    assert (is_true s lit);
    lit

(*****************************************************************************)
(* Reasons *)

(* If we find a clause, cls, with one literal, lit, unassigned, and all
   other literals assigned false, then:
   - lit is assigned true in order to satisfy cls.
   - The reason for lit is set to Propagation(cls).
   - Propagation(cls) cannot be the reason for any other literal.
   - We ensure that lit is at position 0 or 1 of cls.
*)

(* Only used in assertions.  Is lits a valid explanation for lit?
   There must be at least two literals, one of which must be lit, and all
   others must be assigned false. *)
let explanation_valid mem for_all s lit lits =
  let others_false l = is_false s l || l = lit in
    is_not_false s lit && mem lit lits && for_all others_false lits

let unit_propagating_clause s lit =
  match reason s lit with
    | Propagation cls ->
	assert (explanation_valid Array_util.mem Array_util.for_all 
		  s lit cls.literals);
	cls
    | _ -> assert (false)

(* Only used in assertions. *)
let reason_valid s lit r =
  match r with
    | Unassigned -> false
    | Decision -> true
    | External -> true
    | Propagation cls ->
        explanation_valid Array_util.mem Array_util.for_all s lit cls.literals

(*****************************************************************************)
(* Model Update *)

(* Make an assignment so that lit holds for the reason stated.  
   It would be correct, but slow, to now remove the corresponding variable
   from the decision queue. Instead, we remove assigned variables from the
   decision queue lazily (see decide). *)
let assign_lit s lit reason =
  assert (not s.conflict_detected);
  assert (not (is_assigned s lit));
  assert (reason_valid s lit reason);
  let var = variable lit in
  let var_val = ternary_of_bool (is_positive lit) in
    if reason = Decision then s.decision_level <- s.decision_level + 1;
    Variable_map.replace s.reason     var reason;
    Variable_map.replace s.assignment var var_val;
    Variable_map.replace s.level      var s.decision_level;
    (* Future: Make auto-biasing an option to the dpll solver. *)
    (*         Variable_map.replace s.bias var (is_positive lit); *)
    Vec.push s.trace lit

(* Unassign the last assigned literal in the current partial model.
   The variable from the discarded assignment must be added to the
   decision_queue. If the literal could have been assigned earlier
   than it was, then put it on pending_literals for possible
   re-assignment. *)
let unassign_last_lit s =
  assert (not (Vec.is_empty s.trace));
  let lit = Vec.pop s.trace in
  let var = variable lit in
  let reason = Variable_map.find s.reason var in
    assert (is_assigned s lit);
    assert (reason_valid s lit reason);
    assert (level s lit = s.decision_level);
    Variable_map.replace s.reason     var Unassigned;
    Variable_map.replace s.assignment var unknown;
    Variable_queue.enqueue s.decision_queue var;
    if reason = Decision then s.decision_level <- s.decision_level - 1;
    if early_level s lit != invalid_level then begin
      assert (early_level s lit < level s lit);
      Queue.push lit s.pending_literals;
    end

(* Backtrack to a given level, unassigning all variables assigned
   above that level. *)
let backtrack s lev =
  assert (s.decision_level > 0);
  assert (s.decision_level > lev);
  while (s.decision_level != lev) do
    unassign_last_lit s
  done;
  let n = Vec.length s.trace in
    assert (s.next_inference_lit >= n);
    s.next_inference_lit <- n;
    s.next_inference_cls <- 0;
    s.dispatch_lit <- min s.dispatch_lit n 
      
(*****************************************************************************)
(* Heuristic Performance Tuning *) 

let scale = 1.05 (* scale factor for cls_increment and var_increment *)
let shy_of_max = max_float /. scale (* within one scale of overflow *)

(* Rescale decision priorities to avoid arithmetic overflow. *)
let rescale_decision_priorities s = 
  Variable_queue.scale s.decision_queue min_float;
  s.var_increment <- s.var_increment *. min_float

(* Rescale clause activities to avoid arithmetic overflow. *)
let rescale_clause_activities s =
  let rescale c = c.activity <- c.activity *. min_float in
    Vec.iter rescale s.learned;
    s.cls_increment <- s.cls_increment *. min_float

(* Increase decision priority of a literal, rescale if necessary. *)
let inc_decision_priority s lit =
  let var = variable lit in
  let priority = Variable_queue.priority s.decision_queue var in
    if priority > max_float -. s.var_increment then 
      rescale_decision_priorities s;
    Variable_queue.inc_priority s.decision_queue var s.var_increment

(* Increase activity factor of a clause, rescale if necessary.  The
   activity factor is used to prioritize clauses in
   reduced_learned_clauses. *)
let inc_clause_activity s cls =
  if cls.activity > max_float -. s.cls_increment then 
    rescale_clause_activities s;
  cls.activity <- cls.activity +. s.cls_increment
  
(*****************************************************************************)
(* Conflict Management *)

let add_conflict_lit s lit =
  assert (s.conflict_detected);
  let lit_level = level s lit in
    assert (lit_level <= s.conflict_level);
    if not (is_conflict_lit s lit) then begin
      assert (is_false s lit);
      Variable_set.add s.conflict_vars (variable lit);
      inc_decision_priority s lit;
      if lit_level = s.conflict_level then
	s.num_implication_points <- s.num_implication_points + 1 
      else
	s.lower_conflict_lits <- lit::s.lower_conflict_lits
    end

let del_conflict_lit s lit =
  assert (s.conflict_detected); 
  assert (is_conflict_lit s lit);
  Variable_set.remove s.conflict_vars (variable lit)

let literal_to_explain s = 
  assert (conflict_detected s);
  assert (not (unsatisfiable s));
  assert (not (backjump_possible s));
  let lit = Vec.top s.trace in
    assert (is_implication_point s lit);
    assert (is_true s lit);
    assert (not (is_inferred s lit));
    assert (not (is_decided s lit));
    lit

(* Explain last literal on the trace with a generic container of literals. *)
let set_explanation iter s lits =
  let lit = Vec.top s.trace in
    iter (add_conflict_lit s) lits;
    del_conflict_lit s lit;
    s.num_implication_points <- s.num_implication_points - 1;
    unassign_last_lit s;
    assert (conflict_detected s)

let rec explain_inferences s =
  assert (conflict_detected s);
  assert (not (unsatisfiable s));
  while not (is_conflict_lit s (Vec.top s.trace)) do unassign_last_lit s done;
  if s.num_implication_points > 1 && is_inferred s (Vec.top s.trace) then begin
    let cls = unit_propagating_clause s (Vec.top s.trace) in
      inc_clause_activity s cls;
      set_explanation Array.iter s cls.literals;
      explain_inferences s
  end;
  assert (conflict_detected s);
  assert (not (unsatisfiable s));
  assert (is_conflict_lit s (Vec.top s.trace));
  assert
    ((s.num_implication_points = 1) || (not (is_inferred s (Vec.top s.trace))));
  assert
    ((s.num_implication_points = 1) || (not (is_decided s (Vec.top s.trace))))

(* Explain the external literal on the trace with a list of literals. *)
let set_literal_explanation s lits = 
  assert (conflict_detected s);
  assert (not (unsatisfiable s));
  assert (not (backjump_possible s));
  assert (not (is_decided s (literal_to_explain s)));
  assert (not (is_inferred s (literal_to_explain s)));
  assert 
    (explanation_valid List.mem List.for_all s (literal_to_explain s) lits);
  set_explanation List.iter s lits;
  explain_inferences s

(* Initialize a conflict with a generic container of conflict literals. *)
let set_conflict iter fold s lits =
  s.num_conflicts <- s.num_conflicts + 1;
  s.conflict_detected <- true;
  s.conflict_level <- fold (fun acc l -> max acc (level s l)) 0 lits;
  iter (add_conflict_lit s) lits;
  if s.conflict_level != 0 then explain_inferences s

(* Initialize a conflict with an array of literals for the conflict clause. *)
let set_conflict_lits s lits =
  assert (not s.conflict_detected);
  assert (Array_util.for_all (is_false s) lits);
  set_conflict Array.iter Array.fold_left s lits;
  assert (conflict_detected s)

(* Initialize a conflict with a list of literals for the conflict clause. *)
let set_conflict_clause s lits =
  assert (not s.conflict_detected);
  assert (List.for_all (is_false s) lits);
  set_conflict List.iter List.fold_left s lits;
  assert (conflict_detected s);
  assert (invariant s)
	
let clear_conflict s =
  assert (backjump_possible s);
  assert (s.decision_level = s.conflict_level);
  del_conflict_lit s (Vec.top s.trace);
  s.num_implication_points <- 0;
  List.iter (del_conflict_lit s) s.lower_conflict_lits;
  s.lower_conflict_lits <- [];
  s.conflict_detected <- false

(*****************************************************************************)
(* Clause creation *)

let new_clause lit_list = 
  {literals = Array.of_list lit_list; activity = 0.0}

let empty_clause = 
  new_clause []

let new_clause_vec () = 
  Vec.create 0 empty_clause

(*****************************************************************************)
(* Early Implications *)

(* Register an implication that could have been made earlier, if the
   clause had been added earlier. *)
let register_early_implication s lit lev cls = 
  assert (not s.conflict_detected);
  assert (is_true s lit);
  assert (lev < level s lit); 
  assert 
    (explanation_valid Array_util.mem Array_util.for_all s lit cls.literals);
  let old_early_level = early_level s lit in
    if old_early_level = invalid_level || lev < old_early_level then begin
      let var = variable lit in
        Variable_map.replace s.early_reason var cls;
        Variable_map.replace s.early_level  var lev;
    end;
    assert (not s.conflict_detected);
    assert (is_true s lit)

(* Remove early implication information once we've backtracked past the
   level where the implication could have been made. *)
let remove_early_implication s lit =
  assert (not s.conflict_detected);
  assert (early_level s lit > decision_level s);
  let var = variable lit in
    Variable_map.replace s.early_reason var empty_clause;
    Variable_map.replace s.early_level var invalid_level;
    assert (not s.conflict_detected)

(* Attempt to assign a literal based on the clause stored in
   early_reason. *)
let assign_early_implication s lit = 
  assert (not s.conflict_detected);
  let level = early_level s lit in
  let reason = early_reason s lit in
    assert (explanation_valid Array_util.mem Array_util.for_all 
	      s lit reason.literals);
    if level <= decision_level s then begin
      if is_false s lit then begin
        set_conflict_lits s reason.literals;
	inc_clause_activity s reason
      end
      else if not (is_assigned s lit) then
        assign_lit s lit (Propagation reason)
    end
    else begin
      remove_early_implication s lit
    end;
    assert (invariant s)

(*****************************************************************************)
(* Solver Creation, and Problem Setup. *)

(* A new empty solver. *)
let new_solver () =
  let s = 
    {
      vars = Variable.new_generator ();

      clause = new_clause_vec ();
      learned = new_clause_vec ();

      trace = Vec.create 0 Literal.invalid;
      assignment = Variable_map.create unknown;
      reason = Variable_map.create Unassigned;
      level = Variable_map.create invalid_level;
      decision_level = 0;
      decision_queue = Variable_queue.create Variable.invalid;
      bias = Variable_map.create true;

      conflict_detected = false;
      conflict_level = 0;
      conflict_vars = Variable_set.create ();
      num_implication_points = 0;
      lower_conflict_lits = [];

      interface_variables = Variable_set.create ();
      watchers = Literal_map.create (new_clause_vec ());
      next_inference_lit = 0;
      next_inference_cls = 0;

      pending_clauses = Queue.create ();
      early_reason = Variable_map.create empty_clause;
      early_level = Variable_map.create invalid_level;
      pending_literals = Queue.create ();
 
      dispatch_lit = 0;

      var_increment = 1.0;
      cls_increment = 1.0;

      num_decisions  = 0;
      num_conflicts  = 0;
      num_inferences = 0;
      num_externally_set = 0;
    }
  in
    assert (invariant s);
    assert (decision_level s = 0);
    assert (satisfied s);
    assert (not (get_literal_possible s));
    assert (not (conflict_detected s));
    s

(* A fresh variable. *)
let fresh_variable s =
  let v = Variable.fresh s.vars in
  let l = literal v in
    (* notify maps and queues about the new values they may see *)
    Variable_map.add s.assignment v unknown;
    Variable_map.add s.level v invalid_level;
    Variable_map.add s.reason v Unassigned;
    Variable_map.add s.early_reason v empty_clause;
    Variable_map.add s.early_level v invalid_level;
    Variable_map.add s.bias v true;
    (* initially nobody is watching either +ve or -ve literals *)
    Literal_map.add s.watchers l (new_clause_vec ()); 
    Literal_map.add s.watchers (negation l) (new_clause_vec ()); 
    (* all unassigned variables must be in the decision queue *)
    Variable_queue.enqueue s.decision_queue v;
    assert (invariant s);
    assert (not (satisfied s));
    v

(* Future work: Bias the decision heuristic. *)
let set_bias s lit = 
  Variable_map.replace s.bias (variable lit) (is_positive lit)

let remove_known_false s lits =
  let keep lit = not (level s lit = 0 && is_false s lit) in
    List.filter keep lits

(* Only used in assertions.
   
   It is invalid to watch a false literal unless the other literal is
   assigned true.  Further, the other literal must be assigned true at
   equal or lower decision level so that this property is preserved
   after backjumps.

   This property will be temporarily invalidated if a watched literal
   is assigned false, but will be re-established when unit propagation
   processes that assignment. *)
let watch_valid s cls =
  let earliest_level i = 
    if early_level s i != invalid_level then begin
      assert (early_level s i <= level s i);
      early_level s i 
    end
    else begin 
      level s i
    end
  in
  let ok i j =
    let lit = cls.literals.(i) in
    let other_lit = cls.literals.(j) in
      is_not_false s lit 
      || (is_true s other_lit && earliest_level other_lit <= level s lit)
  in
    ok 0 1 && ok 1 0

(* Add an interface variable of interest. *)
let add_interface_variable s var =
  (* Bad things happen if we make an assigned variable an interface
     variable and miss dispatching it. *)
  assert (not (is_assigned_var s var) || (s.dispatch_lit = 0));
  Variable_set.add s.interface_variables var;
  assert (invariant s)

(* Total order on literals, used when adding a clause.

   Literals are compared first by current literal assignment:
   literals assigned truth precede unassigned literals followed by
   literals assigned falsity.

   Then, the assignment level is used for assigned literals:
   literals assigned truth at a lower level precede those assigned later, 
   literals assigned falsity at a higher level precede those assigned earlier.

   Structural comparison is used to remove tie-breaks.
*)
let lit_compare s lit1 lit2 =
  let a1 = assignment s lit1 in
  let a2 = assignment s lit2 in
    if a1 = truth then begin
      if a2 = truth then begin
        let level1 = level s lit1 in
        let level2 = level s lit2 in
          if level1 = level2 then
            Pervasives.compare lit1 lit2
          else
            Pervasives.compare level1 level2
      end
      else
        -1 
    end
    else if a2 = truth then
      1
    else if a1 = unknown then begin
      if a2 = unknown then
        Pervasives.compare lit1 lit2
      else
        -1
    end
    else if a2 = unknown then
      1
    else begin
      let level1 = level s lit1 in
      let level2 = level s lit2 in
        if level1 = level2 then
          Pervasives.compare lit1 lit2
        else
          Pervasives.compare level2 level1
    end

(* Add a new clause to the solver database. *)
let add_clause s lit_list =
  if not (conflict_detected s) then begin
    let norm_lits = 
      remove_known_false s 
	(List_util.remove_sorted_duplicates (lit_compare s) lit_list)
    in 
    let cls = new_clause norm_lits in
    let lits = cls.literals in
    let len = Array.length lits in
      Array.iter (inc_decision_priority s) lits; 
      match len with
	| 0 -> 
	    (* All literals were false at level 0, there is a
	       contradiction. *)
	    set_conflict_lits s empty_clause.literals
	| 1 -> 
	    (* Only 1 literal was not false at level 0, infer it. *)
            let lit = lits.(0) in
              assert (is_not_false s lit || s.decision_level > 0);
              assert (level_valid s lit);
              if is_false s lit then begin
                set_conflict_lits s cls.literals;
		inc_clause_activity s cls
	      end
              else begin
                if not (is_assigned s lit) then
	          assign_lit s lit (Propagation cls);
                assert (is_true s lit);
                if level s lit > 0 then
                  register_early_implication s lit 0 cls
              end
	| _ -> 
	    (* At least 2 literals were not false at level 0, keep
	       the clause and watch it. *)
	    assert (len >= 2);
            let lit0 = lits.(0) in
            let lit1 = lits.(1) in
              assert (is_not_false s lit0 || is_false s lit1);
	      assert (is_not_false s lit1 || s.decision_level > 0);
              assert (level_valid s lit0);
              assert (level_valid s lit1);
	      Vec.push s.clause cls;
              Vec.push (Literal_map.find s.watchers lit0) cls;
              Vec.push (Literal_map.find s.watchers lit1) cls;
              if is_false s lit1 then begin
                if is_false s lit0 then begin
                  set_conflict_lits s cls.literals;
		  inc_clause_activity s cls
		end
                else begin
                  if not (is_assigned s lit0) then
                    assign_lit s lit0 (Propagation cls);
                  assert (is_true s lit0);
                  let lev0 = level s lit0 in
                  let lev1 = level s lit1 in
                    if lev0 > lev1 then
                      register_early_implication s lit0 lev1 cls
                end
              end;
	      assert (s.conflict_detected || watch_valid s cls)
  end
  else begin
    assert (conflict_detected s);
    Queue.push lit_list s.pending_clauses;
  end;
  assert (invariant s)

(*****************************************************************************)
(* Search and Inference. *)

(* Make an arbitrary assignment decision for the highest priority
   unassigned variable. *)
let decide s =
  assert (not (conflict_detected s));
  assert (not (inference_possible s));
  assert (decision_possible s);
  (* Remove assigned variables from the head of the queue, since we
     did not eagerly remove them when assigning the variables. *)
  while is_assigned_var s (Variable_queue.head s.decision_queue) do
    Variable_queue.drop s.decision_queue
  done;
  assert (not (Variable_queue.is_empty s.decision_queue));
  (* An unassigned variable found, decide it (to be true). *)
  let var = Variable_queue.dequeue s.decision_queue in
  let lit = 
    if Variable_map.find s.bias var then literal var 
    else negation (literal var) 
  in
    s.num_decisions <- s.num_decisions + 1;
    assign_lit s lit Decision;
    assert (inference_possible s);
    assert (invariant s);
    assert (decision_level s > 0);
    assert (not (conflict_detected s))

(* Only used in assertions. Does x occur in vector v from i up to (but
   not including) j.  *)
let rec subvec_mem x v i j =
  i < j && (Vec.get v i = x || subvec_mem x v (i+1) j)

(* Find a unit-propagation or a conflict, if one exists. *)
let rec unit_propagate_or_conflict s lit watcher = 
  assert (not (conflict_detected s));
  assert (inference_possible s);
  assert (is_false s lit);
  if s.next_inference_cls < Vec.length watcher then begin
    (* is the next clause in watcher unit or conflicting? *)
    let cls = Vec.get watcher s.next_inference_cls in
    let lits = cls.literals in
      assert (Array.length lits >= 2);
      (* lit is watched by cls: it must occur at position 0 or 1 of lits. *)
      assert (lits.(0) = lit || lits.(1) = lit);
      let (lit_idx, other_idx) = if lits.(0) = lit then (0,1) else (1,0) in
      let other_lit = lits.(other_idx) in
      let other_assign = assignment s other_lit in
        if other_assign = truth then begin
	  (* Other watched literal is true, so clause is satisfied and
	     nothing may be inferred from it. Try next clause. *)
	  s.next_inference_cls <- s.next_inference_cls + 1;
	  unit_propagate_or_conflict s lit watcher
        end
        else begin
          assert (other_assign != truth);
	  let len = Array.length lits in
	  let i = Subarray.first_const is_not_false s lits 2 len in
	    if i != len then begin
	      (* Another non-false literal found, no inference
		 possible while more than one literal is not
		 false. Watch for the negation of that literal now
		 rather than lit. *)
	      lits.(lit_idx) <- lits.(i);
	      lits.(i) <- lit;
	      Vec.swap_out watcher s.next_inference_cls;
	      Vec.push (Literal_map.find s.watchers lits.(lit_idx)) cls;
	      assert (watch_valid s cls 
		      || (subvec_mem (negation other_lit) s.trace 
			    (s.next_inference_lit+1) (Vec.length s.trace)));
	      (* Keep looking. *)
	      unit_propagate_or_conflict s lit watcher
	    end
	    else if other_assign = unknown then begin
	      (* The other watched literal is unassigned, but all
		 others are false. Infer that the other watched
		 literal must be true. *)
	      assign_lit s lits.(other_idx) (Propagation cls);
	      s.num_inferences <- s.num_inferences + 1;
	      s.next_inference_cls <- s.next_inference_cls + 1;
	    end
	    else begin
	      (* All literals of the clause are false, so it is a
		 contradiction. *)
	      set_conflict_lits s cls.literals;
	      inc_clause_activity s cls
            end
        end
  end
  else begin
    assert (s.next_inference_cls = Vec.length watcher);
    (* No more clauses to visit in watcher. Continue with remaining
       literals in trace. *)
    s.next_inference_lit <- s.next_inference_lit + 1;
    s.next_inference_cls <- 0;
    if s.next_inference_lit < Vec.length s.trace then begin
      let lit = negation (Vec.get s.trace s.next_inference_lit) in
      let watcher = Literal_map.find s.watchers lit in
        unit_propagate_or_conflict s lit watcher
    end
  end


(* Attempt to infer a unit propagation or a contradiction. *)
let infer s = 
  assert (not (conflict_detected s));
  assert (inference_possible s);
  if not (Queue.is_empty s.pending_literals) then
    assign_early_implication s (Queue.take s.pending_literals)
  else if not (Queue.is_empty s.pending_clauses) then
    add_clause s (Queue.take s.pending_clauses)
  else begin
    assert (s.next_inference_lit < Vec.length s.trace);
    let lit = negation (Vec.get s.trace s.next_inference_lit) in
    let watcher = Literal_map.find s.watchers lit in
      unit_propagate_or_conflict s lit watcher;
  end;
  assert (invariant s)

(* Assign a literal inferred by an external agent. *)
let set_literal s lit = 
  assert (not (conflict_detected s));
  if not (is_assigned s lit) then begin
    assign_lit s lit External;
    s.num_externally_set <- s.num_externally_set + 1;
  end;
  (* Note: We ignore literals that conflict with DPLL assignments.  It
     is left to the external agent to detect the conflict (when it
     learns of the DPLL assignment through get_literal). *)
  assert (invariant s);
  assert (not (conflict_detected s))

(*****************************************************************************)
(* Conflict Resolution. *)

let backjump s =
  assert (backjump_possible s);
  assert (decision_level s = conflict_level s);
  let lit = Vec.top s.trace in
    assert (is_implication_point s lit);
    let learn_lits = (negation lit)::s.lower_conflict_lits in
    let norm_learn_lits = remove_known_false s learn_lits in
    let cls = new_clause norm_learn_lits in
    let lits = cls.literals in 
    let n = Array.length lits in
      if n >= 2 then begin
	(* learn from the conflict *)
	let lit_level_compare l0 l1 = 
	  Pervasives.compare (level s l1) (level s l0)
	in
          Array.sort lit_level_compare lits;
	  (* Sorting brings (negation lit) to position 0 as it is the
	     unique element assigned at the current level. 
	     
	     We want to backtrack to the next highest level at which a
	     literal was assigned; now the level of the literal at
	     position 1.  

	     This is sufficient to establish watch_valid after the
	     backjump, sorting the literals beyond this is not
	     necessary, but seems to help performance by establishing
	     a good initial order of literals. *)
          assert (lits.(0) = negation lit);
          assert (level s lits.(0) > level s lits.(1));
	  (* learn the clause, and watch it *)
          Vec.push s.learned cls;
          Vec.push (Literal_map.find s.watchers cls.literals.(0)) cls;
          Vec.push (Literal_map.find s.watchers cls.literals.(1)) cls;
          (* backjump *)
  	  clear_conflict s;
	  backtrack s (level s lits.(1));
      end
      else begin 
	assert (Array.length lits = 1);
        (* no clause to learn; just backjump *)
	clear_conflict s;
	backtrack s 0;
      end;
      assert (lits.(0) = negation lit);
      assign_lit s lits.(0) (Propagation cls);
      assert (n = 1 || watch_valid s cls);
      if s.var_increment > shy_of_max then rescale_decision_priorities s;
      s.var_increment <- s.var_increment *. scale;
      if s.cls_increment > shy_of_max then rescale_clause_activities s;
      s.cls_increment <- s.cls_increment *. scale;
      assert (decision_level s < conflict_level s);
      assert (not (conflict_detected s));
      assert (inference_possible s);
      assert (invariant s)


(*****************************************************************************)
(* Forget and Restart. *)

(* Forget a clause (stop watching it). This is safe for a clause if it is
   entailed by the remaining clauses, in particular :
   - all learned clauses.
   - original clauses that we know to be satisfied at decision level 0. *)
let forget s cls =
  let w0 = Literal_map.find s.watchers cls.literals.(0) in
  let w1 = Literal_map.find s.watchers cls.literals.(1) in
    (* Equally valid to use (=) here.  If there are duplicate clauses then
       (=) may remove a duplicate rather than the clause given.  This may
       change the order of clauses in watched sets, with an observable
       difference in runtime. *)
    Vec.swap_out w0 (Vec.first_const (==) cls w0);
    Vec.swap_out w1 (Vec.first_const (==) cls w1)

(* Auxiliary function to prioritize learned clauses when reducing learned
   clauses.  Two literal clauses are all equally, and extremely,
   valuable. All other clauses are ranked by activity factor. *)
let cls_compare cls0 cls1 =
  let len0 = Array.length cls0.literals in
  let len1 = Array.length cls1.literals in
    assert (len0 >= 2);
    assert (len1 >= 2);
    if len0 = 2 then (if len1 = 2 then 0 else (-1))
    else if len1 = 2 then 1
    else
      let n = Pervasives.compare cls1.activity cls0.activity in
        if n = 0 then Pervasives.compare len0 len1 else n

(* Forget some, at most half, of the least valuable learned clauses. *)
let forget_some_learned_clauses s =
  assert (not (conflict_detected s));
  Vec.sort cls_compare s.learned;
  let half = (Vec.length s.learned) / 2 in
  let keep i cls =
    i < half || Array.length cls.literals = 2 ||
    (forget s cls; false)
  in
    Vec.filteri keep s.learned;
    assert (invariant s);
    assert (not (conflict_detected s))

(* Forget clauses that are satisfied at level 0, they have no deductive
   value. *)
let simplify s = 
  assert (not (conflict_detected s));
  assert (s.decision_level = 0);
  let keep cls =
    Subarray.for_all_const is_not_true s cls.literals 
      0 (Array.length cls.literals) 
    || (forget s cls; false)
  in
    Vec.filter keep s.clause;
    Vec.filter keep s.learned;
    assert (s.decision_level = 0);
    assert (not (conflict_detected s))

(* Restart the proof search. Perhaps with the newly learned clauses it
   will take a better path next time. *)
let restart s = 
  assert (not (conflict_detected s));
  if s.decision_level > 0 then backtrack s 0;
  simplify s;
  assert (invariant s);
  assert (not (conflict_detected s));
  assert (s.decision_level = 0)
