(* $Id: dpll_core.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** State and actions of a MiniSat like DPLL solver. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** A low level interface to a DPLL solver with an implementation
    following the algorithm presented in the following paper:

    - Niklas Een, Niklas Sorensson.
    {{:http://dx.doi.org/10.1007/b95238} An extensible SAT-solver}.
    {i Lecture Notes in Computer Science} 2919:502-518, 2004. *)

(** {2 Solver Creation, Problem Setup and Solution} *)

open Dpll_literal

(** Type of DPLL-based SAT solver instances. *)
type solver

val new_solver: unit -> solver
  (** New DPLL solver. 

      - Ensure {!satisfied}[ (new_solver ())] 
      - Ensure {!decision_level}[ (new_solver ()) = 0] 
      - Ensure [not (]{!conflict_detected}[ (new_solver ()))] *)

val fresh_variable: solver -> Variable.t
  (** Fresh variable in the given solver.  

      - Ensure [not (]{!satisfied}[ s)] *)

val add_clause: solver -> Literal.t list -> unit
  (** Adds the clause (disjunction of literals) to the solver. *)

val model: solver -> Literal.t list
  (** A model satisfying the problem.

      - Require {!satisfied}[ s]
      - Ensure {!satisfied}[ s]
      - Ensure [List.length (]{!model}[ s) = ]{!num_vars}[ s] *)

val add_interface_variable: solver -> Variable.t -> unit
  (** [add_interface_variables s v] treats the variable [v] and its
      negation as an interface literal. *)

(** {2 Solver Status} *)

val decision_level: solver -> int
  (** The number of variables assigned arbitrarily by {!decide} in the
      current solver state.

      - Ensure [0 <= ]{!decision_level}[ s] *)

val inference_possible: solver -> bool
  (** Might an inference be made? *)
  
val get_literal_possible: solver -> bool
  (** Is there an interface literal in the current partial model that
      has not been dispatched by {!get_literal} since it was
      assigned? *)

val conflict_detected: solver -> bool
  (** Has the solver deduced that the current (partial) model cannot
      be (extended to) a satisfying variable assignment? 
      
      If a conflict is detected, then either the problem is
      unsatisfiable, or it may be resolved by a backjump, or there is
      a literal assigned using set_literal that must be explained. *)

val decision_possible: solver -> bool
  (** Might further arbitrary variable assignment decisions be
      attempted? *)

val backjump_possible: solver -> bool
  (** Might a backjump be used to resolve a conflict?
      
      - Ensure {!conflict_detected}[ s || not(]
      {!backjump_possible}[ s)]
      - Ensure [not (]{!backjump_possible}[ s && ]
      {!unsatisfiable}[ s)] *)

val satisfied: solver -> bool
  (** Has the solver deduced that the current model is a satisfying
      assignment for every variable?

      - Ensure 
      [satisfied s = not (]
      {!conflict_detected}[ s || ]{!decision_possible}[s || ]
      {!inference_possible}[ s)] *)

val unsatisfiable: solver -> bool
  (** Has the solver deduced that the problem is unsatisfiable?

      - Ensure [not (unsatisfiable s) || (]{!conflict_detected}[ s)] *)

(** {2 Literal Status} *)

(** Functions to access the status of a literal in the current solver
    state. A literal is either unassigned or assigned true or assigned
    false. If a literal is assigned a value (true or false), it is
    either a decision literal, or inferred by unit propagation or else
    externally set. *)

val is_true: solver -> Literal.t -> bool
  (** Is the literal assigned true? *)

val is_decided: solver -> Literal.t -> bool
  (** Has the literal been assigned by arbitrary decision? *)

val is_inferred: solver -> Literal.t -> bool
  (** Has the literal been inferred by unit propagation? *)

(** {2 Search and Inference} *)

val decide: solver -> unit
  (** Makes an arbitrary assignment to an unassigned variable. 

      - Require [not (]{!conflict_detected}[ s)] 
      - Require [not (]{!inference_possible}[ s)]
      - Require {!decision_possible}[ s]
      - Ensure {!decision_level}[ s > 0]
      - Ensure [not (]{!conflict_detected}[ s)] *)

val infer: solver -> unit
  (** Attempt to infer a unit propagation or a conflict.
      
      - Require [not (]{!conflict_detected}[ s)]
      - Require {!inference_possible}[ s] *)

val set_literal: solver -> Literal.t -> unit
  (** Attempt to make a variable assignment inferred by an external
      agent.  The interface literal is ignored if the corresponding
      variable is already assigned in the current partial model.

      - Require [not (]{!conflict_detected}[ s)] *)

val get_literal: solver -> Literal.t
  (** Choose an interface literal assigned true in the current partial
      model that has not been chosen since it was assigned.

      - Require {!get_literal_possible}[ s] 
      - Ensure {!is_true}[ (get_literal s)] *)

(** {2 Conflict Resolution} *)

val set_conflict_clause: solver -> Literal.t list -> unit
  (** [set_conflict_clause ls] initializes the conflict clause with
      the disjoined literals.

      - Require [not (]{!conflict_detected}[ s)]
      - Require 
      {!List.for_all} 
      [(fun l -> ]{!is_true}[ s (]{!Dpll_literal.negation}[ l)) ls]
      - Ensure {!conflict_detected}[ s] *)

val literal_to_explain: solver -> Literal.t
  (** Returns the next conflict literal to explain for conflict
      resolution.
      
      - Require {!conflict_detected}[ s]
      - Require [not (]{!unsatisfiable}[ s)]
      - Require [not (]{!backjump_possible}[ s)]
      - Ensure {!is_true}[ s (literal_to_explain s)]
      - Ensure [not (]{!is_inferred}[ s (literal_to_explain s))]
      - Ensure [not (]{!is_decided}[ s (literal_to_explain s))] *) 

val set_literal_explanation: solver -> Literal.t list -> unit
  (** [set_literal_explanation s ls] removes {!literal_to_explain}
      from the set of conflict literals using the provided explanation
      [ls].

      - Require {!conflict_detected}[ s]
      - Require [not (]{!unsatisfiable}[ s)]
      - Require [not (]{!backjump_possible}[ s)]
      - Require {!List.mem}[ (]{!literal_to_explain}[ s) ls]
      - Require 
      {!List.for_all} 
      [(fun l -> 
      l = (]{!literal_to_explain}[ s) || is_true s (Dpll_literal.negation l)) 
      ls]
      - Ensure {!conflict_detected}[ s]
      - Ensure [not (]{!unsatisfiable}[ s)] *)

val backjump: solver -> unit
  (** Resolve a conflict by backjumping to undo conflicting
      assignments from the current partial model. A learned clause is
      added to the solver instance to prevent the same conflict later.

      - Require {!backjump_possible}[ s]
      - Ensure [not (]{!conflict_detected}[ s)] 
      - Ensure {!inference_possible}[ s] *)

(** {2 Search Optimization} *)

val forget_some_learned_clauses: solver -> unit
  (** Forget some clauses learned by the solver though conflict
      analysis.
      
      - Require [not (]{!conflict_detected}[ s)] 
      - Ensure [not (]{!conflict_detected}[ s)] *)

val restart: solver -> unit
  (** Undo all arbitrary variable assignment decisions.

      - Require [not (]{!conflict_detected}[ s)]
      - Ensure [not (]{!conflict_detected}[ s)] 
      - Ensure {!decision_level}[ s = 0] *)

(** {2 Usage Statistics} *)

val num_vars: solver -> int
  (** Number of variables created. *)

val num_clauses: solver -> int
  (** Number of added clauses that are tracked in current state. *)

val num_learneds: solver -> int
  (** Number of clauses learned through conflict analysis in current
      state. *)

val num_conflicts: solver -> int
  (** Number of conflicts encountered. *)

val num_decisions: solver -> int
  (** Number of arbitrary variable assignment ever tried. *)

val num_inferences: solver -> int
  (** Number of variable assignment ever made by unit propagation. *)

val num_externally_set: solver -> int
  (** Number of variable assignments ever made by {!set_literal}. *)
