(* $Id: solver_dpll.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* DPLL-based SAT Solver *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Log
open Dpll_core

class make (make_dpll: unit -> Solver_api.dpll) =
object (self)

  val dpll = make_dpll () 

  val log_start_level = 2
  val log_stats_level = 3

  val mutable conflict_limit = 100
  val mutable learneds_limit = 0

  method private invariant =
    (* The conflict limit is positive. *)
    (0 < conflict_limit) 
    && (* The learneds limit is non-negative. *)
      (0 <= learneds_limit)

  initializer
    assert (dpll#decision_level = 0);
    assert (not dpll#conflict_detected);
    assert (Term.Term_set.is_empty dpll#variables);
    assert self#invariant

  method private log_start = 
    log_string log_start_level "Variables: "; 
    log_int log_start_level dpll#num_vars; 
    log_newline log_start_level ();
    log_string log_start_level "Clauses: "; 
    log_int log_start_level dpll#num_clauses; 
    log_newline log_start_level ()

  method private log_stats =
    log_string log_stats_level  "Confs: "; 
    log_int log_stats_level dpll#num_conflicts;
    log_string log_stats_level " Decns: "; 
    log_int log_stats_level dpll#num_decisions;
    log_string log_stats_level " Infrs: "; 
    log_int log_stats_level dpll#num_inferences;
    log_string log_stats_level " Lrnts: "; 
    log_int log_stats_level dpll#num_learneds;
    log_newline log_stats_level ()

  method add_clause cls =
    dpll#add_clause cls;
    assert self#invariant

  method private solved =
    dpll#satisfied || dpll#unsatisfiable

  method private search =
    while not self#solved && dpll#num_conflicts < conflict_limit do
      while not dpll#conflict_detected && dpll#inference_possible do
        dpll#infer
      done;
      if dpll#conflict_detected then begin
        if not dpll#unsatisfiable then begin
          dpll#backjump;
          if dpll#num_learneds > learneds_limit then
            dpll#forget_some_learned_clauses;
	end
      end
      else if dpll#decision_possible then
        dpll#decide
      else
	assert dpll#satisfied
    done;
    assert (not dpll#conflict_detected || dpll#unsatisfiable);
    assert self#invariant

  method solve = 
    learneds_limit <- max 1000 (dpll#num_clauses / 3);
    self#log_start;
    while not self#solved do
      dpll#restart;
      self#search;
      conflict_limit <- conflict_limit * 2;
      learneds_limit <- learneds_limit * 12 / 10;
      self#log_stats;
    done;
    assert self#invariant;
    assert self#solved;
    dpll#satisfied

end
