(* $Id: solver_dpllt.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* A DPLL(T) solver *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Term
open Log

(* A Dpllt solver from a dpll solver and a theory solver *)
class make 
  (make_dpll: unit -> Solver_api.dpll) 
  (make_theory: unit -> Solver_api.theory) =
object (self)
  val dpll = make_dpll ()
  val theory = make_theory ()

  val log_start_level = 2
  val log_stats_level = 3

  val mutable conflict_limit = 100
  val mutable learneds_limit = 0

  val uninterpret_tbl = Term_hashtbl.create 0
  val homogenization_tbl = Term_hashtbl.create 0

  method private invariant =
    (* The conflict limit is positive. *)
    (0 < conflict_limit)
    && (* The learneds limit is non-negative. *)
      (0 <= learneds_limit)
    && (* The decision levels of the solvers are syncronized. *)
      (dpll#decision_level = theory#decision_level)
    && (* Any theory conflict is immediately propagated to dpll. *)
      (not theory#conflict_detected) || dpll#conflict_detected

  initializer    
    assert (dpll#decision_level = 0);
    assert (not dpll#conflict_detected);
    assert (Term_set.is_empty dpll#variables);
    assert (theory#decision_level = 0);
    assert (not theory#conflict_detected);
    assert (Term_set.is_empty theory#variables);
    assert self#invariant

  method private log_start =
    log_string log_start_level "Variables: "; 
    log_int log_start_level dpll#num_vars; 
    log_newline log_start_level ();
    log_string log_start_level "Clauses: "; 
    log_int log_start_level dpll#num_clauses;
    log_newline log_start_level ()
    
  method private log_stats =
    log_string log_stats_level  "Confs: "; 
    log_int log_stats_level dpll#num_conflicts;
    log_string log_stats_level " Decns: "; 
    log_int log_stats_level dpll#num_decisions;
    log_string log_stats_level " Infrs: "; 
    log_int log_stats_level dpll#num_inferences;
    log_string log_stats_level " TPrps: "; 
    log_int log_stats_level dpll#num_externally_set;
    log_string log_stats_level " Lrnts: "; 
    log_int log_stats_level dpll#num_learneds;
    log_newline log_stats_level ()

  (* Replace "uninterpreted constants" by fresh variables. *)
  method private uninterpret term =
    try Term_hashtbl.find uninterpret_tbl term
    with Not_found ->
      let uninterpreted_term = match term with
        | Variable _ -> 
            term
        | Constant _ ->
            if dpll#owns_constant term || theory#owns_constant term then term
            else variable (Variable.fresh ())
        | Application (_, f, x) ->
            let f' = self#uninterpret f in
            let x' = self#uninterpret x in
              application f' x'
        | _ -> 
            assert false
      in
        Term_hashtbl.add uninterpret_tbl term uninterpreted_term;
        uninterpreted_term

  (* Purifies a term by replacing alien subterms with their proxies. *)
  method private purify term = 
    let rec proxy = function
      | (0, term) -> dpll#proxy term
      | (_, term) -> theory#proxy term 
    and homogenize term = 
      try Term_hashtbl.find homogenization_tbl term
      with Not_found -> 
        let purified_term = match term with
          | Constant _ ->
              let index = if dpll#owns_constant term then 0 else 1 in
                (index, term)
          | Application (_, f, x) ->
              let (f_solver_index, f') = 
                if is_variable f then (1, f) 
                else homogenize f
              in
              let (x_solver_index, x') as x_i =
                if is_variable x then (f_solver_index, x) 
                else homogenize x
              in
              let x'' = 
                if x_solver_index = f_solver_index then x'
                else proxy x_i
              in
                (f_solver_index, application f' x'')
          | _ -> 
              assert false
        in
          Term_hashtbl.add homogenization_tbl term purified_term;
          purified_term
    in
      if is_variable term then term
      else proxy (homogenize term)

  method add_clause cls = 
    let cls'  = List.map self#uninterpret cls  in
    let cls'' = List.map self#purify cls' in
      dpll#add_clause cls'';
      assert self#invariant

  method private conflict_detected =
    dpll#conflict_detected || theory#conflict_detected

  method private decide =
    assert (not self#conflict_detected);
    assert dpll#decision_possible;
    dpll#decide;
    theory#increment_level;
    assert self#invariant;
    assert (dpll#decision_level > 0);
    assert (not self#conflict_detected)

  method private literal_dispatch =
    assert (not self#conflict_detected);
    assert dpll#get_literal_possible;
    theory#set_literal (dpll#get_literal);
    assert self#invariant;
    assert (not dpll#conflict_detected)

  method private literal_propagate =
    assert (not self#conflict_detected);
    assert theory#get_literal_possible;
    dpll#set_literal (theory#get_literal);
    assert self#invariant;
    assert (not self#conflict_detected)
    
  method private theory_conflict =
    assert (not dpll#conflict_detected);
    assert theory#conflict_detected;
    dpll#set_conflict_clause (theory#conflict_clause);
    assert self#invariant;
    assert dpll#conflict_detected;
    assert theory#conflict_detected

  method private theory_conflict_or_literal_propagate_star =
    assert (not dpll#conflict_detected);
    if theory#conflict_detected then begin
      self#theory_conflict
    end
    else while theory#get_literal_possible do
      self#literal_propagate;
    done;
    assert self#invariant
    
  method private literal_exchange =
    while not (dpll#conflict_detected) && dpll#get_literal_possible do
      self#literal_dispatch;
      self#theory_conflict_or_literal_propagate_star
    done;
    assert self#invariant
    
  method private resolve_conflict =
    assert dpll#conflict_detected;
    assert (dpll#decision_level = theory#decision_level);
    if not dpll#unsatisfiable then begin
      while not dpll#backjump_possible do
        dpll#set_literal_explanation 
	  (theory#literal_explanation dpll#literal_to_explain)
      done;
      dpll#backjump;
      theory#backjump dpll#decision_level;
      assert (not dpll#conflict_detected);
    end;
    assert self#invariant;
    assert (not dpll#conflict_detected || dpll#unsatisfiable)

  method private restart =
    assert (not self#conflict_detected);
    dpll#restart;
    theory#restart;
    assert self#invariant;
    assert (not self#conflict_detected);
    assert (dpll#decision_level = 0)
      
  method private add_interface_variables =
    let interface_variables = Term_set.inter dpll#variables theory#variables in
      Term_set.iter dpll#add_interface_variable   interface_variables;
      Term_set.iter theory#add_interface_variable interface_variables

  method private transition_possible =
    dpll#decision_possible 
    || dpll#inference_possible 
    || theory#inference_possible 
    || dpll#get_literal_possible
    || theory#get_literal_possible

  method private satisfied = 
    let sat = dpll#satisfied 
	&& not dpll#get_literal_possible 
	&& not theory#inference_possible
    in 
      assert (sat = not (self#conflict_detected || self#transition_possible));
      sat

  method private solved =
    dpll#unsatisfiable 
    || self#satisfied 

  method private search =
    while not self#solved && dpll#num_conflicts < conflict_limit do
      while not (dpll#conflict_detected) && dpll#inference_possible do
        dpll#infer;
        self#literal_exchange;
      done;
      assert (not theory#conflict_detected || dpll#conflict_detected);
      if dpll#conflict_detected then begin
        self#resolve_conflict;
        (* Dpll backjump also asserts a literal. Tell theory about that
           literal. *)
        self#literal_exchange;
        if dpll#num_learneds > learneds_limit then
          dpll#forget_some_learned_clauses;
      end
      else begin
        if theory#inference_possible then begin
          theory#infer;
          self#theory_conflict_or_literal_propagate_star;
        end
        else if dpll#decision_possible then begin
          self#decide;
          self#literal_exchange;
        end
        else
          assert dpll#satisfied
      end
    done;
    if dpll#conflict_detected then begin
      self#resolve_conflict;
      if dpll#num_learneds > learneds_limit then
        dpll#forget_some_learned_clauses;
    end;
    assert self#invariant

  method solve =
    learneds_limit <- max 1000 (dpll#num_clauses / 3);
    self#add_interface_variables;
    self#log_start;
    self#literal_exchange;
    while not self#solved do
      self#restart;
      self#search;
      conflict_limit <- conflict_limit * 2;
      learneds_limit <- learneds_limit * 12 / 10;
      self#log_stats;
    done;
    assert self#invariant;
    assert self#solved;
    dpll#satisfied

end
