(* $Id: constant.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** Constants. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Big_int
open Num

(** {2 Abstract constants and general operations on them.} *)

(** Constants *)
type t 

val equal: t -> t -> bool
  (** Are two constants equal? *)

val hash: t -> int
  (** Hash code of a constant. *)

val compare: t -> t -> int
  (** An arbitrary ordering on constants. *)

val arity: t -> int
  (** Arity of a constant.

      - Ensure [arity x >= 0] *)

val string_of: t -> string
  (** String representation of a constant. *)

val pp_print: Format.formatter -> t -> unit
  (** Print a constant. *)

val log: int -> t -> unit
  (** [log level c] is equivalent to 
      [Log.log level pp_print c] *)

(** {2 Symbolic constants.} *)

val symbolic: Symbol.t -> t
  (**  A symbolic constant. *)

val is_symbolic: t -> bool
  (**  Is a constant symbolic?. *)

val symbolic_value: t -> Symbol.t
  (** [symbolic_value c] is the symbol underlying a symbolic constant.

      - Require [is_symbolic c]
      - Ensure [symbolic_value (symbolic s) = s] *)

(** {2 Manifest integer constants.} *)

val integer: big_int -> t
  (**  Arbitrary precision integer constant. *)

val is_integer: t -> bool
  (** Is a constant an integer value? *)

val integer_value: t -> big_int
  (** [integer_value c] is the integer value of c.

      - Require [is_integer c] 
      - Ensure [integer_value (integer i) = i] *)

(** {2 Manifest rational constants.} *)

val rational: num -> t
  (**  Arbitrary precision rational constant. *)
 
val is_rational: t -> bool
  (** Is a constant a rational value *)

val rational_value: t -> num
  (** [rational_value c] is the rational value of c.

      - Require [is_rational c] 
      - Ensure [rational_value (rational r) = r] *)
