(* $Id: term.mli,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(** The structure of terms. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

(** {2 Term data-structures} *)

(** Hash keys for terms *)
type key

(** de Bruijn indices for bound variables *)
type db_index

(** Terms *)
type term = 
    private
  | Bound       of db_index                (** Bound variables *)
  | Variable    of key * Variable.t        (** Free variables *)
  | Constant    of key * Constant.t        (** Constants *)
  | Application of key * term * term       (** Function application *)
  | Abstraction of key * Variable.t * term (** Lambda abstraction *)

val term_equal: term -> term -> bool
  (** Are two terms equal? *)

val term_compare: term -> term -> int
  (** An arbitrary ordering on terms. *)

val term_hash: term -> int
  (** A hash key for the term. *)

(** {2 Variables} *)

val is_variable: term -> bool
  (** Is the term a variable? *)

val variable: Variable.t -> term
  (** A variable. *)

(** {2 Constants} *)

val is_constant: term -> bool
  (** Is the term a constant? *)

val constant: Constant.t -> term
  (** A constant. *)

val constant_value: term -> Constant.t
  (** [constant_value t] is the value of a constant. 

      - Require [is_constant t] *)

(** {3 Symbolic Constants} *)

val is_symbolic_constant: term -> bool
  (** Is a term a symbolic constant. 

      - Ensure [not (is_symbolic_constant t) || is_constant t] *)

val symbolic_constant: Symbol.t -> term
  (** A symbolic constant. *)

val symbolic_value: term -> Symbol.t
  (** [symbolic_value t] is the value of a symbolic constant.

      - Require [is_symbolic_constant t] *)

(** {3 Integer Constants} *)

val is_integer_constant: term -> bool
  (** Is a term an integer constant. 

      - Ensure [not (is_integer_constant t) || is_constant t] *)

val integer_constant: Big_int.big_int -> term
  (** Integer constant. *)

val integer_value: term -> Big_int.big_int
  (** [integer_value t] is the value of an integer constant.

      - Require [is_integer_constant t] *)

(** {3 Rational Constants} *)

val is_rational_constant: term -> bool
  (** Is a term a rational constant. 
  
      - Ensure [not (is_rational_constant t) || is_constant t] *)

val rational_constant: Num.num -> term
  (** Arbitrary precision rational constant. *)
 
val rational_value: term -> Num.num
  (** [rational_value t] is the value of a rational constant.

      - Require {!Term.is_rational_constant}[ t] *)

(** {2 Applications} *)

val is_application: term -> bool
  (** Is a term an application? *)

val application: term -> term -> term
  (** The application of the given terms. *)

val application_function: term -> term 
  (** The function in an application.

      - Require [is_application t] *)

val application_argument: term -> term 
  (** The argument in an application.

      - Require [is_application t] *)

val binary_application: term -> term -> term -> term
  (** [binary_application f x y] is equivalent to [application
      (application f x) y].  *)

val ternary_application: term -> term -> term -> term -> term
  (** [ternary_application f x y z] is equivalent to [application
      (application (application f x) y) z].  *)

val nary_application: term -> term list -> term
  (** [nary_application f args] is equivalent to [List.fold_left
      application f args].  *)

val nary_application_decomposition: term -> (term * term list)
  (** Decompose a term to yield a function and a list of arguments.
      If the term is not an application then the pair of the term and
      the empty list is produced. *)

(** {2 Abstractions} *)

val is_abstraction: term -> bool
  (** Is a term an abstraction? *)

val abstraction: Variable.t -> term -> term
  (** [abstraction v tm] is [\v. tm]. *)

val nary_abstraction: Variable.t list -> term -> term
  (** [nary_abstraction \[v0, ... ,vn\] tm] is [\v0. ... \vn.tm]. *)

val beta_normal_form: term -> term
  (** The term obtained by reducing all beta-redexes in a term.
      Applicative order reduction is used. *)

(** {2 Data-structures.} *)

(** Terms augmented with comparison functions. *)
module Term_compared: Util.ComparedType with type t = term

(** Hash tables keyed off terms. *)
module Term_hashtbl: Hashtbl.S with type key = term

(** Sets of terms. *)
module Term_set: Set.S with type elt = term

(** Maps from terms. *)
module Term_map: Map.S with type key = term

(** {2 Pretty-Printing.} *)

val string_of_term: term -> string
  (** String representation of term. *)

val pp_print_term: Format.formatter -> term -> unit
  (** Pretty-print a term. *)

val print_term: term -> unit
  (** Print a term using [std_formatter]. *)

val log_term: int -> term -> unit
  (** [log_term level tm] is equivalent to 
      [Log.log level pp_print_term tm] *)
