(* $Id: term_logic.ml,v 1.1.1.1 2008/05/22 16:46:07 rupak Exp $ *)
(* Term constructors. *)

(* Copyright 2007 Intel Corporation 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. *)

open Term

module type Constructor = sig

  val negation:     term -> term
  val conjunction:  term -> term -> term
  val disjunction:  term -> term -> term
  val equivalence:  term -> term -> term
  val difference:   term -> term -> term
  val implication:  term -> term -> term
  val conditional:  term -> term -> term -> term 

  val equality:     term -> term -> term
  val ite:          term -> term -> term -> term

end

module Standard = struct

  let negation term = 
    application Term_constant.negation term

  let conjunction term1 term2 = 
    binary_application Term_constant.conjunction term1 term2

  let disjunction term1 term2 = 
    binary_application Term_constant.disjunction term1 term2
      
  let equivalence term1 term2 =
    binary_application Term_constant.equivalence term1 term2

  let difference term1 term2 = 
    binary_application Term_constant.difference term1 term2

  let implication term1 term2 = 
    binary_application Term_constant.implication term1 term2

  let conditional term1 term2 term3 = 
    ternary_application Term_constant.conditional term1 term2 term3

  let equality term1 term2 = 
    binary_application Term_constant.equality term1 term2

  let ite term1 term2 term3 = 
    ternary_application Term_constant.ite term1 term2 term3

end


module Rbc = struct

  let is_negation term = match term with
    | Application (_, f, _) when term_equal f Term_constant.negation -> true
    | _ -> false

  let negation term = match term with
    | Application (_, f, x) when term_equal f Term_constant.negation -> 
        (* (not (not x)) = x *)
        x
    | _ -> 
        if term_equal term Term_constant.falsity then 
          (* (not false) = true *)
          Term_constant.truth
        else if term_equal term Term_constant.truth then 
          (* (not true) = false *)
          Term_constant.falsity
        else
          application Term_constant.negation term

  let conjunction term1 term2 = 
    if
      term_equal term1 Term_constant.falsity    (* false and y *)
      || term_equal term2 Term_constant.falsity (* x and false *)
      || term_equal (negation term1) term2      (* x and (not x) *)
    then
      Term_constant.falsity
    else if term_equal term1 Term_constant.truth then (* true and y *)
      term2
    else if term_equal term2 Term_constant.truth then (* x and true *)
      term1
    else if term_equal term1 term2 then (* x and x *)
      term1
    else if term_compare term1 term2 < 0 then
      binary_application Term_constant.conjunction term1 term2
    else begin
      assert (term_compare term1 term2 > 0);
      binary_application Term_constant.conjunction term2 term1
    end

  let disjunction term1 term2 = 
    negation (conjunction (negation term1) (negation term2))

  let rec equivalence term1 term2 = 
    if is_negation term1 then begin
      if is_negation term2 then (* (not x) <-> (not y) = x <-> y *)
        equivalence (negation term1) (negation term2)
      else (* (not x) <-> y = not (x <-> y) *)
        negation (equivalence term1 term2)
    end
    else if is_negation term2 then (* x <-> (not y) = not (x <-> y) *)
      negation (equivalence term1 term2)
    else if term_equal term1 Term_constant.truth then (* true <-> y *)
      term2
    else if term_equal term1 Term_constant.falsity then (* false <-> y *)
      negation term2
    else if term_equal term2 Term_constant.truth then (* x <-> true *)
      term1
    else if term_equal term2 Term_constant.falsity then (* x <-> false *)
      negation term1
    else if term_equal term1 term2 then (* x <-> x *)
      Term_constant.truth
    else if term_equal (negation term1) term2 then (* x <-> (not x) *)
      Term_constant.falsity
    else if term_compare term1 term2 < 0 then
      binary_application Term_constant.equivalence term1 term2
    else begin
      assert (term_compare term1 term2 > 0);
      binary_application Term_constant.equivalence term2 term1
    end

  let difference term1 term2 = 
    negation (equivalence term1 term2)
        
  let implication term1 term2 = 
    disjunction (negation term1) term2

  let conditional term1 term2 term3 = 
    conjunction
      (implication term1 term2)
      (implication (negation term1) term3)

  let equality term1 term2 = 
    let comparison = term_compare term1 term2 in
      if comparison = 0 then (* term1 = term2 *)
        Term_constant.truth
      else if comparison < 0 then
        binary_application Term_constant.equality term1 term2
      else begin
        assert (comparison > 0);
        binary_application Term_constant.equality term2 term1
      end

  let ite term1 term2 term3 = 
    if term_equal term2 term3 then
      term2
    else if is_negation term1 then
      ternary_application Term_constant.ite (negation term1) term3 term2
    else
      ternary_application Term_constant.ite term1 term2 term3

end

module type Rewriter = sig

  val rewrite: term -> term

end

module Make_rewriter (C: Constructor) = struct

  let rewrite_tbl = Term_hashtbl.create 0

  let rec rewrite term = 
    try
      Term_hashtbl.find rewrite_tbl term
    with Not_found ->
      let term' = match term with
        | Application (_, f, x)
            when term_equal f Term_constant.negation ->
            C.negation (rewrite x)
        | Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.conjunction ->
            C.conjunction (rewrite x) (rewrite y)
        | Application (_, Application (_, f, x), y) 
            when term_equal f Term_constant.disjunction ->
            C.disjunction (rewrite x) (rewrite y)
        | Application (_, Application (_, f, x), y)
            when term_equal f Term_constant.equivalence ->
            C.equivalence (rewrite x) (rewrite y)
        | Application (_, Application (_, f, x), y)
            when term_equal f Term_constant.difference ->
            C.difference (rewrite x) (rewrite y)
        | Application (_, Application (_, f, x), y)
            when term_equal f Term_constant.implication ->
            C.implication (rewrite x) (rewrite y)
        | Application (_, Application (_, (Application (_, f, x)), y), z)
            when term_equal f Term_constant.conditional ->
            C.conditional (rewrite x) (rewrite y) (rewrite z)
        | Application (_, Application (_, f, x), y)
            when term_equal f Term_constant.equality ->
            C.equality (rewrite x) (rewrite y)
        | Application (_, Application (_, (Application (_, f, x)), y), z)
            when term_equal f Term_constant.ite ->
            C.ite (rewrite x) (rewrite y) (rewrite z)
        | _ -> term
      in
        Term_hashtbl.add rewrite_tbl term term';
        term'

end

module Rbc_rewriter = Make_rewriter (Rbc)

