module E = Ast.Expression
module P = Ast.Predicate

module C = Term_constant
module L = Term_logic.Standard
module T = Term

exception AtomInExpression

let symbol_tbl = Hashtbl.create 0
let convert_symbol name = 
  try Hashtbl.find symbol_tbl name 
  with Not_found ->
    let v = T.variable (Variable.fresh ()) in
      Hashtbl.add symbol_tbl name v;
      v

let string_tbl = Hashtbl.create 0
let convert_string name = 
  try Hashtbl.find string_tbl name 
  with Not_found ->
    let v = T.variable (Variable.fresh ()) in
      Hashtbl.add string_tbl name v;
      v

let chlval_tbl = Hashtbl.create 0
let convert_chlval name index = 
  try Hashtbl.find chlval_tbl (name, index)
  with Not_found ->
    let v = T.variable (Variable.fresh ()) in
      Hashtbl.add chlval_tbl (name, index) v;
      v

let dpt_int i = T.constant (Constant.integer (Big_int.big_int_of_int i))
let dpt_zero  = dpt_int 0
 
let sel_var = T.variable (Variable.fresh()) 
let store_var = T.variable (Variable.fresh()) 
let select a i   = T.binary_application  (*C.select*) sel_var a i
let store  a i x = T.ternary_application (*C.store*) store_var a i x
let address_op   = convert_symbol "ADDRESS"
let address_of e = T.application address_op e

let field_offset = T.variable (Variable.fresh ())
let offset = T.variable (Variable.fresh ())

let rec convert_lval e = match e with
  | E.Symbol (name) ->
      convert_symbol name
  | E.Dereference d  -> 
      select (convert_expression d) (dpt_zero)
  | E.Access (E.Dot, e, s) 
  | E.Access (E.Arrow, e, s) ->
      select (convert_expression e) (convert_symbol s)
  | E.Indexed (e1, e2) ->
      select (convert_expression e1) (convert_expression e2) 
  | _ ->
      failwith ("DptInterface: Unsupported lval " ^ (E.lvalToString e))

and convert_expression e = match e with
  | E.Lval l ->
      convert_lval l
  | E.Chlval (lv, s) -> 
      convert_chlval lv s
  | E.Unary (op, e1) ->
      begin match op with
        | E.UnaryMinus -> 
            T.application C.unary_minus (convert_expression e1)
        | E.UnaryPlus -> 
            convert_expression e1
        | E.Address ->
            begin match e1 with
              | E.Lval _ | E.Chlval _ ->
                  address_of (convert_expression e1)
              | _ ->
                  failwith 
                    ("DptInterface: Unsupported address " ^ (E.toString e)) 
            end
        | _ ->
           failwith ("DptInterface: Unsupported address " ^ (E.toString e)) 
      end
  | E.Binary (op, e1, e2) ->
      let f = match op with
        | E.Minus -> C.minus
        | E.Plus  -> C.plus
        | E.Mul   -> C.times
        | E.FieldOffset -> field_offset
        | E.Offset -> offset
        | _ when E.isRelOp op -> 
            raise AtomInExpression
        | _       -> 
           failwith ("DptInterface: Unsupported binop " ^ (E.toString e)) 
      in
        T.binary_application f (convert_expression e1) (convert_expression e2)
  | E.Constant (c) ->
      begin match c with
        | Ast.Constant.Int i -> dpt_int i
        | Ast.Constant.String s -> convert_string s
        | Ast.Constant.Float f -> convert_string (string_of_float f)
        | c ->
            failwith 
              ("DptInterface: Unsupported constant "^(Ast.Constant.toString c))
      end
  | E.Select (e1, e2) -> 
      select 
        (convert_expression e1) 
        (convert_expression e2)
  | E.Store (e1, e2, e3) -> 
      store 
        (convert_expression e1) 
        (convert_expression e2) 
        (convert_expression e3)
  | E.FunctionCall (E.Lval (E.Symbol f), args) ->
      List.fold_left 
        (fun f a -> T.application f (convert_expression a))
        (convert_symbol f)
        args
    | _ -> 
        failwith ("DptInterface: Unsupported Expression " ^ (E.toString e))
	      

and convert_atom a = match a with
  | E.Select _
  | E.Store _
  | E.FunctionCall (E.Lval (E.Symbol _), _) ->
      let e = convert_expression a in
      L.equality e dpt_zero
  | E.Binary (E.Or, e1, e2) ->
      L.disjunction (convert_atom e1) (convert_atom e2)
  | E.Binary (E.And, e1, e2) ->
      L.conjunction (convert_atom e1) (convert_atom e2)
  | E.Binary (op, e1, e2) ->
      let e1' = convert_expression e1 in
      let e2' = convert_expression e2 in
        begin match op with
          | E.Eq  -> L.equality    e1' e2'
          | E.Ne  -> L.negation (L.equality e1' e2')
          | E.Gt  -> T.binary_application C.greater          e1' e2'
          | E.Lt  -> T.binary_application C.less             e1' e2'
          | E.Ge  -> T.binary_application C.greater_or_equal e1' e2'
          | E.Le  -> T.binary_application C.less_or_equal    e1' e2'
          | _ ->
              failwith ("DptInterface: Unsupported relop " ^ (E.toString a))
        end
  | _ ->
      failwith ("DptInterface: Unsupported atom " ^ (E.toString a))

and convert_predicate p = match p with
  | P.True -> 
      C.truth
  | P.False -> 
      C.falsity
  | P.And preds -> 
      let dpt_preds = List.map  convert_predicate preds in
        List.fold_left L.conjunction C.truth dpt_preds
  | P.Or preds ->
      let dpt_preds = List.map  convert_predicate preds in
        List.fold_left L.disjunction C.falsity dpt_preds
  | P.Implies (p1,p2) ->
      let d1 = convert_predicate p1 in
      let d2 = convert_predicate p2 in
        L.implication d1 d2
  | P.Iff (p1,p2) ->
      let d1 = convert_predicate p1 in
      let d2 = convert_predicate p2 in
        L.equivalence d1 d2
  | P.Not p1 -> 
      let d1 = convert_predicate p1 in
        L.negation d1
  | P.Atom a ->
      convert_atom a
  | _ -> 
      failwith ("DptInterface: Unsupported predicate " ^ (P.toString p))

let assumptions_stack = Stack.create ()

let block_reset () =
  Stack.clear assumptions_stack

let block_assert p =
  let t = convert_predicate p in
    Stack.push t assumptions_stack

let block_pop () =
  ignore (Stack.pop assumptions_stack)

let solve solver is_interpreted term =
  let cache = Term.Term_hashtbl.create 0 in
  let term = Term.uninterpret is_interpreted cache term in
    solver#add_clause [term];
    solver#solve

let is_contra () =
  let dpll = new Dpll.make () in
  let cc = new Cc.make () in
  let theory = new  Zimplex.make() in
  let solver =  new Solver_dtc.make dpll cc [theory] in
  let is_interpreted tm =
    dpll#owns_constant tm || cc#owns_constant tm || theory#owns_constant tm
  in
  let term = ref C.truth in
    Stack.iter (fun a -> term := L.conjunction !term a) assumptions_stack;
    not (solve solver is_interpreted !term)

let reset () = 
  ()

let block_check_unsat p =
  block_assert p;
  let result = is_contra () in
    block_pop ();
    result

let queryExp p =
  block_check_unsat (P.Not p)

let queryExpContext a p =
  block_assert a;
  let result = queryExp p in
    block_pop ();
    result

let sat = queryExp

let rec convert_back ptbl f =
  try T.Term_hashtbl.find ptbl f
  with Not_found ->
    match f with
      | T.Application (_, n, x) when T.term_equal n C.negation ->
          P.Not (convert_back ptbl x)
      | T.Application (_, T.Application (_, a, x), y) 
          when T.term_equal a C.conjunction ->
          P.And [(convert_back ptbl x); (convert_back ptbl y)]
      | T.Application (_, T.Application (_, d, x), y)
          when T.term_equal d C.disjunction ->
          P.Or [(convert_back ptbl x); (convert_back ptbl y)]
      | _ when T.term_equal f C.truth -> 
          P.True
      | _ when T.term_equal f C.falsity -> 
          P.False
      | _ -> failwith "Dptinterface reconvert trouble."
      
let get_allsat f preds orig_preds =
  Log.set_log_level 0;
  Log.log_string 0 "Entering Dpt AllSat";
  Log.log_newline 0 ();
  let dpll = new Dpll.make () in
  let cc = new Cc.make () in
  let dl = new Dl.Int_solver.make () in
  let s = new Solver_dtc.make_all_sat dpll cc [dl] in
  let is_interpreted tm =
    dpll#owns_constant tm || cc#owns_constant tm || dl#owns_constant tm
  in
  let cache = T.Term_hashtbl.create 0 in
  let dpt_f = convert_predicate f in
  let dpt_preds = List.map convert_predicate preds in
  let dpt_f' = Term.uninterpret is_interpreted cache dpt_f in
  let dpt_preds' = List.map (Term.uninterpret is_interpreted cache) dpt_preds in
  let ptbl = T.Term_hashtbl.create 0 in
  let phi = 
    try
      s#all_sat dpt_f' dpt_preds'
    with excn -> print_string "DPT exception"; raise excn 
  in
    List.iter2 (fun d p -> T.Term_hashtbl.add ptbl d p) dpt_preds' orig_preds;
    convert_back ptbl phi

let print_stats () =
  failwith "DptInterface: print_stats not implemented"

let whoami () = "DptInterface"
