
(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

(*************************************************************************************)
(*************************************************************************************)
(*                               Interface to FOCI                                   *)
(*************************************************************************************)
(*************************************************************************************)
open Ast
module P = Predicate
module E = Expression
module M = Message
module O = Options

open Printf
open Misc

exception AtomInExpressionException
exception FociSatException
exception FociSatExceptionAsgn of string list

module FociPredicate = FociAst
module FociExpression = FociAst
module FociConstant = FociAst.Constant

let stats_nb_foci_queries = ref 0


(* API *)
let whoami () = 
  M.msg_string M.Normal "This program contains Foci, Copyright 2003 Cadence Berkeley Laboratories," ;
  M.msg_string M.Normal "Cadence Design Systems. All rights reserved.";
  "FOCI"

let interpolate a b =  
    incr stats_nb_foci_queries;
    FociDag.pred_tree_of_dag
    <| try Stats.time "foci interpolate" 
        (fun (x,y) -> 
            match O.getValueOfInt "focimax" with
            | -1 -> Foci.interpolate (x,y)
            | fmax -> List.hd <| FociSplit3.interpolateN [x;y] fmax)
        <| map_pair FociDag.pred_dag_of_tree (a,b)
    with Foci.SAT _ -> raise FociSatException

let interpolateN,
    interpolateN_symm,
    interpolateN_dasdill =

    let itpN interpolateFn ps = 
        let get_pred_symbol fp =
            match fp with FociPredicate.Atom s -> Some s | _ -> None in

        incr stats_nb_foci_queries;

        List.map FociDag.pred_tree_of_dag 
        << Stats.time "compute itpN" (fun () -> 
            try M.msg_string M.Debug "calling into foci"; 

                let interpolant = 
                    Stats.time "foci itpn" interpolateFn
                    <| List.map FociDag.pred_dag_of_tree ps in 

                M.msg_string M.Debug "back from foci";
                interpolant

            with Foci.SAT (ps,_) | FociSplit3.SAT (ps,_) -> begin

                M.msg_string M.Normal 
                    << sprintf "Foci SAT: [ %s ]" 
                    << String.concat ", " 
                    <| List.map FociDag.Predicate.toString ps;

                raise <| 
                   (* if ps = [] then FociSatException else *) 
                   FociSatExceptionAsgn 
                        (map_partial get_pred_symbol
                        <| List.map FociDag.pred_tree_of_dag ps)
            end)
        <| () in

    (* Q: does the value of the focimax option change as we go? *)

    (fun ps -> itpN (match O.getValueOfInt "focimax" with
        | -1 -> Foci.interpolateN
        | fmax -> flip FociSplit3.interpolateN <| fmax)
        ps),

    (fun ps -> itpN (match O.getValueOfInt "focimax" with
        | -1 -> Foci.interpolateN_symm
        | fmax -> flip FociSplit3.interpolateN_symm <| fmax)
        ps),

    (fun ts cs -> itpN 
        (FociDasDill.refine <| List.map FociDag.pred_dag_of_tree ts)
        cs)

let theory_clauses () = 
    let fp_list_list = Foci.theoryClauses () in  
    M.msg_string M.Normal <| sprintf 
        "Foci says: %d theory clauses" 
        (List.length <| List.flatten fp_list_list);
    List.map (List.map FociDag.pred_tree_of_dag) fp_list_list


let foci_expression_toString = FociAst.Expression.toString
let foci_predicate_toString = FociAst.Predicate.toString

type foci_return = Sat | Unsat of P.predicate

let isSat a = match a with
  | Sat     -> true
  | Unsat _ -> false

(* For the benefit of foci_symm -- we need to rename the UFs used with this
 * suffix so that they are disjoint *)

let foci_make_disjoint_flag = ref false
let set_disjoint_flag () = 
  M.msg_string M.Debug "Set Disjoint flag";
  foci_make_disjoint_flag := true

let unset_disjoint_flag () = 
  M.msg_string M.Debug "Unset Disjoint flag";
  foci_make_disjoint_flag := false

let foci_UF_idx = ref 0
let foci_UF_idx_tick () = incr foci_UF_idx
let foci_UF_idx_reset () = foci_UF_idx := 0

let foci_UF_string s =
    if !foci_make_disjoint_flag && not (E.is_interpreted s) 
        then s^"_"^(string_of_int !foci_UF_idx) 
        else s

let clean_foci_UF_string s = 
    if not !foci_make_disjoint_flag then s
    else let pieces = Misc.chop s "_" in
        if List.length pieces > 1 then String.concat "_" (Misc.list_init pieces)
        else s

let foci_var_counter = ref 0 
 
let new_foci_var () = 
    incr foci_var_counter;
    "foci"^(string_of_int !foci_var_counter)

let listOfAddrVarsFoci = ref []
let chlval_table = Hashtbl.create 109
let chlval_map_table  = Hashtbl.create 109
let int_table_to = Hashtbl.create 23
let int_table_from = Hashtbl.create 23
let int_table_seen = Hashtbl.create 23

let max_int = 999999 (* some RANDOM number I made up at 1am -- RJ *)
let make_new_num i = int_of_string (Misc.suffix 2 (string_of_int i))

let sign i = if i > 0 then 1 else if i = 0 then 0 else -1 
 
let int_map_to i =
  (* add i to the list of constants we've passed to foci *)
  Hashtbl.replace int_table_seen i true;
  let i' = abs i in
  let j = (sign i) 
      * (if i' < max_int then i' else begin
          if Hashtbl.mem int_table_to i' 
              then Hashtbl.find int_table_to i' 
              else let key = make_new_num i' in 
                  Hashtbl.add int_table_to i' key; 
                  Hashtbl.add int_table_from key i';
                  key
        end) in
   if (j <> i) then M.msg_string M.Normal "WARNING :int_maps_to overflow!";
   j

let int_map_from i = 
  let i' = abs i in
  let seen = 0::(Misc.hashtbl_keys int_table_seen) in
  let focimax = O.getValueOfInt "focimax" in
  if false (* XXX: what's going on here? -- mje *)
      && (focimax <> -1) 
      && (not (Hashtbl.mem int_table_seen i)) 
      && (i' > focimax) 
      && List.for_all (fun k -> abs ((abs k) - i') > focimax) seen (* TBD:SPEED *)
  then begin
      M.msg_string M.Error ("Invalid Interpolant -- large constant:"^(string_of_int i));
      raise FociSatException
      (* this will trickle up all the way and be caught by whoever *)
  end;
  (try Hashtbl.find int_table_from i' with Not_found -> i') * (sign i)
 
(* in theory one could even use simplify or something else to "simplify" *)
let getConstant e = 
  match e with
    | E.Constant (Constant.Int i) -> Some i
    | E.Unary (E.UnaryMinus,E.Constant (Constant.Int i)) -> Some (-i)
    | E.Unary (E.UnaryPlus,E.Constant (Constant.Int i)) -> Some i
    | _ -> None



let get_string_constant,
    get_const_string =

    let string_const_ctr_ref = ref 0
    and string_const_table : (string,string) Hashtbl.t = Hashtbl.create 101
    and const_string_table : (string,string) Hashtbl.t = Hashtbl.create 101 in

    (fun s ->
    try Hashtbl.find string_const_table s 
    with Not_found -> 
        incr string_const_ctr_ref; 
        let c = string_of_int !string_const_ctr_ref in 
        Hashtbl.replace string_const_table s c; 
        Hashtbl.replace const_string_table c s; 
        c),

    fun c ->
    try Hashtbl.find const_string_table c 
    with Not_found -> failwith "Unknown string constant c"


let rec convertExpFoci e =
  let convert_times e1 e2 = 
    match (getConstant e1, getConstant e2) with
      | (Some i1, Some i2) -> FociExpression.Constant (FociConstant.Int (i1 * i2))
      | (Some i1, None) -> FociExpression.Coeff (FociConstant.Int i1, (convertExpFoci e2))
      | (None, Some i2) -> FociExpression.Coeff (FociConstant.Int i2, (convertExpFoci e1))
      | (None, None) -> begin
	  M.msg_string M.Debug (Printf.sprintf "WARNING! convertExpFoci: times of %s * %s  " 
              (E.toString e1) (E.toString e2));
          FociExpression.Application 
              (foci_UF_string "times", List.map convertExpFoci [e1;e2])
      end in

  let foci_minus e' =
    match getConstant e' with
      | Some i -> FociExpression.Constant (FociConstant.Int (int_map_to (-i)))
      | None -> FociExpression.Coeff (FociConstant.Int (-1), (convertExpFoci e')) in

  let chlval_convert_table e = 
    try FociExpression.Variable (Hashtbl.find chlval_map_table e) 
    with Not_found ->
        let s_e = new_foci_var () in
        Hashtbl.add chlval_table s_e e;
        Hashtbl.add chlval_map_table e s_e;
        FociExpression.Variable s_e in

  let chlval_convert_recursive (l,s) = 
    let ufs = foci_UF_string "chlval" in
    FociExpression.Application (ufs, [ convertLvalFoci l; FociExpression.Variable ("_"^s) ]) in

  match e with
    | E.Lval lv -> convertLvalFoci lv
    | E.Chlval (l, s) -> 
        if O.getValueOfBool "fociUF"
            then chlval_convert_recursive (l,s) 
            else chlval_convert_table e
	  
      (* So the chlval thing didnt work out so well. look at test/foci/ex3.foci 
	 so now we just convert it to string and for reconversion, stuff the result in a hashtbl!
	FociExpression.Application ("chlval", [ convertLvalFoci l; FociExpression.Variable ("_"^s) ])
      *)

    | E.Binary (op, e1, e2) -> begin
	match op with 
	  | E.Minus -> 
	      if (O.getValueOfBool "uif") then
                let ufs = foci_UF_string "UIminus" in
		FociExpression.Application (ufs, [convertExpFoci e1; convertExpFoci e2])
	      else
		FociExpression.Sum [convertExpFoci e1; foci_minus e2]
	  | E.Plus  -> 
	      if (O.getValueOfBool "uif") then
                let ufs = foci_UF_string "UIplus" in
		FociExpression.Application (ufs, [convertExpFoci e1; convertExpFoci e2])
	      else
		FociExpression.Sum[convertExpFoci e1; convertExpFoci e2]
	  | E.Mul   -> convert_times e1 e2

          | E.Offset | E.BitAnd | E.BitOr | E.Div 
	  | E.Rem | E.RShift | E.Xor | E.Comma -> 
              let ufs = foci_UF_string (E.binaryOpToVanillaString op ) in
              FociExpression.Application (ufs, List.map convertExpFoci [e1;e2])
          | E.LShift -> begin
		match e2 with
                  | E.Constant (Constant.Int k) -> 
                          convert_times e1 (E.Constant (Constant.Int (Misc.power 2 k)))
		  | _ -> 
                      let ufs = foci_UF_string (E.binaryOpToVanillaString op ) in 
                      FociExpression.Application (ufs, List.map convertExpFoci [e1;e2])
	  end
	  | E.FieldOffset ->
	     (* the second argument to field offset is a field id *)
	     let v = convertExpFoci e1 in
	     let s = match e2 with
               | E.Lval (E.Symbol s) -> s
	       | _ -> failwith ("convertExp : (FOCI) Second operator of field offset not a field id "^
				 (E.toString e)) in
	     let ufs_fof = foci_UF_string "foffset" in
             let ufs_fld = foci_UF_string ("@field_"^s) in
             FociExpression.Application(ufs_fof, [v; FociExpression.Variable ufs_fld])
	     
	  | _ -> 
              if E.isRelOp op then raise AtomInExpressionException
              else failwith ("convertExp (Foci): This operator should not be handled here "
                          ^(E.toString e))
    end
    | E.Constant (c) -> begin
	match c with
	  | Constant.Int i -> FociExpression.Constant (FociConstant.Int (int_map_to i))
	  | Constant.String s -> begin
	      M.msg_string M.Debug ("convertExpFoci: String Constant! in"^(E.toString e));
	      FociExpression.Variable ("_STRINGCONSTANT"^(get_string_constant s))
          end
	  | Constant.Float f ->
              if (f -. (ceil f) = 0.0) 
                  then FociExpression.Constant (FociConstant.Int (int_of_float f))
                  else FociExpression.Variable "_FLOATINGPTCONSTANT"
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
    end
    | E.Unary (uop, e) -> begin
	match uop with 
	  | E.UnaryMinus -> foci_minus e
	  | E.UnaryPlus -> convertExpFoci e
	  | E.Address -> 
              let v = convertExpFoci e in
              if (List.mem v (!listOfAddrVarsFoci)) 
                  then () 
                  else listOfAddrVarsFoci := v::(!listOfAddrVarsFoci);
              FociExpression.Application (foci_UF_string "addrOf", [v])
	  | _ -> FociExpression.Application (foci_UF_string (E.unaryOpToVanillaString uop), 
                    [convertExpFoci e])
    end
    | E.FunctionCall(E.Lval (E.Symbol f), args) ->
        FociExpression.Application(foci_UF_string f, List.map convertExpFoci args)
    | E.Select (e1,e2) -> FociExpression.Application("select",List.map convertExpFoci [e1;e2])
    | E.Store  (e1,e2,e3) -> FociExpression.Application("store",List.map convertExpFoci [e1;e2;e3])
    | _ -> failwith ("convertExpFoci: " ^ (E.toString e) ^ "not handled")
	
  and convertLvalFoci l =
    match l with
    | E.This -> failwith "This not handled in convertLval"
    | E.Symbol(str) -> FociExpression.Variable (str) 
    | E.Access(E.Dot, e, s) -> (* NEEDS DEBUGGING *) 
	let v = convertExpFoci e in (* hack to get around foci bug! *)
        let ufs = foci_UF_string ("sel@field_"^s) in
	FociExpression.Application( ufs, [v(*; FociExpression.Variable ("@field_"^s)*)])
    | E.Access(E.Arrow, e, s) ->
	failwith "foci doesnt like Arrow"
    | E.Dereference e ->
        let ufs = foci_UF_string "sel" in
	FociExpression.Application(ufs, [convertExpFoci e; FociAst.zero])
    | E.Indexed (e1, e2) -> 
        let ufs = foci_UF_string "sel" in
	FociExpression.Application(ufs, List.map convertExpFoci [e1;e2])

  and convertAtomFoci e = (* this returns a focipredicate *)
      let make_eq e1' e2' = 
        try FociPredicate.Equality (convertExpFoci e1', convertExpFoci e2') 
        with AtomInExpressionException -> begin
          match e1' with
            | E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) -> begin
                let s1 = convertExpFoci e1' in
                let s2 = convertAtomFoci e2' in
                FociPredicate.Or [ 
                    (FociPredicate.And [s2 ; 
                        FociPredicate.Equality (s1, FociExpression.Constant (FociConstant.Int 1))]); 
                    (FociPredicate.And [FociPredicate.Not s2 ; 
                        FociPredicate.Equality (s1, FociAst.zero)]) ]
            end
            | _ -> begin
                match e2' with
                  | E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) -> begin
                      let s1 = convertExpFoci e2' in
                      let s2 = convertAtomFoci e1' in
                      FociPredicate.Or [ 
                          (FociPredicate.And [s2 ; 
                              FociPredicate.Equality (s1, FociExpression.Constant (FociConstant.Int 1))]);
                          (FociPredicate.And [FociPredicate.Not s2 ; 
                              FociPredicate.Equality (s1, FociAst.zero)]) ]
                  end
                  | _ -> failwith ("convertAtomFoci: Cannot handle this expression: "^(E.toString e))
            end 
        end in
      let make_leq e1' e2' = FociPredicate.Leq (convertExpFoci e1', convertExpFoci e2') in
      match e with
        | E.Binary (op, e1, e2) -> begin
            match op with 
              | E.Eq  -> make_eq e1 e2 
              | E.Ne  -> FociPredicate.Not (make_eq e1 e2)
              | E.Le  -> make_leq e1 e2	      
              | E.Lt  -> FociPredicate.Not (make_leq e2 e1)
              | E.Gt  -> FociPredicate.Not (make_leq e1 e2)
              | E.Ge  -> make_leq e2 e1 
              | E.And -> FociPredicate.And [ (convertAtomFoci e1); (convertAtomFoci e2) ] 
              | E.Or -> FociPredicate.Or [ (convertAtomFoci e1); (convertAtomFoci e2) ] 
              | _ -> failwith ("convertAtomFoci: Cannot handle this expression: "^(E.toString e))
        end	
        | E.Unary (E.Not, e1) -> FociPredicate.Not (convertAtomFoci e1)
        | E.FunctionCall(E.Lval (E.Symbol f), args) -> make_eq e (E.Constant (Constant.Int 0))
        | E.Lval (E.Symbol s) -> FociPredicate.Atom s
        | _ -> failwith ("convertAtomFoci: "^(E.toString e)^ " Not a binary relation operator")

and convertPredFoci e =
  match e with
    | P.True -> FociPredicate.True
    | P.False -> FociPredicate.False
    | P.And plist -> 
        FociPredicate.And (List.map convertPredFoci plist)
    | P.Or plist -> 
        FociPredicate.Or (List.map convertPredFoci plist)
    | P.Implies (e1, e2) -> 
        FociPredicate.Or [FociPredicate.Not (convertPredFoci e1); (convertPredFoci e2)]
    | P.Not e1 -> FociPredicate.Not (convertPredFoci e1)
    | P.Atom e ->  (convertAtomFoci e)
    | P.Iff (e1,e2) ->  convertPredFoci 
        (P.Or [(P.And [e1;e2]);(P.And [P.Not e1 ; P.Not e2])])
    | P.All (s,e1) -> FociPredicate.Forall ([s],convertPredFoci e1)  
    | _ -> failwith ("Foci Cannot handle this predicate "^(P.toString e))
	
and convertToFociSyntax (e: P.predicate) =
  listOfAddrVarsFoci := [];
  let v_exp = convertPredFoci (P.constantfold e (Hashtbl.create 1)) in
  let ufs_sel = foci_UF_string "sel" in
  let ufs_aof = foci_UF_string "addrOf" in
  let genAddrOfEq b =
      FociPredicate.Equality(b, FociExpression.Application(ufs_sel, 
          [FociExpression.Application(ufs_aof, [b]); FociAst.zero])) in
  let genAddrPositive b = FociPredicate.Not  (FociPredicate.Leq 
      (FociExpression.Application(ufs_aof, [b]), (FociAst.zero))) in
  begin
    let vv = FociPredicate.And 
        ((v_exp :: (List.map (fun b -> FociPredicate.Not (FociPredicate.Equality 
            ((FociExpression.Application(ufs_aof, [b])), FociAst.zero))) !listOfAddrVarsFoci))
            @ (List.map genAddrOfEq !listOfAddrVarsFoci)
            @ (List.map genAddrPositive !listOfAddrVarsFoci)) in
    vv
  end

(* oh no! now we have to write the convert from foci to BLAST. ugh. *)
let rec convertFociLval l = 
  (* this is weird ... since the arg: "l" is supposed to be a foci expression that is really an lval ! *)
  match convertFociExp l with
    | E.Lval l' -> l'
    | _ -> failwith ("convertFociLval failed on argument"^(foci_expression_toString l))
    
and make_bin op e1 e2 = E.Binary(op,convertFociExp e1, convertFociExp e2)
and make_eq e1 e2 = make_bin (E.Eq) e1 e2
and make_leq e1 e2 = make_bin (E.Le) e1 e2

and gatherFociSum el = 
  match el with
    | [] -> E.Constant (Constant.Int 0)
    | h::tl -> List.fold_left (fun a b -> E.Binary (E.Plus,a,convertFociExp b))
	(convertFociExp h) tl
	
and convertFociExp e = 
  match e with
    | FociExpression.Constant (FociConstant.Int i) -> E.Constant (Constant.Int (int_map_from i))
    | FociExpression.Variable v -> begin
        match (Hashtbl.mem chlval_table v) with
          | true -> Hashtbl.find chlval_table v 
          | false -> 
              if Misc.is_prefix "_STRINGCONSTANT" v then
                  let c = Misc.chop_after_prefix "_STRINGCONSTANT" v in 
                  let s = get_const_string c in
                  E.Constant (Constant.String s)
              else E.Lval (E.Symbol v)
    end
    | FociExpression.Application (f',el) -> begin
        let f = clean_foci_UF_string f' in
        match f with
          | "select" -> begin
              try assert (List.length el = 2);
                  let (e1,e2) = (List.nth el 0, List.nth el 1) in
                  E.Select(convertFociExp e1, convertFociExp e2)
              with _ -> failwith "exn in convertFociExp @ select"
          end
          | "store" -> begin
              try assert (List.length el = 3);
                  let (e1',e2',e3') = 
                      let l' = (List.map convertFociExp el) in 
                      List.nth l' 0, List.nth l' 1, List.nth l' 2 in
                  E.Store(e1',e2',e3')
              with _ -> failwith "exn in convertFociExp @ store"
          end
          | "chlval" -> begin
              match el with
                | [l;(FociExpression.Variable s)] -> 
                        E.Chlval (convertFociLval l, Misc.chop_after_prefix "_" s)
                | _ -> failwith ("unknown chlval args in convertFociExp! "^(foci_expression_toString e))
          end
          | "*" -> begin
              try make_bin E.Mul (List.hd el) (List.nth el 2)
              with _ ->
                  failwith ("unknown * args in convertFociExp! "^(foci_expression_toString e))
          end

          (* Hopefully we wont have to deal with wacko spl cases here! *)
	  
          | "addrOf" -> begin
              try let e1 = Misc.list_only el in E.Unary(E.Address, convertFociExp e1) 
              with _ -> failwith ("unknown & args in convertFociExp! "^(foci_expression_toString e))
          end
	  
          | "sel" -> begin
              try let e1, e2 = Misc.list_only2 el in 
                  begin
                      match e2 with
                        | FociExpression.Variable s -> begin
                              if (Misc.is_prefix "@field_" s)
                                  then E.Lval (E.Access(E.Dot, convertFociExp e1, 
                                      (Misc.chop_after_prefix "@field_" s)))
                                  else E.Lval (E.Indexed(convertFociExp e1,convertFociExp e2))
                        end
                        | _ -> (* JHALA: flaky! TBD:ARRAYCANON? *(x+i), x[i] *)
                            if (e2 = FociAst.zero) 
                                then E.Lval (E.Dereference(convertFociExp e1))
                                else E.Lval (E.Indexed (convertFociExp e1, convertFociExp e2))
                  end
              with _ -> failwith ("unknown sel args in convertFociExp! "^(foci_expression_toString e))
          end
	  
          | "+" -> begin
              M.msg_string M.Normal "foci uses + in app";
              gatherFociSum el
          end
	      
          | "offset" -> 
              let e1, e2 = Misc.list_only2 (List.map convertFociExp el) in
              E.Binary(E.Offset,e1,e2)
          | "UIplus" -> begin
              M.msg_string M.Normal "foci uses UIplus in app";
              gatherFociSum el
          end
	  | "UIminus" ->
              if (List.length el <> 2) then failwith "Bad use of UIminus";
              let e1, e2 = Misc.list_only2 (List.map convertFociExp el) in E.Binary(E.Minus,e1,e2)
          | "foffset" -> begin
              match el with
                | [e1;(FociExpression.Variable s)] -> 
                    let fld = 
                        if Misc.is_prefix "@field_" s 
                            then Misc.chop_after_prefix "@field_" s 
                            else failwith "bad foffset!" in
                    E.Binary (E.FieldOffset, (convertFociExp e1), (E.Lval (E.Symbol fld)))
                | _ -> failwith "bad number of args to foffset!"
          end
          | _ -> begin
              if Misc.is_prefix "sel@field_" f then
                  (* a field "access" thing *)
                  let field_name = Misc.chop_after_prefix "sel@field_" f in
                  try let e = Misc.list_only el in
                      E.Lval (E.Access(E.Dot, convertFociExp e, field_name))
                  with _ -> failwith "multiple args to sel@field_ ! ?"
              else
              try let op = E.binaryOpFromVanillaString f in
                  let e1, e2 = Misc.list_only2 el in
                   E.Binary (op, convertFociExp e1,convertFociExp e2)
              with _ -> E.FunctionCall(E.Lval (E.Symbol f), List.map convertFociExp el)
          end
    end
    | FociExpression.Sum el -> gatherFociSum el
    | FociExpression.Coeff ((FociConstant.Int i),e') ->
        if i = -1 then 
            E.Unary (E.UnaryMinus, (convertFociExp e'))
        else if i < 0 then 
            E.Binary(E.Mul, E.Constant (Constant.Int (int_map_from i)), convertFociExp e')
        else if i = 1 then 
            convertFociExp e'
        else 
            E.Binary(E.Mul, E.Constant (Constant.Int (int_map_from i)), convertFociExp e')
    | _ -> failwith "convertFociExp: unexpected expression"
	      

and convertFociPred p =
  match p with
    | FociPredicate.True -> P.True
    | FociPredicate.False -> P.False
    | FociPredicate.And plist -> P.And (List.map convertFociPred plist)
    | FociPredicate.Or plist -> P.Or (List.map convertFociPred plist)
    | FociPredicate.Not p' -> P.Not (convertFociPred p')
    | FociPredicate.Implies (p1,p2) -> P.Implies (convertFociPred p1,convertFociPred p2)
    | FociPredicate.Equality (e1,e2) -> P.Atom (make_eq e1 e2)
    | FociPredicate.Leq (e1,e2) -> P.Atom (make_leq e1 e2)
    | FociPredicate.Atom s -> P.Atom (E.Lval (E.Symbol s))
    | _ -> failwith "convertFociPred: unexpected predicate"


let foci_plist_toString = 
    String.concat "\n ; \n" 
    << List.map foci_predicate_toString

let fq_dump num_cons =
    let big_fq_count_ref = ref 0 in
    fun cons_l_string ->
    let file = 
        sprintf "%s.%s" (O.getValueOfString "mainsourcename") <| 
        if num_cons > 2 then begin 
            incr big_fq_count_ref; 
            sprintf "%N.fq" !big_fq_count_ref end 
        else "fociqs" in

    M.msg_string M.Normal <| sprintf "writing to: %s" file;
    Misc.write_to_file file <| sprintf "\n\n\n %s\n\n\n" cons_l_string

(* now the somewhat non-trivial bit *)
(* cons_pre and cons_post are the two sets of constraints. This function returns the predicates
appearing in the fwd interpolant  *)


let check_interpolant itp phi1 phi2 =
    (* for now, only check that every symbol in itp is in both phi1 and phi2 *)
    match List.filter (fun x -> not ((List.mem x <| P.allVarExps phi1) 
            && (List.mem x <| P.allVarExps phi2))) 
        <| P.allVarExps itp with
    | [] -> ()
    | exps -> begin
        ignore 
        << map_pair2 (M.msg_string M.Normal) (M.msg_string M.Error) 
        << double <| sprintf 
            "Interpolant issue: %s\n phi1: %s\n phi2: %s\n" 
            (P.toString itp) (P.toString phi1) (P.toString phi1);
        List.iter (M.msg_string M.Normal << E.toString) exps
    end
    
      
let extract_and_process_interpolant itp_process_function (cons_pre,cons_post) =
    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    M.msg_string M.Debug "In extract_and_process_interpolant";

    let cons_l = List.map (fun x -> P.And x) [cons_pre;cons_post] in

    List.iter (M.msg_printer M.Debug P.print) cons_l; 

    let foci_c_pre, foci_c_post = 
        Misc.list_only2 (List.map convertToFociSyntax cons_l) in

    try M.msg_string M.Debug ">>>>>>>>>>>> calling foci";
        M.msg_string M.Debug <| foci_predicate_toString foci_c_pre;
        M.msg_string M.Debug ";";
        M.msg_string M.Debug <| foci_predicate_toString foci_c_post;
        M.msg_string M.Debug ">>>>>>>>>>>> foci done";

        if O.getValueOfBool "fqdump" then 
            fq_dump 2 <| foci_plist_toString [foci_c_pre;foci_c_post];

        let foci_file_string = 
            sprintf "\n\n\n %s\n ;\n  %s\n\n\n" 
                (foci_predicate_toString foci_c_pre) 
                (foci_predicate_toString foci_c_post) in

	if (false) then Misc.append_to_file "fociqs" foci_file_string;

        let fwd_ip = 
            try interpolate foci_c_pre foci_c_post
            with e -> begin
                M.msg_string M.Error <| sprintf
                    "Foci FAILS!(eap_itp) raises exception!! %s"
                    (Printexc.to_string e);
                M.msg_string M.Normal <| sprintf
                    "Foci FAILS!(eap_itp) raises exception!! %s"
                    (Printexc.to_string e);

                Misc.append_to_file "foci_failed_q" foci_file_string;
                FociPredicate.True end in

        M.msg_string M.Debug 
            << sprintf "Foci yields Interpolant: %s" 
            <| foci_predicate_toString fwd_ip;

        let fwd_ip_b = convertFociPred fwd_ip in

        M.msg_string M.Debug 
            << sprintf "Foci yields b_Interpolant: %s" 
            <| P.toString fwd_ip_b;

        uncurry <| check_interpolant fwd_ip_b <| list_only2 cons_l;

	itp_process_function fwd_ip_b

    with FociSatException -> failwith "Foci raised SAT!"



let extract_preds = 
    extract_and_process_interpolant P.getAtoms

let extract_preds_parity = 
    extract_and_process_interpolant <| P.getAtoms_parity true

let extract_interpolant = 
    extract_and_process_interpolant id

let interpolant_sat cons_pre cons_post =
    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    M.msg_string M.Debug "In interpolant_sat:";
    M.msg_string M.Debug <| P.toString cons_pre;
    M.msg_string M.Debug <| P.toString cons_post;
    
    let foci_c_pre = convertToFociSyntax cons_pre
    and foci_c_post = convertToFociSyntax cons_post in
    let cons_l_string = foci_plist_toString [foci_c_pre; foci_c_post] in

    M.msg_string M.Debug ">>>>>>>>>>>> calling foci";
    M.msg_string M.Debug cons_l_string ;
    M.msg_string M.Debug ">>>>>>>>>>>> foci done";

    try let fwd_ip = interpolate foci_c_pre foci_c_post in

        M.msg_string M.Debug <| sprintf 
            "Foci yields Interpolant: %s"
            (foci_predicate_toString fwd_ip);

        let fwd_ip_b = convertFociPred fwd_ip in

        M.msg_string M.Debug <| sprintf 
            "Foci yields b_Interpolant: %s"
            (P.toString fwd_ip_b);
	Unsat fwd_ip_b
    with 
    | FociSatException -> 
        M.msg_string M.Debug "Foci says SAT!";  Sat

    | FociSatExceptionAsgn _ -> 
        M.msg_string M.Debug "Foci says SATa!"; Sat

    | exc -> begin
        M.msg_string M.Error <| sprintf 
            "Foci FAILS!(isat) raises exception!! %s"
            (Printexc.to_string exc);
        Misc.append_to_file "foci_failed_q" <| sprintf 
            "\n\n\n %s\n ;\n  %s\n\n\n" 
            (foci_predicate_toString foci_c_pre) 
            (foci_predicate_toString foci_c_post);
        raise exc
    end
	      	  
let interpolant_implies a b = 
  not << isSat << interpolant_sat a <| P.negate b
  
let interpolant_unsat a =
  not << isSat <| interpolant_sat a P.True

let extract_and_process_interpolantN itp_fun foci_syn_conv itp_process_function cons_l 
        : P.predicate list = 

    Hashtbl.clear chlval_table;
    Hashtbl.clear chlval_map_table;
    M.msg_string M.Debug "In extract_and_process_interpolantN:";
    List.iter (M.msg_printer M.Debug P.print) cons_l;

    let foci_cons_l = List.map foci_syn_conv cons_l in
    let cons_l_string = foci_plist_toString foci_cons_l in

    M.msg_string M.Debug ">>>>>>>>>>>> calling foci";
    M.msg_string M.Debug cons_l_string ;
    M.msg_string M.Debug ">>>>>>>>>>>> foci done";

    if O.getValueOfBool "fqdump" then fq_dump (List.length cons_l) cons_l_string;

    let itp_l = 
        try let rv = itp_fun foci_cons_l in 
            M.msg_string M.Debug "FOCI returns"; 
            rv
        with
          | FociSatException -> begin
              M.msg_string M.Normal "Foci says SAT!"; 
              raise FociSatException end

          | FociSatExceptionAsgn a -> begin
              M.msg_string M.Normal "Foci says SATa!";
              raise (FociSatExceptionAsgn a) end

          | e -> begin
              M.msg_string M.Normal <| sprintf
                  "Foci FAILS!(eap_itpN) raises exception!! %s"
                  (Printexc.to_string e);
              M.msg_string M.Error <| sprintf
                  "Foci FAILS!(eap_itpN) raises exception!! %s"
                  (Printexc.to_string e);
              Misc.append_to_file "foci_failed_q" <| sprintf 
                  "\n\n\n %s\n\n\n"
                  cons_l_string;
              raise FociSatException 
          end in

    M.msg_string M.Debug <| sprintf 
        "Foci yields Interpolants: %s\n"
        (foci_plist_toString itp_l);

    let itp_b_l = List.map convertFociPred itp_l in
    M.msg_string M.Debug 
        << sprintf "Foci yields b_Interpolant: %s"
        << String.concat "\n ; \n" <| List.map P.toString itp_b_l;
    List.map itp_process_function itp_b_l

let conv_fun p = 
    if O.getValueOfBool "fmc" then 
        foci_UF_idx_tick (); 
    convertToFociSyntax p
  
let extract_interpolantN = 
    extract_and_process_interpolantN interpolateN conv_fun id

let extract_interpolantN_noconv =
    extract_and_process_interpolantN interpolateN convertToFociSyntax id
  
let extract_interpolantN_symm arg =
    set_disjoint_flag ();
    try let ps = 
            extract_and_process_interpolantN 
                interpolateN_symm conv_fun id arg in
        unset_disjoint_flag ();
        ps
    with e -> begin
        unset_disjoint_flag (); 
        raise e
    end
  
let extract_minp_interpolantN cs =
    M.msg_string M.Debug "extract_minp_itpN";
    ignore <| extract_interpolantN cs;
    let n = List.length cs in
    let k = binary_search (fun i ->
        M.msg_string M.Debug "extract_minp_itpN:tick";
        try ignore << extract_interpolantN <| take (i+1) cs; 
            true
        with FociSatException | FociSatExceptionAsgn _ -> false)
        0 (n-1) in
    M.msg_string M.Debug <| sprintf "extract_minp_itpN: k/n = %d/%d" k n;
    extract_interpolantN <| take (k+1) cs @ clone P.True (n-k-1)

let extract_interpolantN_dasdill' trel_cons_l cube_cons_l =
    assert (List.length trel_cons_l + 1 = List.length cube_cons_l);

    M.msg_string M.Debug "Calling extract_and_process_interpolantN *for dasdill*:";
    List.iter (M.msg_printer M.Debug P.print) trel_cons_l;

    let foci_trel_cons_l = List.map conv_fun trel_cons_l in
    M.msg_string M.Debug ">>>>>>>>>>>> calling foci:trel";
    M.msg_string M.Debug <| foci_plist_toString foci_trel_cons_l;

    extract_and_process_interpolantN 
        (interpolateN_dasdill foci_trel_cons_l)
        conv_fun 
        id 
        cube_cons_l


(* incremental foci interface -- using foci just as a DP 
0. reset: unit -> unit
1. push: pred-> unit 
2. pop: unit -> unit
3. check_sat: unit -> bool
*)

(* foci 3 stuff *)
let foci_instance_ref = ref (Foci2.new_instance 1 false true, 0)

let inc_reset () =
    foci_instance_ref := Foci2.new_instance 1 false true, 0

let inc_snd r = let inst, ht = !r in r := inst, ht+1
and dec_snd r = let inst, ht = !r in r := inst, ht-1

let inc_pop () = 
    if snd !foci_instance_ref > 0 then begin
        M.msg_string M.Debug "fociinc:pop";
        Foci2.pop <| fst !foci_instance_ref;
        dec_snd foci_instance_ref
    end
  
let inc_push p =
    Foci2.push <| fst !foci_instance_ref;
    Foci2.add (fst !foci_instance_ref) 0 
        << FociDag.pred_dag_of_tree 
        <| convertToFociSyntax p;
    M.msg_string M.Debug "fociinc:push";
    inc_snd foci_instance_ref

let inc_check_sat () =
    try begin
        ignore << Foci2.solve <| fst !foci_instance_ref; 
        M.msg_string M.Debug 
            << sprintf "fociinc: ht=%s."
            << string_of_int 
            <| snd !foci_instance_ref;
        false 
    end
    with Foci2.SAT _ -> true

let inc_check_unsat p = 
    inc_push p;
    let a = inc_check_sat () in
    inc_pop ();
    not a

let inc_is_contra = not << inc_check_sat
 
(******************************  End Interface to FOCI  ******************************)

(* THEOREMPROVER_QUERY interface -- mje *)
let rec queryExp exp = 
    isSat <| interpolant_sat exp P.True
and queryExpContext pred query = 
    not << isSat <| interpolant_sat pred P.True
    (* XXX: can this be right? shouldn't 'not' be omitted *)
and block_assert = inc_push
and block_pop = inc_pop
and block_reset = inc_reset
and is_contra = inc_is_contra
and block_check_unsat p = 
    ignore <| block_assert p; 
    let a = is_contra () in 
    block_pop (); a
and assume = inc_push
and pop = inc_pop
and reset () = ()
and sat pred = failwith "foci:sat not supported"
and print_stats () = failwith "print_stats not supported"
