(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

(*************************************************************************************)
(* Interface to Simplify.                                                            *)
(* Simplify is run as a server and queries are piped to it.                          *)
(* Thanks to Wes Weimer for the code.                                                *)
(*************************************************************************************)
open Ast
module E = Expression
module P = Predicate
module M = Message
module O = Options

exception AtomInExpressionException
exception ChannelException

open Misc

let reset_alarm () =
  (* Sys.set_signal Sys.sigalrm (Sys.Signal_ignore) *) ()

let set_alarm () = ()

(* may raise channel exception ... *)
let secure_output_string oc s = 
  try
    set_alarm ();output_string oc s;
    flush oc; reset_alarm ()
  with ChannelException -> 
    reset_alarm ();
    raise ChannelException

(* may raise channel exception *)
let secure_input_line ic = 
  set_alarm ();
  try
    let rv = input_line ic in
    reset_alarm ();rv
  with ChannelException ->
    reset_alarm () ;
    raise ChannelException


(* Return a "\n" terminated string corresponding to simplify input *)
let clear_varHashtbl_flag = ref true (* the table will get cleared only if this flag is true *)
		    (* moreover -- when simplify_assume is called the flag gets set to false *)
		    (* when pop is called, the flag is again set to true *)

let varHashtbl = Hashtbl.create 53 
let next_var = ref 0 
let chlvalHashtbl = Hashtbl.create 53 

let unknown_var_counter = ref 0

let get_fresh_unknown () = 
  let rv = !unknown_var_counter in
    unknown_var_counter := !unknown_var_counter + 1;
    ("UNK"^(string_of_int rv))

let get_string_constant,
    get_const_string =

    let string_const_ctr_ref = ref 0
    and string_const_table : (string,string) Hashtbl.t = Hashtbl.create 101
    and const_string_table : (string,string) Hashtbl.t = Hashtbl.create 101 in

    (fun s ->
    try Hashtbl.find string_const_table s 
    with Not_found -> 
        incr string_const_ctr_ref; 
        let c = string_of_int !string_const_ctr_ref in 
        Hashtbl.replace string_const_table s c; 
        Hashtbl.replace const_string_table c s; 
        c),

    fun c ->
    try Hashtbl.find const_string_table c 
    with Not_found -> failwith "Unknown string constant c"

let convertToSimplifySyntax exp =
(*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
  (* M.msg_string M.Debug ("convertToSimplifySyntax exp = ");
     M.msg_printer M.Debug P.print exp; *)  (*RuGang*)
  let get_next_var () = next_var := !next_var + 1; !next_var in
  if (!clear_varHashtbl_flag) then 
      (Hashtbl.clear varHashtbl;
       next_var :=0; 
       Hashtbl.clear chlvalHashtbl);
  let fieldId str =
    try Hashtbl.find varHashtbl str 
    with Not_found -> 
	let new_id = "v"^(string_of_int (get_next_var ())) in
	(Hashtbl.add varHashtbl str  new_id; new_id) in
  let id_of_var lv =
    match lv with
      E.Symbol s -> begin
	  try
	    let vid = Hashtbl.find varHashtbl s in
	    int_of_string (Misc.chop_after_prefix "v" vid)
	  with Not_found -> failwith ("id_of_var: (Simplify) variable not found "^(s))
      end
    | _ -> failwith ("id_of_var: (Simplify) lvalue not a symbol. "^(E.lvalToString lv)) in
  (* let id_of_fld s =
    let vid = 
      try Hashtbl.find varHashtbl s with Not_found -> failwith ("id_of_fld: (Simplify) variable not found "^(s)) in
    int_of_string (Misc.chop_after_prefix "v" vid) in *)
  let rec convertExp e =
    let chlval_convert_recursive (lv,s) = Printf.sprintf "(chlval %s %s)" (convertLval lv) (fieldId s) in  
    let chlval_convert_table e = 
      try (Hashtbl.find chlvalHashtbl e) 
      with Not_found ->
	      let s_e = get_next_var () (* E.toString e*) in
              let new_id = "chlval"^(string_of_int s_e) in
		(Hashtbl.add chlvalHashtbl e new_id; new_id) in
    match e with
      | E.Lval lv -> convertLval lv
      | E.Chlval (lv,s) -> if (O.getValueOfBool "clock") then chlval_convert_recursive (lv,s) else chlval_convert_table e
      | E.Binary (op, e1, e2) -> 
	      (match op with 
		  E.Minus -> Printf.sprintf "(- %s %s)" (convertExp e1) (convertExp e2)
                | E.Plus  -> Printf.sprintf "(+ %s %s)" (convertExp e1) (convertExp e2)
		| E.Mul   -> Printf.sprintf "(* %s %s)" (convertExp e1) (convertExp e2)
                | E.Offset 
                | E.BitAnd 
		| E.BitOr 
		| E.Div 
		| E.Rem 
		| E.Xor 
		| E.Comma -> Printf.sprintf "(%s %s %s)" (E.binaryOpToVanillaString op) (convertExp e1) (convertExp e2) 
                | E.RShift -> (* second term / shifted value must be a constant *)
		    (match (e2) with E.Constant (Constant.Int _) | _ ->  
                      Printf.sprintf "(%s %s %s)" (E.binaryOpToVanillaString op) (convertExp e1) (convertExp e2))
                | E.LShift ->
		    (match e2 with E.Constant (Constant.Int k) -> Printf.sprintf "(* %s %d)" (convertExp e1) (Misc.power 2 k)
		     | _ -> Printf.sprintf "(%s %s %s)"
                     (E.binaryOpToVanillaString op) (convertExp e1) (convertExp e2))
                | E.FieldOffset -> (* the second argument to field offset is a field id *)
                    let fld = match e2 with E.Lval (E.Symbol s) -> s
		              | _ -> failwith ("convertExp (simp): field offset 2nd op not field "^(E.toString e)) in
	            Printf.sprintf "(foffset %s %s)" (convertExp  e1) (fieldId fld)
                |	_ -> if E.isRelOp op then raise AtomInExpressionException 
                             else failwith ("convertExp: (Simplify) not here:"^(E.binaryOpToVanillaString op)^" :  "^(E.toString e)))
    | E.Constant (c) -> 
	begin
	  match c with
	    Constant.Int i ->  if (i < 0) then "(- 0 "^(string_of_int (-i))^")" else string_of_int i 
	  | Constant.String s -> 
              let c = get_string_constant s in
              ("(_STRINGCONSTANT "^c^")")
	  | Constant.Float f -> 
	      if (f -. (ceil f) = 0.0) then string_of_int (int_of_float f)
	      else ("FLOATINGPT"^(string_of_int (get_next_var ())))
	  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	end

    | E.Unary (uop, e1) -> 
	begin
	  match uop with 
	    E.UnaryMinus -> "(- 0 "^(convertExp e1)^")"
	  | E.UnaryPlus -> convertExp e1
	  | E.Address -> 
	      begin
		match e1 with
		  E.Lval lv -> 
		    let l = convertLval lv
		    in
		    let id = (id_of_var lv) in
		    Printf.sprintf "(addrOf %s %d)" l id
		| E.Chlval (lv, f) -> 
		    (* let lstr = (convertLval lv) in
		    let fstr = (fieldId f) in
		    let id = (id_of_fld f) in
		       *)
		    let chlv = convertExp e1 in
		    let id = int_of_string (Misc.chop_after_prefix "chlval" chlv) in
		    Printf.sprintf "(addrOf %s %d)" chlv id
		      
		| _ -> failwith ("convertExp: (Simplify) Address: strange expression "^(E.toString e))
	      end
	  | E.BitNot -> Printf.sprintf "(BitNot %s)" (convertExp e1)
	  |	_ -> "("^E.unaryOpToVanillaString uop^" "^(convertExp e1)^")"
	end
    | E.Select (e1,e2) -> Printf.sprintf "(select %s %s)" (convertExp e1) (convertExp e2)
    | E.Store (e1,e2,e3) -> Printf.sprintf "(store %s %s %s)" (convertExp e1) (convertExp e2) (convertExp e3) 
    | E.FunctionCall (E.Lval (E.Symbol f), args) ->
        Printf.sprintf "(%s %s)" f (Misc.string_list_cat " " (List.map convertExp args)) (* TBD:RP *)
    | _ -> failwith ("convertExp (Simplify): " ^ (E.toString e) ^ "not handled")
	 
  and convertLval l =
    match l with
    | E.This -> failwith "This not handled in convertLval"
    | E.Symbol(str) -> 
	(try Hashtbl.find varHashtbl str with Not_found ->
	    let new_id = "v"^(string_of_int (get_next_var ())) in
	    (Hashtbl.add varHashtbl str new_id;new_id))
    | E.Dereference drfe  -> "(select "^(convertExp drfe)^" 0)"
    | E.Access(E.Dot, e, s) -> (Printf.sprintf "(select %s %s)" (convertExp e) (fieldId s))
    | E.Access(E.Arrow, e, s) ->
	let fieldId_s =
(* MAJOR HACK: *) if s = "next" then "next" else
	  try Hashtbl.find varHashtbl s with Not_found -> 
	      (let new_id = "v"^(string_of_int (get_next_var ())) in
	       Hashtbl.add varHashtbl s new_id; new_id) in
	(Printf.sprintf "(select %s %s)" (convertExp e) fieldId_s)
	  (* JHALA: MAJOR BUG!! this should not be (+ e1 e2) surely *)
    | E.Indexed (e1, e2) -> Printf.sprintf "(select %s %s)" (convertExp e1) (convertExp e2)

  and convertAtom e =   
    match e with
      E.Binary (op, e1, e2) -> 
	begin
          match op with
          | E.Or -> 
            begin
              let s1 = convertAtom e1 and
                  s2 = convertAtom e2 in
              Printf.sprintf "(OR %s %s)" s1 s2
            end
          | E.And ->
            begin
              let s1 = convertAtom e1 and
                  s2 = convertAtom e2 in
              Printf.sprintf "(AND %s %s)" s1 s2
            end
          | _ ->
	  let opstring = 
	    match op with 
		E.Eq  -> "EQ" |	E.Ne  -> "NEQ" 
	    |	E.Gt  -> ">"  |	E.Lt  -> "<"
	    |	E.Ge  -> ">=" |	E.Le  -> "<="
	    |	_   -> M.msg_string M.Error ("Strange binop: "^(E.binaryOpToVanillaString op)); "" 
              (**** see below ** failwith ("convertAtom: Cannot handle this expression: "^(E.toString e)) *)
	  in
          if opstring <> "" 
	  then
	    begin
              try Printf.sprintf "(%s %s %s)" opstring (convertExp e1) (convertExp e2)
              with AtomInExpressionException ->
		begin
		  if (opstring = "EQ" or opstring = "NEQ") then
		    begin
                      match e1 with
			E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) ->
			  begin
			    let s1 = convertExp e1 in
			    let s2 = convertAtom e2 in
			    Printf.sprintf "(OR (AND %s  (%s %s  1)) (AND (NOT %s) (%s %s 0)))" s2 opstring s1 s2 opstring s1 
			  end
                      | _ -> 
			  begin
			    match e2 with
			      E.Lval _ | E.Chlval _ | E.Constant (Constant.Int _) ->
				begin
				  let s1 = convertExp e2 in
				  let s2 = convertAtom e1 in
				  Printf.sprintf "(OR (AND %s  (%s %s  1)) (AND (NOT %s) (%s %s 0)))" s2 opstring s1 s2 opstring s1 
				end
			    | _ -> 
				Printf.sprintf "(EQ %s 0)" (get_fresh_unknown ()) 
			  end 
		    end
		  else
		    Printf.sprintf "(EQ %s 0)" (get_fresh_unknown () )
		end
            end
	  else (M.msg_string M.Debug ("failed opthing (simplify): "^(E.toString e));
                Printf.sprintf "(EQ %s 0)" (get_fresh_unknown ()) )
	end
    | E.FunctionCall (E.Lval (E.Symbol f), args) ->
        Printf.sprintf "(EQ (%s %s) 0)" f (Misc.string_list_cat " " (List.map convertExp args))
    | E.Select (e1,e2) -> Printf.sprintf "(EQ (select %s %s) 0)" (convertExp e1) (convertExp e2)
    | E.Store (e1,e2,e3) -> Printf.sprintf "(EQ (store %s %s %s) 0)" (convertExp e1) (convertExp e2) (convertExp e3) 
    | E.Lval (E.Symbol s) when Misc.is_prefix "b_p__BS" s -> s
    | _ -> (M.msg_string M.Debug "failed opthing2 (simplify) ";M.msg_string M.Debug (E.toString e); 
            (* convertExp e *) Printf.sprintf "(EQ %s 0)" (get_fresh_unknown () ) )
	(*** instead i shall simply call convert expression on it *** failwith ("convertAtom: "^(E.toString e)^ " Not a binary relation operator") *)
  
  and convertPred e =
    match e with
      P.True -> "TRUE"
    | P.False -> "FALSE"
    | P.And plist -> 
	List.fold_left (function a -> function b -> "(AND "^(convertPred b)^" "^a^")") 
	  "TRUE" (plist)
    | P.Or plist -> 
	List.fold_left (function a -> function b -> "(OR "^(convertPred b)^" "^a^")") 
	  "FALSE" (plist)
    | P.Implies (e1, e2) -> "(IMPLIES "^(convertPred e1)^" "^(convertPred e2)^")"
    | P.Iff (e1,e2) -> Printf.sprintf "(IFF %s %s)" (convertPred e1) (convertPred e2)
    | P.Not e1 -> "(NOT "^(convertPred e1)^")"
    | P.Atom e ->  (convertAtom e) 
    | P.All (s,e1) -> Printf.sprintf "(FORALL (%s) %s)" (convertLval (E.Symbol s)) (convertPred e1)
    | P.Exist (s,e1) -> Printf.sprintf "(EXISTS (%s) %s)" (convertLval (E.Symbol s)) (convertPred e1)
    | _ -> failwith ("Simplify: Cannot handle this predicate "^(P.toString e))
  in
  (convertPred (P.constantfold exp (Hashtbl.create 1)))^"\n"


let bits =  ref 4


let assertBitVector s =
  let rec assertBitV n s = 
    if (n = 0) then ""
    else 
      begin
	let f = Printf.sprintf "(OR (EQ %s_%d 0) (EQ %s_%d 1)) " s n s n in
	  (f ^ (assertBitV (n-1) s))
      end in
    begin
      ("(BG_PUSH (AND "^(assertBitV !bits s)^"))")
    end
      
let shift_R res a b_num =
  let fu = ref "" in
    for i = 0 to (b_num-1) do
      fu := !fu ^ (Printf.sprintf "(EQ %s_%d 0) " res (!bits - i))
    done;
    
    for i = b_num to (!bits-1) do
      fu := !fu ^ (Printf.sprintf "(EQ %s_%d %s_%d) " res (!bits - i) a (!bits - i + b_num ))
    done;
    !fu

let rShiftVar res a b = 
  let everything = ref "(AND " in
    for i = 0 to !bits do
      everything := !everything ^ (Printf.sprintf "(IFF (EQ %s %d) (AND %s)) " 
				     b i (shift_R res a i))
    done;
    !everything ^ (Printf.sprintf "(IFF (> %s %d) (AND %s)))" 
		     b !bits (shift_R res a !bits))
      
(* let () = print_string ("(BG_PUSH "^(rShiftVar "d" "a" "b")^")\n") *)

let rShiftCons res a b_cons =
  (Printf.sprintf "(AND %s) " (shift_R res a b_cons))

let condAnd res a b = 
  let foo = ref "(AND " in
    foo := !foo ^ (Printf.sprintf "(EQ %s 1) " res);
    foo := !foo ^ (Printf.sprintf "(IFF (EQ %s 0) (OR " res);
    foo := !foo ^ "(AND ";
    for i = 1 to !bits do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " a i)
    done;
    foo := !foo ^ ") ";
    foo := !foo ^ "(AND ";
    for i = 1 to !bits do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " b i)
    done;
    foo := !foo ^ ")))) ";
    !foo


let rec convToBin x =
  if (x = 0) then []
  else (x mod 2)::(convToBin (x/2))
    

(* let () = print_string ("(BG_PUSH "^(condAnd "d" "a" "b") ^ ")\n") *)

let rec bitEqBlist n a b  =
  match b with
      [] -> ""
    | numb::rest -> ((Printf.sprintf "(EQ %s_%d %d) " a n numb) ^ 
	(bitEqBlist (n-1) a rest))

let rec bitNEqBlist n a b  =
  match b with
      [] -> ""
    | numb::rest -> ((Printf.sprintf "(NEQ %s_%d %d) " a n numb) ^ 
	(bitNEqBlist (n-1) a rest))


let bitEqInt a b = 
  let blist = convToBin b in
  let foo = ref "(AND " in
  let i = List.length blist in
    for j = !bits downto (i+1) do
      foo := !foo ^ (Printf.sprintf "(EQ %s_%d 0) " a j)
    done;
    foo := !foo ^ (bitEqBlist i a (List.rev blist));
    !foo ^ ")"


let bitNEqInt a b = 
  let blist = convToBin b in
  let foo = ref "(OR " in
  let i = List.length blist in
    for j = !bits downto (i+1) do
      foo := !foo ^ (Printf.sprintf "(NEQ %s_%d 0) " a j)
    done;
    foo := !foo ^ (bitNEqBlist i a (List.rev blist));
    !foo ^ ")"

    (*let () = print_string ("(BG_PUSH "^(bitEqInt "a" 10) ^ ")\n") *)

let bitEqS a b =
  let foo = ref "(AND " in
  for i = 1 to !bits do
    foo := !foo ^ (Printf.sprintf "(EQ %s_%d %s_%d) " a i b i)
  done;
    !foo ^ ")"
      
let bitNEqS a b =
  let foo = ref "(OR " in
  for i = 1 to !bits do
    foo := !foo ^ (Printf.sprintf "(NEQ %s_%d %s_%d) " a i b i)
  done;
    !foo ^ ")"

let fixed_simplify_axioms = ref 
	    [
(*
             "(DEFPRED (BitAnd x y))\n" ;
	     "(DEFPRED (BitOr x y))\n" ;
	     "(DEFPRED (BitNot x))\n" ;
	     "(DEFPRED (Xor x y))\n" ;
	     "(DEFPRED (addrOf x y))\n" ;
	     (* "(DEFPRED (select y x))\n" ; *)
	     "(DEFPRED (foffset x f))\n" ;
	     "(DEFPRED (RShift x y))\n" ;
	     "(DEFPRED (LShift x y))\n" ;
	     "(DEFPRED (Div x y))\n" ;
	     "(DEFPRED (Rem x y))\n" ;
*)
  (*
  "(BG_PUSH (FORALL (x y z) (IMPLIES (AND (NEQ y 0) (EQ z (BitOr x y))) (NEQ (BitAnd z y) 0))))\n" ;
*)
  "(BG_PUSH (FORALL (x y) (EQ (select (addrOf x y) 0) x)))" ;
  "(BG_PUSH (FORALL (x y d1) (IMPLIES (EQ (foffset x d1) (foffset y d1)) (EQ  x y))))\n" ;
  "(BG_PUSH (FORALL (x y) (NEQ (addrOf x y) 0)))" ;
  "(BG_PUSH (FORALL (x y) (EQ (* (Div x y) y) x )) ) " ;
  "(BG_PUSH (FORALL (x) (NEQ (_STRINGCONSTANT x) 0 ))) " ;

  "(BG_PUSH (FORALL (s t x ) (IMPLIES (EQ s (add t x)) (EQ (in s x) 1))))";
  "(BG_PUSH (FORALL (s x) (IMPLIES (EQ s (emptyset )) (NEQ (in s x) 1))))";
(*
  "(BG_PUSH (FORALL (x y z) (IMPLIES (EQ z (addrOf x)) (EQ (select z 0) x))))\n" ;
  "(BG_PUSH (FORALL (x) (EQ (addrOf (select x 0)) x)))" ;
  "(BG_PUSH (FORALL (x y d1 d2) (IMPLIES (EQ (addrOf (select x d1)) (addrOf (select y d1))) (EQ  (addrOf x) (addrOf y)))))\n"
*)
	]



let simplify_stack_counter = ref 0
let current_simplify_stack = ref [] (* a string list *)
let simplifyServer = ref None
let simplifyServerCount = ref 0 
let simplifyCacheCount = ref 0 

let simplify_query_flag = ref false (* true *)

let set_simplify_query_flag () = 
  simplify_query_flag := (* false *) true


let rec addAxioms axiom_list =
  M.msg_string M.Minor "(Adding axioms)" ;
  let oc  = match (!simplifyServer) with
      None -> snd (getServer ()) 
    | Some(_,oc1,_) -> oc1
  in
    try
      List.iter 
	(function x -> (M.msg_string M.Debug (x^"\n"); 
			secure_output_string oc (x^"\n")))
	axiom_list ;
      if (!simplify_query_flag) then
	List.iter 
	  (function x -> ( 
	     Misc.append_to_file "simp_query" (x^"\n")))
	  axiom_list ;
      ()
    with ChannelException -> failwith "Channel Exception in addAxioms!"

  and  getServer () =
  bits := O.getValueOfInt "bits";
    match !simplifyServer with
       None -> begin
	  M.msg_string M.Normal "Forking Simplify process..." ;
	  let cmd = "Simplify -nosc " in 
	  let ic,oc,ec = Unix.open_process_full cmd 
			   (Unix.environment ()) in
	    simplifyServer := Some(ic,oc,ec) ;
	  addAxioms (!fixed_simplify_axioms);
	  let axfile = O.getValueOfString "ax" in
          if not (axfile = "") then
            begin
	      M.msg_string M.Normal ("Adding additional axioms from " ^ axfile) ;
	      try 
                let axinc = open_in axfile in
                try
                  while (true) do
                    let l = input_line axinc in 
		    M.msg_string M.Debug (l^"\n");
                    secure_output_string oc (l^"\n");
                  done;
                with End_of_file -> close_in axinc
              with Sys_error s -> M.msg_string M.Normal s
            end;
	  M.msg_string M.Normal "done!\n";
	  flush stdout ;
	  at_exit (fun () -> 
		       ignore (kill_server ()) ;  ) ;
	  (ic,oc)
	end
      | Some(a,b,c) -> (a,b)
  and  kill_server () = 
  match !simplifyServer with
      None -> ()
    | Some (ic,oc,ec) -> 
	begin
	  try
	    M.msg_string M.Error "Killing simplify server...";
	    (* ignore (Unix.close_process_full (ic,oc,ec) ); garbage process ... how to collect *)
	    ignore(Sys.command "killall Simplify");
	    simplifyServer := None;
	    M.msg_string M.Error "Done killing ..." ;
	    M.msg_string M.Normal "Done killing ..."
	  with _ -> failwith " failed to kill simplify !"
	end

(* RJ: is this maintained ? *)
let convertToSimplifyBitSyntax exp =
  (*Simplify has issues with variable names starting with "_", so we map all vars to v<num> *)
  
  (*M.msg_string M.Debug ("convertToSimplifySyntax exp = ");
    M.msg_printer M.Debug P.print exp;*)  (*RuGang*)
  
  (* let  isAtom e =   
    match e with
	E.Binary (op, e1, e2) -> 
	  begin
	    match op with 
		E.Eq  | E.Ne | E.Gt 
	      | E.Lt | E.Ge  | E.Le 
		  -> true
	      | _ -> false
	  end
      | _ -> false in *)

  let get_next_var () = next_var := !next_var + 1; !next_var in
    let newVar str = 
      if Hashtbl.mem varHashtbl str then
	(Hashtbl.find varHashtbl str)
      else
	begin
	  let new_id = "v"^(string_of_int (get_next_var ())) in
	  let (_, oc) = getServer () in
	    Hashtbl.add varHashtbl str new_id;
	    (*M.msg_string M.Debug ("NOT in hashtbl:"^str^":"^new_id);*)
	    let s =  ((assertBitVector new_id)^"\n") in
	      M.msg_string M.Debug ("new bitvector: "^s);
	      let _ = current_simplify_stack := s :: !current_simplify_stack in
	      let _ = simplify_stack_counter := !simplify_stack_counter + 1 in
		output_string oc s;
		new_id
	end in

    let rec  convertExp e =
      let chlval_convert_table e = 
	try (Hashtbl.find chlvalHashtbl e) 
	with 
	    Not_found ->
	      begin
		let s_e = get_next_var () (* E.toString e*) in
		let new_id = "chlval"^(string_of_int s_e) in
		  Hashtbl.add chlvalHashtbl e new_id;
		  new_id
	      end
      in
	match e with
	  | E.Lval lv -> convertLval lv
	  | E.Chlval (lv,s)  -> 
	      chlval_convert_table e
		
	  | E.Constant (c) -> 
	      begin
		match c with
		    Constant.Int i ->  if (i < 0) then "(- 0 "^(string_of_int (-i))^")" else string_of_int i 
		  | _ -> failwith ("Cannot handle constant "^(Constant.toString c))
	      end
		
	  | E.Unary (uop, e1) -> 
	      begin
		match uop with 
		    E.UnaryMinus -> "(- 0 "^(convertExp e1)^")"
		  | E.UnaryPlus -> convertExp e1
		  | E.Address -> failwith ("unimplemented unary")
		  | E.BitNot -> failwith ("unimplemented unary")
		  |	_ -> failwith ("unimplemented unary")
	      end
          | E.Select (e1,e2) -> Printf.sprintf "(select %s %s)" (convertExp e1) (convertExp e2)
          | E.Store (e1,e2,e3) -> Printf.sprintf "(store %s %s %s)" (convertExp e1) (convertExp e2) (convertExp e3) 
          | E.FunctionCall (E.Lval (E.Symbol f), args) ->
               Printf.sprintf "(%s %s)" f (Misc.string_list_cat " " (List.map convertExp args)) (* TBD:RP *)
	  | _ -> failwith ("convertExp (Simplify): " ^ (E.toString e) ^ "not handled")
	      
    and convertLval l =
      match l with
	| E.Symbol(str) -> 
	    newVar str
	| _ ->failwith ("unimplemented lval")  
	    
    and convertAtom e =   
    match e with
	E.Binary (E.Eq, e1, e2 ) -> 
	  begin
	    match e1 with
		E.Lval _ | E.Chlval _ ->
		  let s1 = convertExp e1 in
		    begin 
		      match e2 with
			  E.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  (bitEqS s1 newRes) (convertExpRes newRes e2)		  
			| E.Lval _
			| E.Chlval _ -> (bitEqS s1 (convertExp e2))
			| E.Constant (Constant.Int i) -> 
			    (bitEqInt s1 i)
			| _ -> failwith ("not implemented "^(E.toString e2))
		    end
	      | E.Constant (Constant.Int j) ->
		  begin 
		    match e2 with
			E.Binary _ ->
			  let newRes = newVar "result" in
			    Printf.sprintf "(AND %s %s)"  (bitEqInt newRes j) (convertExpRes newRes e2)		  
		      | E.Lval _
		      | E.Chlval _ -> Printf.sprintf "%s"  (bitEqInt (convertExp e2) j)
		      | E.Constant (Constant.Int i) -> 
			  if (i = j) then "TRUE" else "FALSE"
		      | _ -> failwith ("not implemented "^(E.toString e2))
		  end
	      | E.Binary _ -> 
		  begin
		    match e2 with
			E.Binary _ -> failwith ("not implemented "^(E.toString e2))
		      | _ -> convertAtom (E.Binary (E.Eq, e2, e1))
		  end
	      | _ ->  failwith ("not implemented")
	  end
      | E.Binary (E.Ne, e1,  e2) -> 
	  begin
	    match e1 with
		E.Lval _ | E.Chlval _ ->
		  let s1 = convertExp e1 in
		    begin 
		      match e2 with
			  E.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  
				(bitNEqS s1 newRes) (convertExpRes newRes e2) 
			| E.Lval _
			| E.Chlval _ -> (bitNEqS s1 (convertExp e2))
			| E.Constant (Constant.Int i) -> 
			    (bitNEqInt s1 i)
			| _ -> failwith ("not implemented "^(E.toString e2))
		    end
	      | E.Constant (Constant.Int j) ->
		    begin 
		      match e2 with
			  E.Binary _ ->
			    let newRes = newVar "result" in
			      Printf.sprintf "(AND %s %s)"  (bitNEqInt newRes j) (convertExpRes newRes e2)
			| E.Lval _
			| E.Chlval _ -> Printf.sprintf "%s"  (bitNEqInt (convertExp e2) j)
			| E.Constant (Constant.Int i) -> 
			    if (j = i) then "TRUE" else "FALSE" 
			| _ -> failwith ("not implemented "^(E.toString e2))
		    end
	      | E.Binary _ -> 
		  begin
		    match e2 with
			E.Binary _ -> failwith ("not implemented "^(E.toString e2))
		      | _ -> convertAtom (E.Binary (E.Ne, e2, e1))
		  end
	      | _ ->  failwith ("not implemented "^(E.toString e1))
	  end
      | _ -> failwith ("not implemented "^(E.toString e))
  
	  
    and convertExpRes res e = 
      match e with 
      | E.FunctionCall (E.Lval (E.Symbol f), args) ->
          Printf.sprintf "(EQ (%s %s) 0)" f (Misc.string_list_cat " " (List.map convertExp args))
      |	E.Binary (op, e1, e2) -> 
	    begin
	      match op with 
		| E.RShift -> 
		    begin
		      match e2 with
			E.Constant (Constant.Int i) ->
			  begin 
			    match e1 with
			      E.Binary _ ->
				let newRes = newVar "result" in
				let s = convertExpRes newRes e1 in
				(Printf.sprintf "(AND %s %s)" (rShiftCons res newRes i) s)
				| _ -> (* e1 is just a lval? *)
				    let s = convertExp e1 in
				    (Printf.sprintf "%s" (rShiftCons res s i))
			  end
		      | E.Lval _ | E.Chlval _  ->
			    let s2 = convertExp e2 in
			    begin 
			      match e1 with
				E.Binary _ ->
				  let newRes = newVar "result" in
				  let s = convertExpRes newRes e1 in
				  (Printf.sprintf "(AND %s %s)" (rShiftVar res newRes s2) s)
			      | _ -> (* e1 is just a lval? *)
				  let s = convertExp e1 in
				  (Printf.sprintf "%s" (rShiftVar res s s2))
			    end
			| _ -> failwith ("not implemented "^(E.toString e2))
		    end
		| E.Minus 
		| E.Plus   
		| E.Mul   
		| E.BitAnd 
		| E.BitOr 
		| E.Div 
		| E.Rem 
		| E.Xor 
		| E.FieldOffset 
		| E.Comma -> failwith ("not implemented")
		| _ -> 
		    if E.isRelOp op then raise AtomInExpressionException
		    else failwith ("convertExp: (Simplify) invalid op "^(E.binaryOpToVanillaString op)^" in "^(E.toString e))
	    end
     | _ -> failwith "match failure : convertExpRes"

    and convertPred e =
      match e with
	  P.True -> "TRUE"
	| P.False -> "FALSE"
	| P.And plist -> 
	    List.fold_left (function a -> function b -> "(AND "^(convertPred b)^" "^a^")") 
	      "TRUE" (plist)
	| P.Or plist -> 
	    List.fold_left (function a -> function b -> "(OR "^(convertPred b)^" "^a^")") 
	      "FALSE" (plist)
	| P.Implies (e1, e2) -> "(IMPLIES "^(convertPred e1)^" "^(convertPred e2)^")"
        | P.Iff (e1,e2) -> Printf.sprintf "(IFF %s %s)" (convertPred e1) (convertPred e2)
        | P.Not e1 -> "(NOT "^(convertPred e1)^")"
	| P.Atom e ->  (convertAtom e)
        | P.All (s,e1) -> "(FORALL ("^s ^")"^(convertPred e1)^")" 
        | P.Exist (s,e1) -> "(EXISTS ("^s ^")"^(convertPred e1)^")" 
    in
      (convertPred (P.constantfold exp (Hashtbl.create 1)))^"\n"


(*
let simplify_line_tap = ref [] (* string list *)
*)

let dump_line_tap () = ()
(*
  let oc = open_out "line_tap" in
    Marshal.to_channel oc ( List.rev !simplify_line_tap) []; 
    close_out oc;
    M.msg_string M.Error "Done writing line_tap";
    failwith "simplify choked!"
 *) 






let restart_simplify () = 
  M.msg_string M.Error "restarting simplify ...";
  M.msg_string M.Normal "restarting simplify ...";
  kill_server ();
  let (ic,oc) = getServer () in
    try
      List.iter (secure_output_string oc) (List.rev !current_simplify_stack)
    with ChannelException -> failwith "Simplify fails on restart!"
      
   
let assume pred = 
(* deadchange *)
  (* M.msg_string M.Debug ("simplify_assume");
     M.msg_string M.Debug ("sstk:"^(string_of_int
     !simplify_stack_counter));*)
  let simplifyInput = 
    match (O.getValueOfString "s") with
      | "simplify-bits" ->
          let simplifyBitInput = Bstats.time "convertToSimplifySyntax" convertToSimplifyBitSyntax pred in
          let simplifyOldInput = convertToSimplifySyntax pred in
          (Printf.sprintf "(AND %s %s)" simplifyBitInput simplifyOldInput) 
      | _ -> Bstats.time "convertToSimplifySyntax" convertToSimplifySyntax pred
  in
    clear_varHashtbl_flag := false;
    M.msg_string M.Debug ("Simplify Assuming: "^simplifyInput);
    flush stdout;
    flush stderr;
    let s = ("(BG_PUSH "^simplifyInput^" )\n") in
    let _ = current_simplify_stack := s :: !current_simplify_stack in
    let _ = simplify_stack_counter := !simplify_stack_counter + 1 in
    let (_,oc) = getServer () in
    try
      secure_output_string oc s ;
      flush oc ;     
      (* M.msg_string M.Debug "Done Assume"; *)
      flush stdout;
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" ("\n"^s^"\n")
    with  ChannelException ->
      begin
	M.msg_string M.Error "ChannelException: in simp_assume";
	restart_simplify ()
      end
	

let pop () = 
  let _ = 
    if (!simplify_stack_counter = 0) then failwith "bad simplify stack!"
    else simplify_stack_counter := !simplify_stack_counter - 1 in
  (* let top_of_stack = List.hd !current_simplify_stack in *)
  let _ = current_simplify_stack := List.tl !current_simplify_stack in
    (* M.msg_string M.Debug ("Simplify Popping"); *)
    let (_,oc) = getServer () in
    let s = "(BG_POP)\n" in
    clear_varHashtbl_flag := true;
    try
      secure_output_string oc s;
      (* M.msg_string M.Debug "Done pop";*)
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" ("\n"^s^"\n");
      flush oc
    with ChannelException ->
      begin
	M.msg_string M.Error "ChannelException: in simp_pop";
	restart_simplify ()
      end

(* doesnt clear the varHashtbl *)
let pop_noflush () =  
  let _ = 
    if (!simplify_stack_counter = 0) then failwith "bad simplify stack_nf!"
    else simplify_stack_counter := !simplify_stack_counter - 1 in
  let (_,oc) = getServer () in
  let s = "(BG_POP)\n" in
  let _ = current_simplify_stack := List.tl !current_simplify_stack in
    try
      secure_output_string oc s;
      if (!simplify_query_flag) then Misc.append_to_file "simp_query" s;
 flush oc
 with ChannelException ->
   begin
     M.msg_string M.Error "ChannelException: in simp_pop_pnf";
     restart_simplify ()
   end
 


let rec isValid ic = 
(* deadchange *)
    M.msg_string M.Debug "Simplify : in isValid";
    flush stdout;
    let line = secure_input_line ic in (* any failure here will be caught further up *)
    M.msg_string M.Debug ("Simplify says: "^line);
    if Misc.is_substring line "Bad input" then begin
	M.msg_string M.Error "Simplify poisoned!";
        exit 1 end;
    if String.contains line 'V' then true
    else if String.contains line 'I' then false
    else isValid ic
      
let ques_flag = ref false (* to see if simplify is failing forever *)

let rec queryExp pred = 
    let simplifyInput = 
        (if O.getValueOfString "s" = "simplify-bits" then 
             fun x -> (Printf.sprintf "(AND %s %s)" 
                 (Bstats.time "convertToSimplifySyntax" convertToSimplifyBitSyntax pred) x) 
         else Misc.id)
        (Bstats.time "convertToSimplifySyntax" convertToSimplifySyntax pred) in 

    M.msg_string M.Debug ("Simplify: "^simplifyInput);
    flush stdout;
    flush stderr;

    let result = try
        let ic, oc = getServer () in
	secure_output_string oc simplifyInput;
	if !simplify_query_flag then Misc.append_to_file "simp_query" ("\n"^simplifyInput^"\n");
        flush oc;
        isValid ic
    with 
      | ChannelException -> begin
          if !ques_flag then failwith "Simplify fails again!";
          ques_flag := true;
          restart_simplify ();
          queryExp pred 
      end
      | e -> failwith (
          Printf.sprintf "Simplify raised exception %s for %s. Check that Simplify is in your path." 
          (Printexc.to_string e) (P.toString pred)) in

    M.msg_string M.Debug (Printf.sprintf "%B" result);
    ques_flag := false;
    result



(* mje: THEOREMPROVER_QUERY interface *)
let stack_ctr = ref 0
      
let rec queryExpContext pred = queryExp
and block_assert p = 
    stack_ctr := !stack_ctr + 1;
    assume p
and block_pop () =
    assert (!stack_ctr > 0);
    stack_ctr := !stack_ctr - 1;
    pop_noflush ()
and block_reset () = 
    repeat pop !stack_ctr;
    stack_ctr := 0
and is_contra () = 
  begin
    M.msg_string M.Debug "simplify check contra";
    flush stdout;
    try
        let rv = queryExp (P.False) in 
        M.msg_string M.Debug ("Contra ? "^(string_of_bool rv)); 
        rv
    with e -> 
        failwith ("Simplify raised exception "^(Printexc.to_string e) 
        ^" in simplify_is_contra")
  end
and block_check_unsat p =
    ignore (block_assert p); 
    let rv = is_contra () in ignore (block_pop ()); 
    rv
and reset = kill_server
and sat = queryExp
and print_stats () = failwith "SimplifyInterface: print_stats unimplemented"

let get_allsat _ _ = failwith "SimplifyInterface: get_allsat unimplemented"

let whoami () = "Simplify"
