(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

open Ast
module S = Symbol
module C = Constant
module E = Expression
module P = Predicate
module M = Message
module O = Options

open Misc
open Printf

exception AtomInExpressionException 
exception ChannelException

type yices_expr
type yices_type
type yices_var_decl
type yices_context 
type assertion_id
type yices_model
type yices_var_decl_iterator

type lbool = int

let lbool2bool = function -1 -> Some false | 1 -> Some true | _ -> None

type mpq_t
type mpz_t

external yices_set_verbosity: int -> unit = "ml_yices_set_verbosity"
external yices_version: unit -> string = "ml_yices_version"
external yices_set_max_num_conflicts_in_maxsat_iteration: int -> unit =
    "ml_yices_set_max_num_conflicts_in_maxsat_iteration"
external yices_enable_type_checker: int -> unit = "ml_yices_enable_type_checker"
external yices_set_max_num_iteration_in_maxsat: int -> unit =
    "ml_yices_set_max_num_iterations_in_maxsat"
external yices_set_maxsat_initial_cost: int -> unit =
    "ml_yices_set_maxsat_initial_cost"
external yices_set_arith_only: int -> unit = "ml_yices_set_arith_only"
external yices_enable_log_file: string -> unit = "ml_yices_enable_log_file"

external yices_mk_context: unit -> yices_context = "ml_yices_mk_context"
external yices_del_context: yices_context -> unit = "ml_yices_del_context"
external yices_reset: yices_context -> unit = "ml_yices_reset"
external yices_dump_context: yices_context -> unit = "ml_yices_dump_context"
external yices_push: yices_context -> unit = "ml_yices_push"
external yices_pop: yices_context -> unit = "ml_yices_pop"
external yices_assert: yices_context -> yices_expr -> unit = "ml_yices_assert"

external yices_assert_weighted: yices_context -> yices_expr -> int ->
    assertion_id = "ml_yices_assert_weighted"
external yices_assert_retractable: yices_context -> yices_expr ->
    assertion_id = "ml_yices_assert_retractable"

external yices_retract: yices_context -> assertion_id -> int = "ml_yices_retract"
external yices_inconsistent: yices_context -> int = "ml_yices_inconsistent"
external yices_check: yices_context -> lbool = "ml_yices_check"

external yices_find_weighted_model: yices_context -> int -> lbool =
    "ml_yices_find_weighted_model"
external yices_evaluate_in_model: yices_model -> yices_expr -> lbool =
    "ml_yices_evaluate_in_model"
external yices_max_sat: yices_context -> lbool = "ml_yices_max_sat"
external yices_max_sat_cost_leq: yices_context -> int -> lbool =
    "ml_yices_max_sat_cost_leq"

external yices_get_model: yices_context -> yices_model = "ml_yices_get_model"
external yices_get_value: yices_model -> yices_var_decl -> lbool =
    "ml_yices_get_value"
external yices_get_int_value: yices_model -> yices_var_decl -> int * int =
    "ml_yices_get_int_value"
external yices_get_arith_value: yices_model -> yices_var_decl -> int * int * int
    = "ml_yices_get_arith_value"
external yices_get_double_value: yices_model -> yices_var_decl -> int * float =
    "ml_yices_get_double_value"
external yices_get_mpq_value: yices_model -> yices_var_decl -> int * mpq_t =
    "ml_yices_get_mpq_value"
external yices_get_mpz_value: yices_model -> yices_var_decl -> int * mpz_t =
    "ml_yices_get_mpz_value"
external yices_get_bitvector_value: yices_model -> yices_var_decl -> int * int *
    int array = "ml_yices_get_bitvector_value"
external yices_get_assertion_value: yices_model -> assertion_id -> int =
    "ml_yices_get_assertion_value"

external yices_display_model: yices_model -> unit = "ml_yices_display_model"

external yices_get_cost: yices_model -> int = "ml_yices_get_cost"
external yices_get_cost_as_double: yices_model -> float =
    "ml_yices_get_cost_as_double"

external yices_mk_true: yices_context -> yices_expr = "ml_yices_mk_true"
external yices_mk_false: yices_context -> yices_expr = "ml_yices_mk_false"
external yices_mk_bool_var: yices_context -> string -> yices_expr = "ml_yices_mk_bool_var"
external yices_mk_fresh_bool_var: yices_context -> yices_expr = "ml_yices_mk_fresh_bool_var"
external yices_get_var_decl: yices_expr -> yices_var_decl = "ml_yices_get_var_decl"
external yices_mk_bool_var_decl: yices_expr -> string -> yices_var_decl = "ml_yices_mk_bool_var_decl"
external yices_get_var_decl_name: yices_var_decl -> string = "ml_yices_get_var_decl_name"
external yices_mk_bool_from_var_decl: yices_context -> yices_var_decl -> yices_expr = "ml_yices_mk_bool_from_var_decl"

external yices_mk_or: yices_context -> yices_expr list -> yices_expr = "ml_yices_mk_or"
external yices_mk_and: yices_context -> yices_expr list -> yices_expr = "ml_yices_mk_and"
external yices_mk_eq: yices_context -> yices_expr -> yices_expr -> yices_expr = "ml_yices_mk_eq"
external yices_mk_diseq: yices_context -> yices_expr -> yices_expr -> yices_expr = "ml_yices_mk_diseq"
external yices_mk_ite: yices_context -> yices_expr -> yices_expr -> yices_expr -> yices_expr = "ml_yices_mk_ite"
external yices_mk_not: yices_context -> yices_expr -> yices_expr = "ml_yices_mk_not"

external yices_create_var_decl_iterator: yices_context ->
    yices_var_decl_iterator = "ml_yices_create_var_decl_iterator"
external yices_iterator_has_next: yices_var_decl_iterator -> int =
    "ml_yices_iterator_has_next"
external yices_iterator_next: yices_var_decl_iterator -> yices_var_decl =
    "ml_yices_iterator_next"
external yices_iterator_reset: yices_var_decl_iterator -> unit =
    "ml_yices_iterator_reset"
external yices_del_iterator: yices_var_decl_iterator -> unit =
    "ml_yices_del_iterator"

external yices_mk_type: yices_context -> string -> yices_type =
    "ml_yices_mk_type"
external yices_mk_function_type: yices_context -> yices_type list -> yices_type
    -> yices_type = "ml_yices_mk_function_type"
external yices_mk_bitvector_type: yices_context -> int -> yices_type =
    "ml_yices_mk_bitvector_type"
external yices_mk_tuple_type: yices_context -> yices_type list -> yices_type =
    "ml_yices_mk_tuple_type"

external yices_mk_var_decl: yices_context -> string -> yices_type ->
    yices_var_decl = "ml_yices_mk_var_decl"
external yices_get_var_decl_from_name: yices_context -> string -> yices_var_decl =
    "ml_yices_get_var_decl_from_name"
external yices_mk_var_from_decl: yices_context -> yices_var_decl -> yices_expr =
    "ml_yices_mk_var_from_decl"

external yices_mk_app: yices_context -> yices_expr -> yices_expr list ->
    yices_expr = "ml_yices_mk_app"
external yices_mk_num: yices_context -> int -> yices_expr = "ml_yices_mk_num"
external yices_mk_num_from_string: yices_context -> string -> yices_expr = "ml_yices_mk_num_from_string"
external yices_mk_num_from_mpz: yices_context -> mpz_t -> yices_expr = "ml_yices_mk_num_from_mpz"
external yices_mk_num_from_mpq: yices_context -> mpq_t -> yices_expr = "ml_yices_mk_num_from_mpq"

external yices_mk_sum: yices_context -> yices_expr list -> yices_expr =
    "ml_yices_mk_sum"
external yices_mk_sub: yices_context -> yices_expr list -> yices_expr =
    "ml_yices_mk_sub"
external yices_mk_mul: yices_context -> yices_expr list -> yices_expr =
    "ml_yices_mk_mul"
external yices_mk_lt: yices_context -> yices_expr -> yices_expr -> yices_expr =
    "ml_yices_mk_lt"
external yices_mk_le: yices_context -> yices_expr -> yices_expr -> yices_expr =
    "ml_yices_mk_le"
external yices_mk_gt: yices_context -> yices_expr -> yices_expr -> yices_expr =
    "ml_yices_mk_gt"
external yices_mk_ge: yices_context -> yices_expr -> yices_expr -> yices_expr =
    "ml_yices_mk_ge"
external yices_mk_bv_constant: yices_context -> int -> int ->
    yices_expr = "ml_yices_mk_bv_constant"
external yices_mk_bv_constant_from_array: yices_context -> int -> int array ->
    yices_expr = "ml_yices_mk_bv_constant_from_array"
external yices_mk_bv_add: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_add"
external yices_mk_bv_sub: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_sub"
external yices_mk_bv_mul: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_mul"
external yices_mk_bv_minus: yices_context -> yices_expr -> yices_expr =
    "ml_yices_mk_bv_minus"
external yices_mk_bv_concat: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_concat"
external yices_mk_bv_and: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_and"
external yices_mk_bv_or: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_or"
external yices_mk_bv_xor: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_xor"
external yices_mk_bv_not: yices_context -> yices_expr -> yices_expr =
    "ml_yices_mk_bv_not"
external yices_mk_bv_extract: yices_context -> int -> int -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_extract"
external yices_mk_bv_sign_extend: yices_context -> yices_expr -> int ->
    yices_expr = "ml_yices_mk_bv_sign_extend"
external yices_mk_bv_shift_left0: yices_context -> yices_expr -> int ->
    yices_expr = "ml_yices_mk_bv_shift_left0"
external yices_mk_bv_shift_left1: yices_context -> yices_expr -> int ->
    yices_expr = "ml_yices_mk_bv_shift_left1"
external yices_mk_bv_shift_right0: yices_context -> yices_expr -> int ->
    yices_expr = "ml_yices_mk_bv_shift_right0"
external yices_mk_bv_shift_right1: yices_context -> yices_expr -> int ->
    yices_expr = "ml_yices_mk_bv_shift_right1"
external yices_mk_bv_lt: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_lt"
external yices_mk_bv_le: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_le"
external yices_mk_bv_gt: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_gt"
external yices_mk_bv_ge: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_ge"
external yices_mk_bv_slt: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_slt"
external yices_mk_bv_sle: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_sle"
external yices_mk_bv_sgt: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_sgt"
external yices_mk_bv_sge: yices_context -> yices_expr -> yices_expr ->
    yices_expr = "ml_yices_mk_bv_sge"

external yices_pp_expr: yices_expr -> unit = "ml_yices_pp_expr"


let symbols = Hashtbl.create 100
let yicesSymbols = Hashtbl.create 100
let yicesSymbolDefStack = Stack.create ()
let _ = Stack.push [] yicesSymbolDefStack
let unknownExprs = Hashtbl.create 100

let stackCounter = ref 0
let yicesContext = ref None
let contextUnsat = ref false
let contextUnsatPushes = ref 0

let ptr_type ctx = yices_mk_type ctx "ptr-type"
and int_type ctx = yices_mk_type ctx "int"
and field_type ctx = yices_mk_type ctx "field"
and set_type ctx = yices_mk_type ctx "set"
and bool_type ctx = yices_mk_type ctx "bool"

let mk_arrow ctx ls = yices_mk_function_type ctx (list_init ls) (list_last ls)

let yicesDefinitions ctx = [
    Left ("ptr-type", None);
    Right ("null", ptr_type ctx, None);
    Right ("sel", mk_arrow ctx [ptr_type ctx; int_type ctx; int_type ctx], None);
    Right ("addr", mk_arrow ctx [int_type ctx; ptr_type ctx], None);

    Left ("field", None);
    Right ("offset", mk_arrow ctx [ptr_type ctx; field_type ctx; ptr_type ctx], None); 

    Left ("set", None);
    Right ("empty", set_type ctx, None);
    Right ("add", mk_arrow ctx [set_type ctx; int_type ctx; set_type ctx], None);
    Right ("in", mk_arrow ctx [int_type ctx; set_type ctx; bool_type ctx], None);
]

and yicesAxioms = [
    "(forall (x::int) (= (sel (addr x) 0) x))";
    "(forall (p::ptr-type q::ptr-type f::field) (=> (= (offset p f) (offset q f)) (= p q)))";
    "(forall (x::int) (/= (addr x) null))";
    "(forall (x::int y::int) (= (* (div x y) y) x))";
    "(forall (s::set x::int) (in x (add s x)))";
    (* XXX: "(forall (s::set t::set x::int) (=> (= s (add t x)) (in s x)))"; *)
    "(forall (x::int) (not (in x empty)))";
    (* XXX: "(forall (s::set x::int) (=> (= s empty) (not (in x s))))"; *)
]

let rec getContext () =
    maybe'' (fun () ->
        let ctx = yices_mk_context () in
        yicesContext := Some ctx;
        match lbool2bool <| yices_check ctx with
        | Some false -> failwith ("yices library failed sanity check");
        | _ -> ();
        List.iter 
            (either (uncurry <| yicesDefineType ctx) (uncurry3 <| yicesDefine ctx))
            <| yicesDefinitions ctx;
        (* List.iter yicesAssert yicesAxioms; *)
        ctx

    ) id !yicesContext

and deleteContext () =
    maybe () yices_del_context !yicesContext

and renewContext () =
    deleteContext ();
    ignore <| getContext ()

and isDefined sym =
    Hashtbl.mem yicesSymbols sym

and yicesDefineType ctx typ def_m =
    if not !contextUnsat then begin
        M.msg_string M.Debug <|
            sprintf "YIe: telling yices '(define-type %s ...)'" typ;
        match def_m with 
          | Some def -> ignore <| yices_mk_type ctx typ
          | None     -> ignore <| yices_mk_type ctx typ
    end

and yicesDefine ctx sym typ def_m = 
    if not !contextUnsat then begin
        M.msg_string M.Debug <| 
            sprintf "YIe: telling yices '(define %s ...)'" sym;
        match def_m with
          | Some def -> ignore <| yices_mk_var_decl ctx sym typ
          | None     -> ignore <| yices_mk_var_decl ctx sym typ;
        Hashtbl.add yicesSymbols sym typ;
        Stack.push (sym::(Stack.pop yicesSymbolDefStack)) yicesSymbolDefStack
    end

and yicesAssert (expr : yices_expr) = 
    let ctx = getContext () in
    if not !contextUnsat then begin
        M.msg_string M.Debug "YIe: telling yices '(assert ...)'";
        ignore <| yices_assert ctx expr
    end

and yicesPush () = 
    let ctx = getContext () in 
    Stack.push [] yicesSymbolDefStack;
    if !contextUnsat then incr contextUnsatPushes
    else begin
        M.msg_string M.Debug "YIe: telling yices '(push)'";
        ignore <| yices_push ctx
    end;
    incr stackCounter

and yicesPop () = 
    let ctx = getContext () in
    assert (!stackCounter > 0);
    List.iter 
        (Hashtbl.remove yicesSymbols) 
        (Stack.pop yicesSymbolDefStack);
    if !contextUnsat && !contextUnsatPushes > 0 then
        decr contextUnsatPushes
    else begin 
        if !contextUnsat then contextUnsat := false;
        M.msg_string M.Debug "YIe: telling yices '(pop)'";
        ignore <| yices_pop ctx
    end;
    decr stackCounter

and yicesPopAll () =
    repeat yicesPop !stackCounter

and yicesCheck () = 
    let ctx = getContext () in
    if !contextUnsat then false
    else begin
        M.msg_string M.Debug "YIe: telling yices '(check)'";
        match lbool2bool <| yices_check ctx with
        | Some outcome -> begin
            M.msg_string M.Debug 
                << sprintf "YIe: yices says %s" 
                <| if outcome then "sat" else "unsat"; 
            outcome
        end
        | None -> 
            M.msg_string M.Debug "YIe: yices says unknown (we'll say sat)";
            true
        (* | None -> begin
            M.msg_string M.Debug "YIe: yices doesn't know (sat)";
            true
        end *)
    end

and yicesReset () = 
    let ctx = getContext () in 
    contextUnsat := false;
    contextUnsatPushes := 0;
    stackCounter := 0;
    M.msg_string M.Debug "YIe: telling yices '(reset)'";
    yices_reset ctx
    
(**
 * Conversion to yices input syntax
 *)

let cannotHandle str where =
    failwith (sprintf "YIe: cannot handle '%s' in %s" str where)

and sayCannotHandle str where =
    M.msg_string M.Debug (sprintf "YIe: cannot handle '%s' in %s" str where)

let get_var ctx name = 
    yices_mk_var_from_decl ctx <| 
    yices_get_var_decl_from_name ctx name

let mk_app ctx name args =
    yices_mk_app ctx (get_var ctx name) args

let zero ctx = yices_mk_num ctx 0
and one ctx = yices_mk_num ctx 1

let freshUnknownSymbol =
    let freshUnknownSymbolCounter = ref 0 in
    fun () ->
        let sym = sprintf "unknown_%d" !freshUnknownSymbolCounter in
        incr freshUnknownSymbolCounter;
        sym

let rec predToYicesExpr ctx pred =
    match pred with
      | P.True -> yices_mk_true ctx
      | P.False -> yices_mk_false ctx
      | P.Atom e -> atomToYicesExpr ctx e
      | P.And [] -> yices_mk_true ctx
      | P.And plist -> yices_mk_and ctx <| List.map (predToYicesExpr ctx) plist
      | P.Or [] -> yices_mk_false ctx
      | P.Or plist -> yices_mk_or ctx <| List.map (predToYicesExpr ctx) plist
      | P.Implies (e, e') -> predToYicesExpr ctx <| P.Or [ P.Not e; e' ]
      | P.Not e -> yices_mk_not ctx <| predToYicesExpr ctx e
      | P.Iff (e, e') -> predToYicesExpr ctx <| P.And [ P.Implies (e,e'); P.Implies (e',e) ]
      | _ -> cannotHandle (P.toString pred) "predToYicesExpr"

and atomToYicesExpr ctx expr =

    match expr with
      | E.Binary (op, e, e') -> begin
          try (match op with
                | E.Eq -> yices_mk_eq
                | E.Ne -> yices_mk_diseq
                | E.Le -> yices_mk_le
                | E.Lt -> yices_mk_lt
                | E.Gt -> yices_mk_gt
                | E.Ge -> yices_mk_ge
                | _ -> cannotHandle (E.binaryOpToString op) "atomToYicesExpr")
            ctx
            (exprToYicesExpr ctx e) 
            (exprToYicesExpr ctx e')

          with AtomInExpressionException -> begin
              match op with E.Eq | E.Ne -> begin
                  match e, e' with
                  | E.Lval _, _ | E.Chlval _, _ | E.Constant (Constant.Int _), _ ->
                      let s1 = exprToYicesExpr ctx e
                      and s2 = atomToYicesExpr ctx e' in
                      yices_mk_or ctx [
                          yices_mk_and ctx [ s2; yices_mk_eq ctx s1 (one ctx) ] ;
                          yices_mk_and ctx [ yices_mk_not ctx s2; yices_mk_eq ctx s1 (zero ctx)]
                      ]

                  | _, E.Lval _ | _, E.Chlval _ | _, E.Constant (Constant.Int _) ->
                      let s1 = exprToYicesExpr ctx e'
                      and s2 = atomToYicesExpr ctx e in
                      yices_mk_or ctx [ 
                          yices_mk_and ctx [ s2; yices_mk_eq ctx s1 (one ctx) ] ; 
                          yices_mk_and ctx [ yices_mk_not ctx s2; yices_mk_eq ctx s1 (zero ctx)] 
                      ]

                  | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (*)"
              end
              | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (**)"
          end
          | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (***)"
      end

      | E.FunctionCall (E.Lval (E.Symbol _), args) ->
          yices_mk_eq ctx (exprToYicesExpr ctx expr) (zero ctx)

      | E.Lval (E.Symbol s) -> symbolToYicesExpr ctx s

      | _ -> cannotHandle (E.toString expr) "atomToYicesExpr"

and exprToYicesExpr ctx expr =
    match expr with
      | E.Lval lv -> lvalToYicesExpr ctx lv
      | E.Binary (op, e, e') -> 
          (match op with
            | E.Minus -> yices_mk_sub
            | E.Plus -> yices_mk_sum
            | E.Mul -> yices_mk_mul
            | E.FieldOffset -> (fun ctx -> mk_app ctx "offset")
            | _ when E.isRelOp op ->
                    raise AtomInExpressionException
            | _ -> cannotHandle (E.binaryOpToString op) 
                "exprToYicesExpr(Binary)")
          ctx
          [ exprToYicesExpr ctx e ; exprToYicesExpr ctx e' ]

      | E.Constant c -> begin
          match c with
            | Constant.Int i -> yices_mk_num ctx i
            | _ -> cannotHandle (C.toString c) "exprToYicesExpr(Constant)"
      end

      | E.Unary (op, e) ->
          (match op with
            | E.UnaryMinus -> (fun es -> yices_mk_sub ctx ((zero ctx)::es))
            | E.UnaryPlus -> List.hd
            | E.Address -> mk_app ctx "addr"
            | _ -> cannotHandle (E.unaryOpToString op) 
                "exprToYicesExpr(Unary)")
          [ exprToYicesExpr ctx e ]

      | E.FunctionCall (E.Lval (E.Symbol f), args) ->
          mk_app ctx "fn_call" <|
              symbolToYicesExpr ctx f ::
              List.map (exprToYicesExpr ctx) args

      | E.Select (e,e') -> 
          mk_app ctx "sel" [ 
              exprToYicesExpr ctx e ; 
              exprToYicesExpr ctx e' ]

      | E.Store (e,e',e'') ->
          mk_app ctx "store" [
              exprToYicesExpr ctx e; 
              exprToYicesExpr ctx e'; 
              exprToYicesExpr ctx e'' ]

      | _ -> 
          sayCannotHandle (E.toString expr) "exprToYicesExpr";
          unknownExprSymbol ctx expr

and lvalToYicesExpr ctx lval =
    match lval with
      | E.Symbol s -> symbolToYicesExpr ctx s

      | E.Access (E.Dot, e, s) -> 
          mk_app ctx "offset" [ 
              exprToYicesExpr ctx e ; 
              symbolToYicesExpr ctx s ]

      | E.Dereference e -> 
          mk_app ctx "sel" [ exprToYicesExpr ctx e ; zero ctx ]

      | E.Indexed (e, e') -> 
          mk_app ctx "sel" [
              exprToYicesExpr ctx e ;
              exprToYicesExpr ctx e' ]

      | _ -> cannotHandle (E.lvalToString lval) "lvalToYicesExpr"

and unknownExprSymbol ctx expr = 
    symbolToYicesExpr ctx (
        if Hashtbl.mem unknownExprs expr then
            Hashtbl.find unknownExprs expr
        else begin
            let s = freshUnknownSymbol () in
            Hashtbl.add unknownExprs expr s;
            s
        end)

and symbolToYicesExpr ctx sym =
    let yicesSym = 
        if Hashtbl.mem symbols sym then 
            Hashtbl.find symbols sym 
        else begin 
            let s = S.toString sym in 
            Hashtbl.add symbols sym s; 
            s
        end in

    if not (isDefined yicesSym) then
        yicesDefine ctx yicesSym (int_type ctx) None;

    get_var ctx yicesSym

(**
 * Implementation of the THEOREM_PROVER signature
 *)

let block_reset () = 
    yicesPopAll ()

and block_assert pred = 
    let ctx = getContext () in
    yicesPush ();
    yicesAssert << predToYicesExpr ctx <| pred;

and block_check_unsat pred = 
    let ctx = getContext () in
    yicesPush ();
    yicesAssert << predToYicesExpr ctx <| pred;
    let outcome = yicesCheck () in
    yicesPop ();
    outcome = false

and is_contra () =
    let outcome = yicesCheck () in
    let negOutcome = if outcome then false else true in
    negOutcome


let block_pop () = 
    yicesPop ()

let reset () = 
    renewContext ()

let queryExp exp =
    block_check_unsat (P.Not exp)
    (* yicesPush ();
    let ctx = getContext () in
    yicesAssert (predToYicesExpr ctx exp);
    let outcome = yicesCheck () in
    yicesPop ();
    outcome <> "unsat" *)

let queryExpContext pred query = 
    let ctx = getContext () in
    yicesPush ();
    (* yicesAssert (predToYicesExpr pred); *)
    yicesAssert (yices_mk_not ctx <| predToYicesExpr ctx query);
    let outcome = not <| yicesCheck () in
    yicesPop ();
    outcome

let print_stats () = 
    failwith "YicesInterface2.print_stats not implemented"

and sat pred = 
    let ctx = getContext () in
    yicesPush ();
    yicesAssert << predToYicesExpr ctx <| pred;
    let c = yicesCheck () in
    yicesPop ();
    c

and assume pred =
    let ctx = getContext () in
    yicesAssert << predToYicesExpr ctx <| pred

and pop () = 
    yicesPop ()

and getValue name =
    let ctx = getContext () in
    let mdl = yices_get_model ctx in
    lbool2bool << yices_get_value mdl <| yices_get_var_decl_from_name ctx name

and getIntValue name =
    let ctx = getContext () in
    let mdl = yices_get_model ctx in
    snd << yices_get_int_value mdl <| yices_get_var_decl_from_name ctx name

let get_allsat p plist = failwith "Yices: allsat: To do" 
let whoami () = 
  Message.msg_string Message.Normal "Yices. Copyright SRI." ; "Yices"
