(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

open Ast
module S = Symbol
module C = Constant
module E = Expression
module P = Predicate
module M = Message
module O = Options

open Misc
open Printf

exception AtomInExpressionException 
exception ChannelException

type yices_expr
type yices_type
type yices_var_decl
type yices_context 
type assertion_id
type yices_model
type yices_var_decl_iterator

type lbool = int

let lbool2bool = function -1 -> Some false | 1 -> Some true | _ -> None

type mpq_t
type mpz_t

external yices_set_verbosity: int -> unit = "ml_yices_set_verbosity"
external yices_version: unit -> string = "ml_yices_version"
external yices_set_max_num_conflicts_in_maxsat_iteration: int -> unit =
    "ml_yices_set_max_num_conflicts_in_maxsat_iteration"
external yices_enable_type_checker: int -> unit = "ml_yices_enable_type_checker"
external yices_set_max_num_iteration_in_maxsat: int -> unit =
    "ml_yices_set_max_num_iterations_in_maxsat"
external yices_set_maxsat_initial_cost: int -> unit =
    "ml_yices_set_maxsat_initial_cost"
external yices_set_arith_only: int -> unit = "ml_yices_set_arith_only"
external yices_enable_log_file: string -> unit = "ml_yices_enable_log_file"

external yices_mk_context: unit -> yices_context = "ml_yices_mk_context"
external yices_del_context: yices_context -> unit = "ml_yices_del_context"
external yices_reset: yices_context -> unit = "ml_yices_reset"
external yices_dump_context: yices_context -> unit = "ml_yices_dump_context"
external yices_push: yices_context -> unit = "ml_yices_push"
external yices_pop: yices_context -> unit = "ml_yices_pop"
external yices_assert: yices_context -> yices_expr -> unit = "ml_yices_assert"

external yices_assert_weighted: yices_context -> yices_expr -> int ->
    assertion_id = "ml_yices_assert_weighted"
external yices_assert_retractable: yices_context -> yices_expr ->
    assertion_id = "ml_yices_assert_retractable"

external yices_retract: yices_context -> assertion_id -> int = "ml_yices_retract"
external yices_inconsistent: yices_context -> int = "ml_yices_inconsistent"
external yices_check: yices_context -> lbool = "ml_yices_check"

external yices_find_weighted_model: yices_context -> int -> lbool =
    "ml_yices_find_weighted_model"
external yices_evaluate_in_model: yices_model -> yices_expr -> lbool =
    "ml_yices_evaluate_in_model"
external yices_max_sat: yices_context -> lbool = "ml_yices_max_sat"
external yices_max_sat_cost_leq: yices_context -> int -> lbool =
    "ml_yices_max_sat_cost_leq"

external yices_get_model: yices_context -> yices_model = "ml_yices_get_model"
external yices_get_value: yices_model -> yices_var_decl -> lbool =
    "ml_yices_get_value"
external yices_get_int_value: yices_model -> yices_var_decl -> int * int =
    "ml_yices_get_int_value"
external yices_get_arith_value: yices_model -> yices_var_decl -> int * int * int
    = "ml_yices_get_arith_value"
external yices_get_double_value: yices_model -> yices_var_decl -> int * float =
    "ml_yices_get_double_value"
external yices_get_mpq_value: yices_model -> yices_var_decl -> int * mpq_t =
    "ml_yices_get_mpq_value"
external yices_get_mpz_value: yices_model -> yices_var_decl -> int * mpz_t =
    "ml_yices_get_mpz_value"
external yices_get_bitvector_value: yices_model -> yices_var_decl -> int * int *
    int array = "ml_yices_get_bitvector_value"
external yices_get_assertion_value: yices_model -> assertion_id -> int =
    "ml_yices_get_assertion_value"

external yices_display_model: yices_model -> unit = "ml_yices_display_model"

external yices_get_cost: yices_model -> int = "ml_yices_get_cost"
external yices_get_cost_as_double: yices_model -> float =
    "ml_yices_get_cost_as_double"

external yices_create_var_decl_iterator: yices_context ->
    yices_var_decl_iterator = "ml_yices_create_var_decl_iterator"
external yices_iterator_has_next: yices_var_decl_iterator -> int =
    "ml_yices_iterator_has_next"
external yices_iterator_next: yices_var_decl_iterator -> yices_var_decl =
    "ml_yices_iterator_next"
external yices_iterator_reset: yices_var_decl_iterator -> unit =
    "ml_yices_iterator_reset"
external yices_del_iterator: yices_var_decl_iterator -> unit =
    "ml_yices_del_iterator"

external yices_mk_fresh_bool_var : yices_context -> yices_expr = "ml_yices_mk_fresh_bool_var"

external yices_mk_var_decl: yices_context -> string -> yices_type ->
    yices_var_decl = "ml_yices_mk_var_decl"
external yices_get_var_decl_from_name: yices_context -> string -> yices_var_decl =
    "ml_yices_get_var_decl_from_name"
external yices_mk_var_from_decl: yices_context -> yices_var_decl -> yices_expr =
    "ml_yices_mk_var_from_decl"

external yices_pp_expr: yices_expr -> unit = "ml_yices_pp_expr"

type yicesl_context
external yicesl_read: yicesl_context -> string -> int = "ml_yicesl_read"
external yices_get_lite_context: yices_context -> yicesl_context =
    "ml_yices_get_lite_context"


let symbols = Hashtbl.create 100
let yicesSymbols = Hashtbl.create 100
let yicesSymbolDefStack = Stack.create ()
let _ = Stack.push [] yicesSymbolDefStack
let unknownExprs = Hashtbl.create 100

let stackCounter = ref 0
let yicesContext = ref None
let yicesLiteContext = ref None
let contextUnsat = ref false
let contextUnsatPushes = ref 0

(* let _ = 
    yices_set_verbosity 10;
    yices_enable_log_file "YICES.LOG" *)

let yicesDefinitions = [
    Left ("ptr-type", None);
    Right ("null", "ptr-type", None);
    Right ("sel", "(-> ptr-type int int)", None);
    Right ("addr", "(-> int ptr-type)", None);

    Left ("field", None);
    Right ("offset", "(-> ptr-type field ptr-type)", None);

    Left ("set", None);
    Right ("empty", "set", None);
    Right ("add", "(-> set int set)", None);
    Right ("in", "(-> int set bool)", None);
]

let yicesAxioms = [ 
    "(forall (x::int) (= (sel (addr x) 0) x))";
    "(forall (p::ptr-type q::ptr-type f::field) (=> (= (offset p f) (offset q f)) (= p q)))";
    "(forall (x::int) (/= (addr x) null))";
    (* "(forall (x::int y::int) (= (* (div x y) y) x))"; *)
    "(forall (s::set x::int) (in x (add s x)))";
    (* XXX: "(forall (s::set t::set x::int) (=> (= s (add t x)) (in s x)))"; *)
    "(forall (x::int) (not (in x empty)))";
    (* XXX: "(forall (s::set x::int) (=> (= s empty) (not (in x s))))"; *)
]
    

let rec getContext () =
    maybe'' (fun () ->
        let ctx = yices_mk_context () in
        yicesContext := Some ctx;

        yicesLiteContext := Some (yices_get_lite_context ctx);

        yices_push ctx;

        match lbool2bool <| yices_check ctx with
        | Some false -> failwith ("yices library failed sanity check");
        | _ -> ();
        List.iter 
            (either (uncurry <| yicesDefineType) (uncurry3 <| yicesDefine))
            <| yicesDefinitions;
        List.iter yicesAssert yicesAxioms;
        ctx

    ) id !yicesContext

and deleteContext () =
    maybe () yices_del_context !yicesContext;
    yicesContext := None;
    yicesLiteContext := None;
    contextUnsat := false;
    contextUnsatPushes := 0;
    stackCounter := 0;

and renewContext () =
    deleteContext ();
    ignore <| getContext ()

and yicesLiteCommand cmd_s =
    maybe'' (ignore << getContext) ignore !yicesLiteContext;
    let ctx = some id !yicesLiteContext in

    M.msg_string M.Debug <| sprintf "YIe: telling yices '%s'" cmd_s;

    if yicesl_read ctx cmd_s = 0 then 
        failwith "YIe: something bad happened";

and isDefined sym =
    Hashtbl.mem yicesSymbols sym

and yicesDefineType typ def_m =
    if not !contextUnsat then begin
        yicesLiteCommand <| 
            match def_m with 
            | Some def -> sprintf "(define-type %s %s)" typ def
            | None     -> sprintf "(define-type %s)" typ
    end

and yicesDefine sym typ def_m = 
    if not !contextUnsat then begin
        yicesLiteCommand (
            match def_m with 
            | Some def -> sprintf "(define %s::%s %s)" sym typ def
            | None     -> sprintf "(define %s::%s)" sym typ);

        Hashtbl.add yicesSymbols sym typ;
        Stack.push (sym::(Stack.pop yicesSymbolDefStack)) yicesSymbolDefStack
    end

and yicesAssert expr_s = 
    if not !contextUnsat then 
        yicesLiteCommand <| sprintf "(assert %s)" expr_s

and yicesPush () = 
    let ctx = getContext () in 
    Stack.push [] yicesSymbolDefStack;
    if !contextUnsat then incr contextUnsatPushes
    else begin
        M.msg_string M.Debug "YIe: telling yices '(push)'";
        yices_push ctx
    end;
    incr stackCounter

and yicesPop () = 
    let ctx = getContext () in
    assert (!stackCounter > 0);
    List.iter 
        (Hashtbl.remove yicesSymbols) 
        (Stack.pop yicesSymbolDefStack);
    if !contextUnsat && !contextUnsatPushes > 0 then
        decr contextUnsatPushes
    else begin 
        if !contextUnsat then contextUnsat := false;
        M.msg_string M.Debug "YIe: telling yices '(pop)'";
        yices_pop ctx
    end;
    decr stackCounter

and yicesPopAll () =
    repeat yicesPop !stackCounter

and yicesCheck () = 
    let ctx = getContext () in
    if !contextUnsat then false
    else begin
        M.msg_string M.Debug "YIe: telling yices '(check)'";
        match lbool2bool <| yices_check ctx with
        | Some outcome -> begin
            M.msg_string M.Debug 
                << sprintf "YIe: yices says %s" 
                <| if outcome then "sat" else "unsat"; 
            outcome
        end
        | None -> 
            M.msg_string M.Debug "YIe: yices says unknown (we'll say sat)";
            true
    end

and yicesReset () = 
    let ctx = getContext () in 
    contextUnsat := false;
    contextUnsatPushes := 0;
    stackCounter := 0;
    M.msg_string M.Debug "YIe: telling yices '(reset)'";
    yices_reset ctx
    
(**
 * Conversion to yices input syntax
 *)

let cannotHandle str where =
    failwith (sprintf "YIe: cannot handle '%s' in %s" str where)

and sayCannotHandle str where =
    M.msg_string M.Debug (sprintf "YIe: cannot handle '%s' in %s" str where)

let freshUnknownSymbol =
    let freshUnknownSymbolCounter = ref 0 in
    fun () ->
        let sym = sprintf "unknown_%d" !freshUnknownSymbolCounter in
        incr freshUnknownSymbolCounter;
        sym

let rec predToYicesExpr pred =
    match pred with
      | P.True -> "true"
      | P.False -> "false"
      | P.Atom e -> atomToYicesExpr e

      | P.And [] -> "true"
      | P.And plist -> 
          sprintf "(and %s)" << String.concat " " << List.map predToYicesExpr
          <| plist

      | P.Or [] -> "false"
      | P.Or plist ->
          sprintf "(or %s)" << String.concat " " << List.map predToYicesExpr
          <| plist

      | P.Implies (e, e') ->
          sprintf "(=> %s %s)" (predToYicesExpr e) (predToYicesExpr e');

      | P.Not e -> sprintf "(not %s)" << predToYicesExpr <| e

      | P.Iff (e, e') -> 
          let s, s' = predToYicesExpr e, predToYicesExpr e' in
          sprintf "(and (=> %s %s) (=> %s %s))" s s' s' s

      | P.All (s,e) -> 
          sprintf "(forall (%s::%s) %s)"
              s 
              "int" 
              (predToYicesExpr e)

      | P.Exist (s,e) -> 
          sprintf "(exists (%s::%s) %s)"
              s 
              "int" 
              (predToYicesExpr e)
      | _ -> cannotHandle (P.toString pred) "predToYicesExpr"

and atomToYicesExpr expr =
    match expr with
      | E.Binary(E.And, e, e') -> sprintf "(and %s %s)" (atomToYicesExpr e) (atomToYicesExpr e')
      | E.Binary(E.Or, e, e')  -> sprintf "(or %s %s)" (atomToYicesExpr e) (atomToYicesExpr e')
      | E.Unary(E.Not, e)      -> sprintf "(not %s)" (atomToYicesExpr e) 
      | E.Binary (op, e, e')   -> begin
          try sprintf "(%s %s %s)" (match op with
                | E.Eq -> "="
                | E.Ne -> "/="
                | E.Le -> "<="
                | E.Lt -> "<"
                | E.Gt -> ">"
                | E.Ge -> ">="
                | _ -> cannotHandle (E.binaryOpToString op) "atomToYicesExpr")
              (exprToYicesExpr e) (exprToYicesExpr e')

          with AtomInExpressionException -> begin
              match op with E.Eq | E.Ne -> begin
                  match e, e' with
                  | E.Lval _, _ | E.Chlval _, _ | E.Constant (Constant.Int _), _ ->
                      let s1 = exprToYicesExpr e
                      and s2 = atomToYicesExpr e' in
                      sprintf "(or (and %s (= %s 1)) (and (not %s) (= %s 0)))" s2 s1 s2 s1
                  | _, E.Lval _ | _, E.Chlval _ | _, E.Constant (Constant.Int _) ->
                      let s1 = exprToYicesExpr e'
                      and s2 = atomToYicesExpr e in
                      sprintf "(or (and %s (= %s 1)) (and (not %s) (= %s 0)))" s2 s1 s2 s1
                  | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (*)"
              end
              | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (**)"
          end
          | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (***)"
      end

      | E.FunctionCall (E.Lval (E.Symbol _), args) ->
          sprintf "(= %s 0)" (exprToYicesExpr expr)

      | E.Lval (E.Symbol s) -> symbolToYicesExpr s

      | _ -> cannotHandle (E.toString expr) "atomToYicesExpr"

and exprToYicesExpr expr =
    match expr with
      | E.Lval lv -> lvalToYicesExpr lv
      | E.Binary (op, e, e') -> 
          sprintf "(%s %s %s)" (match op with
            | E.Minus -> "-"
            | E.Plus -> "+"
            | E.Mul -> "*"
            | E.Div -> "/"
            | E.Rem -> "mod"
            | E.FieldOffset -> "offset"
            | _ when E.isRelOp op ->
                    raise AtomInExpressionException
            | _ -> cannotHandle (E.binaryOpToString op) 
                "exprToYicesExpr(Binary)")
          (exprToYicesExpr e) (exprToYicesExpr e')

      | E.Constant c -> begin
          match c with
            | Constant.Int i -> sprintf "%d" i
            | Constant.String s -> sprintf "\"%s\"" s
            | Constant.Float f -> sprintf "%f" f
            | _ -> cannotHandle (C.toString c) "exprToYicesExpr(Constant)"
      end

      | E.Unary (op, e) ->
          sprintf "(%s %s)" (match op with
            | E.UnaryMinus -> "-"
            | E.UnaryPlus -> "+"
            | E.Address -> "addr"
            | _ -> cannotHandle (E.unaryOpToString op) 
                "exprToYicesExpr(Unary)")
          (exprToYicesExpr e)

      | E.FunctionCall (E.Lval (E.Symbol f), args) ->
          sprintf "(fn_call %s %s)" 
              (symbolToYicesExpr f)
              (String.concat " " (List.map exprToYicesExpr args))

      | E.Select (e,e') -> 
          sprintf "(sel %s %s)" 
              (exprToYicesExpr e)
              (exprToYicesExpr e')

      | E.Store (e,e',e'') ->
          sprintf "(store %s %s %s)" 
              (exprToYicesExpr e)
              (exprToYicesExpr e')
              (exprToYicesExpr e'')

      | _ -> 
          sayCannotHandle (E.toString expr) "exprToYicesExpr";
          unknownExprSymbol expr

and lvalToYicesExpr lval =
    match lval with
      | E.Symbol s -> symbolToYicesExpr s

      | E.Access (E.Dot, e, s) -> 
          sprintf "(offset %s %s)"
              (exprToYicesExpr e)
              (symbolToYicesExpr s)

      | E.Dereference e -> 
          sprintf "(sel %s 0)" 
              (exprToYicesExpr e)

      | E.Indexed (e, e') -> 
          sprintf "(sel %s %s)" 
              (exprToYicesExpr e) 
              (exprToYicesExpr e')

      | _ -> cannotHandle (E.lvalToString lval) "lvalToYicesExpr"

and unknownExprSymbol expr = 
    symbolToYicesExpr (
        if Hashtbl.mem unknownExprs expr then
            Hashtbl.find unknownExprs expr
        else begin
            let s = freshUnknownSymbol () in
            Hashtbl.add unknownExprs expr s;
            s
        end)

and symbolToYicesExpr sym =
    let yicesSym = 
        if Hashtbl.mem symbols sym then 
            Hashtbl.find symbols sym 
        else begin 
            let s = S.toString sym in 
            Hashtbl.add symbols sym s; 
            s
        end in

    if not (isDefined yicesSym) then
        yicesDefine yicesSym "int" None;

    yicesSym

(**
 * Implementation of the THEOREM_PROVER signature
 *)

let block_reset () = 
    yicesPopAll ()

and block_assert pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;

and block_check_unsat pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;
    let outcome = yicesCheck () in
    yicesPop ();
    outcome = false

let is_contra () =
    let outcome = yicesCheck () in
    let negOutcome = if outcome then false else true in
    negOutcome


let block_pop () = 
    yicesPop ()

let reset () = 
    renewContext ()

let queryExp exp =
    block_check_unsat (P.Not exp)

let queryExpContext pred query = 
    yicesPush ();
    (* yicesAssert (predToYicesExpr pred); *)
    yicesAssert << sprintf "(not %s)" <| predToYicesExpr query;
    let outcome = not <| yicesCheck () in
    yicesPop ();
    outcome

let print_stats () = 
    failwith "YicesInterface2.print_stats not implemented"

and sat pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;
    let c = yicesCheck () in
    yicesPop ();
    c

and assume pred =
    yicesAssert << predToYicesExpr <| pred

and pop () = 
    yicesPop ()

and getValue name =
    let ctx = getContext () in
    let mdl = yices_get_model ctx in
    lbool2bool << yices_get_value mdl <| yices_get_var_decl_from_name ctx name

and getIntValue name =
    let ctx = getContext () in
    let mdl = yices_get_model ctx in
    snd << yices_get_int_value mdl <| yices_get_var_decl_from_name ctx name

let rec convert_back ptbl f =
  try Hashtbl.find ptbl f
  with Not_found ->
   failwith "Yices::allsat: reconvert trouble"

let allsat f preds orig_plist =
  block_reset ();
  yicesPush();
  let pvars = List.map 
    (fun _ ->
      let s = freshUnknownSymbol () in
      let yicesSym = S.toString s in
      yicesDefine yicesSym "bool" None;
      yicesSym
    ) preds in
  let is_still_sat = ref true in
  let res = ref P.False in
  yicesAssert f;
  List.iter2 
    (fun v p -> let eqv = sprintf "(and (=> %s %s) (=> %s %s))" v p p v in
      yicesAssert eqv) pvars preds;
  while (!is_still_sat) do
    is_still_sat :=  yicesCheck() ;
    if !is_still_sat then begin
(*
      let m = yices_get_model ctx in
*)
      let vals = List.map (fun v -> getValue v) pvars in
      let minterm = List.map2 (fun v pred ->  match v with
	Some true -> pred | Some false -> P.Not pred | _ -> assert false ) vals orig_plist in
      let blocking = 
        let ycmd' = List.fold_left2 (fun sofar thisvar thisval -> 
         let s = match thisval with 
          Some true -> sprintf " (not %s) " thisvar 
         | Some false -> sprintf " %s " thisvar 
         | _ -> "" in
         sofar ^ s) 
         "(or " pvars 
         vals in
        ycmd' ^ ")"
      in 
      yicesAssert blocking ;
      let conj = P.And minterm in
      res := P.Or [conj  ; !res];
    end; 
  done; 
  yicesPop();
  block_reset ();
  !res

let get_allsat p plist orig_plist = 
  let y_p = predToYicesExpr p in
  let y_preds = List.map predToYicesExpr plist in
  let ptbl = Hashtbl.create 13 in
  let phi = allsat y_p y_preds orig_plist
  in
(*
  List.iter2 (fun d p -> Hashtbl.add ptbl d p) y_preds orig_plist;
  convert_back ptbl phi
*) 
  phi
  


let whoami () = 
  Message.msg_string Message.Normal "Yices. Copyright SRI." ; "YicesExt"

