(* Copyright 2008  The Regents of the University of California. 

   Licensed under the Apache License, Version 2.0 (the "License"); you
   may not use this file except in compliance with the License.  You
   may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
   implied.  See the License for the specific language governing
   permissions and limitations under the License. 
*)

open Ast
module S = Symbol
module C = Constant
module E = Expression
module P = Predicate
module M = Message
module O = Options

open Misc
open Printf

exception AtomInExpressionException 
exception ChannelException

type yicesl_context 
external yicesl_mk_context: unit -> yicesl_context = "ml_yicesl_mk_context"
external yicesl_del_context: yicesl_context -> unit = "ml_yicesl_del_context"
external yicesl_read: yicesl_context -> string -> int = "ml_yicesl_read"
external yicesl_inconsistent: yicesl_context -> int = "ml_yicesl_inconsistent"

let symbols = Hashtbl.create 100
let yicesSymbols = Hashtbl.create 100
let yicesSymbolDefStack = Stack.create ()
let _ = Stack.push [] yicesSymbolDefStack
let unknownExprs = Hashtbl.create 100

let stackCounter = ref 0
let yicesContext = ref None
let contextUnsat = ref false
let contextUnsatPushes = ref 0

let yicesDefinitions = [
    Left ("ptr-type", None);
    Right ("null", "ptr-type", None);
    Right ("sel", "(-> ptr-type int int)", None);
    Right ("addr", "(-> int ptr-type)", None);

    Left ("field", None);
    Right ("offset", "(-> ptr-type field ptr-type)", None);

    Left ("set", None);
    Right ("empty", "set", None);
    Right ("add", "(-> set int set)", None);
    Right ("in", "(-> int set bool)", None);
]

let yicesAxioms = [ 
    "(forall (x::int) (= (sel (addr x) 0) x))";
    "(forall (p::ptr-type q::ptr-type f::field) (=> (= (offset p f) (offset q f)) (= p q)))";
    "(forall (x::int) (/= (addr x) null))";
    "(forall (x::int y::int) (= (* (div x y) y) x))";
    "(forall (s::set x::int) (in x (add s x)))";
    (* XXX: "(forall (s::set t::set x::int) (=> (= s (add t x)) (in s x)))"; *)
    "(forall (x::int) (not (in x empty)))";
    (* XXX: "(forall (s::set x::int) (=> (= s empty) (not (in x s))))"; *)
]

(**
 * Yices communication via IPC
 *
 *
let rec yicesCommand =
    let yicesProcess = ref None in
    function cmd_s -> 

    let rec startYices () =
        try M.msg_string M.Debug "YI[L]: starting yices process...";
            yicesProcess := Some (
                Unix.open_process_full "yices" (Unix.environment ()));

            at_exit quitYices;

            if yicesCheck () <> "sat" then
                failwith "yices process failed sanity check"

            else M.msg_string M.Debug "YI[L]: yices process started.";

            List.iter 
                (either (uncurry yicesDefineType) (uncurry3 yicesDefine))
                yicesDefinitions;
            List.iter yicesAssert yicesAxioms

        with e -> begin
            M.msg_string M.Debug ("YI[L]: (1) " ^ (Printexc.to_string e));
            failwith "YI[L]: failed to start yices"
        end

    and quitYices () = 
        M.msg_string M.Debug "YI[L]: quitting yices process...";
        maybe () (compose ignore Unix.close_process_full) 
            !yicesProcess;
        M.msg_string M.Debug "YI[L]: yices process closed.";

    (* and restartYices () =
        quitYices ();
        startYices () --- unused so far *)

    and take_line ic =
        match fst3 (Unix.select [Unix.descr_of_in_channel ic] [] [] 0.005) with 
        | [] -> None
        | _  -> Some (input_line ic) in

    let ic, oc, _ = 
        maybe' (Lazy.lazy_from_fun startYices) ignore !yicesProcess;
        maybe' (lazy (assert false)) id !yicesProcess in

    try M.msg_string M.Debug ("YI[L]: yices command: " ^ cmd_s);
        output_string oc (cmd_s ^ "\n");
        flush oc;

        match take_line ic with
          | None -> 
              M.msg_string M.Debug ("YI[L]: yices didn't reply."); 
              ""
          | Some line ->
              M.msg_string M.Debug ("YI[L]: yices replied: '" ^ line ^"'");
              if line = "unsat" then contextUnsat := true;
              line

    with 
      | e -> begin
          M.msg_string M.Debug ("YI[L]: (2) " ^ (Printexc.to_string e));
          failwith "YI[L]: command error"
      end 
 *
 *)

let rec createContext () =
    yicesContext := Some (yicesl_mk_context ());
    let c = yicesCheck () in
    if c <> "sat" then
        failwith ("yices library failed sanity check---"^c);
    List.iter 
        (either (uncurry yicesDefineType) (uncurry3 yicesDefine))
        yicesDefinitions;
    List.iter yicesAssert yicesAxioms

and deleteContext () =
    maybe () yicesl_del_context !yicesContext

and renewContext () =
    deleteContext ();
    createContext ()

and yicesCommand cmd_s =
    maybe' (Lazy.lazy_from_fun createContext) ignore !yicesContext;
    assert (isSome !yicesContext);

    M.msg_string M.Debug (sprintf "YI[L]: telling yices '%s'" cmd_s);

    if some yicesl_read !yicesContext cmd_s = 0 then 
        failwith "YI[L]: something bad happened";

    let answer = 
        if some yicesl_inconsistent !yicesContext = 0 then "sat" 
        else begin 
            contextUnsat := true; 
            M.msg_string M.Debug (sprintf "YI[L]: yices says unsat");
            "unsat" 
        end in
        
    if cmd_s = "(check)" then begin
        M.msg_string M.Debug (sprintf "YI[L]: yices says %s" answer);
        Some answer end
    else None

and isDefined sym =
    Hashtbl.mem yicesSymbols sym

and yicesDefineType typ def_m =
    if not !contextUnsat then begin
        ignore << yicesCommand <|
            (match def_m with 
              | Some def -> sprintf "(define-type %s %s)" typ def
              | None     -> sprintf "(define-type %s)" typ)
    end

and yicesDefine sym typ def_m = 
    if not !contextUnsat then begin
        ignore << yicesCommand <|
            (match def_m with
              | Some def -> sprintf "(define %s::%s %s)" sym typ def
              | None     -> sprintf "(define %s::%s)" sym typ);
        Hashtbl.add yicesSymbols sym typ;
        Stack.push (sym::(Stack.pop yicesSymbolDefStack)) yicesSymbolDefStack
    end

and yicesAssert expr = 
    if not !contextUnsat then begin
        ignore << yicesCommand << sprintf "(assert %s)" <| expr
    end

and yicesPush () = 
    Stack.push [] yicesSymbolDefStack;
    if !contextUnsat then incr contextUnsatPushes
    else ignore (yicesCommand "(push)");
    incr stackCounter

and yicesPop () = 
    assert (!stackCounter > 0);
    List.iter 
        (Hashtbl.remove yicesSymbols) 
        (Stack.pop yicesSymbolDefStack);
    if !contextUnsat && !contextUnsatPushes > 0 then
        decr contextUnsatPushes
    else begin 
        if !contextUnsat then contextUnsat := false;
        ignore (yicesCommand "(pop)")
    end;
    decr stackCounter

and yicesPopAll () =
    repeat yicesPop !stackCounter

and yicesCheck () = 
    if !contextUnsat then "unsat"
    else some id (yicesCommand "(check)")

and yicesReset () = 
    contextUnsat := false;
    contextUnsatPushes := 0;
    stackCounter := 0;
    ignore (yicesCommand "(reset)")
    
(**
 * Conversion to yices input syntax
 *)

let cannotHandle str where =
    failwith (sprintf "YI[L]: cannot handle '%s' in %s" str where)

and sayCannotHandle str where =
    M.msg_string M.Debug (sprintf "YI[L]: cannot handle '%s' in %s" str where)

let freshUnknownSymbol =
    let freshUnknownSymbolCounter = ref 0 in
    fun () ->
        let sym = sprintf "unknown_%d" !freshUnknownSymbolCounter in
        incr freshUnknownSymbolCounter;
        sym

let rec predToYicesExpr pred =
    match pred with
      | P.True -> "true"
      | P.False -> "false"
      | P.Atom e -> atomToYicesExpr e

      | P.And [] -> "true"
      | P.And plist -> 
          sprintf "(and %s)" << String.concat " " << List.map predToYicesExpr
          <| plist

      | P.Or [] -> "false"
      | P.Or plist ->
          sprintf "(or %s)" << String.concat " " << List.map predToYicesExpr
          <| plist

      | P.Implies (e, e') ->
          sprintf "(=> %s %s)" (predToYicesExpr e) (predToYicesExpr e');

      | P.Not e -> sprintf "(not %s)" << predToYicesExpr <| e

      | P.Iff (e, e') -> 
          let s, s' = predToYicesExpr e, predToYicesExpr e' in
          sprintf "(and (=> %s %s) (=> %s %s))" s s' s' s

      | P.All (s,e) -> 
          sprintf "(forall (%s::%s) %s)"
              "?" 
              "?" 
              (predToYicesExpr e)

      | _ -> cannotHandle (P.toString pred) "predToYicesExpr"

and atomToYicesExpr expr =
    match expr with
      | E.Binary (op, e, e') -> begin
          try sprintf "(%s %s %s)" (match op with
                | E.Eq -> "="
                | E.Ne -> "/="
                | E.Le -> "<="
                | E.Lt -> "<"
                | E.Gt -> ">"
                | E.Ge -> ">="
                | _ -> cannotHandle (E.binaryOpToString op) "atomToYicesExpr")
              (exprToYicesExpr e) (exprToYicesExpr e')

          with AtomInExpressionException -> begin
              match op with E.Eq | E.Ne -> begin
                  match e, e' with
                  | E.Lval _, _ | E.Chlval _, _ | E.Constant (Constant.Int _), _ ->
                      let s1 = exprToYicesExpr e
                      and s2 = atomToYicesExpr e' in
                      sprintf "(or (and %s (= %s 1)) (and (not %s) (= %s 0)))" s2 s1 s2 s1
                  | _, E.Lval _ | _, E.Chlval _ | _, E.Constant (Constant.Int _) ->
                      let s1 = exprToYicesExpr e'
                      and s2 = atomToYicesExpr e in
                      sprintf "(or (and %s (= %s 1)) (and (not %s) (= %s 0)))" s2 s1 s2 s1
                  | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (*)"
              end
              | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (**)"
          end
          | _ -> cannotHandle (E.toString expr) "atomToYicesExpr (***)"
      end

      | E.FunctionCall (E.Lval (E.Symbol _), args) ->
          sprintf "(= %s 0)" (exprToYicesExpr expr)

      | E.Lval (E.Symbol s) -> symbolToYicesExpr s

      | _ -> cannotHandle (E.toString expr) "atomToYicesExpr"

and exprToYicesExpr expr =
    match expr with
      | E.Lval lv -> lvalToYicesExpr lv
      | E.Binary (op, e, e') -> 
          sprintf "(%s %s %s)" (match op with
            | E.Minus -> "-"
            | E.Plus -> "+"
            | E.Mul -> "*"
            | E.Div -> "/"
            | E.Rem -> "mod"
            | E.FieldOffset -> "offset"
            | _ when E.isRelOp op ->
                    raise AtomInExpressionException
            | _ -> cannotHandle (E.binaryOpToString op) 
                "exprToYicesExpr(Binary)")
          (exprToYicesExpr e) (exprToYicesExpr e')

      | E.Constant c -> begin
          match c with
            | Constant.Int i -> sprintf "%d" i
            | Constant.String s -> sprintf "\"%s\"" s
            | Constant.Float f -> sprintf "%f" f
            | _ -> cannotHandle (C.toString c) "exprToYicesExpr(Constant)"
      end

      | E.Unary (op, e) ->
          sprintf "(%s %s)" (match op with
            | E.UnaryMinus -> "-"
            | E.UnaryPlus -> "+"
            | E.Address -> "addr"
            | _ -> cannotHandle (E.unaryOpToString op) 
                "exprToYicesExpr(Unary)")
          (exprToYicesExpr e)

      | E.FunctionCall (E.Lval (E.Symbol f), args) ->
          sprintf "(fn_call %s %s)" 
              (symbolToYicesExpr f)
              (String.concat " " (List.map exprToYicesExpr args))

      | E.Select (e,e') -> 
          sprintf "(sel %s %s)" 
              (exprToYicesExpr e)
              (exprToYicesExpr e')

      | E.Store (e,e',e'') ->
          sprintf "(store %s %s %s)" 
              (exprToYicesExpr e)
              (exprToYicesExpr e')
              (exprToYicesExpr e'')

      | _ -> 
          sayCannotHandle (E.toString expr) "exprToYicesExpr";
          unknownExprSymbol expr

and lvalToYicesExpr lval =
    match lval with
      | E.Symbol s -> symbolToYicesExpr s

      | E.Access (E.Dot, e, s) -> 
          sprintf "(offset %s %s)"
              (exprToYicesExpr e)
              (symbolToYicesExpr s)

      | E.Dereference e -> 
          sprintf "(sel %s 0)" 
              (exprToYicesExpr e)

      | E.Indexed (e, e') -> 
          sprintf "(sel %s %s)" 
              (exprToYicesExpr e) 
              (exprToYicesExpr e')

      | _ -> cannotHandle (E.lvalToString lval) "lvalToYicesExpr"

and unknownExprSymbol expr = 
    symbolToYicesExpr (
        if Hashtbl.mem unknownExprs expr then
            Hashtbl.find unknownExprs expr
        else begin
            let s = freshUnknownSymbol () in
            Hashtbl.add unknownExprs expr s;
            s
        end)

and symbolToYicesExpr sym =
    let yicesSym = 
        if Hashtbl.mem symbols sym then 
            Hashtbl.find symbols sym 
        else begin 
            let s = S.toString sym in 
            Hashtbl.add symbols sym s; 
            s
        end in

    if not (isDefined yicesSym) then
        yicesDefine yicesSym "int" None;

    yicesSym

(**
 * Implementation of the THEOREM_PROVER signature
 *)

let block_reset () = 
    yicesPopAll ()

and block_assert pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;

and block_check_unsat pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;
    let outcome = yicesCheck () in
    yicesPop ();
    outcome = "unsat"

and is_contra () =
    let outcome = (yicesCheck () = "unsat") in 
    M.msg_string M.Debug (sprintf "YI[L]: is_contra '%b'" outcome);
    outcome

and block_pop () = 
    yicesPop ()

and reset () = 
    renewContext ()

and queryExp exp =
    failwith "YI.queryExp: let me know if somebody uses this."
    (* yicesPush ();
    yicesAssert (predToYicesExpr exp);
    let outcome = yicesCheck () in
    yicesPop ();
    outcome <> "unsat" *)

and queryExpContext pred query = 
    yicesPush ();
    (* yicesAssert (predToYicesExpr pred); *)
    yicesAssert (sprintf "(not %s)" (predToYicesExpr query));
    let outcome = yicesCheck () in
    yicesPop ();
    outcome = "unsat"

and print_stats () = 
    M.msg_string M.Normal ("yices status: " ^ 
        (maybe "" id (yicesCommand "(status)")))

and sat pred = 
    yicesPush ();
    yicesAssert << predToYicesExpr <| pred;
    let outcome = yicesCheck () in
    yicesPop ();
    outcome = "sat"

and assume pred =
    yicesAssert << predToYicesExpr <| pred

and pop () = 
    yicesPop ()
let get_allsat p plist = failwith "Yices: allsat: To do"
let whoami () = 
  Message.msg_string Message.Normal "Yices. Copyright SRI." ; "YicesLite"

