(*
 * Copyright  1990-2002 The Regents of the University of California. All rights reserved. 
 *
 * Permission is hereby granted, without written agreement and without 
 * license or royalty fees, to use, copy, modify, and distribute this 
 * software and its documentation for any purpose, provided that the 
 * above copyright notice and the following two paragraphs appear in 
 * all copies of this software. 
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY 
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES 
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN 
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY 
 * OF SUCH DAMAGE. 
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY 
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS 
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION 
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *)

module O = Options
module M = Message
module P = Ast.Predicate
module E = Ast.Expression
module C_SD = BlastCSystemDescr.C_System_Descr
module Cmd = BlastCSystemDescr.C_Command
module S = Bstats
module AA = AliasAnalyzer
module A = Abstraction.C_Abstraction
module TP = (YicesVsSimplifyInterface : BlastArch.THEOREMPROVER_QUERY)
module CFBMC = CfbLazyModelChecker.Make_Lazy_Model_Checker(A)
module FMC = FociModelChecker.Make_Foci_Model_Checker(A)

let final_string_ref = ref ""
let preds_ref = ref P.True

let read_preds fn =
    if fn = "" then []
    else try
        M.msg_string M.Normal "Reading in seed predicates...";
        let preds =
            let lexbuf = Errormsg.startParsing fn in
            Inputparse.main Inputflex.token lexbuf in
        Errormsg.finishParsing();
        M.msg_string M.Normal "Seed predicates read.";
        preds
    with Sys_error _ | Errormsg.Error -> begin
        M.msg_string M.Error ("Error raised reading preds in: "^fn);
        []
    end

let runmodelcheck () =
  M.msg_string M.Normal "****** Now running foci model checker ******";flush stdout;
  let _ = A.initialize_abstraction () in
  let r0 = A.create_region false !preds_ref P.True in 
  let err = A.create_region true P.True P.True in
  let mc = if O.getValueOfBool "fmc" then FMC.model_check else CFBMC.model_check in
  try 
    let rv = S.time "Model Checking" mc r0 err in 
    A.dump_abs (); rv
  with e -> (A.dump_abs (); A.print_stats (); raise e)

let print_header () = 
  M.msg_string M.Normal "B2 v0.1. Copyright 2006 Regents of the University of California.";
  M.msg_string M.Debug "Blast was run with the following command line:";
  Array.iter (fun s -> M.msg_string M.Debug (s^" ")) Sys.argv ;
  Abs_config.whoami()

let read_options () = 
  Ast.Counter.set_max (O.getValueOfInt "maxcounter"); 
  (if O.getValueOfBool "msvc" then Cil.msvcMode := true);
(*
  (if (O.getValueOfBool "simpquery") then TP.set_simplify_query_flag ());
*)
 ()

let parse_files () = 
  let srcs = O.getSourceFiles () in
  (if srcs = [] then (Misc.error "no input sources")); 
  M.msg_string M.Normal ("Begin Parsing: " ^ (Misc.strList srcs));
  C_SD.initialize_blast (C_SD.initialize_cil srcs);
  M.msg_string M.Normal ("done Parsing");
  M.msg_string M.Debug "Printing System Description:"; 
  M.msg_printer M.Debug C_SD.print_system ()

let alias_analysis () = 
  S.time "Alias DB" AA.constructAliasDatabase (); 
  M.msg_string M.Normal ("done aliasDB");
  AA.constructMustAliasDatabase (); 
  M.msg_string M.Normal ("done mustAliasDB");
  S.time "Mods DB" (C_SD.post_alias_analysis AA.get_lval_aliases_scope) ();
  M.msg_string M.Normal ("done modsDB");
  AA.add_alias_override (P.conjoinL (read_preds (O.getValueOfString "aliasfile"))); 
  M.msg_string M.Normal "done overRideDB"

let call_graph () = 
  C_SD.CallGraph.compute_callgraph ();
  let fn = O.getValueOfString "callgraph" in
  (if fn <> "" then C_SD.CallGraph.output_callgraph fn);
  C_SD.CallGraph.output_call_paths ()

let finish () = 
  let memstats = Gc.stat () in
  M.msg_string M.Normal ("Minor Words : "^(string_of_float memstats.Gc.minor_words)) ;
  M.msg_string M.Normal ("Major Words : "^(string_of_float memstats.Gc.major_words)) ;
  M.msg_string M.Normal ("Total size of heap in words : "^(string_of_int memstats.Gc.heap_words)) ;
  (if not (O.getValueOfBool "quiet") then S.print stderr "Timings:\n");
  M.msg_string M.Normal !final_string_ref

let _ =
try 
  M.set_verbosity M.Default;  
  O.buildOptionsDatabase (); O.dumpOptionsDatabase ();
  S.reset ();
  print_header (); 
  Cil.initCIL (); 
  parse_files (); 
  alias_analysis (); call_graph ();
  preds_ref := P.conjoinL (read_preds (O.getValueOfString "pred")) ;
  runmodelcheck ();
  finish ()
with e -> (M.msg_string M.Error (O.abort_msg^(Printexc.to_string e)); raise e)
